@@ -64,7 +64,6 @@ int    min_width  = 103;
 pipe_item *pipe_chain = NULL;
 
 char  *qualify_domain = NULL;
-char  *queue_buffer = NULL;
 int    queue_depth = 200;
 char  *queue_font = NULL;
 int    queue_max_addresses = 10;
@@ -109,9 +108,13 @@ FILE  *debug_file             = NULL;
 int    debug_level            = 0;
 BOOL   debug_trace_memory     = FALSE;
 int    deliver_datafile       = -1;
+BOOL   deliver_firsttime      = FALSE;
 BOOL   deliver_freeze         = FALSE;
 int    deliver_frozen_at      = 0;
 BOOL   deliver_manual_thaw    = FALSE;
+char  *dsn_envid              = 0;
+int    dsn_ret                = 0;
+
 char  *exim_path              = BIN_DIRECTORY "/exim"
 			"\0<---------------Space to patch exim_path->";
 
@@ -142,9 +145,10 @@ uid_t  originator_uid;
 BOOL   print_topbitchars      = FALSE;
 
 int    received_count         = 0;
+char  *received_protocol      = NULL;
 int    received_time          = 0;
 int    recipients_count       = 0;
-char **recipients_list        = NULL;
+recipient_item *recipients_list = NULL;
 int    recipients_list_max    = 0;
 
 char  *sender_address         = NULL;
@@ -159,6 +163,8 @@ volatile BOOL sigalrm_seen;
 BOOL   split_spool_directory  = FALSE;
 char  *spool_directory        = SPOOL_DIRECTORY
 			  "\0<--------------Space to patch spool_directory->";
+int    store_max_malloc       = 0;
+int    store_pool             = POOL_MAIN;
 
 tree_node *tree_duplicates    = NULL;
 tree_node *tree_nonrecipients = NULL;
