@@ -211,7 +211,7 @@ skip_item *sk = queue_skip;
 while (sk != NULL)
   {
   skip_item *next = sk->next;
-  free(sk);
+  store_free(sk);
   sk = next;
   }
 queue_skip = NULL;
@@ -377,7 +377,7 @@ switch(dialog_action)
 
   /* Add item to skip queue */
 
-  sk = (skip_item *)malloc(sizeof(skip_item) + (int)strlen(s));
+  sk = (skip_item *)store_malloc(sizeof(skip_item) + (int)strlen(s));
   sk->next = queue_skip;
   queue_skip = sk;
   strcpy(sk->text, s);
@@ -451,7 +451,7 @@ while (p != NULL)
     {
     close(p->fd);
     *pp = p->next;
-    free(p);
+    store_free(p);
     /* If configured, cause display update */
     if (action_queue_update) tick_queue_accumulator = 999999;
     }
@@ -483,7 +483,6 @@ Widget stripchart_form_widget,
        quit_widget,
        resize_widget;
 
-
 /* The exim global message_id needs to get set */
 
 message_id_external = message_id_option + 1;
@@ -491,10 +490,11 @@ message_id = message_id_external + 1;
 message_subdir[1] = 0;
 
 /* Some store needs getting for big_buffer, which is used for
-constructing file names and things. */
+constructing file names and things. This call will initialize
+the store_get() function. */
 
 big_buffer_size = 1024;
-big_buffer = (char *)malloc(big_buffer_size);
+big_buffer = store_get(big_buffer_size);
 
 /* Set up the version string and date and output them */
 
@@ -502,10 +502,6 @@ version_init();
 printf("\nExim Monitor version %s (compiled %s) initializing\n",
   version_string, version_date);
 
-/* Get the large buffer from free store */
-
-queue_buffer = (char *)malloc(queue_buffer_size);
-
 /* Initialize various things from the environment and arguments. */
 
 init(argc, argv);
@@ -516,7 +512,7 @@ signal(SIGCHLD, sigchld_handler);
 
 /* Get the buffer for storing the string for the log display. */
 
-log_display_buffer = (char *)malloc(log_buffer_size);
+log_display_buffer = (char *)store_malloc(log_buffer_size);
 log_display_buffer[0] = 0;
 
 /* Initialize the data structures for the stripcharts */
