@@ -339,7 +339,12 @@ if (!delivery)
   if (rc == 0 && strcmp(action + (int)strlen(action) - 4, "-Mes") == 0)
     {
     queue_item *q = find_queue(id, queue_noop, 0);
-    if (q != NULL) q->sender = string_copy(address_arg);
+    if (q != NULL)
+      {
+      if (q->sender != NULL) store_free(q->sender);
+      q->sender = store_malloc((int)strlen(address_arg) + 1);
+      strcpy(q->sender, address_arg);
+      }
     }
 
   /* If configured, cause a display update and return */
@@ -382,7 +387,7 @@ if ((pid = fork()) == 0)
 
 if (pid < 0) text_showf(text, "Failed to fork: %s\n", strerror(pid)); else
   {
-  pipe_item *p = (pipe_item *)malloc(sizeof(pipe_item));
+  pipe_item *p = (pipe_item *)store_malloc(sizeof(pipe_item));
 
   if (p == NULL)
     {
@@ -637,6 +642,11 @@ char buffer[256];
 header_line *h, *next;
 Widget text = text_create((char *)client_data, text_depth);
 
+/* Remember the point in the dynamic store so we can recover to it afterwards.
+Then use Exim's function to read the header. */
+
+void *reset_point = store_get(0);
+
 sprintf(buffer, "%s-H", (char *)client_data);
 if (spool_read_header(buffer, TRUE, FALSE) != spool_read_OK)
   {
@@ -650,6 +660,7 @@ if (spool_read_header(buffer, TRUE, FALSE) != spool_read_OK)
     else text_showf(text, "Format error in spool file %s\n", buffer);
     }
   else text_showf(text, "Read error for spool file %s\n", buffer);
+  store_reset(reset_point);
   return;
   }
 
@@ -657,7 +668,6 @@ if (sender_address != NULL)
   {
   text_showf(text, "%s sender: <%s>\n", sender_local? "Local" : "Remote",
     sender_address);
-  free(sender_address);
   }
 
 if (recipients_list != NULL)
@@ -667,23 +677,20 @@ if (recipients_list != NULL)
   for (i = 0; i < recipients_count; i++)
     {
     text_showf(text, "  %s %s\n",
-      (tree_search(tree_nonrecipients, recipients_list[i]) == NULL)? " ":"*",
-        recipients_list[i]);
-    free(recipients_list[i]);
+      (tree_search(tree_nonrecipients, recipients_list[i].address) == NULL)?
+        " ":"*", recipients_list[i].address);
     }
   text_show(text, "\n");
-  free(recipients_list);
   }
 
-tree_free(tree_nonrecipients);
-
 for (h = header_list; h != NULL; h = next)
   {
   next = h->next;
   text_showf(text, "%c ", h->type);   /* Don't push h->text through a %s */
   text_show(text, h->text);           /* expansion as it may be v large */
-  free(h);
   }
+
+store_reset(reset_point);
 }
 
 
