@@ -61,7 +61,7 @@ while (*d != NULL)
     if (action != dest_remove) return *d;
     dd = *d;
     *d = dd->next;
-    free(dd);
+    store_free(dd);
     return NULL;
     }
   d = &((*d)->next);
@@ -69,7 +69,7 @@ while (*d != NULL)
 
 if (action != dest_add) return NULL;
 
-dd = (dest_item *)malloc(sizeof(dest_item) + (int)strlen(name));
+dd = (dest_item *)store_malloc(sizeof(dest_item) + (int)strlen(name));
 strcpy(dd->address, name);
 dd->next = NULL;
 dd->parent = NULL;
@@ -89,11 +89,11 @@ dest_item *dd = p->destinations;
 while (dd != NULL)
   {
   dest_item *next = dd->next;
-  free(dd);
+  store_free(dd);
   dd = next;
   }
-if (p->sender != NULL) free(p->sender);
-free(p);
+if (p->sender != NULL) store_free(p->sender);
+store_free(p);
 }
 
 
@@ -105,8 +105,9 @@ static queue_item *set_up(char *name, int dir_char)
 {
 int i, rc, save_errno;
 struct stat statdata;
+void *reset_point;
 char *p;
-queue_item *q = (queue_item *)malloc(sizeof(queue_item));
+queue_item *q = (queue_item *)store_malloc(sizeof(queue_item));
 char buffer[256];
 
 /* Initialize the block */
@@ -120,12 +121,17 @@ q->dir_char = dir_char;
 q->sender = NULL;
 q->size = 0;
 
-/* Read the header file from the spool; if there is a failure it
-might mean inaccessibility as a result of protections.
-A successful read will have caused sender_address to get set and
-the recipients fields to be initialized. If there's a format error
-in the headers, we can still display info from the envelope. */
+/* Read the header file from the spool; if there is a failure it might mean
+inaccessibility as a result of protections. A successful read will have caused
+sender_address to get set and the recipients fields to be initialized. If
+there's a format error in the headers, we can still display info from the
+envelope.
 
+Before reading the header remember the position in the dynamic store so that
+we can recover the store into which the header is read. All data read by
+spool_read_header that is to be preserved is copied into malloc store. */
+
+reset_point = store_get(0);
 message_size = 0;
 message_subdir[0] = dir_char;
 sprintf(buffer, "%s-H", name);
@@ -142,12 +148,12 @@ if (rc != spool_read_OK && rc != spool_read_hdrerror)
   q->update_time = q->input_time = t;
   }
 
-/* Envelope read; get input time and sender. */
+/* Envelope read; get input time and remove qualify_domain from sender address,
+if it's there. */
 
 else
   {
   q->update_time = q->input_time = received_time;
-  /* Remove qualify_domain from sender address, and then save it. */
   if ((p = strstric(sender_address+1, qualify_domain, FALSE)) != NULL &&
     *(--p) == '@') *p = 0;
   }
@@ -174,7 +180,6 @@ if (rc != spool_read_OK)
   if (rc == spool_read_hdrerror)
     {
     (void)find_dest(q, msg, dest_add);
-    free(msg);
     }
   else
     {
@@ -189,11 +194,14 @@ q->frozen = deliver_freeze;
 
 if (user_null_sender && sender_address != NULL && sender_address[0] != 0)
   {
-  q->sender = malloc((int)strlen(sender_address) + 6);
+  q->sender = store_malloc((int)strlen(sender_address) + 6);
   sprintf(q->sender, "<> (%s)", sender_address);
-  free(sender_address);
   }
-else q->sender = sender_address;
+else
+  {
+  q->sender = store_malloc((int)strlen(sender_address) + 1);
+  strcpy(q->sender, sender_address);
+  }
 
 sender_address = NULL;
 
@@ -209,7 +217,7 @@ if (recipients_list != NULL)
   for (i = 0; i < recipients_count; i++)
     {
     BOOL delivered = TRUE;
-    char *r = recipients_list[i];
+    char *r = recipients_list[i].address;
     if (tree_search(tree_nonrecipients, r) == NULL)
       {
       char *rr = r;
@@ -225,12 +233,12 @@ if (recipients_list != NULL)
         *(--p) == '@') *p = 0;
       (void)find_dest(q, r, dest_add);
       }
-    free(r);
     }
-  free(recipients_list);
   }
 
-tree_free(tree_nonrecipients);
+/* Recover the dynamic store used by spool_read_header(). */
+
+store_reset(reset_point);
 return q;
 }
 
@@ -555,6 +563,7 @@ static void update_recipients(queue_item *p)
 {
 int i;
 FILE *jread;
+void *reset_point;
 struct stat statdata;
 char buffer[1024];
 
@@ -569,10 +578,16 @@ if (jread == NULL)
     return;
   }
 
-/* Get the contents of the header file; if any problem, just give up. */
+/* Get the contents of the header file; if any problem, just give up.
+Arrange to recover the dynamic store afterwards. */
 
+reset_point = store_get(0);
 sprintf(buffer, "%s-H", p->name);
-if (spool_read_header(buffer, FALSE, TRUE) != spool_read_OK) return;
+if (spool_read_header(buffer, FALSE, TRUE) != spool_read_OK)
+  {
+  store_reset(reset_point);
+  return;
+  }
 
 /* If there's a journal file, add its contents to the non-recipients tree */
 
@@ -596,7 +611,7 @@ if (recipients_list != NULL)
   for (i = 0; i < recipients_count; i++)
     {
     char *pp;
-    char *r = recipients_list[i];
+    char *r = recipients_list[i].address;
     tree_node *node = tree_search(tree_nonrecipients, r);
 
     if (node == NULL)
@@ -615,20 +630,18 @@ if (recipients_list != NULL)
       (void)find_dest(p, r, dest_add);
     else
       (void)find_dest(p, r, dest_remove);
-    free(r);
     }
-  free(recipients_list);
   }
 
 /* We also need to scan the tree of non-recipients, which might
 contain child addresses that are not in the recipients list, but
 which may have got onto the address list as a result of eximon
-noticing an == line in the log. Then free the tree, remember the
-update time, and we are done. */
+noticing an == line in the log. Then remember the update time,
+recover the dynamic store, and we are done. */
 
 scan_tree(p, tree_nonrecipients);
-tree_free(tree_nonrecipients);
 p->update_time = statdata.st_mtime;
+store_reset(reset_point);
 }
 
 
@@ -667,6 +680,16 @@ while (p != NULL)
       {
       u = 'd';
       t = (t + 12)/24;
+      if (t > 99)                    /* someone had > 99 days */
+        {
+        u = 'w';
+        t = (t + 3)/7;
+        if (t > 99)                  /* so, just in case */
+          {
+          u = 'y';
+          t = (t + 26)/52;
+          }
+        }
       }
     }
 
@@ -697,7 +720,7 @@ while (p != NULL)
       while (sk != NULL && now >= sk->reveal)
         {
         *skp = sk->next;
-        free(sk);
+        store_free(sk);
         sk = *skp;
         if (queue_skip == NULL)
           {
