@@ -19,16 +19,18 @@ happen to be convenient for this program. */
 
 /* Unpick a variable-length argument list and set up an
 appropriate call to XtSetValues. To make it reasonably
-efficient, we keep a working Arg structure of length 10;
-anything setting more than that uses malloc/free. */
+efficient, we keep a working Arg structure of length 15;
+the largest call in eximon sets 11 values. The code uses
+malloc/free if more, just in case there is ever a longer
+one that gets overlooked. */
 
-static Arg xs_temparg[10];
+static Arg xs_temparg[15];
 
 void xs_SetValues(Widget w, Cardinal num_args, ...)
 {
 int i;
 va_list ap;
-Arg *aa = (num_args > 10)? (Arg *)malloc(num_args*sizeof(Arg)) : xs_temparg;
+Arg *aa = (num_args > 15)? (Arg *)malloc(num_args*sizeof(Arg)) : xs_temparg;
 va_start(ap, num_args);
 for (i = 0; i < num_args; i++)
   {
@@ -36,8 +38,7 @@ for (i = 0; i < num_args; i++)
   aa[i].value = va_arg(ap, XtArgVal);
   }
 XtSetValues(w, aa, num_args);
-if (num_args > 10) free(aa);
+if (num_args > 15) free(aa);
 }
 
 /* End of em_xs.c */
-
