@@ -36,6 +36,7 @@ typedef struct {
 } address_setup;
 
 host_setup host_setup_list[] = {
+  { &helo_accept_junk_hosts,     &helo_accept_junk_hostlist },
   { &helo_verify_except_hosts,   &helo_verify_except_hostlist },
   { &sender_verify_except_hosts, &sender_verify_except_hostlist },
   { &rfc1413_except_hosts,       &rfc1413_except_hostlist },
@@ -45,6 +46,7 @@ host_setup host_setup_list[] = {
   { &sender_host_reject_except,  &sender_host_reject_except_hosts },
   { &sender_host_reject_recipients, &sender_host_reject_recipients_hosts },
   { &sender_host_reject_relay,   &sender_host_reject_relay_hosts },
+  { &sender_host_reject_relay_except, &sender_host_reject_relay_except_hosts },
   { &smtp_etrn_hosts,            &smtp_etrn_hostlist },
   { &smtp_expn_hosts,            &smtp_expn_hostlist },
   { &smtp_reserve_hosts,         &smtp_reserve_hostlist },
@@ -55,6 +57,7 @@ host_setup host_setup_list[] = {
 };
 
 net_setup net_setup_list[] = {
+  { &helo_accept_junk_nets,      &helo_accept_junk_netlist },
   { &helo_verify_except_nets,    &helo_verify_except_netlist },
   { &host_lookup_nets,           &host_lookup_netlist },
   { &rbl_except_nets,            &rbl_except_netlist },
@@ -66,6 +69,7 @@ net_setup net_setup_list[] = {
   { &sender_net_reject_except,   &sender_net_reject_except_nets },
   { &sender_net_reject_recipients, &sender_net_reject_recipients_nets },
   { &sender_net_reject_relay,    &sender_net_reject_relay_nets },
+  { &sender_net_reject_relay_except, &sender_net_reject_relay_except_nets },
   { &smtp_etrn_nets,             &smtp_etrn_netlist },
   { &smtp_reserve_nets,          &smtp_reserve_netlist },
   { &receiver_unqualified_nets,  &receiver_unqualified_netlist },
@@ -125,7 +129,7 @@ static void
 daemon_sigalrm_handler(int sig)
 {
 daemon_sigalrm_seen = TRUE;
-signal(SIGALRM, daemon_sigalrm_handler);
+os_non_restarting_signal(SIGALRM, daemon_sigalrm_handler);
 }
 
 
@@ -179,23 +183,26 @@ signal(SIGCHLD, SIG_DFL);
 
 /* This function is called when an SMTP connection has been accepted.
 If there are too many, give an error message and close down. Otherwise
-spin off a sub-process to handle the call. The listening socket is
-required so that it can be closed in the sub-process.
+spin off a sub-process to handle the call. The list of listening sockets
+is required so that they can be closed in the sub-process. Take care not to
+leak store in this process - reset the stacking pool at the end.
 
 Arguments:
-  listen_socket     socket which is listening for incoming calls
-  accept_socket     socket of the current accepted call
-  accepted          socket information about the current call
+  listen_sockets        sockets which are listening for incoming calls
+  listen_socket_count   count of listening sockets
+  accept_socket         socket of the current accepted call
+  accepted              socket information about the current call
 
 Returns:            nothing
 */
 
 static void
-handle_smtp_call(int listen_socket, int accept_socket,
-  struct sockaddr *accepted)
+handle_smtp_call(int *listen_sockets, int listen_socket_count,
+  int accept_socket, struct sockaddr *accepted)
 {
 pid_t pid;
 int dup_accept_socket = dup(accept_socket);
+void *reset_point = store_get(0);
 
 /* Make the address available in ASCII representation. */
 
@@ -208,7 +215,8 @@ socket worked, and if so, make an input file. */
 smtp_out = fdopen(accept_socket, "w");
 if (dup_accept_socket < 0)
   {
-  DEBUG(1) debug_printf("Couldn't dup socket descriptor\n");
+  log_write(0, LOG_MAIN, "Couldn't dup socket descriptor: %s\n",
+    strerror(errno));
   DEBUG(1) debug_printf("421 %s: Connection setup failed: %s\n",
     primary_hostname, strerror(errno));
   fprintf(smtp_out, "421 %s: Connection setup failed: %s\r\n",
@@ -265,17 +273,18 @@ pid = fork();
 
 if (pid == 0)
   {
+  int i;
+  int message_count = 0;
   BOOL local_queue_only = queue_only;
-  BOOL smtp_first = TRUE;
 
-  /* Close the listening socket, and set the SIGCHLD handler to SIG_IGN.
+  /* Close the listening sockets, and set the SIGCHLD handler to SIG_IGN.
   This is not the same as SIG_DFL, despite the fact that documentation often
   lists the default as "ignore". At least on some systems, setting SIG_IGN
   causes child processes that complete simply to go away without ever becoming
   defunct. You can't therefore wait for them - but in this process we don't
   want to wait for them as they are doing independent deliveries. */
 
-  close(listen_socket);
+  for (i = 0; i < listen_socket_count; i++) close(listen_sockets[i]);
   signal(SIGCHLD, SIG_IGN);
 
   /* Attempt to get an id from the sending machine via the RFC 1413
@@ -292,26 +301,24 @@ if (pid == 0)
 
   /* If there are too many child processes for immediate delivery,
   set the local_queue_only flag, which is initialized from the
-  configured value and may therefore already be TRUE. */
+  configured value and may therefore already be TRUE. Leave logging
+  till later so it will have a message id attached. */
 
   if (smtp_accept_queue > 0 && smtp_accept_count >= smtp_accept_queue)
-    {
-    DEBUG(1) debug_printf(
-      "Too many connections (%d) for immediate delivery (max %d)\n",
-        smtp_accept_count, smtp_accept_queue);
     local_queue_only = TRUE;
-    }
 
-  /* Now loop, accepting incoming messages from the SMTP connection. The
-  end will come at the QUIT command, when accept_msg() returns FALSE.
-  A break in the connection causes the process to die (see accept.c). */
+  /* Handle the start of the SMTP session, then loop, accepting incoming
+  messages from the SMTP connection. The end will come at the QUIT command,
+  when accept_msg() returns FALSE. A break in the connection causes the
+  process to die (see accept.c). */
 
-  sender_local = sender_host_unknown = FALSE;
+  if (!smtp_start_session()) _exit(EXIT_SUCCESS);
 
   for (;;)
     {
     int rc;
-    message_id[0] = 0;   /* Clear out any previous message_id */
+    message_id[0] = 0;            /* Clear out any previous message_id */
+    reset_point = store_get(0);   /* Save current store high water point */
 
     set_process_info("handling incoming connection from %s",
       sender_fullhost);
@@ -324,9 +331,8 @@ if (pid == 0)
     retry if no message was accepted (can happen for invalid message
     parameters). */
 
-    if ((rc = smtp_setup_msg(smtp_first)) > 0)
+    if ((rc = smtp_setup_msg()) > 0)
       {
-      smtp_first = FALSE;
       (void) accept_msg(smtp_in, smtp_out, FALSE);
       if (message_id[0] == 0) continue;
       }
@@ -343,7 +349,7 @@ if (pid == 0)
         {
         debug_printf("Recipients:\n");
         for (i = 0; i < recipients_count; i++)
-          debug_printf("  %s\n", recipients_list[i]);
+          debug_printf("  %s\n", recipients_list[i].address);
         }
       }
 
@@ -358,28 +364,44 @@ if (pid == 0)
     while ((pid = waitpid(-1, NULL, WNOHANG)) > 0)
       DEBUG(1) debug_printf("child %d ended\n", pid);
 
-    /* Tidy up the store used in accepting this message */
+    /* Reclaim up the store used in accepting this message */
 
-    store_free(sender_address);
-    accept_free_recipients();
+    store_reset(reset_point);
 
     /* If queue_only is set or if there are too many incoming connections in
-    existence, local_queue_only will be TRUE. If it is not, and queue_only_load
-    is set, check that the load average is below it. If it is not, set
-    local_queue_only TRUE. Note that it then remains this way for any
-    subsequent messages on the same SMTP connection. This is a deliberate
-    choice; even though the load average may fall, it doesn't seem right to
-    deliver later messages on the same call when not delivering earlier ones. */
+    existence, local_queue_only will be TRUE. If it is not, check whether we
+    have received too many messages in this session for immediate delivery. If
+    not, and queue_only_load is set, check that the load average is below it.
+    Note that, once set, local_queue_only remains set for any subsequent
+    messages on the same SMTP connection. This is a deliberate choice; even
+    though the load average may fall, it doesn't seem right to deliver later
+    messages on the same call when not delivering earlier ones. */
 
-    if (!local_queue_only && queue_only_load >= 0)
+    if (!local_queue_only)
       {
-      local_queue_only = (load_average = os_getloadavg()) > queue_only_load;
-      DEBUG(1)
+      if (smtp_accept_queue_per_connection > 0 &&
+          ++message_count > smtp_accept_queue_per_connection)
+        {
+        local_queue_only = TRUE;
+        log_write(4, LOG_MAIN, "no immediate delivery: more than %d messages "
+          "received in one connection\n", smtp_accept_queue_per_connection);
+        }
+      else if (queue_only_load >= 0)
+        {
+        local_queue_only = (load_average = os_getloadavg()) > queue_only_load;
         if (local_queue_only)
-          debug_printf("no delivery: load average = %.2f\n",
+          log_write(4, LOG_MAIN, "no immediate delivery: load average = %.2f\n",
             (double)load_average/1000.0);
+        }
       }
 
+    /* Log the queueing here, when it will get a message id attached, but
+    not if queue_only is set. */
+
+    else if (!queue_only) log_write(4, LOG_MAIN,
+      "no immediate delivery: too many connections (%d, max %d)\n",
+        smtp_accept_count, smtp_accept_queue);
+
     /* If a delivery attempt is required, spin off a new process to handle it.
     If we are not root, we have to re-exec exim unless root can be regained by
     the use of seteuid. */
@@ -392,7 +414,7 @@ if (pid == 0)
       if (security_level >= 2)
         {
         int i = 0;
-        char *argv[8 + MAX_CLMACROS];
+        char *argv[9 + MAX_CLMACROS];
         argv[i++] = exim_path;
         if (clmacro_count > 0)
           {
@@ -401,6 +423,7 @@ if (pid == 0)
           }
         if (debug_level > 0)
           argv[i++] = string_sprintf("-d%d", debug_level);
+        if (debug_trace_memory) argv[i++] = "-dm";
         if (dont_deliver) argv[i++] = "-N";
         if (config_changed)
           {
@@ -430,8 +453,9 @@ remember the pid for ticking off when the child completes. */
 
 if (pid < 0)
   {
-  DEBUG(1) debug_printf("fork of SMTP accept process failed\n");
-  fprintf(smtp_out, "421: %s Connection refused: %s\r\n", primary_hostname,
+  DEBUG(1) debug_printf("fork of SMTP accept process failed: %s\n",
+    strerror(errno));
+  fprintf(smtp_out, "421 %s: Fork failed: %s\r\n", primary_hostname,
     strerror(errno));
   }
 else
@@ -464,9 +488,10 @@ if (smtp_in != NULL)
   }
 fclose(smtp_out);
 
-/* Free the store used for holding the incoming host address */
+/* Release any store used in this process, including the store used for holding
+the incoming host address. */
 
-store_free(sender_host_address);
+store_reset(reset_point);
 sender_host_address = NULL;
 }
 
@@ -479,27 +504,26 @@ sender_host_address = NULL;
 
 /* The daemon can do two jobs, either of which is optional:
 
-   (1) Listens for incoming SMTP calls and spawns off a sub-process to handle
-   each one. This is requested by the -bd option, with -oX specifying the SMTP
-   port on which to listen (for testing).
+(1) Listens for incoming SMTP calls and spawns off a sub-process to handle
+each one. This is requested by the -bd option, with -oX specifying the SMTP
+port on which to listen (for testing).
 
-   (2) Spawns a queue-running process every so often. This is controlled by the
-   -q option with a an interval time. (If no time is given, a single queue run
-   is done from the main function, and control doesn't get here.)
+(2) Spawns a queue-running process every so often. This is controlled by the
+-q option with a an interval time. (If no time is given, a single queue run
+is done from the main function, and control doesn't get here.)
 
 Root privilege is required in order to attach to port 25. Some systems require
 it when calling socket() rather than bind(). To cope with all cases, we run as
-root for both socket() and bind(). When a listener is required, this function
-is entered as root for security levels 0 or 2; otherwise seteuid to exim.
+root for both socket() and bind(). Some systems also require root in order to
+write to the pid file directory. This function is entered as root for security
+levels 0 or 2; otherwise seteuid to exim.
 
-Once the socket is bound, root privilege is given up if there is an exim uid,
-either by seteuid for security level 1, or by setuid for higher levels. In the
-latter case, a re-exec is subsequently required to cause a received message to
-be delivered.
+Once the socket is bound and the pid file written, root privilege is given up
+if there is an exim uid, either by seteuid for security level 1, or by setuid
+for higher levels. In the latter case, a re-exec is subsequently required to
+cause a received message to be delivered.
 
-When listening is not required - i.e. the daemon is simply a means of
-periodically starting a queue runner - this function is entered with the
-appropriate uid already set. There are no arguments, and it never returns. */
+There are no arguments to this function, and it never returns. */
 
 void
 daemon_go(void)
@@ -581,6 +605,11 @@ if (debug_level <= 0 && !debug_trace_memory)
   (void)setsid();
   }
 
+/* Ensure root privilege. It will only not exist at this stage if seteuid
+can be used to regain it. */
+
+if (geteuid() != root_uid) mac_seteuid(root_uid);
+
 /* If SMTP listening is requested, set up a socket on the SMTP port or
 a given port, and compile the verification acceptance data, if any, so
 that it is available to all spawned processes. If local_interfaces is set,
@@ -658,7 +687,7 @@ if (daemon_listen)
 
   else
     {
-    addresses = store_malloc(sizeof(ip_address_item));
+    addresses = store_get(sizeof(ip_address_item));
     addresses->next = NULL;
     addresses->address[0] = 0;
     listen_socket_count = 1;
@@ -666,12 +695,7 @@ if (daemon_listen)
 
   /* Get a vector to remember all the sockets in */
 
-  listen_sockets = store_malloc(sizeof(int *) * listen_socket_count);
-
-  /* Ensure root privilege. It will only not exist at this stage if seteuid
-  can be used to regain it. */
-
-  if (geteuid() != root_uid) mac_seteuid(root_uid);
+  listen_sockets = store_get(sizeof(int *) * listen_socket_count);
 
   /* If any option requiring a load average to be available during the
   reception of a message is set, call os_getloadavg() while we are root
@@ -778,25 +802,6 @@ if (daemon_listen)
       }
     }
 
-  /* If exim_uid is set, give up root privilege at this point, using setuid or
-  seteuid as appopriate. The macros expand to -1 on systems that don't have
-  the sete{g,u}id functions, but the security level cannot be set to values
-  implying the use of these functions on such systems. */
-
-  if (exim_uid_set)
-    {
-    if (security_level >= 2)
-      {
-      setgid(exim_gid);
-      setuid(exim_uid);
-      }
-    else
-      {
-      mac_setegid(exim_gid);
-      mac_seteuid(exim_uid);
-      }
-    }
-
   /* Do a sanity check on the max connects value; zero means no limit,
   in which case we don't need to keep a list of them. */
 
@@ -819,7 +824,7 @@ if (daemon_listen)
   if (smtp_pid_slotcount > 0)
     {
     int i;
-    smtp_pid_slots = store_malloc(smtp_pid_slotcount * sizeof(pid_t));
+    smtp_pid_slots = store_get(smtp_pid_slotcount * sizeof(pid_t));
     for (i = 0; i < smtp_pid_slotcount; i++) smtp_pid_slots[i] = 0;
     }
 
@@ -830,6 +835,41 @@ if (daemon_listen)
     listen(listen_sockets[sk], smtp_connect_backlog);
   }
 
+/* Write the pid to a known file for assistance in identification. Make it
+read-only. We do this before giving up root privilege, because on some systems
+it is necessary to be root in order to write into the pid file directory. */
+
+f = fopen(buff, "w");
+if (f != NULL)
+  {
+  fprintf(f, "%d\n", (int)getpid());
+  fchmod(fileno(f), 0644);
+  fclose(f);
+  DEBUG(2) debug_printf("pid written to %s\n", buff);
+  }
+else
+  DEBUG(2) debug_printf("%s\n", string_open_failed(errno, "pid file %s", buff));
+
+/* If exim_uid is set, give up root privilege at this point, using setuid or
+seteuid as appopriate. The macros expand to -1 on systems that don't have
+the sete{g,u}id functions, but the security level cannot be set to values
+implying the use of these functions on such systems. */
+
+if (exim_uid_set)
+  {
+  if (security_level >= 2)
+    {
+    setgid(exim_gid);
+    setuid(exim_uid);
+    }
+  else
+    {
+    mac_setegid(exim_gid);
+    mac_seteuid(exim_uid);
+    }
+  DEBUG(2) debug_printf("uid=%d gid=%d euid=%d egid=%d\n",
+    getuid(), getgid(), geteuid(), getegid());
+  }
 
 /* Get somewhere to keep the list of queue-runner pids if we are keeping track
 of them (and also if we are doing queue runs). */
@@ -840,7 +880,7 @@ if (queue_interval > 0)
   if (queue_pid_slotcount > 0)
     {
     int i;
-    queue_pid_slots = store_malloc(queue_pid_slotcount * sizeof(pid_t));
+    queue_pid_slots = store_get(queue_pid_slotcount * sizeof(pid_t));
     for (i = 0; i < queue_pid_slotcount; i++) queue_pid_slots[i] = 0;
     }
   }
@@ -891,20 +931,6 @@ else
   set_process_info("daemon: no queue runs, port %d", smtp_port);
   }
 
-/* Write the pid to a known file for assistance in identification. Make it
-read-only. */
-
-f = fopen(buff, "w");
-if (f != NULL)
-  {
-  fprintf(f, "%d\n", (int)getpid());
-  fchmod(fileno(f), 0644);
-  fclose(f);
-  DEBUG(2) debug_printf("pid written to %s\n", buff);
-  }
-else
-  DEBUG(2) debug_printf("%s\n", string_open_failed(errno, "pid file %s", buff));
-
 /* Close the log so it can be renamed and moved. This process doesn't write
 to the log again, unless it is about to die or exec and in the latter case
 it closes the log first. */
@@ -1023,7 +1049,7 @@ for (;;)
     /* Reset the alarm time */
 
     daemon_sigalrm_seen = FALSE;
-    signal(SIGALRM, daemon_sigalrm_handler);
+    os_non_restarting_signal(SIGALRM, daemon_sigalrm_handler);
     alarm(queue_interval);
     }
 
@@ -1031,7 +1057,7 @@ for (;;)
   /* Sleep till a connection happens if listening, and handle the connection if
   that is why we woke up. The FreeBSD operating system requires the use of
   select() before accept() because the latter function is not interrupted by
-  a signal, and we want to wake up for SIGCHLD and SIGLARM signals. Some other
+  a signal, and we want to wake up for SIGCHLD and SIGALRM signals. Some other
   OS do notice signals in accept() but it does no harm to have the select()
   in for all of them - and it won't then be a lurking problem for ports to
   new OS. In fact, the later addition of listening on specific interfaces only
@@ -1134,7 +1160,7 @@ for (;;)
       /* If select/accept succeeded, deal with the connection. */
 
       if (accept_socket >= 0)
-        handle_smtp_call(listen_sockets[sk], accept_socket,
+        handle_smtp_call(listen_sockets, listen_socket_count, accept_socket,
 	  (struct sockaddr *)&accepted);
       }
     }
