@@ -90,7 +90,7 @@ created, and we also try to make the directory if the initial open fails. */
 sprintf(buffer, "%s/db/%s.lockfile", spool_directory, name);
 if ((dbblock->lockfd = open(buffer, O_RDWR|O_CREAT, DB_LOCKFILE_MODE)) < 0)
   {
-  directory_make(spool_directory, "db", DB_DIRECTORY_MODE);
+  directory_make(spool_directory, "db", DB_DIRECTORY_MODE, TRUE);
   dbblock->lockfd = open(buffer, O_RDWR|O_CREAT, DB_LOCKFILE_MODE);
   }
 
@@ -118,7 +118,7 @@ lock_data.l_type = read_only? F_RDLCK : F_WRLCK;
 lock_data.l_whence = lock_data.l_start = lock_data.l_len = 0;
 
 sigalrm_seen = FALSE;
-signal(SIGALRM, sigalrm_handler);
+os_non_restarting_signal(SIGALRM, sigalrm_handler);
 alarm(DB_LOCK_TIMEOUT);
 rc = fcntl(dbblock->lockfd, F_SETLKW, &lock_data);
 alarm(0);
@@ -243,7 +243,7 @@ if (setjmp(read_env) == 0)
   found = EXIM_DBGET(dbblock->dbptr, key_datum, result_datum);
   signal(SIGSEGV, SIG_DFL);
   if (!found) return NULL;
-  yield = (void *)store_malloc(EXIM_DATUM_SIZE(result_datum));
+  yield = store_get(EXIM_DATUM_SIZE(result_datum));
   memcpy(yield, EXIM_DATUM_DATA(result_datum), EXIM_DATUM_SIZE(result_datum));
   if (length != NULL) *length = EXIM_DATUM_SIZE(result_datum);
   return yield;
