@@ -17,10 +17,20 @@
 data blocks and hence have the opt_public flag set. */
 
 optionlist optionlist_directors[] = {
+  { "*expand_group",      opt_stringptr | opt_hidden | opt_public,
+                 (void *)(offsetof(director_instance, expand_gid)) },
   { "*expand_transport",  opt_stringptr|opt_public|opt_hidden,
                  (void *)offsetof(director_instance, expand_transport) },
+  { "*expand_user",       opt_stringptr | opt_hidden | opt_public,
+                 (void *)(offsetof(director_instance, expand_uid)) },
+  { "*set_group",         opt_bool | opt_hidden | opt_public,
+                 (void *)(offsetof(director_instance, gid_set)) },
+  { "*set_user",          opt_bool | opt_hidden | opt_public,
+                 (void *)(offsetof(director_instance, uid_set)) },
   { "condition",          opt_stringptr|opt_public,
                  (void *)offsetof(director_instance, condition) },
+  { "debug_print",        opt_stringptr | opt_public,
+                 (void *)offsetof(director_instance, debug_string) },
   { "domains",            opt_stringptr|opt_public,
                  (void *)offsetof(director_instance, domains) },
   { "driver",             opt_stringptr|opt_public,
@@ -41,10 +51,16 @@ optionlist optionlist_directors[] = {
                  (void *)offsetof(director_instance, fail_verify_recipient) },
   { "fail_verify_sender", opt_bool|opt_public,
                  (void *)offsetof(director_instance, fail_verify_sender) },
+  { "fallback_hosts",     opt_stringptr|opt_public,
+                 (void *)offsetof(director_instance, fallback_hosts) },
+  { "group",              opt_expand_gid | opt_public,
+                 (void *)(offsetof(director_instance, gid)) },
   { "headers_add",        opt_stringptr|opt_public,
                  (void *)offsetof(director_instance, extra_headers) },
   { "headers_remove",     opt_stringptr|opt_public,
                  (void *)offsetof(director_instance, remove_headers) },
+  { "initgroups",         opt_bool | opt_public,
+                 (void *)(offsetof(director_instance, initgroups)) },
   { "local_parts",        opt_stringptr|opt_public,
                  (void *)offsetof(director_instance, local_parts) },
   { "more",               opt_bool|opt_public,
@@ -67,6 +83,8 @@ optionlist optionlist_directors[] = {
                  (void *)offsetof(director_instance, transport) },
   { "unseen",             opt_bool|opt_public,
                  (void *)offsetof(director_instance, unseen) },
+  { "user",               opt_expand_uid | opt_public,
+                 (void *)(offsetof(director_instance, uid)) },
   { "verify",             opt_bool_verify | opt_hidden|opt_public,
                  (void *)offsetof(director_instance, verify_sender) },
   { "verify_only",        opt_bool|opt_public,
@@ -83,6 +101,120 @@ int optionlist_directors_size =
 
 
 
+/*************************************************
+*      Set uid/gid from block into address       *
+*************************************************/
+
+/* This function copies any set uid or gid from a ugid block into an
+address.
+
+Arguments:
+  addr        the address
+  ugid        the ugid block
+
+Returns:      nothing
+*/
+
+void
+direct_set_ugid(address_item *addr, ugid_block *ugid)
+{
+if (ugid->uid_set)
+  {
+  addr->uid = ugid->uid;
+  addr->uid_set = TRUE;
+  }
+
+if (ugid->gid_set)
+  {
+  addr->gid = ugid->gid;
+  addr->gid_set = TRUE;
+  addr->initgroups = ugid->initgroups;
+  }
+}
+
+
+
+
+
+/*************************************************
+*          Get uid/gid for a director/router     *
+*************************************************/
+
+/* This function is called by both directors and routers to sort out the
+uid/gid values which are passed with an address for use by local transports.
+The fields it uses are identical in both director and router blocks.
+
+Arguments:
+  dblock       the director/router block
+  driver_type  "director" or "router"
+  addr         the address being worked on
+  ugid         pointer to a ugid block to fill in
+
+Returns:       TRUE if all goes well, else FALSE
+*/
+
+BOOL
+direct_get_ugid(director_instance *dblock, char *driver_type,
+  address_item *addr, ugid_block *ugid)
+{
+struct passwd *upw = NULL;
+
+/* Initialize from fixed values */
+
+ugid->uid = dblock->uid;
+ugid->gid = dblock->gid;
+ugid->uid_set = dblock->uid_set;
+ugid->gid_set = dblock->gid_set;
+ugid->initgroups = dblock->initgroups;
+
+/* If there is no fixed uid set, see if there's a dynamic one that can
+be expanded and possibly looked up. */
+
+if (!ugid->uid_set && dblock->expand_uid != NULL)
+  {
+  if (direct_find_expanded_user(dblock->expand_uid, dblock->name, driver_type,
+    &upw, &(ugid->uid), &(addr->message))) ugid->uid_set = TRUE;
+  else
+    {
+    addr->special_action = SPECIAL_FREEZE;
+    return FALSE;
+    }
+  }
+
+/* Likewise for the gid */
+
+if (!ugid->gid_set && dblock->expand_gid != NULL)
+  {
+  if (direct_find_expanded_group(dblock->expand_gid, dblock->name, driver_type,
+    &(ugid->gid), &(addr->message))) ugid->gid_set = TRUE;
+  else
+    {
+    addr->special_action = SPECIAL_FREEZE;
+    return FALSE;
+    }
+  }
+
+/* If a uid is set, then a gid must also be available; use one from the passwd
+lookup if it happened. */
+
+if (ugid->uid_set && !ugid->gid_set)
+  {
+  if (upw != NULL)
+    {
+    ugid->gid = upw->pw_gid;
+    ugid->gid_set = TRUE;
+    }
+  else
+    log_write(0, LOG_PANIC_DIE, "User set without group for %s %s",
+      dblock->name, driver_type);
+  }
+
+return TRUE;
+}
+
+
+
+
 /*************************************************
 *    Get errors address for a director/router    *
 *************************************************/
@@ -142,8 +274,10 @@ if (verify) *errors_to = s; else
     save1 = sender_address[0];
     sender_address[0] = 0;
     }
+  DEBUG(2) debug_printf("------ Verifying errors address %s ------\n", s);
   if (verify_address(s, NULL, NULL, &snew, vopt_is_recipient | vopt_local)
     == OK) *errors_to = snew;
+  DEBUG(2) debug_printf("------ End verifying errors address %s ------\n", s);
   if (sender_address != NULL) sender_address[0] = save1;
   }
 
@@ -181,55 +315,66 @@ if (dblock->extra_headers == NULL)
   *extra_headers = addr->extra_headers;
 else
   {
-  int slen;
-  BOOL addnl = FALSE;
   header_line *h;
   char *s = expand_string(dblock->extra_headers);
+
   if (s == NULL)
     {
-    addr->message = string_sprintf("%s %s failed to expand \"%s\": %s",
-      dblock->name, type, dblock->extra_headers, expand_string_message);
-    if (search_find_defer) return DEFER;
-    addr->special_action = SPECIAL_FREEZE;
-    return ERROR;
+    if (!expand_string_forcedfail)
+      {
+      addr->message = string_sprintf("%s %s failed to expand \"%s\": %s",
+        dblock->name, type, dblock->extra_headers, expand_string_message);
+      if (search_find_defer) return DEFER;
+      addr->special_action = SPECIAL_FREEZE;
+      return ERROR;
+      }
     }
-  slen = (int)strlen(s);
-  if (slen == 0 || s[slen-1] != '\n')
+
+  /* Expand succeeded */
+
+  else
     {
-    addnl = TRUE;
-    slen++;
+    char *addnl = "";
+    int slen = (int)strlen(s);
+    if (slen > 0)
+      {
+      if (s[slen-1] != '\n')
+        {
+        addnl = "\n";
+        slen++;
+        }
+      h = store_get(sizeof(header_line));
+      h->text = string_sprintf("%s%s", s, addnl);
+      h->next = addr->extra_headers;
+      h->type = htype_other;
+      h->slen = slen;
+      *extra_headers = h;
+      }
     }
-  h = store_malloc(sizeof(header_line) + slen);
-  strcpy(h->text, s);
-  if (addnl) strcpy(h->text + slen - 1, "\n");
-  h->next = addr->extra_headers;
-  h->type = htype_other;
-  h->slen = slen;
-  *extra_headers = h;
-  store_free(s);
   }
 
 if (dblock->remove_headers == NULL)
   *remove_headers = addr->remove_headers;
 else
   {
-  int slen;
   char *s = expand_string(dblock->remove_headers);
   if (s == NULL)
     {
-    addr->message = string_sprintf("%s %s failed to expand \"%s\": %s",
-      dblock->name, type, dblock->remove_headers, expand_string_message);
-    if (search_find_defer) return DEFER;
-    addr->special_action = SPECIAL_FREEZE;
-    return ERROR;
+    if (!expand_string_forcedfail)
+      {
+      addr->message = string_sprintf("%s %s failed to expand \"%s\": %s",
+        dblock->name, type, dblock->remove_headers, expand_string_message);
+      if (search_find_defer) return DEFER;
+      addr->special_action = SPECIAL_FREEZE;
+      return ERROR;
+      }
     }
-
-  if (addr->remove_headers == NULL)
-    *remove_headers = s;
-  else
+  else if (*s != 0)
     {
-    *remove_headers = string_sprintf("%s:%s", addr->remove_headers, s);
-    store_free(s);
+    if (addr->remove_headers == NULL)
+      *remove_headers = s;
+    else
+      *remove_headers = string_sprintf("%s:%s", addr->remove_headers, s);
     }
   }
 
@@ -266,6 +411,8 @@ direct_check_fsc(director_instance *dblock, address_item *addr)
 {
 int rc;
 
+debug_print_string(dblock->debug_string);
+
 switch (match_exists(dblock->require_files))
   {
   case FAIL:
@@ -457,7 +604,6 @@ BOOL
 direct_find_expanded_user(char *string, char *driver_name,
   char *driver_type, struct passwd **pw, uid_t *uid, char **errmsg)
 {
-BOOL yield = TRUE;
 char *user = expand_string(string);
 
 if (user == NULL)
@@ -468,16 +614,12 @@ if (user == NULL)
   return FALSE;
   }
 
-if (!direct_finduser(user, pw, uid))
-  {
-  *errmsg = string_sprintf("Failed to find user \"%s\" from expanded string "
-    "\"%s\" from the %s %s", user, string, driver_name, driver_type);
-  log_write(0, LOG_MAIN|LOG_PANIC, "%s", *errmsg);
-  yield = FALSE;
-  }
+if (direct_finduser(user, pw, uid)) return TRUE;
 
-store_free(user);
-return yield;
+*errmsg = string_sprintf("Failed to find user \"%s\" from expanded string "
+  "\"%s\" from the %s %s", user, string, driver_name, driver_type);
+log_write(0, LOG_MAIN|LOG_PANIC, "%s", *errmsg);
+return FALSE;
 }
 
 
@@ -522,7 +664,6 @@ if (!direct_findgroup(group, gid))
   yield = FALSE;
   }
 
-store_free(group);
 return yield;
 }
 
@@ -550,8 +691,7 @@ readconf_driver_init("director",
   optionlist_directors,                 /* generic options */
   optionlist_directors_size);
 
-/* A director may not have more=false and unseen=true. Also, check that
-any settings of new_director are valid, and set up the actual pointers. */
+/* Check various options for consistency */
 
 for (d = directors; d != NULL; d = d->next)
   {
@@ -560,6 +700,8 @@ for (d = directors; d != NULL; d = d->next)
       "the combination of \"unseen\" and \"no_more\" is not permitted on "
       "a director,\n  but was set for the %s director", d->name);
 
+  /* Check new_director is valid, and set actual pointers */
+
   if (d->new_director != NULL)
     {
     director_instance *dd;
@@ -576,6 +718,16 @@ for (d = directors; d != NULL; d = d->next)
         "new_director \"%s\" not found for \"%s\" director", d->new_director,
           d->name);
     }
+
+  /* If a fixed uid field is set, then a gid field must also be set. */
+
+  if (d->uid_set && !d->gid_set && d->expand_gid == NULL)
+    log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
+      "user set without group for the %s director", d->name);
+
+  /* Build a host list if fallback hosts is set */
+
+  host_build_hostlist(&(d->fallback_hostlist), d->fallback_hosts);
   }
 }
 
@@ -919,7 +1071,7 @@ for (d = (addr->start_director == NULL)? directors : addr->start_director;
         }
       else if (d->unseen)
         {
-        address_item *new = deliver_make_addr(string_copy(addr->orig));
+        address_item *new = deliver_make_addr(addr->orig, TRUE);
         new->parent = addr;
         new->ignore_error |= addr->ignore_error;
         new->errors_address = addr->errors_address;
@@ -970,21 +1122,8 @@ for (d = (addr->start_director == NULL)? directors : addr->start_director;
   if (yield == FORCEFAIL) break;
   }
 
-/* Ensure store for domain and local part data is freed */
-
-if (domain_data != NULL)
-  {
-  store_free(domain_data);
-  domain_data = NULL;
-  }
-
-if (local_part_data != NULL)
-  {
-  store_free(local_part_data);
-  local_part_data = NULL;
-  }
-
-/* No directors accepted this address; fail it. */
+/* No directors accepted this address; fail it, setting a default message
+if there is nothing already set. */
 
 if (addr->message == NULL)
   addr->message = string_sprintf("unknown local-part \"%s\" in domain \"%s\"",
