@@ -14,16 +14,8 @@
 /* Options specific to the aliasfile director. */
 
 optionlist aliasfile_director_options[] = {
-  { "*expand_group",      opt_stringptr | opt_hidden,
-      (void *)(offsetof(aliasfile_director_options_block, expand_gid)) },
-  { "*expand_user",       opt_stringptr | opt_hidden,
-      (void *)(offsetof(aliasfile_director_options_block, expand_uid)) },
   { "*partial_match", opt_int | opt_hidden,
       (void *)(offsetof(aliasfile_director_options_block, partial_match)) },
-  { "*set_group",         opt_bool | opt_hidden,
-      (void *)(offsetof(aliasfile_director_options_block, gid_set)) },
-  { "*set_user",          opt_bool | opt_hidden,
-      (void *)(offsetof(aliasfile_director_options_block, uid_set)) },
   { "check_ancestor",     opt_bool,
       (void *)(offsetof(aliasfile_director_options_block, check_ancestor)) },
   { "current_directory",     opt_stringptr,
@@ -46,16 +38,14 @@ optionlist aliasfile_director_options[] = {
       (void *)(offsetof(aliasfile_director_options_block, forbid_pipe)) },
   { "freeze_missing_include", opt_bool,
       (void *)(offsetof(aliasfile_director_options_block, freeze_missing_include)) },
-  { "group",              opt_expand_gid,
-      (void *)(offsetof(aliasfile_director_options_block, gid)) },
   { "home_directory",     opt_stringptr,
       (void *)(offsetof(aliasfile_director_options_block, home_dir)) },
   { "include_domain",     opt_bool,
       (void *)(offsetof(aliasfile_director_options_block, include_domain)) },
-  { "initgroups",         opt_bool,
-      (void *)(offsetof(aliasfile_director_options_block, initgroups)) },
   { "modemask",           opt_octint,
       (void *)(offsetof(aliasfile_director_options_block, modemask)) },
+  { "one_time",           opt_bool,
+      (void *)(offsetof(aliasfile_director_options_block, one_time)) },
   { "optional",           opt_bool,
       (void *)(offsetof(aliasfile_director_options_block, optional)) },
   { "owners",             opt_uidlist,
@@ -77,9 +67,7 @@ optionlist aliasfile_director_options[] = {
   { "skip_syntax_errors", opt_bool,
       (void *)(offsetof(aliasfile_director_options_block, skip_syntax_errors)) },
   { "syntax_errors_to",   opt_stringptr,
-      (void *)(offsetof(aliasfile_director_options_block, syntax_errors_to)) },
-  { "user",               opt_expand_uid,
-      (void *)(offsetof(aliasfile_director_options_block, uid)) }
+      (void *)(offsetof(aliasfile_director_options_block, syntax_errors_to)) }
 };
 
 /* Size of the options list. An extern variable has to be used so that its
@@ -97,26 +85,20 @@ aliasfile_director_options_block aliasfile_director_option_defaults = {
   NULL,     /* home_dir */
   NULL,     /* current_dir */
   NULL,     /* syntax_errors_to */
-  NULL,     /* expand_uid */
-  NULL,     /* expand_gid */
-  -1,       /* uid */
-  -1,       /* gid */
   022,      /* modemask */
   -1,       /* partial_match */
   -1,       /* search_type */
   NULL,     /* owners */
   NULL,     /* owngroups */
-  FALSE,    /* uid_set */
-  FALSE,    /* gid_set */
   FALSE,    /* include_domain */
   FALSE,    /* check_ancestor */
-  FALSE,    /* initgroups */
   FALSE,    /* expand */
   FALSE,    /* optional */
   FALSE,    /* qualify_preserve_domain */
   FALSE,    /* forbid_file */
   FALSE,    /* forbid_pipe */
   TRUE,     /* freeze_missing_include */
+  FALSE,    /* one_time */
   TRUE,     /* rewrite */
   FALSE,    /* skip_syntax_errors */
   NULL,     /* directory_transport */
@@ -147,12 +129,6 @@ aliasfile_director_init(director_instance *dblock)
 aliasfile_director_options_block *ob =
   (aliasfile_director_options_block *)(dblock->options_block);
 
-/* If a fixed uid field is set, then a gid field must also be set. */
-
-if (ob->uid_set && !ob->gid_set && ob->expand_gid == NULL)
-  log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
-    "user set without group for the %s director", dblock->name);
-
 /* A search type is mandatory */
 
 if (ob->search_type < 0)
@@ -198,6 +174,13 @@ else
     log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s director:\n  "
       "only one of \"query\" or \"queries\" may be given", dblock->name);
   }
+
+/* Onetime aliases can only be real addresses */
+
+if (ob->one_time && (!ob->forbid_pipe || !ob->forbid_file))
+  log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s director:\n  "
+    "forbid_pipe and forbid_file must be set when one_time is set",
+      dblock->name);
 }
 
 
@@ -232,11 +215,9 @@ void *handle;
 header_line *extra_headers;
 char *remove_headers;
 error_block *eblock = NULL;
-uid_t uid = ob->uid;
-gid_t gid = ob->gid;
-struct passwd *upw = NULL;
-BOOL uid_set = ob->uid_set;
-BOOL gid_set = ob->gid_set;
+ugid_block ugid;
+void *reset_point;
+int yield = OK;
 int  extracted, rc;
 
 /* Perform file existence and sender verification checks. */
@@ -371,51 +352,12 @@ if (ob->expand)
     addr->special_action = SPECIAL_FREEZE;
     return ERROR;
     }
-  store_free(aliastext);
   aliastext = newtext;
   }
 
-/* If there is no fixed uid set, see if there's a dynamic one that can
-be expanded and possibly looked up. */
+/* Sort out the fixed or dynamic uid/gid */
 
-if (!uid_set && ob->expand_uid != NULL)
-  {
-  if (direct_find_expanded_user(ob->expand_uid, dblock->name, "director",
-    &upw, &uid, &(addr->message))) uid_set = TRUE;
-  else
-    {
-    addr->special_action = SPECIAL_FREEZE;
-    return ERROR;
-    }
-  }
-
-/* Likewise for the gid */
-
-if (!gid_set && ob->expand_gid != NULL)
-  {
-  if (direct_find_expanded_group(ob->expand_gid, dblock->name, "director",
-    &gid, &(addr->message))) gid_set = TRUE;
-  else
-    {
-    addr->special_action = SPECIAL_FREEZE;
-    return ERROR;
-    }
-  }
-
-/* If a uid is set, then a gid must also be available; use one from the passwd
-lookup if it happened. */
-
-if (uid_set && !gid_set)
-  {
-  if (upw != NULL)
-    {
-    gid = upw->pw_gid;
-    gid_set = TRUE;
-    }
-  else
-    log_write(0, LOG_PANIC_DIE, "User set without group for %s director",
-      dblock->name);
-  }
+if (!direct_get_ugid(dblock, "director", addr, &ugid)) return ERROR;
 
 /* Set up the errors address for the children of this address. */
 
@@ -443,17 +385,7 @@ if (dblock->transport != NULL)
   addr->extra_headers = extra_headers;
   addr->remove_headers = remove_headers;
 
-  if (uid_set)
-    {
-    addr->uid = uid;
-    addr->uid_set = TRUE;
-    }
-  if (gid_set)
-    {
-    addr->gid = gid;
-    addr->gid_set = TRUE;
-    addr->initgroups = ob->initgroups;
-    }
+  direct_set_ugid(addr, &ugid);
 
   if (addr->transport->info->local)
     {
@@ -465,6 +397,7 @@ if (dblock->transport != NULL)
     {
     addr->next = *addr_remote;
     *addr_remote = addr;
+    addr->fallback_hosts = dblock->fallback_hostlist;
     }
 
   DEBUG(2)
@@ -479,6 +412,12 @@ if (dblock->transport != NULL)
   return OK;
   }
 
+/* If qualify_preserve_domain is set, we temporarily reset the value of
+qualify_domain_recipient to the current domain, so that any unqualified
+addresses get qualified with the same domain as the incoming address. */
+
+if (ob->qualify_preserve_domain) qualify_domain_recipient = addr->domain;
+
 /* There is a common function for use by forwarding and aliasing directors that
 extracts a list of addresses from a text string. Setting the fourth-last
 argument FALSE makes generating no addresses an error. However, setting the
@@ -487,32 +426,42 @@ generate nothing without causing an error. It also permits :defer: and :fail:
 to occur. Setting the last argument NULL causes no checks to be made on
 :include: files.
 
-If qualify_preserve_domain is set, we temporarily reset the value of
-qualify_domain_recipient to the current domain, so that any unqualified
-addresses get qualified with the same domain as the incoming address. */
+It is possible that a fair amount of store might be used, especially if there
+is an :include: file. Therefore, switch to the temporary pool for the duration.
+Note that addresses built by the extractor are forced into the MAIN pool. */
 
-if (ob->qualify_preserve_domain) qualify_domain_recipient = addr->domain;
+store_pool = POOL_TEMP;
+reset_point = store_get(0);
 
 extracted = parse_extract_addresses(aliastext, addr->domain, &generated,
   &error, FALSE, TRUE, ob->rewrite, NULL,
   ob->skip_syntax_errors? &eblock : NULL);
 
-store_free(aliastext);
+/* Restore the qualify domain, and reset to the main pool for the generation of
+any error texts. */
+
 qualify_domain_recipient = save_qualify_domain_recipient;
+store_pool = POOL_MAIN;
 
 /* If extraction failed, return error and freeze, unless it was a missing
 include file and no_freeze_missing_include is set, or unless it was an explicit
-:defer: or :fail: that was encountered. */
+:defer: or :fail: that was encountered. If verifying, set a flag that causes
+the messages from :defer: and :fail: to be passed back to the caller - in
+particular, this gets them into the response to an SMTP command. */
 
 if (extracted != 0) switch (extracted)
   {
   case EXTRACTED_DEFER:
   addr->message = string_sprintf("forced defer: %s", error);
-  return DEFER;
+  if (verify && *error != 0) verify_forced_errmsg = error;
+  yield = DEFER;
+  break;
 
   case EXTRACTED_FAIL:
   addr->message = string_sprintf("forced failure: %s", error);
-  return FORCEFAIL;
+  if (verify && *error != 0) verify_forced_errmsg = error;
+  yield =  FORCEFAIL;
+  break;
 
   default:
   addr->basic_errno = ERRNO_BADALIAS;
@@ -522,7 +471,8 @@ if (extracted != 0) switch (extracted)
     string_sprintf("<%s> - error in alias lookup: %s", addr->orig, error);
   if (extracted > 0 && !ob->freeze_missing_include) return DEFER;
   addr->special_action = SPECIAL_FREEZE;
-  return ERROR;
+  yield = ERROR;
+  break;
   }
 
 /* If skip_syntax_errors was set and there were syntax errors in the list,
@@ -530,13 +480,24 @@ error messages will be present in eblock. Log them, and sent them off in
 a mail message if so configured. A common function is used by aliasfile
 and forwardfile. */
 
-if (eblock != NULL && !moan_skipped_syntax_errors(dblock->name, filename,
-    eblock, verify? NULL : ob->syntax_errors_to))
+if (yield == OK && eblock != NULL &&
+    !moan_skipped_syntax_errors(dblock->name, filename,
+      eblock, verify? NULL : ob->syntax_errors_to))
   {
   addr->special_action = SPECIAL_FREEZE;
-  return ERROR;
+  yield = ERROR;
   }
 
+/* We are now through with the store in the temporary pool; if there was an
+error, we are done. */
+
+store_pool = POOL_TEMP;
+store_reset(reset_point);
+store_pool = POOL_MAIN;
+
+if (yield != OK) return yield;
+
+
 /* Add the new addresses to the list of new addresses, copying in the
 uid, gid and permission flags for use by pipes and files, setting
 the parent, and or-ing its ignore_error flag. Also record the setting
@@ -553,19 +514,26 @@ delivered. */
 
 while (generated != NULL)
   {
+  address_item *parent;
   address_item *next = generated;
+
   generated = next->next;
   next->parent = addr;
   next->ignore_error |= addr->ignore_error;
   next->start_director = dblock->new;
   addr->child_count++;
 
+  if (ob->one_time)
+    {
+    for (parent = addr; parent->parent != NULL; parent = parent->parent);
+    next->onetime_parent = parent->orig;
+    }
+
   next->next = *addr_new;
   *addr_new = next;
 
   if (ob->check_ancestor)
     {
-    address_item *parent;
     for (parent = addr; parent != NULL; parent = parent->parent)
       {
       if (strcmp(next->orig, parent->orig) == 0)
@@ -584,21 +552,11 @@ while (generated != NULL)
   if (next->pfr)
     {
     next->director = dblock;
-    if (uid_set)
-      {
-      next->uid = uid;
-      next->uid_set = TRUE;
-      }
-    if (gid_set)
-      {
-      next->gid = gid;
-      next->gid_set = TRUE;
-      next->initgroups = ob->initgroups;
-      }
     next->home_dir = ob->home_dir;
     next->home_dir = ob->current_dir;
     next->allow_pipe = !ob->forbid_pipe;
     next->allow_file = !ob->forbid_file;
+    direct_set_ugid(next, &ugid);
 
     /* Aliasfile can produce only pipes or files; if the transport
     setting is null, the global setting will get used later. */
