@@ -14,14 +14,6 @@
 /* Options specific to the forwardfile director. */
 
 optionlist forwardfile_director_options[] = {
-  { "*expand_group",      opt_stringptr | opt_hidden,
-      (void *)(offsetof(forwardfile_director_options_block, expand_gid)) },
-  { "*expand_user",       opt_stringptr | opt_hidden,
-      (void *)(offsetof(forwardfile_director_options_block, expand_uid)) },
-  { "*set_group",         opt_bool | opt_hidden,
-      (void *)(offsetof(forwardfile_director_options_block, gid_set)) },
-  { "*set_user",          opt_bool | opt_hidden,
-      (void *)(offsetof(forwardfile_director_options_block, uid_set)) },
   { "allow_system_actions", opt_bool,
       (void *)(offsetof(forwardfile_director_options_block, allow_system_actions)) },
   { "check_ancestor",     opt_bool,
@@ -60,18 +52,16 @@ optionlist forwardfile_director_options[] = {
       (void *)(offsetof(forwardfile_director_options_block, forbid_reply)) },
   { "freeze_missing_include", opt_bool,
       (void *)(offsetof(forwardfile_director_options_block, freeze_missing_include)) },
-  { "group",              opt_expand_gid,
-      (void *)(offsetof(forwardfile_director_options_block, gid)) },
   { "home_directory",     opt_stringptr,
       (void *)(offsetof(forwardfile_director_options_block, home_dir)) },
   { "ignore_eacces",     opt_bool,
       (void *)(offsetof(forwardfile_director_options_block, ignore_eacces)) },
   { "ignore_enotdir",     opt_bool,
       (void *)(offsetof(forwardfile_director_options_block, ignore_enotdir)) },
-  { "initgroups",         opt_bool,
-      (void *)(offsetof(forwardfile_director_options_block, initgroups)) },
   { "modemask",           opt_octint,
       (void *)(offsetof(forwardfile_director_options_block, modemask)) },
+  { "one_time",           opt_bool,
+      (void *)(offsetof(forwardfile_director_options_block, one_time)) },
   { "owners",             opt_uidlist,
       (void *)(offsetof(forwardfile_director_options_block, owners)) },
   { "owngroups",          opt_gidlist,
@@ -89,9 +79,7 @@ optionlist forwardfile_director_options[] = {
   { "skip_syntax_errors", opt_bool,
       (void *)(offsetof(forwardfile_director_options_block, skip_syntax_errors)) },
   { "syntax_errors_to",   opt_stringptr,
-      (void *)(offsetof(forwardfile_director_options_block, syntax_errors_to)) },
-  { "user",               opt_expand_uid,
-      (void *)(offsetof(forwardfile_director_options_block, uid)) }
+      (void *)(offsetof(forwardfile_director_options_block, syntax_errors_to)) }
 };
 
 /* Size of the options list. An extern variable has to be used so that its
@@ -108,15 +96,9 @@ forwardfile_director_options_block forwardfile_director_option_defaults = {
   NULL,     /* current_dir */
   NULL,     /* file */
   NULL,     /* syntax_errors_to */
-  NULL,     /* expand_uid */
-  NULL,     /* expand_gid */
-  -1,       /* uid */
-  -1,       /* gid */
   022,      /* modemask */
   NULL,     /* owners */
   NULL,     /* owngroups */
-  FALSE,    /* uid_set */
-  FALSE,    /* gid_set */
   FALSE,    /* check_ancestor */
   TRUE,     /* check_local_user */
   FALSE,    /* check_group */
@@ -129,7 +111,7 @@ forwardfile_director_options_block forwardfile_director_option_defaults = {
   FALSE,    /* forbid_reply */
   FALSE,    /* ignore_eacces */
   FALSE,    /* ignore_enotdir */
-  FALSE,    /* initgroups */
+  FALSE,    /* one_time */
   FALSE,    /* qualify_preserve_domain */
   FALSE,    /* use_seteuid */
   TRUE,     /* freeze_missing_include */
@@ -157,12 +139,6 @@ void forwardfile_director_init(director_instance *dblock)
 forwardfile_director_options_block *ob =
   (forwardfile_director_options_block *)(dblock->options_block);
 
-/* If a fixed uid field is set, then a gid field must also be set. */
-
-if (ob->uid_set && !ob->gid_set && ob->expand_gid == NULL)
-  log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
-    "user set without group for the %s director", dblock->name);
-
 /* A file name is mandatory */
 
 if (ob->file == NULL)
@@ -194,8 +170,8 @@ if (ob->use_seteuid && !have_seteuid)
     "seteuid requested but not configured as available",
     dblock->name);
 
-if (ob->use_seteuid && !ob->check_local_user && !ob->uid_set &&
-    ob->expand_uid == NULL)
+if (ob->use_seteuid && !ob->check_local_user && !dblock->uid_set &&
+    dblock->expand_uid == NULL)
   log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s director:\n  "
     "seteuid requested without check_local_user or explicit uid", dblock->name);
 
@@ -204,6 +180,16 @@ if (ob->use_seteuid && !ob->check_local_user && !ob->uid_set &&
 if (dblock->transport != NULL || dblock->expand_transport != NULL)
   log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s director:\n  "
     "a transport is not allowed for this director", dblock->name);
+
+/* Onetime aliases can only be real addresses. If filter is not set,
+force forbid_reply, since a non-filter file can't generate one anyway. */
+
+if (!ob->filter) ob->forbid_reply = TRUE;
+
+if (ob->one_time && (!ob->forbid_pipe || !ob->forbid_file || !ob->forbid_reply))
+  log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s director:\n  "
+    "forbid_pipe, forbid_file, and forbid_reply must be set when one_time is set",
+      dblock->name);
 }
 
 
@@ -234,21 +220,18 @@ char *filebuf;
 char *error;
 char *s, *tag;
 char *remove_headers;
+void *reset_point = NULL;
 header_line *extra_headers;
 struct stat statbuf;
 struct passwd *pw;
-struct passwd *upw = NULL;
 error_block *eblock = NULL;
+ugid_block ugid;
 FILE *fwd;
-uid_t uid = ob->uid;
-gid_t gid = ob->gid;
 int   yield = OK;
 uid_t saved_euid = 0;
 gid_t saved_egid = 0;
 int   extracted = 0;
 int   rc;
-BOOL  uid_set = ob->uid_set;
-BOOL  gid_set = ob->gid_set;
 BOOL  uid_ok = FALSE;
 BOOL  gid_ok = !ob->check_group;
 BOOL  restore_uid = FALSE;
@@ -339,52 +322,13 @@ if (filename[0] != '/')
   return ERROR;
   }
 
-/* If there is no fixed explicit uid set, see if there's a dynamic one that can
-be expanded and possibly looked up. This uid is used (a) for reading the file
-if seteuid is on and (b) for running the transports for generated file and pipe
-addresses. It is not (necessarily) the same as the uids that may own the file.
-Exim panics if an expanded string is not a number and can't be found in the
-password file. */
+/* Sort out the fixed or dynamic uid/gid. This uid is used (a) for reading the
+file if seteuid is on and (b) for running the transports for generated file and
+pipe addresses. It is not (necessarily) the same as the uids that may own the
+file. Exim panics if an expanded string is not a number and can't be found in
+the password file. Other errors set the freezing bit. */
 
-if (!uid_set && ob->expand_uid != NULL)
-  {
-  if (direct_find_expanded_user(ob->expand_uid, dblock->name, "director",
-    &upw, &uid, &(addr->message))) uid_set = TRUE;
-  else
-    {
-    addr->special_action = SPECIAL_FREEZE;
-    return ERROR;
-    }
-  }
-
-/* Likewise, set up an expanded gid if necessary. */
-
-if (!gid_set && ob->expand_gid != NULL)
-  {
-  if (direct_find_expanded_group(ob->expand_gid, dblock->name, "director",
-    &gid, &(addr->message))) gid_set = TRUE;
-  else
-    {
-    addr->special_action = SPECIAL_FREEZE;
-    return ERROR;
-    }
-  }
-
-/* If a uid was specified, a gid to go with it must also be available. Take it
-from the passwd entry if one was looked up for the uid; otherwise it is a
-disaster. */
-
-if (uid_set && !gid_set)
-  {
-  if (upw != NULL)
-    {
-    gid = upw->pw_gid;
-    gid_set = TRUE;
-    }
-  else
-    log_write(0, LOG_PANIC_DIE, "User set without group for %s director",
-      dblock->name);
-  }
+if (!direct_get_ugid(dblock, "director", addr, &ugid)) return ERROR;
 
 /* If the seteuid option is on and a uid is available (either explicit or as a
 result of check_local_user), use seteuid() to become the local user. This is
@@ -397,26 +341,26 @@ Even if the seteuid option is not on, we need to do this when the security
 level is 1 or 3, because in those cases this code is running seteuid to exim.
 But only when a uid is available. */
 
-if ((uid_set || ob->check_local_user) &&
+if ((ugid.uid_set || ob->check_local_user) &&
     (ob->use_seteuid || security_level == 1 || security_level == 3))
   {
   saved_euid = geteuid();
   saved_egid = getegid();
   if (saved_euid != root_uid) mac_seteuid(root_uid);
 
-  if (uid_set)
+  if (ugid.uid_set)
     {
-    if (ob->initgroups)
+    if (dblock->initgroups)
       {
-      struct passwd *ppw = getpwuid(uid);
-      if (ppw != NULL) initgroups(ppw->pw_name, gid);
+      struct passwd *ppw = getpwuid(ugid.uid);
+      if (ppw != NULL) initgroups(ppw->pw_name, ugid.gid);
       }
-    mac_setegid(gid);
-    mac_seteuid(uid);
+    mac_setegid(ugid.gid);
+    mac_seteuid(ugid.uid);
     }
   else
     {
-    if (ob->initgroups) initgroups(pw->pw_name, pw->pw_gid);
+    if (dblock->initgroups) initgroups(pw->pw_name, pw->pw_gid);
     mac_setegid(pw->pw_gid);
     mac_seteuid(pw->pw_uid);
     }
@@ -437,25 +381,40 @@ directory or to a specified directory), and in that case we try to do a bit
 better by statting the directory first. If it cannot be statted, assume there
 is some mounting problem, and defer the delivery.
 
+It seems that in Solaris 2.6, statting an entry in an indirect map that is
+currently unmounted does not cause the mount to happen. Instead, dummy data is
+returned, which defeats the whole point of this test. However, if a stat() is
+done on some object inside the directory, such as the "." back reference to
+itself, then the mount does occur. The stat() may, however, fail as a result
+of a lack of search permission for the directory. This of course means that
+opening .forward is also going to fail, but what that failure means is
+controlled by an option below, so it seems that the safest plan is to do the
+"mounting" stat, and then the directory stat as before if the first one fails
+with EACCES.
+
 If an NFS host is taken offline, it is possible for the stat() to get stuck
 until it comes back. To guard against this, stick a timer round it. */
 
 if (directory != NULL && stat_directory)
   {
-  int rc;
+  int rc, saved_errno;
   sigalrm_seen = FALSE;
-  signal(SIGALRM, sigalrm_handler);
+  os_non_restarting_signal(SIGALRM, sigalrm_handler);
   alarm(30);
-  rc = stat(directory, &statbuf);
+  sprintf(big_buffer, "%.*s/.", big_buffer_size, directory);
+  rc = stat(big_buffer, &statbuf);
+  if (rc != 0 && errno == EACCES && !sigalrm_seen)
+    rc = stat(directory, &statbuf);
+  saved_errno = errno;
   alarm(0);
   signal(SIGALRM, SIG_DFL);
 
   if (sigalrm_seen || rc != 0)
     {
-    DEBUG(2) debug_printf("%s director failed to stat %s: deferred\n",
-      dblock->name, directory);
-    addr->message = string_sprintf("%s director failed to stat %s",
-      dblock->name, directory);
+    addr->message = string_sprintf("%s director failed to stat %s (%s)",
+      dblock->name, directory,
+      sigalrm_seen? "timeout" : strerror(saved_errno));
+    DEBUG(2) debug_printf("%s\n", addr->message);
     yield = DEFER;
     goto RESTORE_UID;             /* skip forward */
     }
@@ -582,12 +541,22 @@ if (!uid_ok || !gid_ok)
   goto CLOSE_RESTORE_UID;         /* skip forward */
   }
 
+/* Processing a .forward file could use a fair bit of store, especially if it
+is a filter file. Therefore it is a good idea to recover the store used
+afterwards. In order to do this, we switch to the temporary stacking pool, and
+recover that. Any new addresses that are created are put in the main pool, so
+as to remain around afterwards. */
+
+store_pool = POOL_TEMP;
+reset_point = store_get(0);
+
 /* Read the .forward file and generate new addresses for each entry therein.
 We read the file in one go in order to minimize the time we have it open. */
 
-filebuf = store_malloc(statbuf.st_size + 1);
+filebuf = store_get(statbuf.st_size + 1);
 if (fread(filebuf, 1, statbuf.st_size, fwd) != statbuf.st_size)
   {
+  store_pool = POOL_MAIN;
   addr->basic_errno = errno;
   addr->message =
     string_sprintf("<%s> - error while reading forward file (%s director)\n",
@@ -626,10 +595,9 @@ for (; *tag != 0; s++, tag++)
   }
 if (*tag == 0) filter_header = TRUE;
 
-/* If qualify_preserve_domain is set, temporarily reset
-qualify_domain_recipient to the current domain so that any unqualified
-addresses will get qualified with the same domain as the incoming
-address. */
+/* If qualify_preserve_domain is set, temporarily reset qualify_domain_
+recipient to the current domain so that any unqualified addresses will get
+qualified with the same domain as the incoming address. */
 
 if (ob->qualify_preserve_domain) qualify_domain_recipient = addr->domain;
 
@@ -653,20 +621,37 @@ if (filter_header && ob->filter)
       restore_uid? 0 : 3,              /* else permit only if uid set */
     ob->rewrite)? 0 : -1;
 
+  store_pool = POOL_MAIN;              /* Reset for error strings */
+
   /* Handle special action (will be SPECIAL_NONE unless permitted otherwise).
   Like the system filter, freeze does not happen after a manual thaw. */
 
   if (action == SPECIAL_FREEZE && !deliver_manual_thaw)
     {
-    addr->message = (filter_fmsg == NULL)? "frozen by filter" : filter_fmsg;
+    addr->message = (filter_fmsg == NULL)?
+      "frozen by filter" : string_copy(filter_fmsg);
     addr->special_action = SPECIAL_FREEZE;
     yield = ERROR;
     }
   else if (action == SPECIAL_FAIL)
     {
-    addr->message = (filter_fmsg == NULL)? "rejected by filter" : filter_fmsg;
+    addr->message = (filter_fmsg == NULL)?
+      "rejected by filter" : string_copy(filter_fmsg);
     yield = FORCEFAIL;
     }
+
+  /* If extraction from a filter file failed, it is a "probable user error", to
+  use a good old IBM term. Just defer delivery and let the user clean things
+  up. */
+
+  if (extracted != 0)
+    {
+    addr->basic_errno = ERRNO_BADFORWARD;
+    addr->message =
+      string_sprintf("<%s> - error in filter file: %s", addr->orig,
+        error);
+    yield = DEFER;
+    }
   }
 
 /* Otherwise it's a vanilla .forward file; call parse_extract_addresses()
@@ -703,12 +688,30 @@ else
     (directory == NULL)? "*" :                 /* if no directory, lock out */
     directory,                                 /* else restrain to directory */
     ob->skip_syntax_errors? &eblock : NULL);
+
+  store_pool = POOL_MAIN;                      /* Reset for error strings */
+
+  /* If extraction from a .forward file failed, freeze and yield ERROR if
+  it was a missing :include: file and freeze_missing_include is TRUE. Other-
+  wise just DEFER and hope things get fixed eventually. */
+
+  if (extracted != 0)
+    {
+    addr->basic_errno = ERRNO_BADFORWARD;
+    addr->message =
+      string_sprintf("<%s> - error in forward file%s: %s", addr->orig,
+        filter_header? " (filtering not enabled)" : "", error);
+    if (extracted > 0 && ob->freeze_missing_include)
+      {
+      addr->special_action = SPECIAL_FREEZE;
+      yield = ERROR;
+      }
+    else yield = DEFER;
+    }
   }
 
-/* The store for holding the forward file is now finished with, and we can
-reset the general qualify domain. */
+/* Reset the general qualify domain. */
 
-store_free(filebuf);
 qualify_domain_recipient = save_qualify_domain_recipient;
 
 /* At this point we are finished with the .forward file. Close it, and, if
@@ -724,66 +727,44 @@ RESTORE_UID:
 if (restore_uid)
   {
   mac_seteuid(root_uid);
-  if (ob->initgroups) setgroups(0, NULL);
+  if (dblock->initgroups) setgroups(0, NULL);
   mac_setegid(saved_egid);
   mac_seteuid(saved_euid);
   DEBUG(2) debug_printf("%s director restored euid=%d egid=%d\n", dblock->name,
     (int)geteuid(), (int)getegid());
   }
 
+
+/* If yield is ok, but skip_syntax_errors was set and there were syntax errors
+in the list, error messages will be present in eblock. Log them and send a
+message if so configured, using a function that is common to aliasfile and
+forwardfile. We cannot do this earlier, because the error message must be sent
+as Exim, not as the local user. */
+
+if (yield == OK && eblock != NULL &&
+    !moan_skipped_syntax_errors(dblock->name, filename,
+      eblock, verify? NULL : ob->syntax_errors_to))
+  {
+  addr->special_action = SPECIAL_FREEZE;
+  yield = ERROR;
+  }
+
+
+/* Now we are finished with the temporary store, if it was in use. */
+
+if (reset_point != NULL)
+  {
+  store_pool = POOL_TEMP;
+  store_reset(reset_point);
+  store_pool = POOL_MAIN;
+  }
+
 /* If there has been an error, return the error value now. Subsequently we
 can just return directly on error, since there is no further need to mess with
 the uid or close the file. */
 
 if (yield != OK) return yield;
 
-/* Extraction failed */
-
-if (extracted != 0)
-  {
-  /* If extraction from a filter file failed, it is a "probable user error", to
-  use a good old IBM term. Just defer delivery and let the user clean things
-  up. */
-
-  if (is_filter)
-    {
-    addr->basic_errno = ERRNO_BADFORWARD;
-    addr->message =
-      string_sprintf("<%s> - error in filter file: %s", addr->orig,
-        error);
-    return DEFER;
-    }
-
-  /* If extraction from a .forward file failed, freeze and yield ERROR if
-  it was a missing :include: file and freeze_missing_include is TRUE. Other-
-  wise just DEFER and hope things get fixed eventually. */
-
-  else
-    {
-    addr->basic_errno = ERRNO_BADFORWARD;
-    addr->message =
-      string_sprintf("<%s> - error in forward file%s: %s", addr->orig,
-        filter_header? " (filtering not enabled)" : "",
-        error);
-    if (extracted > 0 && ob->freeze_missing_include)
-      {
-      addr->special_action = SPECIAL_FREEZE;
-      return ERROR;
-      }
-    else return DEFER;
-    }
-  }
-
-/* If skip_syntax_errors was set and there were syntax errors in the list,
-error messages will be present in eblock. Log them and send a message if
-so configured, using a function that is common to aliasfile and forwardfile. */
-
-if (eblock != NULL && !moan_skipped_syntax_errors(dblock->name, filename,
-    eblock, verify? NULL : ob->syntax_errors_to))
-  {
-  addr->special_action = SPECIAL_FREEZE;
-  return ERROR;
-  }
 
 /* Set up the errors address for the children of this address. */
 
@@ -797,9 +778,9 @@ rc = direct_get_munge_headers(addr, dblock, "director", &extra_headers,
   &remove_headers);
 if (rc != OK) return rc;
 
-/* Add the new addresses to the list of new addresses, copying in the
-uid, gid and permission flags for use by pipes and files and autoreplies,
-setting the parent, and or-ing its ignore_error flag.
+/* Add the new addresses to the list of new addresses, copying in the uid, gid
+and permission flags for use by pipes and files and autoreplies, setting the
+parent, and or-ing its ignore_error flag.
 
 If the generated address is the same as one of its ancestors, and the
 check_ancestor flag is set, do not use this generated address, but replace it
@@ -811,7 +792,9 @@ delivered. */
 
 while (generated != NULL)
   {
+  address_item *parent;
   address_item *next = generated;
+
   generated = next->next;
   next->parent = addr;
   next->ignore_error |= addr->ignore_error;
@@ -820,9 +803,14 @@ while (generated != NULL)
   next->next = *addr_new;
   *addr_new = next;
 
+  if (ob->one_time)
+    {
+    for (parent = addr; parent->parent != NULL; parent = parent->parent);
+    next->onetime_parent = parent->orig;
+    }
+
   if (ob->check_ancestor)
     {
-    address_item *parent;
     for (parent = addr; parent != NULL; parent = parent->parent)
       {
       if (strcmp(next->orig, parent->orig) == 0)
@@ -841,27 +829,28 @@ while (generated != NULL)
   if (next->pfr)
     {
     next->director = dblock;
-    if (uid_set)
+
+    /* If there are no explicit uid/gid settings on the driver,
+    use those for the local user if we've looked one up. */
+
+    direct_set_ugid(next, &ugid);
+
+    if (ob->check_local_user)
       {
-      next->uid = uid;
-      next->uid_set = TRUE;
+      if (!ugid.uid_set)
+        {
+        next->uid = pw->pw_uid;
+        next->uid_set = TRUE;
+        }
+      if (!ugid.gid_set)
+        {
+        next->gid = pw->pw_gid;
+        next->gid_set = TRUE;
+        }
       }
-    else if (ob->check_local_user)
-      {
-      next->uid = pw->pw_uid;
-      next->uid_set = TRUE;
-      }
-    if (gid_set)
-      {
-      next->gid = gid;
-      next->gid_set = TRUE;
-      }
-    else if (ob->check_local_user)
-      {
-      next->gid = pw->pw_gid;
-      next->gid_set = TRUE;
-      }
-    next->initgroups = ob->initgroups;
+
+    /* Remaining options */
+
     if (ob->home_dir != NULL) next->home_dir = ob->home_dir;
       else if (directory != NULL) next->home_dir = string_copy(directory);
     next->current_dir = ob->current_dir;
@@ -919,7 +908,7 @@ if (is_filter)
     {
     if (addr->child_count <= 0) return FAIL;
 
-    next = deliver_make_addr(addr->orig);
+    next = deliver_make_addr(addr->orig, FALSE);
     next->parent = addr;
     next->ignore_error |= addr->ignore_error;
     next->pfr = addr->pfr;
