@@ -13,15 +13,9 @@ typedef struct {
   char *current_dir;
   char *file;
   char *syntax_errors_to;
-  char *expand_uid;
-  char *expand_gid;
-  uid_t uid;
-  gid_t gid;
   int   modemask;
   uid_t *owners;
   gid_t *owngroups;
-  BOOL  uid_set;
-  BOOL  gid_set;
   BOOL  check_ancestor;
   BOOL  check_local_user;
   BOOL  check_group;
@@ -34,7 +28,7 @@ typedef struct {
   BOOL  forbid_reply;
   BOOL  ignore_eacces;
   BOOL  ignore_enotdir;
-  BOOL  initgroups;
+  BOOL  one_time;
   BOOL  qualify_preserve_domain;
   BOOL  use_seteuid;
   BOOL  freeze_missing_include;
