@@ -2,14 +2,23 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 1997 */
+/* Copyright (c) University of Cambridge 1995 - 1998 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Private structure for the private options. */
 
 typedef struct {
   char *new_address;
-  BOOL panic_expansion_fail;
+
+#ifdef ALLOW_EXPIRE
+  char *expire_pattern;
+  char *expire_subject;
+  char *expire_time;
+  pcre *expire_pattern_re;
+  BOOL  expire_force_fail;
+#endif
+
+  BOOL  panic_expansion_fail;
 } smartuser_director_options_block;
 
 /* Data for reading the private options. */
