@@ -234,7 +234,8 @@ Arguments:
   type      type of DNS record required (T_A, T_MX, etc)
 
 Returns:    DNS_SUCCEED   successful lookup
-            DNS_NOMATCH   name not found, or no data found for the given type
+            DNS_NOMATCH   name not found, or no data found for the given type,
+                          or name contains illegal characters (if checking)
             DNS_AGAIN     soft failure, try again later
             DNS_FAIL      DNS failure
 */
@@ -242,6 +243,30 @@ Returns:    DNS_SUCCEED   successful lookup
 int
 dns_basic_lookup(char *name, int type)
 {
+
+/* If configured, check the hygene of the name passed to lookup. Otherwise,
+although DNS lookups may give REFUSED at the lower level, some resolvers
+turn this into TRY_AGAIN, which is silly. Give a NOMATCH return, since such
+domains cannot be in the DNS. */
+
+if (check_dns_names)
+  {
+  int c;
+  char *ptr = name;
+
+  while ((c = *ptr) != 0 && (isalnum(c) || c == '.' || c == '-')) ptr++;
+
+  if (c != 0)
+    {
+    DEBUG(3)
+      debug_printf("DNS name check: %s (%s) includes illegal character \"%c\"\n",
+	name, dns_text_type(type), c);
+    return DNS_NOMATCH;
+    }
+  }
+
+/* Call the resolver */
+
 dns_answerlen = res_search(name, C_IN, type, dns_answer, MAXPACKET);
 
 if (dns_answerlen < 0) switch (h_errno)
