@@ -34,6 +34,10 @@ always present, but the "open" and "find" functions are set to NULL for those
 that are not compiled into the binary. The "check" and "close" functions can
 be NULL for methods that don't need them. */
 
+#ifdef LOOKUP_CDB
+#include "lookups/cdb.h"
+#endif
+
 #ifdef LOOKUP_DBM
 #include "lookups/dbmdb.h"
 #endif
@@ -74,6 +78,22 @@ be NULL for methods that don't need them. */
 
 lookup_info lookup_list[] = {
 
+/* cdb lookup in single file */
+
+  {
+  "cdb",		       /* lookup name */
+  lookup_absfile,	       /* uses absolute file name */
+#ifdef LOOKUP_CDB
+  cdb_open,		       /* open function */
+  cdb_check,		       /* check function */
+  cdb_find,		       /* find function */
+  cdb_close,		       /* close function */
+  NULL			       /* no tidy function */
+#else
+  NULL, NULL, NULL, NULL, NULL /* lookup not present */
+#endif
+  },
+
 /* DBM file lookup; called "dbm" because that is the name in Exim,
 but the code is called dbmdb to avoid name clashes. */
 
