@@ -14,12 +14,51 @@
 
 
 
+/*************************************************
+*      Function interface to store functions     *
+*************************************************/
+
+/* We need some real functions to pass to the PCRE regular expression library
+for store allocation via Exim's store manager. The normal calls are actually
+macros that pass over location information to make tracing easier. These
+functions just interface to the standard macro calls. A good compiler will
+optimize out the tail recursion and so not make them too expensive. There
+are two sets of functions; one for use when we want to retain the compiled
+regular expression for a long time; the other for short-term use. */
+
+static void *
+function_store_get(size_t size)
+{
+return store_get((int)size);
+}
+
+static void
+function_dummy_free(void *block) {}
+
+static void *
+function_store_malloc(size_t size)
+{
+return store_malloc((int)size);
+}
+
+static void
+function_store_free(void *block)
+{
+store_free(block);
+}
+
+
+
+
 /*************************************************
 *  Compile regular expression and panic on fail  *
 *************************************************/
 
 /* This function is called when failure to compile a regular expression leads
-to a panic exit. In other cases, pcre_compile() is called directly.
+to a panic exit. In other cases, pcre_compile() is called directly. In all
+cases where this function is used, the results of the compilation are to be
+placed in long-lived store, so we temporarily reset the store management
+functions that PCRE uses.
 
 Argument:   the pattern to compile
 Returns:    pointer to the compiled pattern
@@ -31,7 +70,11 @@ regex_must_compile(char *pattern)
 int offset;
 pcre *yield;
 const char *error;
+pcre_malloc = function_store_malloc;
+pcre_free = function_store_free;
 yield = pcre_compile(pattern, PCRE_COPT, &error, &offset);
+pcre_malloc = function_store_get;
+pcre_free = function_dummy_free;
 if (yield == NULL)
   log_write(0, LOG_PANIC_DIE, "regular expression error: %s at offset %d "
     "while compiling %s", error, offset, pattern);
@@ -62,7 +105,7 @@ Returns:      TRUE or FALSE
 BOOL
 regex_match_and_setup(pcre *re, char *subject, int options, int setup)
 {
-int ovector[2*(EXPAND_MAXN+1)];
+int ovector[3*(EXPAND_MAXN+1)];
 int n = pcre_exec(re, NULL, subject, (int)strlen(subject), PCRE_EOPT | options,
   ovector, sizeof(ovector)/sizeof(int));
 BOOL yield = n >= 0;
@@ -158,32 +201,6 @@ va_end(ap);
 
 
 
-/*************************************************
-*      Function interface to store functions     *
-*************************************************/
-
-/* We need some real functions to pass to the PCRE regular expression library
-for store allocation via Exim's store manager. The normal usage in exim, viz
-store_malloc() and store_free() are actually macros that pass over location
-information to make tracing easier. These functions just interface to the
-standard macro calls. A good compiler will optimize out the tail recursion and
-so not make them too expensive. */
-
-void *
-function_store_malloc(size_t size)
-{
-return store_malloc(size);
-}
-
-void
-function_store_free(void *block)
-{
-store_free(block);
-}
-
-
-
-
 /*************************************************
 *          Entry point and high-level code       *
 *************************************************/
@@ -212,9 +229,10 @@ int  filter_fd = -1;
 int  group_count;
 int  i;
 int  list_queue_option = 0;
+int  message_count = 0;
 int  msg_action = 0;
 int  msg_action_arg = -1;
-int  namelen = (int)strlen(argv[0]);
+int  namelen = (argv[0] == NULL)? 0 : (int)strlen(argv[0]);
 int  recipients_arg = argc;
 int  sender_address_domain = 0;
 int  test_retry_arg = -1;
@@ -236,7 +254,6 @@ BOOL one_msg_action = FALSE;
 BOOL queue_only_set = FALSE;
 BOOL queue_remote_set = FALSE;
 BOOL queue_smtp_set = FALSE;
-BOOL smtp_first = TRUE;
 BOOL synchronous_delivery = FALSE;
 BOOL verify_as_sender = FALSE;
 BOOL version_printed = FALSE;
@@ -250,6 +267,7 @@ char *ftest_prefix = NULL;
 char *ftest_suffix = NULL;
 BOOL ftest_system = FALSE;
 transport_instance *ti;
+void *reset_point;
 
 struct passwd *pw;
 struct stat statbuf;
@@ -272,21 +290,24 @@ if (log_buffer == NULL)
   exit(EXIT_FAILURE);
   }
 
-/* Set log_stderr to stderr. This gets reset to NULL when the daemon is
-run. We have to use this indirection, because some systems don't allow writing
-to the variable "stderr". */
+/* Set log_stderr to stderr, provided that stderr exists. This gets reset to
+NULL when the daemon is run and all the file descriptors are closed. We have to
+use this indirection, because some systems don't allow writing to the variable
+"stderr". */
 
-log_stderr = stderr;
+if (fstat(fileno(stderr), &statbuf) >= 0) log_stderr = stderr;
 
 /* Arrange for the PCRE regex library to use our store functions. Note that
-store_malloc() is actually a macro that calls store_malloc3() for logging
-purposes, so we have to construct a couple of interface functions. Let's hope
-that compilers do the tail optimization right. */
+the normal calls are actually macros that add additional arguments for
+debugging purposes so we have to assign specially constructed functions here.
+The default is to use store in the stacking pool, but this is overridden in the
+regex_must_compile() function. */
 
-pcre_malloc = function_store_malloc;
-pcre_free = function_store_free;
+pcre_malloc = function_store_get;
+pcre_free = function_dummy_free;
 
-/* Ensure there is a big buffer for temporary use in several places. */
+/* Ensure there is a big buffer for temporary use in several places. It is put
+in malloc store so that it can be freed for enlargement if necessary. */
 
 big_buffer = store_malloc(big_buffer_size);
 
@@ -341,14 +362,17 @@ step with the code that generates ids in the accept.c module. We need to do
 this here, because the -M options check their arguments for syntactic validity
 using mac_ismsgid, which uses this. */
 
-regex_ismsgid = regex_must_compile(
-  "^[a-zA-Z0-9]{6}-[a-zA-Z0-9]{6}-\\d\\d$");
+regex_ismsgid = regex_must_compile("^(?:[^\\W_]{6}-){2}[^\\W_]{2}$");
 
-/* Precompile the regular expression for matching a SIZE parameter in the
-response to an EHLO command. */
+/* Precompile the regular expression for matching SIZE and DSN parameters in
+the response to an EHLO command. */
 
 regex_SIZE = regex_must_compile("\\n250[\\s\\-]SIZE(\\s|\\n|$)");
 
+#ifdef SUPPORT_DSN
+regex_DSN  = regex_must_compile("\\n250[\\s\\-]DSN(\\s|\\n|$)");
+#endif
+
 /* If the program is called as "mailq" treat it as equivalent to "exim -bp";
 this seems to be a generally accepted convention, since one finds symbolic
 links called "mailq" in standard OS configurations. */
@@ -394,6 +418,30 @@ if ((namelen == 4 && strcmp(argv[0], "runq") == 0) ||
   called_as = "-runq";
   }
 
+/* If the program is called as "newaliases" treat it as equivalent to
+"exim -bi"; this is a sendmail convention. */
+
+if ((namelen == 10 && strcmp(argv[0], "newaliases") == 0) ||
+    (namelen  > 10 && strncmp(argv[0] + namelen - 11, "/newaliases", 11) == 0))
+  {
+  bi_option = TRUE;
+  called_as = "-newaliases";
+  }
+
+/* Get the real uid and gid. If the caller is root, force the effective
+uid/gid to be the same as the real ones. This makes a difference only if Exim
+is setuid (or setgid) to something other than root, which could be the case in
+some special configurations. */
+
+real_uid = getuid();
+real_gid = getgid();
+
+if (real_uid == root_uid)
+  {
+  mac_seteuid(real_uid);
+  mac_setegid(real_gid);
+  }
+
 /* Scan the program's arguments. Some can be dealt with right away; others are
 simply recorded for checking and handling afterwards. Do a high-level switch
 on the second character (the one after '-'), to save some effort. */
@@ -670,7 +718,7 @@ for (i = 1; i < argc; i++)
           mlast = m;
           }
 
-        m = store_malloc(sizeof(macro_item) + (int)strlen(name));
+        m = store_get(sizeof(macro_item) + (int)strlen(name));
         m->next = NULL;
         m->command_line = TRUE;
         if (mlast == NULL) macros = m; else mlast->next = m;
@@ -691,7 +739,11 @@ for (i = 1; i < argc; i++)
       set up stderr to a file, or set the drop_cr option. */
 
       case 'd':
-      if (strcmp(argrest, "m") == 0) debug_trace_memory = TRUE;
+      if (strcmp(argrest, "m") == 0)
+        {
+        debug_trace_memory = TRUE;
+        if (debug_level < 1) debug_level = 1;
+        }
       else if (strcmp(argrest, "ropcr") == 0) drop_cr = TRUE;
       else if (strcmp(argrest, "f") == 0)
         {
@@ -710,8 +762,8 @@ for (i = 1; i < argc; i++)
       if (debug_level > 0 && !badarg)
         {
         debug_file = stderr;
-        debug_printf("Exim version %s debug level set to %d\n",
-          version_string, debug_level);
+        debug_printf("Exim version %s debug level %d uid=%d gid=%d\n",
+          version_string, debug_level, real_uid, real_gid);
 
         #ifdef DB_VERSION_STRING
         debug_printf("Berkeley DB: %s\n", DB_VERSION_STRING);
@@ -875,6 +927,17 @@ for (i = 1; i < argc; i++)
         break;
         }
 
+      /* -MCD: set the smtp_use_dsn flag; this is useful only when it
+      precedes -MC (see above) */
+
+      #ifdef SUPPORT_DSN
+      else if (strcmp(argrest, "CD") == 0)
+        {
+        smtp_use_dsn = TRUE;
+        break;
+        }
+      #endif
+
       /* -MCS: set the smtp_use_size flag; this is useful only when it
       precedes -MC (see above) */
 
@@ -899,6 +962,9 @@ for (i = 1; i < argc; i++)
          -Mmd  mark recipients(s) delivered
          -Mes  edit sender
          -Meb  edit body
+         -Mvb  show body
+         -Mvh  show header
+         -Mvl  show log
       */
 
       else if (*argrest == 0)
@@ -939,6 +1005,21 @@ for (i = 1; i < argc; i++)
         }
       else if (strcmp(argrest, "rm") == 0) msg_action = MSG_REMOVE;
       else if (strcmp(argrest, "t") == 0)  msg_action = MSG_THAW;
+      else if (strcmp(argrest, "vb") == 0)
+        {
+        msg_action = MSG_SHOW_BODY;
+        one_msg_action = TRUE;
+        }
+      else if (strcmp(argrest, "vh") == 0)
+        {
+        msg_action = MSG_SHOW_HEADER;
+        one_msg_action = TRUE;
+        }
+      else if (strcmp(argrest, "vl") == 0)
+        {
+        msg_action = MSG_SHOW_LOG;
+        one_msg_action = TRUE;
+        }
       else { badarg = TRUE; break; }
 
       /* All the -Mxx options require at least one message id. */
@@ -1373,24 +1454,18 @@ if ((
   }
 
 
-/* Get the real uid and gid. */
-
-real_uid = getuid();
-real_gid = getgid();
-
-/* Exim is normally entered as root (except when called from inetd under its
-own uid in order to receive a message, with security level >= 2). The security
-level controls how it uses set{u,g}id and/or sete{g,u}id to reduce its
-privilege when not necessary. However, it always spins off sub-processes that
-set their uid and gid as required for local delivery. We don't want to pass on
-any extra groups that root may belong to, so we want to get rid of them all at
-this point.
+/* Exim is normally entered as root (but some special configurations are
+possible that don't do this). The security level controls how it uses
+set{u,g}id and/or sete{g,u}id to reduce its privilege when not necessary.
+However, it always spins off sub-processes that set their uid and gid as
+required for local delivery. We don't want to pass on any extra groups that
+root may belong to, so we want to get rid of them all at this point.
 
 We need to obey setgroups() at this stage, before possibly giving up root
-privilege for a changed configuration file, but later on we might need to check
-on the additional groups for the admin user privilege - can't do that till
-after reading the config, which might specify the exim gid. Therefore, save the
-group list here first. */
+privilege for a changed configuration file, but later on we might need to
+check on the additional groups for the admin user privilege - can't do that
+till after reading the config, which might specify the exim gid. Therefore,
+save the group list here first. */
 
 group_count = getgroups(NGROUPS_MAX, group_list);
 
@@ -1428,10 +1503,10 @@ if (setgroups(0, NULL) != 0)
 command line (either a new file name or a macro definition) and the caller is
 not root or the exim user, or if this is a filter testing run, remove any
 setuid privilege the program has, and run as the underlying user. Otherwise,
-set the real ids to the effective values (should be root unless run from inetd,
-which it can either be root or the exim uid, if one is configured). When
-running with the security options set, root privilege will subsequently be
-relinquished as soon as possible. */
+set the real ids to the effective values (should be root unless run from
+inetd, which it can either be root or the exim uid, if one is configured).
+When running with the security options set, root privilege will subsequently
+be relinquished as soon as possible. */
 
 if (((config_changed || macros != NULL) && real_uid != root_uid &&
     (!exim_uid_set || real_uid != exim_uid)) ||
@@ -1491,7 +1566,7 @@ if (log_arguments && really_exim && !list_variables && !checking)
       p = big_buffer + 3;
       }
     sprintf(p, " %.*s", big_buffer_size - (p - big_buffer) - 2,
-      string_printing(argv[i], FALSE));
+      string_printing(argv[i]));
     while (*p) p++;
     }
   log_write(0, LOG_MAIN, "%s", big_buffer);
@@ -1533,7 +1608,7 @@ if (bi_option)
 
 /* Build the argument list for calling Exim to send an error message. */
 
-mailer_argv = store_malloc((12 + MAX_CLMACROS) * (sizeof(char *)));
+mailer_argv = store_get((12 + MAX_CLMACROS) * (sizeof(char *)));
 mailer_argv[0] = exim_path;
 mailer_argv[1] = "-t";
 mailer_argv[2] = "-oem";
@@ -1606,17 +1681,14 @@ as "trusted user" status. We don't fail all message actions immediately if not
 admin_user, since some actions can be performed by others. Instead, set
 admin_user for later interrogation. */
 
-if (msg_action_arg > 0 || daemon_listen || queue_interval >= 0)
+if (real_uid == root_uid ||
+  (exim_uid_set && real_uid == exim_uid) ||
+  (exim_gid_set && real_gid == exim_gid))
+    admin_user = TRUE;
+else if (exim_gid_set)
   {
-  if (real_uid == root_uid ||
-    (exim_uid_set && real_uid == exim_uid) ||
-    (exim_gid_set && real_gid == exim_gid))
-      admin_user = TRUE;
-  else if (exim_gid_set)
-    {
-    while (--group_count >= 0)
-      if (group_list[group_count] == exim_gid) { admin_user = TRUE; break; }
-    }
+  while (--group_count >= 0)
+    if (group_list[group_count] == exim_gid) { admin_user = TRUE; break; }
   }
 
 
@@ -1639,7 +1711,7 @@ now. */
 
 if (chdir(spool_directory) != 0)
   {
-  directory_make(spool_directory, "", SPOOL_DIRECTORY_MODE);
+  directory_make(spool_directory, "", SPOOL_DIRECTORY_MODE, TRUE);
   (void) chdir(spool_directory);
   }
 
@@ -1751,8 +1823,8 @@ if (exim_uid_set)
 
   else
     {
-    if (queue_interval != 0 &&          /* no queue run, or periodic queue run */
-        !daemon_listen &&               /* and not starting listener */
+    if (queue_interval < 0 &&           /* not running the queue */
+        !daemon_listen &&               /* and not starting the daemon */
         (msg_action_arg < 0 ||
           msg_action != MSG_DELIVER) && /* and not delivering */
         (!checking || host_checking))   /* and not checking address */
@@ -1794,12 +1866,12 @@ if (msg_action_arg > 0 && msg_action != MSG_DELIVER)
   if (!one_msg_action)
     {
     for (i = msg_action_arg; i < argc; i++)
-      if (!queue_action(argv[i], msg_action, stdout, NULL, 0, 0))
+      if (!queue_action(argv[i], msg_action, NULL, 0, 0))
         yield = EXIT_FAILURE;
     }
 
-  else if (!queue_action(argv[msg_action_arg], msg_action, stdout, argv,
-    argc, recipients_arg)) yield = EXIT_FAILURE;
+  else if (!queue_action(argv[msg_action_arg], msg_action, argv, argc,
+    recipients_arg)) yield = EXIT_FAILURE;
 
   exit(yield);
   }
@@ -1919,22 +1991,6 @@ if (test_retry_arg >= 0)
   return EXIT_SUCCESS;
   }
 
-
-/* Handle the -brw option, which is for checking out rewriting
-rules. Cause log writes (on errors) to go to stderr instead. */
-
-if (test_rewrite_arg >= 0)
-  {
-  really_exim = FALSE;
-  if (test_rewrite_arg >= argc)
-    {
-    printf("-brw needs an address argument\n");
-    return EXIT_FAILURE;
-    }
-  rewrite_test(argv[test_rewrite_arg]);
-  return EXIT_SUCCESS;
-  }
-
 /* Handle a request to list one or more configuration variables */
 
 if (list_variables)
@@ -2128,7 +2184,7 @@ for (i = 1; i <= 10; i++)
         if (gecos_pattern != NULL && gecos_name != NULL)
           {
           pcre *re;
-          re = regex_must_compile(gecos_pattern);
+          re = regex_must_compile(gecos_pattern);  /* Uses malloc store */
 
           if (regex_match_and_setup(re, name, 0, -1))
             {
@@ -2145,6 +2201,7 @@ for (i = 1; i <= 10; i++)
             }
           else DEBUG(4) debug_printf("gecos_pattern \"%s\" did not match "
             "gecos field \"%s\"\n", gecos_pattern, name);
+          store_free(re);
           }
         user_name = string_copy(name);
         }
@@ -2191,6 +2248,22 @@ originator_login = user_login;
 originator_uid = real_uid;
 originator_gid = real_gid;
 
+/* Handle the -brw option, which is for checking out rewriting rules. Cause log
+writes (on errors) to go to stderr instead. Can't do this earlier, as want the
+originator_* variables set. */
+
+if (test_rewrite_arg >= 0)
+  {
+  really_exim = FALSE;
+  if (test_rewrite_arg >= argc)
+    {
+    printf("-brw needs an address argument\n");
+    return EXIT_FAILURE;
+    }
+  rewrite_test(argv[test_rewrite_arg]);
+  return EXIT_SUCCESS;
+  }
+
 /* Run in daemon and/or queue-running mode. The function daemon_go() never
 returns. */
 
@@ -2233,14 +2306,8 @@ interface, no -f argument). */
 
 if (sender_address != NULL && sender_address_domain == 0 &&
     sender_address[0] != 0 && sender_address[0] != '@')
-  {
-  char *new_sender_address =
-    store_malloc((int)strlen(sender_address) +
-      (int)strlen(qualify_domain_sender) + 2);
-  sprintf(new_sender_address, "%s@%s", sender_address, qualify_domain_sender);
-  sender_address = new_sender_address;
-  }
-
+      sender_address = string_sprintf("%s@%s", sender_address,
+        qualify_domain_sender);
 
 /* Handle a request to verify a list of addresses. Must follow the setting of
 the sender address, since drivers can now be predicated upon the sender. */
@@ -2340,20 +2407,24 @@ be tested. */
 
 if (host_checking)
   {
-  BOOL first = TRUE;
-  sender_host_unknown = FALSE;
-  sender_local = FALSE;
   host_build_sender_fullhost();
   smtp_input = TRUE;
   smtp_in = stdin;
   smtp_out = stdout;
+  sender_local = FALSE;
   debug_file = stderr;
   fprintf(stdout, "\n**** SMTP testing session as if from host %s\n"
     "**** Not for real!\n\n", sender_host_address);
-  while (smtp_setup_msg(first) > 0)
+
+  if (smtp_start_session())
     {
-    first = FALSE;
-    (void)accept_msg(smtp_in, smtp_out, FALSE);
+    reset_point = store_get(0);
+    for (;;)
+      {
+      store_reset(reset_point);
+      if (smtp_setup_msg() <= 0) break;
+      (void)accept_msg(smtp_in, smtp_out, FALSE);
+      }
     }
   exit(EXIT_SUCCESS);
   }
@@ -2383,11 +2454,13 @@ doing immediate delivery, we fork a delivery process for each received
 message, except for the last one, where we can save a process switch. */
 
 /* If the sender ident has not been set (by a trusted caller) set it to
-the caller. This will get overwritten below for an inetd call. */
+the caller. This will get overwritten below for an inetd call. If a trusted
+caller has set it empty, unset it. */
 
 if (sender_ident == NULL)
   sender_ident = (user_login != NULL)? user_login :
     string_sprintf("uid%d", (int)real_uid);
+else if (sender_ident[0] == 0) sender_ident = NULL;
 
 /* It is only in non-smtp mode that error_handling is allowed to be changed from
 its default of ERRORS_SENDER by argument. (Idle thought: are any of the
@@ -2411,16 +2484,21 @@ else if (is_inetd)
 
 /* If the sender host address has been set, build sender_fullhost if it hasn't
 already been done (which it will have been for inetd). This caters for the
-case when it is forced by -oMa. */
+case when it is forced by -oMa. However, we must flag that is isn't a socket,
+so that the test for IP options is skipped for -bs input. */
 
-if (sender_host_address != NULL && sender_fullhost != NULL)
+if (sender_host_address != NULL && sender_fullhost == NULL)
+  {
   host_build_sender_fullhost();
+  set_process_info("handling incoming connection from %s via -oMa",
+    sender_fullhost);
+  sender_host_notsocket = TRUE;
+  }
 
-/* Otherwise, set the sender host as unknown. This prevents host checking in
-the case of -bs not from inetd and also for -bS. */
-
-else sender_host_unknown = TRUE;
+/* Otherwise, set the sender host as unknown except for inetd calls. This
+prevents host checking in the case of -bs not from inetd and also for -bS. */
 
+else if (!is_inetd) sender_host_unknown = TRUE;
 
 /* If stdout does not exist, then dup stdin to stdout. This can happen
 if exim is started from inetd. In this case fd 0 will be set to the socket,
@@ -2428,7 +2506,6 @@ but fd 1 will not be set. */
 
 if (fstat(1, &statbuf) < 0) dup2(0, 1);
 
-
 /* Set up the incoming protocol name and the state of the program. Root
 is allowed to force received protocol via the -oMr option above, and if we are
 in a non-local SMTP state it means we have come via inetd and the process info
@@ -2470,19 +2547,37 @@ the sender verification options. */
 if (smtp_batched_input && !sender_verify_batch)
   sender_verify = sender_try_verify = FALSE;
 
-/* Loop for several messages when reading SMTP input. If we fork any child
-processes, we don't want to wait for them unless synchronous delivery is
-requested, so set SIGCHLD to SIG_IGN in that case. This is not the same as
-SIG_DFL, despite the fact that documentation often lists the default as
-"ignore". At least on some systems, setting SIG_IGN causes child processes that
-complete simply to go away without ever becoming defunct. You can't therefore
-wait for them - but we don't want to wait for them in the non-synchronous
-delivery case. */
+/* If this is smtp input of any kind, handle the start of the SMTP
+session. */
+
+if (smtp_input)
+  {
+  smtp_in = stdin;
+  smtp_out = stdout;
+  if (!smtp_start_session()) exit(EXIT_SUCCESS);
+  }
+
+/* Loop for several messages when reading SMTP input. Reset the pool store
+each time. If we fork any child processes, we don't want to wait for them
+unless synchronous delivery is requested, so set SIGCHLD to SIG_IGN in that
+case. This is not the same as SIG_DFL, despite the fact that documentation
+often lists the default as "ignore". At least on some systems, setting SIG_IGN
+causes child processes that complete simply to go away without ever becoming
+defunct. You can't therefore wait for them - but we don't want to wait for them
+in the non-synchronous delivery case. */
 
 if (!synchronous_delivery) signal(SIGCHLD, SIG_IGN);
 
+/* Save the current store pool point, for resetting at the start of
+each message, and save the real sender address, if any. */
+
+reset_point = store_get(0);
+real_sender_address = sender_address;
+
 while (more)
   {
+  store_reset(reset_point);
+
   /* In the SMTP case, we have to handle the initial SMTP input and
   build the recipients list, before calling accept_msg to read the
   message proper. Whatever sender address is actually given in the
@@ -2491,37 +2586,16 @@ while (more)
   process or a -f argument provided by a root caller. This is indicated
   by the existence of sender_address, which will be NULL only for a trusted
   caller when no -f was given. It is saved in real_sender_address, which is
-  global so that the error message generator can find it if necessary. The
-  function smtp_setup_msg() throws away any previous recipients list that
-  may be lying around from a previous message. */
+  global so that the error message generator can find it if necessary. */
 
   if (smtp_input)
     {
     int rc;
-    real_sender_address = sender_address;
-    smtp_in = stdin;
-    smtp_out = stdout;
-
-    if ((rc = smtp_setup_msg(smtp_first)) > 0)
+    if ((rc = smtp_setup_msg()) > 0)
       {
-      smtp_first = FALSE;
-
       if (real_sender_address != NULL)
-        {
-        store_free(sender_address);
-        store_free(raw_sender);
-        sender_address = real_sender_address;
-        raw_sender = string_copy(sender_address);
-        }
+        sender_address = raw_sender = real_sender_address;
       more = accept_msg(stdin, stdout, extract_recipients);
-      if (real_sender_address == NULL)
-        {
-        store_free(sender_address);
-        sender_address = NULL;
-        store_free(raw_sender);
-        raw_sender = NULL;
-        }
-
       if (message_id[0] == 0)
         {
         if (more) continue;
@@ -2591,21 +2665,21 @@ while (more)
           if (error_handling == ERRORS_STDERR)
             {
             fprintf(stderr, "exim: bad address \"%s\": %s\n",
-              string_printing(list[i], FALSE), errmess);
+              string_printing(list[i]), errmess);
             return EXIT_FAILURE;
             }
           else
             {
             error_block eblock;
             eblock.next = NULL;
-            eblock.text1 = string_printing(list[i], FALSE);
+            eblock.text1 = string_printing(list[i]);
             eblock.text2 = errmess;
             return
               moan_to_sender(ERRMESS_BADARGADDRESS, &eblock, NULL, stdin, TRUE)?
                 errors_sender_rc : EXIT_FAILURE;
             }
           }
-        accept_add_recipient(receiver);
+        accept_add_recipient(receiver, NULL, 0, 0);
         s = ss;
         if (!finished) while (*(++s) != 0 && (*s == ',' || isspace(*s)));
         }
@@ -2621,7 +2695,7 @@ while (more)
         {
         debug_printf("Recipients:\n");
         for (i = 0; i < recipients_count; i++)
-          debug_printf("  %s\n", recipients_list[i]);
+          debug_printf("  %s\n", recipients_list[i].address);
         }
       }
 
@@ -2662,21 +2736,31 @@ while (more)
     }
 
   /* Else act on the result of message reception. We should not get here unless
-  message_id[0] is non-zero. If queue_only is set local_queue_only will be
-  TRUE. If it is not, and queue_only_load is set, check that the load average
-  is below it. If it is not, set local_queue_only TRUE. Note that it then
-  remains this way for any subsequent messages on the same SMTP connection.
+  message_id[0] is non-zero. If queue_only is set, local_queue_only will be
+  TRUE. If it is not, check on the number of messages received in this
+  connection. If that's OK and queue_only_load is set, check that the load
+  average is below it. If it is not, set local_queue_only TRUE. Note that it
+  then remains this way for any subsequent messages on the same SMTP connection.
   This is a deliberate choice; even though the load average may fall, it
   doesn't seem right to deliver later messages on the same call when not
   delivering earlier ones. */
 
-  if (!local_queue_only && queue_only_load >= 0)
+  if (!local_queue_only)
     {
-    local_queue_only = (load_average = os_getloadavg()) > queue_only_load;
-    DEBUG(1)
+    if (smtp_accept_queue_per_connection > 0 &&
+        ++message_count > smtp_accept_queue_per_connection)
+      {
+      local_queue_only = TRUE;
+        log_write(4, LOG_MAIN, "no immediate delivery: more than %d messages "
+          "received in one connection\n", smtp_accept_queue_per_connection);
+      }
+    else if (queue_only_load >= 0)
+      {
+      local_queue_only = (load_average = os_getloadavg()) > queue_only_load;
       if (local_queue_only)
-        debug_printf("no delivery: load average = %.2f\n",
+        log_write(4, LOG_MAIN, "no immediate delivery: load average = %.2f\n",
           (double)load_average/1000.0);
+      }
     }
 
   if (!local_queue_only)
@@ -2685,7 +2769,15 @@ while (more)
     is not setuid to root (an exceptional case) call the delivery function
     directly for synchronous delivery, or fork a clone of this process to
     deliver in the background. A fork failure is not a disaster, as the
-    delivery will eventually happen on a subsequent queue run. */
+    delivery will eventually happen on a subsequent queue run.
+
+    We want any smtp input streams to be closed. However, it has been
+    observed that using fclose() here causes trouble. When reading in -bS
+    input, duplicate copies of messages have been seen. The files will be
+    sharing a file pointer with the parent process, and it seems that
+    fclose() (at least on some systems - I saw this on Solaris 2.5.1)
+    messes with that file pointer, at least sometimes. Hence we go for
+    closing the underlying file descriptors. */
 
     if (security_level <= 1 || original_euid != root_uid)
       {
@@ -2695,8 +2787,8 @@ while (more)
         {
         if (smtp_input)
           {
-          fclose(smtp_in);
-          fclose(smtp_out);
+          close(fileno(smtp_in));    /* See comment above */
+          close(fileno(smtp_out));
           smtp_in = NULL;
           }
         (void)deliver_message(message_id, FALSE, FALSE, FALSE);
@@ -2710,16 +2802,16 @@ while (more)
     else
       {
       pid_t pid;
+
       if ((pid = fork()) == 0)
         {
         int i = 0;
-        char *argv[11 + MAX_CLMACROS];
+        char *argv[12 + MAX_CLMACROS];
 
         if (smtp_input)
           {
-          fclose(smtp_in);
-          fclose(smtp_out);
-          smtp_in = NULL;
+          close(fileno(smtp_in));   /* See comment above */
+          close(fileno(smtp_out));
           }
 
         argv[i++] = exim_path;
@@ -2730,6 +2822,7 @@ while (more)
           }
         if (debug_level > 0)
           argv[i++] = string_sprintf("-d%d", debug_level);
+        if (debug_trace_memory) argv[i++] = "-dm";
         if (dont_deliver) argv[i++] = "-N";
         if (queue_smtp) argv[i++] = "-odqs";
         if (queue_remote) argv[i++] = "-odqr";
