@@ -50,13 +50,29 @@ are needed by any of the other headers, including system headers. */
 #include <grp.h>
 #include <syslog.h>
 
-#ifndef NO_SYSEXITS      /* some OS don't have this */
+#ifndef NO_SYSEXITS        /* some OS don't have this */
 #include <sysexits.h>
 #endif
 
+/* Ensure that the sysexits we reference are defined */
+
+#ifndef EX_UNAVAILABLE
+#define EX_UNAVAILABLE 69        /* service unavailable; used for execv fail */
+#endif
+#ifndef EX_CANTCREAT
+#define EX_CANTCREAT   73        /* can't create file: treat as temporary */
+#endif
+#ifndef EX_TEMPFAIL
+#define EX_TEMPFAIL    75        /* temp failure; user is invited to retry */
+#endif
+
 #include <sys/time.h>
 #include <sys/param.h>
+
+#ifndef NO_SYS_RESOURCE_H  /* QNX doesn't have this */
 #include <sys/resource.h>
+#endif
+
 #include <sys/socket.h>
 #include <sys/ioctl.h>
 
@@ -67,14 +83,18 @@ block size), so we use a macro to get that instead.
 
 Also arrange to be able to cut it out altogether for way-out OS that don't have
 anything. I've indented a bit here to try to make the mess a bit more
-intelligible. */
+intelligible. Note that simply defining one name to be another when
+HAVE_SYS_STATVFS_H is not set will not work if the system has a statvfs macro
+or a macro with entries f_frsize and f_bsize. */
 
 #ifdef HAVE_STATFS
   #ifdef HAVE_SYS_STATVFS_H
     #include <sys/statvfs.h>
+    #define STATVFS statvfs
+    #define F_FRSIZE f_frsize
   #else
-    #define statvfs statfs
-    #define f_frsize f_bsize
+    #define STATVFS statfs
+    #define F_FRSIZE f_bsize
     #ifdef HAVE_SYS_VFS_H
       #include <sys/vfs.h>
       #ifdef HAVE_SYS_STATFS_H
@@ -183,7 +203,7 @@ Its os.h file defines SELECT_ARG2_TYPE. For the rest, define a default here. */
 #endif
 
 /* One operating system uses a different type for the 4th argument of
-dn_expand(). It's os.h file defines DN_EXPAND_ARG4_TYPE. For the rest, define a
+dn_expand(). Its os.h file defines DN_EXPAND_ARG4_TYPE. For the rest, define a
 default here. */
 
 #ifndef DN_EXPAND_ARG4_TYPE
