@@ -143,7 +143,8 @@ while (fgets(line, maxsize, f) != NULL)
       continue;
       }
     while (isspace(*s)) s++;
-    strcpy (bptr, --s);
+    *(--s) = ' ';
+    strcpy (bptr, s);
     bptr += p - s;
     }
 
@@ -196,8 +197,12 @@ while (fgets(line, maxsize, f) != NULL)
     keybuffer[i] = 0;
     started = 1;
 
-    if (*s == ':') s++;
     while (isspace(*s))s++;
+    if (*s == ':')
+      {
+      s++;
+      while (isspace(*s))s++;
+      }
     if (*s != 0)
       {
       strcpy(bptr, s);
