@@ -49,14 +49,20 @@ a pipe! */
 extern char *spool_directory;
 char *spool_directory;
 
+extern int store_max_malloc;
+int store_max_malloc = 0;
+
+extern int store_pool;
+int store_pool = POOL_MAIN;
+
 extern volatile int sigalrm_seen;
 volatile int sigalrm_seen;
 
 extern int debug_level;
-int debug_level;
+int debug_level = 0;
 
 extern BOOL debug_trace_memory;
-BOOL debug_trace_memory;
+BOOL debug_trace_memory = 0;
 
 extern BOOL print_topbitchars;
 BOOL print_topbitchars = FALSE;
@@ -213,10 +219,13 @@ while (key != NULL)
       case type_reject:
       reject = (dbdata_reject *)value;
       if (length < sizeof(dbdata_reject))
-        printf("%s %s\n", print_time(reject->time_stamp), keybuffer);
-      else
-        printf("%s %c %s\n", print_time(reject->time_stamp),
+        printf("%s %c   0.0 %s\n", print_time(reject->time_stamp),
           reject->rejected_mail_from? 'T':'F', keybuffer);
+      else
+        {
+        printf("%s %c %5.1f %s\n", print_time(reject->time_stamp),
+          reject->rejected_mail_from? 'T':'F', reject->temp_rate, keybuffer);
+        }
       break;
 
       case type_serialize:
@@ -224,7 +233,7 @@ while (key != NULL)
       printf("%s %s\n", print_time(serialize->time_stamp), keybuffer);
       break;
       }
-    store_free(value);
+    store_reset(value);
     }
   key = dbfn_scan(dbm, FALSE, &cursor);
   }
