@@ -80,6 +80,7 @@ static var_entry var_table[] = {
   { "local_part_data",     vtype_stringptr,   &local_part_data },
   { "local_part_prefix",   vtype_stringptr,   &deliver_localpart_prefix },
   { "local_part_suffix",   vtype_stringptr,   &deliver_localpart_suffix },
+  { "localhost_number",    vtype_int,         &host_number },
   { "message_body",        vtype_msgbody,     &message_body },
   { "message_headers",     vtype_msgheaders,  NULL },
   { "message_id",          vtype_stringptr,   &message_id },
@@ -97,6 +98,8 @@ static var_entry var_table[] = {
   { "n9",                  vtype_filter_int,  &filter_n[9] },
   { "original_domain",     vtype_stringptr,   &deliver_domain_orig },
   { "original_local_part", vtype_stringptr,   &deliver_localpart_orig },
+  { "originator_gid",      vtype_int,         &originator_gid },
+  { "originator_uid",      vtype_int,         &originator_uid },
   { "primary_hostname",    vtype_stringptr,   &primary_hostname },
   { "prohibition_reason",  vtype_stringptr,   &prohibition_reason },
   { "qualify_domain",      vtype_stringptr,   &qualify_domain_sender },
@@ -157,7 +160,8 @@ static char *hashcodes = "abcdefghijklmnopqrtsuvwxyz"
 *************************************************/
 
 /* This function is called to expand a string, and test the result for a "true"
-or "false" value. Failure of the expansion is panic-worthy.
+or "false" value. Failure of the expansion is panic-worthy. All store used by
+the function can be released on exit.
 
 Arguments:
   condition     the condition string
@@ -171,6 +175,7 @@ BOOL
 expand_check_condition(char *condition, char *m1, char *m2)
 {
 int rc;
+void *reset_point = store_get(0);
 char *ss = expand_string(condition);
 if (ss == NULL && !expand_string_forcedfail)
   log_write(0, LOG_MAIN|LOG_PANIC_DIE, "failed to expand condition\"%s\" "
@@ -178,7 +183,7 @@ if (ss == NULL && !expand_string_forcedfail)
 if (ss == NULL) return FALSE;
 rc = ss[0] != 0 && strcmp(ss, "0") != 0 && strcmpic(ss, "no") != 0 &&
   strcmpic(ss, "false") != 0;
-store_free(ss);
+store_reset(reset_point);
 return rc;
 }
 
@@ -242,7 +247,7 @@ read_header_name(char *name, int max, char *s)
 int prelen = strchr(name, '_') - name + 1;
 int ptr = strlen(name) - prelen;
 if (ptr > 0) memmove(name, name+prelen, ptr);
-while (isgraph(*s) && *s != ':')
+while (mac_isgraph(*s) && *s != ':')
   {
   if (ptr < max-1) name[ptr++] = *s;
   s++;
@@ -333,7 +338,7 @@ while (*s != 0)
 
   if (length == subkeylength && strncmp(key, subkey, length) == 0)
     {
-    char *yield = store_malloc(datalength + 1);
+    char *yield = store_get(datalength + 1);
 
     if (*data == '\"')
       {
@@ -437,16 +442,13 @@ Arguments:
               of the returned data are to be ignored (typically '\n' at the
               end of header lines); it should normally be set zero before
               calling this function
-  freestore   a pointer to a BOOL which is set TRUE if the value is built
-              in store which should ultimately be freed; it should be set
-              FALSE before calling this function
 
 Returns:      NULL if the variable does not exist, or
               a pointer to the variable's contents
 */
 
 static char *
-find_variable(char *name, int *choplen, BOOL *freestore)
+find_variable(char *name, int *choplen)
 {
 int first = 0;
 int last = var_table_size;
@@ -501,7 +503,7 @@ while (last > first)
       int ptr = 0;
       int size = 100;
       header_line *h;
-      s = store_malloc(size);
+      s = store_get(size);
       for (h = header_list; h != NULL; h = h->next)
         {
         if (h->type == '*') continue;
@@ -509,7 +511,6 @@ while (last > first)
         }
       s[ptr] = 0;
       }
-    *freestore = TRUE;
     return s;
 
     case vtype_msgbody:                        /* Pointer to msgbody string */
@@ -519,7 +520,7 @@ while (last > first)
       char *body;
       int len = message_body_visible;
       if (len > message_size) len = message_size;
-      *ss = body = store_malloc(len+1);
+      *ss = body = store_get(len+1);
       body[0] = 0;
       lseek(deliver_datafile, data_start_offset, SEEK_SET);
       len = read(deliver_datafile, body, len);
@@ -567,12 +568,12 @@ while (last > first)
       int size = 128;
       int ptr = 0;
       int i;
-      s = store_malloc(size);
+      s = store_get(size);
       for (i = 0; i < recipients_count; i++)
         {
         if (i != 0) s = string_cat(s, &size, &ptr, ", ", 2);
-        s = string_cat(s, &size, &ptr, recipients_list[i],
-          (int)strlen(recipients_list[i]));
+        s = string_cat(s, &size, &ptr, recipients_list[i].address,
+          (int)strlen(recipients_list[i].address));
         }
       s[ptr] = 0;                             /* string_cat leaves room */
       }
@@ -596,17 +597,13 @@ return NULL;
 /*
 Arguments:
   name       the name of the header, without the leading $header_ or $h_
-  freestore  a pointer to a BOOL which is set TRUE if the result is a
-             concatenation of several headers in a new bit of store which
-             the caller must subsequently free; it should be set FALSE on
-             entry
 
 Returns:     NULL if the header does not exist, else
              a pointer to the header's contents
 */
 
 static char *
-find_header(char *name, BOOL *freestore)
+find_header(char *name)
 {
 int len = (int)strlen(name);
 char *yield = NULL;
@@ -625,10 +622,9 @@ for (h = header_list; h != NULL; h = h->next)
         }
       else
         {
-        char *newyield = store_malloc((int)strlen(yield) + h->slen - len + 1);
+        char *newyield = store_get((int)strlen(yield) + h->slen - len + 1);
         strcpy(newyield, yield);
         strcat(newyield, h->text + len);
-        if (*freestore) store_free(yield); else *freestore = TRUE;
         yield = newyield;
         }
       }
@@ -676,7 +672,6 @@ header */
 if (strcmp(name, "def") == 0 && *s == ':')
   {
   char *value;
-  BOOL freestore = FALSE;
 
   s = read_name(name, 256, s+1, "_");
 
@@ -685,7 +680,7 @@ if (strcmp(name, "def") == 0 && *s == ':')
   if (strncmp(name, "header_", 7) == 0 || strncmp(name, "h_", 2) == 0)
     {
     s = read_header_name(name, 256, s);
-    value = find_header(name, &freestore);
+    value = find_header(name);
     if (yield != NULL) *yield = (value != NULL) == testfor;
     }
 
@@ -693,7 +688,7 @@ if (strcmp(name, "def") == 0 && *s == ':')
 
   else
     {
-    value = find_variable(name, NULL, &freestore);
+    value = find_variable(name, NULL);
     if (value == NULL)
       {
       expand_string_message = string_sprintf("unknown variable: %s", name);
@@ -703,7 +698,6 @@ if (strcmp(name, "def") == 0 && *s == ':')
       *yield = (value[0] != 0 && strcmp(value, "0") != 0) == testfor;
     }
 
-  if (freestore) store_free(value);
   return s;
   }
 
@@ -720,7 +714,6 @@ else if (strcmp(name, "exists") == 0)
   if (*s++ != '}') goto COND_FAILED_CURLY;
   if (yield != NULL)
     *yield = (stat(filename, &statbuf) == 0) == testfor;
-  store_free(filename);
   return s;
   }
 
@@ -748,20 +741,13 @@ else if (strcmp(name, "eq") == 0 || strcmp(name, "match") == 0)
 
   /* Result not required */
 
-  if (yield == NULL)
-    {
-    store_free(sub[0]);
-    store_free(sub[1]);
-    return s;
-    }
+  if (yield == NULL) return s;
 
   /* Straight comparison */
 
   if (name[0] == 'e')
     {
     *yield = (strcmp(sub[0], sub[1]) == 0) == testfor;
-    store_free(sub[0]);
-    store_free(sub[1]);
     return s;
     }
 
@@ -776,9 +762,6 @@ else if (strcmp(name, "eq") == 0 || strcmp(name, "match") == 0)
     }
 
   *yield = regex_match_and_setup(re, sub[0], 0, -1) == testfor;
-
-  store_free(re);
-  store_free(sub[1]);   /* Not sub[0] 'cause things point into it */
   return s;
   }
 
@@ -841,8 +824,8 @@ return NULL;
 *                 Expand string                  *
 *************************************************/
 
-/* Returns the expanded string in malloc'ed store. Interpreted
-sequences are:
+/* Returns either an unchanged string, or the expanded string in stacking pool
+store. Interpreted sequences are:
 
    \c                    where c is any character -> copies c literally
    $name                 substitutes the variable
@@ -911,6 +894,12 @@ The skipping flag is set true when expanding a substring that isn't actually
 going to be used (after "if" or "lookup") and it prevents lookups from
 happening lower down.
 
+Store usage: At start, a store block of the length of the input plus 64
+is obtained. This is expanded as necessary by string_cat(), which might have to
+get a new block, or might be able to expand the original. At the end of the
+function we can release any store above that portion of the yield block that
+was actually used. In many cases this will be optimal.
+
 Arguments:
   string         the string to be expanded
   ket_ends       true if expansion is to stop at }
@@ -920,7 +909,7 @@ Arguments:
                  to be used (to allow for optimisation)
 
 Returns:         NULL if expansion fails:
-                   expand_string_forcefail is set TRUE if failure was forced
+                   expand_string_forcedfail is set TRUE if failure was forced
                    expand_string_message contains a textual error message
                  a pointer to the expanded string on success
 */
@@ -929,9 +918,10 @@ static char *
 expand_string_internal(char *string, BOOL ket_ends, char **left, BOOL skipping)
 {
 int ptr = 0;
-int size = (int)strlen(string) + 50;
-char *yield = (char *)store_malloc(size);
+int size = (int)strlen(string) + 64;
 char *s = string;
+char *yield = store_get(size);
+char *original_yield = yield;
 
 expand_string_forcedfail = FALSE;
 expand_string_message = "";
@@ -976,18 +966,17 @@ while (*s != 0)
 
   if (isalpha(*(++s)))
     {
-    BOOL freestore = FALSE;
     int choplen = 0;
     s = read_name(name, sizeof(name), s, "_");
     if (strncmp(name, "header_", 7) == 0 || strncmp(name, "h_", 2) == 0)
       {
       s = read_header_name(name, sizeof(name), s);
-      value = find_header(name, &freestore);
+      value = find_header(name);
       choplen = 1;
       }
     else
       {
-      value = find_variable(name, &choplen, &freestore);
+      value = find_variable(name, &choplen);
       if (value == NULL)
         {
         expand_string_message = string_sprintf("unknown variable: %s", name);
@@ -999,7 +988,6 @@ while (*s != 0)
       int len = (int)strlen(value) - choplen;
       if (len < 0) len = 0;
       yield = string_cat(yield, &size, &ptr, value, len);
-      if (freestore) store_free(value);
       }
     continue;
     }
@@ -1123,9 +1111,6 @@ while (*s != 0)
     else if (sub2 != NULL)
       yield = string_cat(yield, &size, &ptr, sub2, (int)strlen(sub2));
 
-    store_free(sub1);
-    if (sub2 != NULL) store_free(sub2);
-
     /* Restore external setting of expansion variables */
 
     expand_nmax = save_expand_nmax;
@@ -1282,12 +1267,11 @@ while (*s != 0)
       filename = NULL;
       }
 
-    /* If skipping, don't do the next bit - it has the result of
-    leaving lookup_value == NULL, as if the entry was not found. The
-    search_close() function is not called here, leaving any files open.
-    This is deliberate; at suitable places in higher logic, search_tidyup()
-    is called to tidy all open files. This can save opening the same file
-    several times.
+    /* If skipping, don't do the next bit - it has the result of leaving
+    lookup_value == NULL, as if the entry was not found. There is no
+    explicit search_close() function. This is deliberate; at suitable places
+    in higher logic, search_tidyup() is called to tidy all open files. This
+    can save opening the same file several times.
 
     Request that a partial search sets up $1 and maybe $2 by passing
     expand_setup containing zero. If its value changes, reset expand_nmax,
@@ -1311,11 +1295,6 @@ while (*s != 0)
       if (expand_setup > 0) expand_nmax = expand_setup;
       }
 
-    /* The file name string is no longer needed; the key might be because
-    numerical variable values may point into it. */
-
-    if (filename != NULL) store_free(filename);
-
     /* Expand the first substring; $value will get the looked up value. */
 
     while (isspace(*s)) s++;
@@ -1380,12 +1359,8 @@ while (*s != 0)
       expand_nlength[i] = save_expand_nlength[i];
       }
 
-    /* Free store and continue */
+    /* Restore lookup_value and continue */
 
-    store_free(key);
-    store_free(sub1);
-    if (sub2 != NULL) store_free(sub2);
-    if (lookup_value != NULL) store_free(lookup_value);
     lookup_value = save_lookup_value;
     continue;
     }
@@ -1419,9 +1394,7 @@ while (*s != 0)
 
     /* If there are three arguments, this is a numbered field extraction,
     and the first argument must be numerical. If there are two arguments,
-    this is a keyed field extraction. Note that the result of
-    expand_gettokened() is a pointer into a modified sub[2], so it must not
-    itself be freed in that case. */
+    this is a keyed field extraction. */
 
     if (sub[2] != NULL)
       {
@@ -1441,13 +1414,6 @@ while (*s != 0)
 
     if (sub[3] != NULL)
       yield = string_cat(yield, &size, &ptr, sub[3], (int)strlen(sub[3]));
-
-    /* Free temporary store */
-
-    store_free(sub[0]);
-    store_free(sub[1]);
-    if (sub[2] != NULL) store_free(sub[2]); else store_free(sub[3]);
-
     continue;
     }
 
@@ -1472,8 +1438,6 @@ while (*s != 0)
         goto EXPAND_FAILED;
         }
       yield = string_cat(yield, &size, &ptr, expanded, (int)strlen(expanded));
-      store_free(expanded);
-      store_free(sub);
       continue;
       }
 
@@ -1485,7 +1449,6 @@ while (*s != 0)
       char *t = sub - 1;
       while (*(++t) != 0) { *t = tolower(*t); count++; }
       yield = string_cat(yield, &size, &ptr, sub, count);
-      store_free(sub);
       continue;
       }
 
@@ -1510,7 +1473,6 @@ while (*s != 0)
           yield = string_cat(yield, &size, &ptr, sub+domain, end-domain);
           }
         }
-      store_free(sub);
       continue;
       }
 
@@ -1540,7 +1502,6 @@ while (*s != 0)
         yield = string_cat(yield, &size, &ptr, "\"", 1);
         }
       else yield = string_cat(yield, &size, &ptr, sub, (int)strlen(sub));
-      store_free(sub);
       continue;
       }
 
@@ -1555,7 +1516,6 @@ while (*s != 0)
         if (!isalnum(*t)) yield = string_cat(yield, &size, &ptr, "\\", 1);
         yield = string_cat(yield, &size, &ptr, t, 1);
         }
-      store_free(sub);
       continue;
       }
 
@@ -1654,8 +1614,8 @@ while (*s != 0)
       else if (len < 0) len = sublen;
 
       /* Cut length down to maximum possible and if non-zero, get the
-      required characters or compute the hash, free the temporary string,
-      and continue. For a hash, offset is always zero. */
+      required characters or compute the hash and continue. For a hash,
+      offset is always zero. */
 
       if (sublen < offset + len) len = sublen - offset;
       if (len > 0)
@@ -1676,7 +1636,6 @@ while (*s != 0)
           }
         yield = string_cat(yield, &size, &ptr, sub + offset, len);
         }
-      store_free(sub);
       continue;
       }
 
@@ -1691,17 +1650,16 @@ while (*s != 0)
 
   if (*s++ == '}')
     {
-    BOOL freestore = FALSE;
     int choplen = 0;
     if (strncmp(name, "header_", 7) == 0 || strncmp(name, "h_", 2) == 0)
       {
       s = read_header_name(name, 256, s);
-      value = find_header(name, &freestore);
+      value = find_header(name);
       choplen = 1;
       }
     else
       {
-      value = find_variable(name, &choplen, &freestore);
+      value = find_variable(name, &choplen);
       if (value == NULL)
         {
         expand_string_message = string_sprintf("unknown variable: %s", name);
@@ -1712,7 +1670,6 @@ while (*s != 0)
       {
       yield = string_cat(yield, &size, &ptr, value,
         (int)strlen(value) - choplen);
-      if (freestore) store_free(value);
       }
     continue;
     }
@@ -1724,11 +1681,17 @@ while (*s != 0)
   goto EXPAND_FAILED;
   }
 
-/* Expansion succeeded; add a terminating zero and return the
-expanded string. If left != NULL, return a pointer to the terminator. */
+/* Expansion succeeded; add a terminating zero, and if left != NULL, return a
+pointer to the terminator. */
 
 yield[ptr] = 0;
 if (left != NULL) *left = s;
+
+/* Any stacking store that was used above the final string is no longer needed.
+In many cases the final string will be the one that was got at the start and so
+there will be optimal store usage. */
+
+store_reset(yield + ptr + 1);
 return yield;
 
 /* This is the failure exit: easiest to program with a goto. */
@@ -1737,12 +1700,13 @@ EXPAND_FAILED_CURLY:
 expand_string_message = "missing or misplaced { or }";
 
 EXPAND_FAILED:
-store_free(yield);
+store_reset(yield);
 return NULL;
 }
 
 
-/* This is the external function call.
+/* This is the external function call. Do a quick check for any expansion
+metacharacters, and if there are none, just return the input string.
 
 Argument: the string to be expanded
 Returns:  the expanded string, or NULL if expansion failed; if failure was
@@ -1754,10 +1718,28 @@ expand_string(char *string)
 {
 lookup_value = NULL;
 search_find_defer = FALSE;
-return expand_string_internal(string, FALSE, NULL, FALSE);
+return (strpbrk(string, "$\\") == NULL)? string :
+  expand_string_internal(string, FALSE, NULL, FALSE);
 }
 
 
+/* Now and again we want to expand a string and be sure that the result is in a
+new bit of store. This function does that.
+
+Argument: the string to be expanded
+Returns:  the expanded string, always in a new bit of store, or NULL
+*/
+
+char *
+expand_string_copy(char *string)
+{
+char *yield = expand_string(string);
+if (yield == string) yield = string_copy(string);
+return yield;
+}
+
+
+
 
 /*************************************************
 **************************************************
@@ -1771,7 +1753,7 @@ return expand_string_internal(string, FALSE, NULL, FALSE);
 BOOL
 regex_match_and_setup(pcre *re, char *subject, int options, int setup)
 {
-int ovector[2*(EXPAND_MAXN+1)];
+int ovector[3*(EXPAND_MAXN+1)];
 int n = pcre_exec(re, NULL, subject, (int)strlen(subject), PCRE_EOPT|options,
   ovector, sizeof(ovector)/sizeof(int));
 BOOL yield = n >= 0;
@@ -1796,9 +1778,18 @@ int main(int argc, char **argv)
 char buffer[256];
 debug_level = 1;
 debug_file = stderr;
+debug_trace_memory = 0;
 big_buffer = malloc(big_buffer_size);
 
-if (argc > 1) debug_level = atoi(argv[1]);
+if (argc > 1)
+  {
+  if (argv[1][0] == '+')
+    {
+    debug_trace_memory = 1;
+    argv[1]++;
+    }
+  debug_level = atoi(argv[1]);
+  }
 
 printf("Testing string expansion: debug_level = %d\n\n", debug_level);
 
@@ -1808,11 +1799,12 @@ expand_nmax = 1;
 
 while (fgets(buffer, 256, stdin) != NULL)
   {
+  void *reset_point = store_get(0);
   char *yield = expand_string(buffer);
   if (yield != NULL)
     {
     printf("%s\n", yield);
-    store_free(yield);
+    store_reset(reset_point);
     }
   else
     {
