@@ -6,7 +6,12 @@
 /* See the file NOTICE for conditions of use and distribution. */
 
 
-/* Code for mail filtering functions. */
+/* Code for mail filtering functions. In order to recover the store used,
+store_pool is set to POOL_TEMP while running the filter code in normal
+situations. However, it must arrange that any created addresses are placed in
+POOL_MAIN, because that data lives on beyond the running of the filter. Text
+placed in error pointers does not behave like this: it is used immediately by
+the caller. */
 
 
 #include "exim.h"
@@ -26,7 +31,6 @@ typedef struct condition_block {
   BOOL testfor;
   void *left;
   void *right;
-  void *save;
 } condition_block;
 
 typedef struct alias {
@@ -61,13 +65,15 @@ of them. */
 enum { cond_and, cond_or, cond_personal, cond_begins, cond_BEGINS,
        cond_ends, cond_ENDS, cond_is, cond_IS, cond_matches,
        cond_MATCHES, cond_contains, cond_CONTAINS, cond_delivered,
-       cond_above, cond_below, cond_errormsg, cond_foranyaddress };
+       cond_above, cond_below, cond_errormsg, cond_firsttime,
+       cond_manualthaw, cond_foranyaddress };
 
 static char *cond_names[] = {
   "and", "or", "personal",
   "begins", "BEGINS", "ends", "ENDS",
   "is", "IS", "matches", "MATCHES", "contains",
-  "CONTAINS", "delivered", "above", "below", "error_message", "foranyaddress" };
+  "CONTAINS", "delivered", "above", "below", "error_message",
+  "first_delivery", "manually_thawed", "foranyaddress" };
 
 static char *cond_not_names[] = {
   "", "", "not personal",
@@ -76,7 +82,7 @@ static char *cond_not_names[] = {
   "is not", "IS not", "does not match",
   "does not MATCH", "does not contain", "does not CONTAIN",
   "not delivered", "not above", "not below", "not error_message",
-  "not foranyaddress" };
+  "not first_delivery", "not manually_thawed", "not foranyaddress" };
 
 /* Tables of binary condition words and their corresponding types. Not easy
 to amalgamate with the above because of the different variants. */
@@ -147,7 +153,10 @@ static char *command_list[] = {
 
 static int command_list_count = sizeof(command_list)/sizeof(char *);
 
-/* The arguments for the mail command. All the text ones come first. */
+/* The arguments for the mail command. All the text ones come first, with
+the header line ones (that are permitted to include \n followed by white space)
+first, and then the body text one last. It can have \n anywhere. The file names
+may not contain \n. */
 
 static char *mailargs[] = {  /* "to" must be first, and */
   "to",                      /* "cc" and "bcc" must follow */
@@ -163,9 +172,14 @@ static char *mailargs[] = {  /* "to" must be first, and */
 /* This defines the offsets for the arguments; first the string ones, and
 then the non-string ones. The order must be as above. */
 
-enum { mailarg_index_to, mailarg_index_cc, mailarg_index_bcc,
-       mailarg_index_subject, mailarg_index_text, mailarg_index_file,
-       mailarg_index_log, mailarg_index_once,
+enum { mailarg_index_to,
+       mailarg_index_cc,
+       mailarg_index_bcc,
+       mailarg_index_subject,
+       mailarg_index_text,         /* text is first after headers */
+       mailarg_index_file,         /* between text and expand are filenames */
+       mailarg_index_log,
+       mailarg_index_once,
        mailarg_index_expand,       /* first non-string argument */
        mailarg_index_return,
        mailargs_total              /* total number of arguments */
@@ -315,7 +329,7 @@ while (*(++ptr) != 0 && *ptr != '\"' && *ptr != '\n')
 if (*ptr == '\"') ptr++;
   else if (*error_pointer == NULL)
     *error_pointer = string_sprintf("quote missing at end of string "
-      "in filter file line %d", line_number);
+      "in line %d", line_number);
 
 *bp = 0;
 return nextsigchar(ptr, TRUE);
@@ -450,8 +464,8 @@ for (;;)
 
     /* Build a condition block from the specific word. */
 
-    c = store_malloc(sizeof(condition_block));
-    c->left = c->right = c->save = NULL;
+    c = store_get(sizeof(condition_block));
+    c->left = c->right = NULL;
     c->testfor = testfor;
     testfor = TRUE;
 
@@ -459,6 +473,8 @@ for (;;)
 
     if (strcmp(buffer, "delivered") == 0) c->type = cond_delivered;
     else if (strcmp(buffer, "error_message") == 0) c->type = cond_errormsg;
+    else if (strcmp(buffer, "first_delivery") == 0) c->type = cond_firsttime;
+    else if (strcmp(buffer, "manually_thawed") == 0) c->type = cond_manualthaw;
 
     /* Personal can be followed by any number of aliases */
 
@@ -478,7 +494,7 @@ for (;;)
           }
         ptr = nextitem(ptr, buffer, sizeof(buffer), TRUE);
         if (*error_pointer != NULL) break;
-        a = store_malloc(sizeof(alias));
+        a = store_get(sizeof(alias));
         a->name = string_copy(buffer);
         a->next = (alias *)(c->left);
         c->left = (void *)a;
@@ -637,13 +653,12 @@ for (;;)
 
     else if (strcmp(buffer, "and") == 0)
       {
-      condition_block *andc = store_malloc(sizeof(condition_block));
+      condition_block *andc = store_get(sizeof(condition_block));
       andc->parent = current_parent;
       andc->type = cond_and;
       andc->testfor = TRUE;
       andc->left = c;
       andc->right = NULL;    /* insurance */
-      andc->save = NULL;
       *current = andc;
       current = (condition_block **)(&(andc->right));
       current_parent = andc;
@@ -656,7 +671,7 @@ for (;;)
 
     else if (strcmp(buffer, "or") == 0)
       {
-      condition_block *orc = store_malloc(sizeof(condition_block));
+      condition_block *orc = store_get(sizeof(condition_block));
       condition_block *or_parent = NULL;
 
       if (current_parent != NULL)
@@ -678,7 +693,6 @@ for (;;)
       orc->testfor = TRUE;
       orc->left = (current_parent == NULL)? c : current_parent;
       orc->right = NULL;   /* insurance */
-      orc->save = NULL;
       current = (condition_block **)(&(orc->right));
       current_parent = orc;
       }
@@ -734,6 +748,8 @@ switch(c->type)
   case cond_personal:
   case cond_delivered:
   case cond_errormsg:
+  case cond_firsttime:
+  case cond_manualthaw:
   debug_printf("%s", name);
   break;
 
@@ -848,7 +864,7 @@ switch (command)
   if (seen_force || noerror_force)
     {
     *error_pointer = string_sprintf("\"seen\", \"unseen\", or \"noerror\" "
-      "found before an \"%s\" command near line %d in filter file",
+      "found before an \"%s\" command near line %d",
         command_list[command], line_number);
     yield = FALSE;
     }
@@ -950,7 +966,7 @@ switch (command)
 
     if (*error_pointer != NULL) yield = FALSE; else
       {
-      new = store_malloc(sizeof(filter_cmd) + sizeof(void *));
+      new = store_get(sizeof(filter_cmd) + sizeof(void *));
       new->next = NULL;
       **lastcmdptr = new;
       *lastcmdptr = &(new->next);
@@ -1017,7 +1033,7 @@ switch (command)
   /* Finish has no arguments; fmsg defaults to NULL */
 
   case finish_command:
-  new = store_malloc(sizeof(filter_cmd));
+  new = store_get(sizeof(filter_cmd));
   new->next = NULL;
   **lastcmdptr = new;
   *lastcmdptr = &(new->next);
@@ -1034,14 +1050,14 @@ switch (command)
   if (seen_force || noerror_force)
     {
     *error_pointer = string_sprintf("\"seen\", \"unseen\", or \"noerror\" "
-      "found before an \"if\" command near line %d in filter file",
+      "found before an \"if\" command near line %d",
         line_number);
     yield = FALSE;
     }
 
   /* Set up the command block for if */
 
-  new = store_malloc(sizeof(filter_cmd) + 4 * sizeof(void *));
+  new = store_get(sizeof(filter_cmd) + 4 * sizeof(void *));
   new->next = NULL;
   **lastcmdptr = new;
   *lastcmdptr = &(new->next);
@@ -1068,7 +1084,7 @@ switch (command)
     while (had_else_endif == had_elif)
       {
       filter_cmd *newnew =
-        store_malloc(sizeof(filter_cmd) + 4 * sizeof(void *));
+        store_get(sizeof(filter_cmd) + 4 * sizeof(void *));
       new->args[2] = newnew;
       new = newnew;
       new->next = NULL;
@@ -1121,7 +1137,7 @@ switch (command)
 
   case mail_command:
   case vacation_command:
-  new = store_malloc(sizeof(filter_cmd) + mailargs_total * sizeof(void *));
+  new = store_get(sizeof(filter_cmd) + mailargs_total * sizeof(void *));
   new->next = NULL;
   new->command = command;
   new->seen = seen_force? seen_value : FALSE;
@@ -1198,32 +1214,7 @@ switch (command)
       yield = FALSE;
       break;
       }
-    else
-      {
-      char *s = expand_string(buffer);
-      char *p = s;
-      if (s == NULL)
-        {
-        *error_pointer = string_sprintf("failed to expand \"%s\" near "
-          "line %d of filter file: %s", buffer, line_number,
-            expand_string_message);
-        yield = FALSE;
-        break;
-        }
-      for (p = buffer; *p != 0; p++)
-        {
-        int c = (uschar)*p;
-        if (!mac_isprint(c) && c != '\t' && (c != '\n' ||
-            i != mailarg_index_text))
-          {
-          *error_pointer = string_sprintf("non-printing character (%d) "
-            "near line %d of filter file", c, line_number);
-          yield = FALSE;
-          break;
-          }
-        }
-      new->args[i] = s;
-      }
+    else new->args[i] = string_copy(buffer);
     }
 
   /* If this is the vacation command, apply some default settings to
@@ -1242,24 +1233,6 @@ switch (command)
       new->args[mailarg_index_once] = string_copy(".vacation");
     }
 
-  /* We have now set up all the arguments. Check that the the current
-  user is not one of the recipients; this is to prevent loops. */
-
-  /**** This isn't so easy to do - what to do if it does match? Just remove
-  the offending address? Then must check that there is at least one address
-  left...
-
-  sprintf(buffer, "%s@%s", deliver_localpart, deliver_domain);
-
-  for (i = mailarg_index_to; i <= mailarg_index_bcc; i++)
-    {
-    if (new->args[i] != NULL && strstric(new->args[i], buffer, FALSE) != NULL)
-      {
-
-      }
-    }
-  ****** needs some thought ******/
-
   /* Join the address on to the chain of generated addresses */
 
   **lastcmdptr = new;
@@ -1274,7 +1247,7 @@ switch (command)
   if (seen_force)
     {
     *error_pointer = string_sprintf("\"seen\" or \"unseen\" repeated "
-      "near line %d in filter file", line_number);
+      "near line %d", line_number);
     yield = FALSE;
     }
   seen_value = (command == seen_command);
@@ -1493,19 +1466,6 @@ switch (c->type)
             test_header("subject:", FALSE, NULL, 1, "circular") &&
             test_header("precedence:", FALSE, NULL, 1, "bulk") &&
             (sender_address == NULL || sender_address[0] != 0);
-
-    /* Tidy up */
-
-    if (self_from != self) store_free(self_from);
-    if (self_to   != self) store_free(self_to);
-    store_free(self);
-
-    if (psself != NULL)
-      {
-      if (psself_from != psself) store_free(psself_from);
-      if (psself_to   != psself) store_free(psself_to);
-      store_free(psself);
-      }
     }
   break;
 
@@ -1521,6 +1481,20 @@ switch (c->type)
     (sender_address == NULL || sender_address[0] == 0 || user_null_sender);
   break;
 
+  /* Only FALSE if a message is actually being processed; TRUE for address
+  and filter testing and verification. */
+
+  case cond_firsttime:
+  yield = filter_test != NULL || message_id[0] == 0 || deliver_firsttime;
+  break;
+
+  /* Only TRUE if a message is actually being processed; FALSE for address
+  testing and verification. */
+
+  case cond_manualthaw:
+  yield = message_id[0] != 0 && deliver_manual_thaw;
+  break;
+
   /* The foranyaddress condition loops through a list of addresses */
 
   case cond_foranyaddress:
@@ -1562,16 +1536,10 @@ switch (c->type)
     if (saveend == 0) break;
     pp = p + 1;
     }
-
-  store_free(ss);
   break;
 
   /* All other conditions have left and right values that need expanding;
-  on error, it doesn't matter what value is returned. We need to save the
-  expansion of the first string because if the second is a regular expression,
-  there may be pointers into the first that are saved. There is a slot in the
-  condition block for hanging on a string to be saved until the whole data
-  structure is freed. */
+  on error, it doesn't matter what value is returned. */
 
   default:
   p = (char *)(c->left);
@@ -1587,8 +1555,6 @@ switch (c->type)
     p = (char *)(c->right);
     }
 
-  c->save = exp[0];
-
   /* Inner switch for the different cases */
 
   switch(c->type)
@@ -1641,12 +1607,11 @@ switch (c->type)
     if (re == NULL)
       {
       *error_pointer = string_sprintf("error while compiling "
-        "regular expression \"%s\" in filter file: %s at offset %d",
+        "regular expression \"%s\": %s at offset %d",
         exp[1], regcomp_error, regcomp_error_offset);
       return FALSE;
       }
     yield = regex_match_and_setup(re, exp[0], PCRE_EOPT, -1);
-    store_free(re);
     break;
 
     /* For above and below, convert the strings to numbers */
@@ -1658,16 +1623,14 @@ switch (c->type)
       val[i] = get_number(exp[i], &yield);
       if (!yield)
         {
-        *error_pointer = string_sprintf("malformed numerical string \"%s\" "
-          "in filter file", exp[i]);
+        *error_pointer = string_sprintf("malformed numerical string \"%s\"",
+          exp[i]);
         return FALSE;
         }
       }
     yield = (c->type == cond_above)? (val[0] > val[1]) : (val[0] < val[1]);
     break;
     }
-
-  store_free(exp[1]);
   break;
   }
 
@@ -1686,127 +1649,6 @@ return yield == c->testfor;
 
 
 
-/*************************************************
-*         Free a condition block                 *
-*************************************************/
-
-/*
-Argument:  pointer to the condition block
-Returns:   nothing
-*/
-
-static void
-free_condition(condition_block *c)
-{
-alias *a, *aa;
-if (c == NULL) return;
-switch (c->type)
-  {
-  case cond_foranyaddress:
-  store_free(c->left);
-  free_condition(c->right);
-  break;
-
-  case cond_and:
-  case cond_or:
-  free_condition(c->left);
-  free_condition(c->right);
-  break;
-
-  case cond_contains:
-  case cond_CONTAINS:
-  case cond_matches:
-  case cond_MATCHES:
-  case cond_is:
-  case cond_IS:
-  case cond_begins:
-  case cond_BEGINS:
-  case cond_ends:
-  case cond_ENDS:
-  store_free(c->left);
-  store_free(c->right);
-  break;
-
-  case cond_personal:
-  for (a = (alias *)(c->left); a != NULL; a = aa)
-    {
-    aa = a->next;
-    store_free(a->name);
-    store_free(a);
-    }
-  break;
-
-  case cond_delivered:
-  case cond_errormsg:
-  case cond_above:
-  case cond_below:
-  break;
-  }
-
-if (c->save != NULL) store_free(c->save);
-store_free(c);
-}
-
-
-
-/*************************************************
-*         Free chain of command blocks           *
-*************************************************/
-
-/*
-Argument:  pointer to the command block
-Returns:   nothing
-*/
-
-static void
-free_commands(filter_cmd *c)
-{
-while (c != NULL)
-  {
-  filter_cmd *cc = c;
-  c = c->next;
-  switch(cc->command)
-    {
-    case deliver_command:
-    case fail_command:
-    case freeze_command:
-    case finish_command:
-    case headers_command:
-    case logwrite_command:
-    case logfile_command:
-    case pipe_command:
-    case save_command:
-    case testprint_command:
-    if (cc->args[0] != NULL) store_free(cc->args[0]);
-    break;
-
-    case add_command:
-    if (cc->args[0] != NULL) store_free(cc->args[0]);
-    if (cc->args[1] != NULL) store_free(cc->args[1]);
-    break;
-
-    case if_command:
-    free_condition((condition_block *)(cc->args[0]));
-    if (cc->args[1] != NULL) free_commands((filter_cmd *)(cc->args[1]));
-    if (cc->args[2] != NULL) free_commands((filter_cmd *)(cc->args[2]));
-    break;
-
-    case mail_command:
-    case vacation_command:
-      {
-      int i;
-      for (i = 0; i < mailargs_string_count; i++)
-        if (cc->args[i] != NULL) store_free(cc->args[i]);
-      }
-    break;
-    }
-  store_free(cc);
-  }
-}
-
-
-
-
 /*************************************************
 *          Interpret chain of commands           *
 *************************************************/
@@ -1831,6 +1673,7 @@ interpret_commands(filter_cmd *commands, address_item **generated,
 {
 char *s;
 int mode;
+int pool_reset = store_pool;
 address_item *addr;
 BOOL condition_value;
 
@@ -1870,7 +1713,6 @@ while (commands != NULL)
           "command", ss);
         return FALSE;
         }
-      store_free(s);
       }
 
     filter_n[n[1]] += n[0];
@@ -1908,8 +1750,6 @@ while (commands != NULL)
         }
       }
 
-
-
     /* Test case: report what would happen */
 
     if (filter_test != NULL)
@@ -1951,26 +1791,24 @@ while (commands != NULL)
         DEBUG(10) debug_printf("Filter: %ssave message to: %s%s\n",
           (commands->seen)? "" : "unseen ", s,
           commands->noerror? " (noerror)" : "");
-        if (s[0] != '/')
-          {
-          char *ss = string_sprintf("%s/%s", deliver_home, s);
-          store_free(s);
-          s = ss;
-          }
+        if (s[0] != '/') s = string_sprintf("%s/%s", deliver_home, s);
         pfr = TRUE;
         break;
         }
 
       /* Create the new address and add it to the chain, setting the
       ignore_error flag if necessary, the mode value for save, and an
-      appropriate pfr flag. */
+      appropriate pfr flag. Ensure that the address is placed in the
+      MAIN storage pool. */
 
-      addr = deliver_make_addr(s);
+      store_pool = POOL_MAIN;
+      addr = deliver_make_addr(s, TRUE);  /* TRUE => copy s */
       addr->pfr = pfr;
       addr->mode = mode;
       addr->ignore_error = commands->noerror;
       addr->next = *generated;
       *generated = addr;
+      store_pool = pool_reset;
       }
     break;
 
@@ -1988,19 +1826,16 @@ while (commands != NULL)
       DEBUG(10) debug_printf("Filter: %spipe message to: %s%s\n",
         (commands->seen)? "" : "unseen ", s,
         commands->noerror? " (noerror)" : "");
-      if (s[0] != '|')
-        {
-        char *ss = string_sprintf("|%s", s);
-        store_free(s);
-        s = ss;
-        }
+      if (s[0] != '|') s = string_sprintf("|%s", s);
 
       /* Create the new address and add it to the chain, setting the
       ignore_error flag if necessary. Set the expand_pipe flag so that
       each command argument is expanded in the transport after the command
-      has been split up into separate arguments. */
+      has been split up into separate arguments. Ensure the address is
+      placed in the MAIN storage pool */
 
-      addr = deliver_make_addr(s);
+      store_pool = POOL_MAIN;
+      addr = deliver_make_addr(s, TRUE);  /* TRUE => copy s */
       addr->pfr = TRUE;
       addr->ignore_error = commands->noerror;
       addr->next = *generated;
@@ -2013,12 +1848,14 @@ while (commands != NULL)
       if (expand_nmax >= 0)
         {
         int i;
-        char **ss = store_malloc(sizeof(char *) * (expand_nmax + 2));
+        char **ss = store_get(sizeof(char *) * (expand_nmax + 2));
         addr->pipe_expandn = ss;
         for (i = 0; i <= expand_nmax; i++)
           *ss++ = string_copyn(expand_nstring[i], expand_nlength[i]);
         *ss = NULL;
         }
+
+      store_pool = pool_reset;
       }
     break;
 
@@ -2062,7 +1899,7 @@ while (commands != NULL)
       {
       indent();
       printf("%sogwrite \"%s\"\n", (commands->seen)? "Seen l" : "L",
-        string_printing(s, FALSE));
+        string_printing(s));
       }
 
     /* Attempt to write to a log file only if configured as permissible and
@@ -2095,7 +1932,7 @@ while (commands != NULL)
         {
         if (log_filename == NULL)
           {
-          *error_pointer = "attempt to obey \"logwrite\" command in filter file "
+          *error_pointer = "attempt to obey \"logwrite\" command "
             "without a previous \"logfile\"";
           return FALSE;
           }
@@ -2127,6 +1964,7 @@ while (commands != NULL)
 
       if (s == NULL)
         {
+        if (expand_string_forcedfail) break;
         *error_pointer = string_sprintf("failed to expand \"%s\": %s",
           (char *)(commands->args[0]), expand_string_message);
         return FALSE;
@@ -2134,9 +1972,17 @@ while (commands != NULL)
 
       if (filter_test != NULL)
         printf("Headers %s \"%s\"\n", adding? "add" : "remove",
-          string_printing(s, FALSE));
+          string_printing(s));
 
-      else if (adding) header_add(htype_other, "%s", s);
+      else if (adding)
+        {
+        while (isspace(*s)) s++;
+        if (s[0] != 0)
+          {
+          header_add(htype_other, "%s%s", s, (s[(int)strlen(s)-1] == '\n')?
+            "" : "\n");
+          }
+        }
 
       else
         {
@@ -2160,12 +2006,13 @@ while (commands != NULL)
           h->type = htype_old;
           }
         }
-      store_free(s);
       }
     break;
 
     /* Freeze and fail are available only when there is somewhere to pass
-    back the action. These commands are rejected at parse time otherwise. */
+    back the action. These commands are rejected at parse time otherwise.
+    The message must be placed in the MAIN storage pool, since it lives
+    on beyond the running of the filter. */
 
     case fail_command:
     case freeze_command:
@@ -2219,19 +2066,6 @@ while (commands != NULL)
     condition_value =
       test_condition((condition_block *)(commands->args[0]), delivered, TRUE);
     if (*error_pointer != NULL) return FALSE;
-
-    /**** This printing happens for each sub-condition now
-    if (filter_test != NULL)
-      {
-      indent();
-      debug_printf("\"If\" condition is %s: %s\n", condition_value?
-        "true" : "false", commands->args[3]);
-      }
-    else
-      DEBUG(10) debug_printf("Filter: condition is %s: %s\n",
-        condition_value? "true" : "false", commands->args[3]);
-    ****/
-
     output_indent += 2;
     if (!interpret_commands((filter_cmd *)
       (commands->args[condition_value? 1:2]),
@@ -2256,12 +2090,62 @@ while (commands != NULL)
       break;
       }
 
+    /* Expand all the string arguments of the command */
+
+    for (i = 0; i < mailargs_string_count; i++)
+      {
+      char *s = (char *)(commands->args[i]);
+      char *p;
+
+      if (s == NULL) continue;
+      s = expand_string((char *)(commands->args[i]));
+
+      if (s == NULL)
+        {
+        *error_pointer = string_sprintf("failed to expand \"%s\" in "
+          "%s command: %s", (char *)(commands->args[i]),
+          command_list[commands->command],
+          expand_string_message);
+        return FALSE;
+        }
+
+      /* Check the contents of the strings. The type of string can be deduced
+      from the value of i. If it is > mailarg_index_text, we are dealing with a
+      file name, which cannot contain non-printing characters. If it is less
+      than mailarg_index_text we are dealing with something that will go in a
+      message header line, where newlines must be followed by white space. If
+      i is equal to mailarg_index_text its a text string for the body, where
+      anything goes. */
+
+      for (p = s; *p != 0; p++)
+        {
+        int c = (uschar)*p;
+        if (i > mailarg_index_text && !mac_isprint(c))
+          {
+          *error_pointer = string_sprintf("non-printing character in \"%s\" "
+            "in %s command", string_printing(s),
+            command_list[commands->command]);
+          return FALSE;
+          }
+        else if (i < mailarg_index_text && c == '\n' && !isspace(p[1]))
+          {
+          *error_pointer = string_sprintf("\\n not followed by space in \"%s\" "
+            "in %s command", string_printing(s),
+            command_list[commands->command]);
+          return FALSE;
+          }
+        }
+
+      /* The string is OK */
+
+      commands->args[i] = s;
+      }
+
     /* Proceed with mail or vacation command */
 
     if (commands->seen) *delivered = TRUE;
     if (filter_test != NULL)
       {
-      int i;
       char *to = (char *)(commands->args[mailarg_index_to]);
       indent();
       printf("%sail to: %s%s%s\n", (commands->seen)? "Seen m" : "M",
@@ -2276,7 +2160,7 @@ while (commands != NULL)
           int len = (int)strlen(mailargs[i]);
           int indent = (debug_level > 0)? output_indent : 0;
           while (len++ < 7 + indent) printf(" ");
-          printf("%s: %s%s\n", mailargs[i], string_printing(arg, FALSE),
+          printf("%s: %s%s\n", mailargs[i], string_printing(arg),
             (commands->args[mailarg_index_expand] != NULL &&
               strcmp(mailargs[i], "file") == 0)? " (expanded)" : "");
           }
@@ -2286,10 +2170,8 @@ while (commands != NULL)
       }
     else
       {
-      int i;
       char *to = (char *)(commands->args[mailarg_index_to]);
-      to = (to == NULL)? expand_string("$reply_address") :
-         string_copy(to);
+      if (to == NULL) to = expand_string("$reply_address");
       while (isspace(*to)) to++;
 
       DEBUG(10)
@@ -2306,22 +2188,25 @@ while (commands != NULL)
             {
             int len = (int)strlen(mailargs[i]);
             while (len++ < 15) debug_printf(" ");
-            debug_printf("%s: %s%s\n", mailargs[i], string_printing(arg,FALSE),
+            debug_printf("%s: %s%s\n", mailargs[i], string_printing(arg),
               (commands->args[mailarg_index_expand] != NULL &&
                 strcmp(mailargs[i], "file") == 0)? " (expanded)" : "");
             }
           }
         }
 
-      addr = deliver_make_addr(string_sprintf(">%s", to));
+      /* Ensure the address is created in the MAIN storage pool. */
+
+      store_pool = POOL_MAIN;
+      addr = deliver_make_addr(string_sprintf(">%s", to), FALSE);
       addr->pfr = TRUE;
       addr->ignore_error = commands->noerror;
       addr->next = *generated;
       *generated = addr;
-      addr->reply = store_malloc(sizeof(reply_item));
+      addr->reply = store_get(sizeof(reply_item));
       addr->reply->headers = NULL;    /* Can't set that from here (yet?) */
       addr->reply->from = NULL;
-      addr->reply->to = to;
+      addr->reply->to = string_copy(to);
       addr->reply->file_expand =
         commands->args[mailarg_index_expand] != NULL;
       addr->reply->return_message =
@@ -2332,6 +2217,7 @@ while (commands != NULL)
         *((char **)(((char *)(addr->reply)) + reply_offsets[i])) =
           (ss == NULL)? NULL : string_copy(ss);
         }
+      store_pool = pool_reset;
       }
     break;
 
@@ -2348,9 +2234,9 @@ while (commands != NULL)
         }
 
       if (filter_test == NULL)
-        debug_printf("Filter: testprint: %s\n", string_printing(s, TRUE));
+        debug_printf("Filter: testprint: %s\n", string_printing(s));
       else
-        printf("Testprint: %s\n", string_printing(s, TRUE));
+        printf("Testprint: %s\n", string_printing(s));
       }
     }
 
@@ -2400,6 +2286,7 @@ global flag that is always TRUE while filtering, and zero the variables. */
 expect_endif = 0;
 output_indent = 0;
 filter_running = TRUE;
+filter_fmsg = NULL;
 for (i = 0; i < FILTER_VARIABLE_COUNT; i++) filter_n[i] = 0;
 
 /* To save having to pass this about all the time, make it static, and
@@ -2436,19 +2323,14 @@ if (read_command_list(&ptr, &lastcmdptr, FALSE))
     }
   }
 
-/* Close the log file if it was opened. */
+/* Close the log file if it was opened, and kill off any numerical variables
+before returning. */
 
 if (log_fd >= 0) close(log_fd);
-
-/* Free up the store that was used in building data structures before
-returning, and kill off any numerical variables. */
-
-free_commands(commands);
-
-DEBUG(9) debug_printf("Filter: end of processing\n");
-
 expand_nmax = -1;
 filter_running = FALSE;
+
+DEBUG(9) debug_printf("Filter: end of processing\n");
 return *error_pointer == NULL;
 }
 
@@ -2519,7 +2401,7 @@ if (fstat(fileno(filter), &statbuf) != 0)
   goto RESTORE_UID;
   }
 
-filebuf = store_malloc(statbuf.st_size + 1);
+filebuf = store_get(statbuf.st_size + 1);
 size_read = fread(filebuf, 1, statbuf.st_size, filter);
 fclose(filter);
 
@@ -2539,7 +2421,6 @@ system_filtering = TRUE;
 yield = filter_interpret(filebuf, generated, delivered, action, error,
   0, TRUE);
 system_filtering = FALSE;
-store_free(filebuf);
 
 /* Restore the original effective uid/gid before returning. */
 
@@ -2599,7 +2480,7 @@ if (fstat(fd, &statbuf) != 0)
   return FALSE;
   }
 
-filebuf = store_malloc(statbuf.st_size + 1);
+filebuf = store_get(statbuf.st_size + 1);
 rc = read(fd, filebuf, statbuf.st_size);
 close(fd);
 
@@ -2665,7 +2546,7 @@ we might as well set up the message_body variable at the same time (when
 out here rather than having options in read_message_data, in order to keep that
 function as efficient as possible. */
 
-message_body = body = store_malloc(message_body_visible + 1);
+message_body = body = store_get(message_body_visible + 1);
 body_len = 0;
 
 if (!dot_ended && !feof(stdin))
