@@ -28,6 +28,14 @@ static void dummy(void) { dummy(); }
 /* For comments on these variables, see globals.h. I'm too idle to
 duplicate them here... */
 
+#ifdef SUPPORT_DSN
+BOOL   dsn                    = FALSE;
+char  *dsn_envid              = NULL;
+int    dsn_ret                = 0;
+pcre  *regex_DSN              = NULL;
+BOOL   smtp_use_dsn           = FALSE;
+#endif
+
 BOOL   accept_8bitmime        = FALSE;
 BOOL   accept_call_bombout    = FALSE;
 int    accept_timeout         = 0;
@@ -50,6 +58,7 @@ char  *big_buffer             = NULL;
 int    big_buffer_size        = BIG_BUFFER_SIZE;
 int    body_linecount         = 0;
 
+BOOL   check_dns_names        = TRUE;
 BOOL   check_relay            = FALSE;
 int    check_log_inodes       = 0;
 int    check_log_space        = 0;
@@ -84,6 +93,7 @@ char  *deliver_out_buffer     = NULL;
 int    deliver_datafile       = -1;
 char  *deliver_domain         = NULL;
 char  *deliver_domain_orig    = NULL;
+BOOL   deliver_firsttime      = FALSE;
 BOOL   deliver_force          = FALSE;
 BOOL   deliver_freeze         = FALSE;
 int    deliver_frozen_at      = 0;
@@ -114,6 +124,9 @@ director_instance director_defaults = {
    NULL,                      /* errors_to */
    NULL,                      /* extra_headers */
    NULL,                      /* remove_headers */
+   NULL,                      /* debug_string */
+   NULL,                      /* fallback_hosts */
+   NULL,                      /* fallback_hostlist */
    NULL,                      /* re_domains */
    NULL,                      /* re_except_domains */
    NULL,                      /* re_local_parts */
@@ -124,6 +137,13 @@ director_instance director_defaults = {
    TRUE,                      /* verify_recipient */
    NULL,                      /* transport */
    NULL,                      /* expanded_transport */
+   NULL,                      /* expand_uid */
+   NULL,                      /* expand_gid */
+   (uid_t)(-1),               /* uid */
+   (gid_t)(-1),               /* gid */
+   FALSE,                     /* uid_set */
+   FALSE,                     /* gid_set */
+   FALSE,                     /* initgroups */
    NULL,                      /* condition */
    NULL,                      /* prefix */
    NULL,                      /* suffix */
@@ -186,6 +206,7 @@ int    expand_nmax            = -1;
 char  *expand_nstring[EXPAND_MAXN+1];
 BOOL   expand_string_forcedfail = FALSE;
 char  *expand_string_message;
+BOOL   extract_addresses_remove_arguments = TRUE;
 
 char  *filter_fmsg            = NULL;
 int    filter_n[FILTER_VARIABLE_COUNT];
@@ -234,6 +255,10 @@ BOOL   headers_check_syntax   = FALSE;
 BOOL   headers_checks_fail    = TRUE;
 BOOL   headers_sender_verify  = FALSE;
 BOOL   headers_sender_verify_errmsg = FALSE;
+char  *helo_accept_junk_hosts = NULL;
+host_item *helo_accept_junk_hostlist = NULL;
+char  *helo_accept_junk_nets  = NULL;
+ip_net_item *helo_accept_junk_netlist = NULL;
 BOOL   helo_strict_syntax     = FALSE;
 BOOL   helo_verify            = FALSE;
 char  *helo_verify_except_hosts = NULL;
@@ -246,6 +271,9 @@ BOOL   host_checking          = FALSE;
 char  *host_findbyaddr_msg    = "";
 char  *host_lookup_nets       = NULL;
 ip_net_item *host_lookup_netlist = NULL;
+int    host_number            = 0;
+char  *host_number_string     = NULL;
+char  *hosts_treat_as_local   = NULL;
 
 BOOL   ignore_errmsg_errors   = FALSE;
 int    ignore_errmsg_errors_after = 0;
@@ -275,7 +303,7 @@ BOOL   log_received_recipients = FALSE;
 BOOL   log_refused_recipients = FALSE;
 BOOL   log_rewrites           = FALSE;
 BOOL   log_smtp_confirmation  = FALSE;
-FILE  *log_stderr;
+FILE  *log_stderr             = NULL;
 BOOL   log_subject            = FALSE;
 char  *lookup_key             = NULL;
 
@@ -336,6 +364,7 @@ char  *prohibition_reason     = NULL;
 char  *qualify_domain_recipient = NULL;
 char  *qualify_domain_sender  = NULL;
 BOOL   queue_2stage           = FALSE;
+BOOL   queue_list_requires_admin = TRUE;
 BOOL   queue_run_force        = FALSE;
 BOOL   queue_run_in_order     = FALSE;
 BOOL   queue_run_local        = FALSE;
@@ -361,6 +390,7 @@ char **raw_recipients         = NULL;
 re_block *re_dns_again_means_nonexist = NULL;
 re_block *re_hold_domains     = NULL;
 re_block *re_hold_domains_except = NULL;
+re_block *re_hosts_treat_as_local = NULL;
 re_block *re_local_domains    = NULL;
 re_block *re_percent_hack_domains = NULL;
 re_block *re_queue_remote_except = NULL;
@@ -396,8 +426,8 @@ char  *received_header_text   =
     "by ${primary_hostname} "
     "${if def:received_protocol {with ${received_protocol}}} "
     "(Exim ${version_number} #${compile_number})\n\t"
-    "${if def:received_for {for $received_for\n\t}}"
     "id ${message_id}"
+    "${if def:received_for {\n\tfor $received_for}}"
     "\0<---------------Space to patch received_header_text->";
 
 int    received_headers_max   = 30;
@@ -417,7 +447,7 @@ ip_net_item *receiver_verify_except_netlist = NULL;
 char  *receiver_verify_senders = NULL;
 char  *receiver_verify_senders_except = NULL;
 int    recipients_count       = 0;
-char **recipients_list        = NULL;
+recipient_item *recipients_list = NULL;
 int    recipients_list_max    = 0;
 int    recipients_max         = 0;
 BOOL   recipients_max_reject  = FALSE;
@@ -446,6 +476,7 @@ host_item *rfc1413_except_hostlist = NULL;
 char  *rfc1413_except_nets    = NULL;
 ip_net_item *rfc1413_except_netlist = NULL;
 int    rfc1413_query_timeout  = 30;
+BOOL   rfc821_domains         = FALSE;
 uid_t  root_uid               = ROOT_UID;
 char  *route_option           = NULL;
 
@@ -463,6 +494,9 @@ router_instance router_defaults = {
    NULL,                      /* errors_to */
    NULL,                      /* extra_headers */
    NULL,                      /* remove_headers */
+   NULL,                      /* debug_string */
+   NULL,                      /* fallback_hosts */
+   NULL,                      /* fallback_hostlist */
    NULL,                      /* re_domains */
    NULL,                      /* re_except_domains */
    NULL,                      /* re_local_parts */
@@ -473,6 +507,13 @@ router_instance router_defaults = {
    TRUE,                      /* verify_recipient */
    NULL,                      /* transport instance */
    NULL,                      /* expanded_transport */
+   NULL,                      /* expand_uid */
+   NULL,                      /* expand_gid */
+   (uid_t)(-1),               /* uid */
+   (gid_t)(-1),               /* gid */
+   FALSE,                     /* uid_set */
+   FALSE,                     /* gid_set */
+   FALSE,                     /* initgroups */
    NULL,                      /* condition */
    NULL,                      /* senders */
    NULL,                      /* except_senders */
@@ -497,6 +538,7 @@ char  *sender_accept_recipients = NULL;
 char  *sender_address         = NULL;
 BOOL   sender_address_forced  = FALSE;
 char  *sender_address_relay   = NULL;
+BOOL   sender_address_rewritten = FALSE;
 char  *sender_fullhost        = NULL;
 char  *sender_helo_name       = NULL;
 char  *sender_host_accept     = NULL;
@@ -505,6 +547,7 @@ char  *sender_host_accept_relay = NULL;
 host_item *sender_host_accept_relay_hosts = NULL;
 char  *sender_host_address    = NULL;
 char  *sender_host_name       = NULL;
+BOOL   sender_host_notsocket  = FALSE;
 char  *sender_host_reject     = NULL;
 char  *sender_host_reject_except = NULL;
 host_item *sender_host_reject_hosts = NULL;
@@ -513,6 +556,8 @@ char  *sender_host_reject_recipients = NULL;
 host_item *sender_host_reject_recipients_hosts = NULL;
 char  *sender_host_reject_relay = NULL;
 host_item *sender_host_reject_relay_hosts = NULL;
+char  *sender_host_reject_relay_except = NULL;
+host_item *sender_host_reject_relay_except_hosts = NULL;
 BOOL   sender_host_unknown    = FALSE;
 char  *sender_ident           = NULL;
 BOOL   sender_local           = FALSE;
@@ -528,6 +573,8 @@ char  *sender_net_reject_recipients = NULL;
 ip_net_item *sender_net_reject_recipients_nets = NULL;
 char  *sender_net_reject_relay = NULL;
 ip_net_item *sender_net_reject_relay_nets = NULL;
+char  *sender_net_reject_relay_except = NULL;
+ip_net_item *sender_net_reject_relay_except_nets = NULL;
 char  *sender_rcvhost         = NULL;
 char  *sender_reject          = NULL;
 char  *sender_reject_except   = NULL;
@@ -546,11 +593,13 @@ ip_net_item *sender_verify_except_netlist = NULL;
 BOOL   sender_verify_fixup    = FALSE;
 BOOL   sender_verify_log_details = FALSE;
 BOOL   sender_verify_reject   = TRUE;
+int    sender_verify_max_retry_rate = 12;
 volatile BOOL sigalrm_seen    = FALSE;
 char **sighup_argv            = NULL;
 int    smtp_accept_count      = 0;
 int    smtp_accept_max        = 20;
 int    smtp_accept_queue      = 0;
+int    smtp_accept_queue_per_connection = 10;
 int    smtp_accept_reserve    = 0;
 char  *smtp_banner            = "${primary_hostname} ESMTP "
                             "Exim ${version_number} #${compile_number} "
@@ -561,9 +610,12 @@ int    smtp_connect_backlog   = 5;
 FILE  *smtp_in                = NULL;
 BOOL   smtp_input             = FALSE;
 int    smtp_load_reserve      = -1;
+BOOL   smtp_log_connections   = FALSE;
 FILE  *smtp_out               = NULL;
 int    smtp_port              = -1;
+BOOL   smtp_reading_data      = FALSE;
 int    smtp_receive_timeout   = 5*60;
+char  *smtp_etrn_command      = NULL;
 host_item *smtp_etrn_hostlist = NULL;
 char  *smtp_etrn_hosts        = NULL;
 ip_net_item *smtp_etrn_netlist = NULL;
@@ -585,6 +637,8 @@ char  *spool_directory        = SPOOL_DIRECTORY
 char  *stderr_filename        = STDERR_FILE
                           "\0<--------------Space to patch stderr_filename->";
 
+int    store_max_malloc       = 0;
+int    store_pool             = POOL_MAIN;
 BOOL   strip_excess_angle_brackets = FALSE;
 BOOL   strip_trailing_dot     = FALSE;
 
@@ -614,8 +668,8 @@ transport_instance transport_defaults = {
    100,                      /* batch_max */
    FALSE,                    /* uid_set */
    FALSE,                    /* gid_set */
-   -1,                       /* uid */
-   -1,                       /* gid */
+   (uid_t)(-1),              /* uid */
+   (gid_t)(-1),              /* gid */
    NULL,                     /* expand_uid */
    NULL,                     /* expand_gid */
    NULL,                     /* home_dir */
@@ -629,6 +683,10 @@ transport_instance transport_defaults = {
    NULL,                     /* filter_command */
    NULL,                     /* add_headers */
    NULL,                     /* remove_headers */
+   NULL,                     /* debug_string */
+   FALSE,                    /* delivery_date_add */
+   FALSE,                    /* envelope_to_add */
+   FALSE,                    /* return_path_add */
    FALSE,                    /* return_output */
    FALSE,                    /* return_fail_output */
    FALSE,                    /* log_output */
@@ -639,8 +697,8 @@ transport_instance transport_defaults = {
 };
 
 int    transport_count;
-int    transport_chunk_timeout= 0;
 char **transport_filter_argv  = NULL;
+int    transport_write_timeout= 0;
 
 tree_node *tree_duplicates    = NULL;
 tree_node *tree_nonrecipients = NULL;
@@ -685,6 +743,7 @@ char  *warnmsg_delay          = NULL;
 char  *warnmsg_file           = NULL;
 char  *warnmsg_recipients     = NULL;
 
+char  *verify_forced_errmsg   = NULL;
 BOOL   verify_only            = FALSE;
 char  *version_copyright      = "Copyright (c) University of Cambridge 1998";
 char  *version_date           = "?";
