@@ -9,10 +9,18 @@
 they are easy to find. First put any specials that are required for some
 operating systems. */
 
-#ifdef  NEED_H_ERRNO
+#ifdef NEED_H_ERRNO
 extern int h_errno;
 #endif
 
+#ifdef SUPPORT_DSN
+extern BOOL   dsn;                    /* FALSE if DSN not to be used */
+extern char  *dsn_envid;              /* DSN envid string */
+extern int    dsn_ret;                /* DSN ret type*/
+extern pcre  *regex_DSN;              /* For recognizing DSN settings */
+extern BOOL   smtp_use_dsn;           /* Global for passed connections */
+#endif
+
 extern BOOL   accept_8bitmime;        /* Allow *BITMIME incoming */
 extern BOOL   accept_call_bombout;    /* Flag for crashing log */
 extern int    accept_timeout;         /* For non-SMTP acceptance */
@@ -37,6 +45,7 @@ extern char  *big_buffer;             /* Used for various temp things */
 extern int    big_buffer_size;        /* Current size (can expand) */
 extern int    body_linecount;         /* Line count in body */
 
+extern BOOL   check_dns_names;        /* True for preliminary syntax check */
 extern BOOL   check_relay;            /* True if relay checking required */
 extern int    check_log_inodes;       /* Minimum for message acceptance */
 extern int    check_log_space;        /* Minimum for message acceptance */
@@ -70,6 +79,7 @@ extern char  *deliver_out_buffer;
 extern int    deliver_datafile;       /* FD for data part of message */
 extern char  *deliver_domain;         /* The local domain for delivery */
 extern char  *deliver_domain_orig;    /* The original local domain for delivery */
+extern BOOL   deliver_firsttime;      /* True for first delivery attempt */
 extern BOOL   deliver_force;          /* TRUE if delivery was forced */
 extern BOOL   deliver_freeze;         /* TRUE if delivery is frozen */
 extern int    deliver_frozen_at;      /* Time of freezing */
@@ -118,6 +128,7 @@ extern int    expand_nmax;            /* Max numerical value */
 extern char  *expand_nstring[];       /* Numbered strings */
 extern BOOL   expand_string_forcedfail; /* TRUE if failure was "expected" */
 extern char  *expand_string_message;  /* Error info for failing expansion */
+extern BOOL   extract_addresses_remove_arguments; /* Controls -t behaviour */
 
 extern char  *filter_fmsg;            /* Message for freeze or fail */
 extern int    filter_n[FILTER_VARIABLE_COUNT]; /* filter variables */
@@ -143,6 +154,10 @@ extern BOOL   headers_check_syntax;   /* Check From, To, etc. */
 extern BOOL   headers_checks_fail;    /* If FALSE, just warn */
 extern BOOL   headers_sender_verify;  /* TRUE to verify sender from headers */
 extern BOOL   headers_sender_verify_errmsg; /* ditto, only if errmsg */
+extern char  *helo_accept_junk_hosts; /* Allowed to use junk args */
+extern host_item *helo_accept_junk_hostlist;
+extern char  *helo_accept_junk_nets;
+extern ip_net_item *helo_accept_junk_netlist;
 extern BOOL   helo_strict_syntax;     /* Allows _ in name if false */
 extern BOOL   helo_verify;            /* Check HELO argument */
 extern char  *helo_verify_except_hosts; /* Except for ... */
@@ -155,6 +170,9 @@ extern BOOL   host_checking;          /* Set when checking a host */
 extern char  *host_findbyaddr_msg;    /* Text for why it failed */
 extern char  *host_lookup_nets;       /* Nets for which IP addresses are always looked up */
 extern ip_net_item *host_lookup_netlist;
+extern int    host_number;            /* For sharing spools */
+extern char  *host_number_string;     /* For expanding */
+extern char  *hosts_treat_as_local;   /* For routing */
 
 extern BOOL   ignore_errmsg_errors;   /* As it says */
 extern int    ignore_errmsg_errors_after; /* Or keep them for this time. */
@@ -254,6 +272,7 @@ extern char  *prohibition_reason;     /* Set while expanding message */
 extern char  *qualify_domain_recipient; /* Domain to qualify recipients with */
 extern char  *qualify_domain_sender;  /* Domain to qualify senders with */
 extern BOOL   queue_2stage;           /* Run queue in 2-stage manner */
+extern BOOL   queue_list_requires_admin; /* TRUE if -bp requires admin */
 extern BOOL   queue_run_force;        /* TRUE to force during queue run */
 extern BOOL   queue_run_local;        /* Local deliveries only in queue run */
 extern BOOL   queue_running;          /* TRUE when running the queue */
@@ -272,6 +291,7 @@ extern char  *queue_smtp_except;      /* Exception list */
 extern re_block *re_dns_again_means_nonexist;
 extern re_block *re_hold_domains;
 extern re_block *re_hold_domains_except;
+extern re_block *re_hosts_treat_as_local;
 extern re_block *re_local_domains;
 extern re_block *re_percent_hack_domains;
 extern re_block *re_queue_remote_except;
@@ -322,7 +342,7 @@ extern ip_net_item *receiver_verify_except_netlist;
 extern char  *receiver_verify_senders;/* Verify only from these senders */
 extern char  *receiver_verify_senders_except; /* Don't verify from these senders */
 extern int    recipients_count;       /* Number of recipients */
-extern char **recipients_list;        /* List of recipient addresses */
+extern recipient_item *recipients_list; /* List of recipient addresses */
 extern int    recipients_list_max;    /* Maximum number fitting in list */
 extern int    recipients_max;         /* Max permitted */
 extern int    recipients_max_reject;  /* If TRUE, reject whole message */
@@ -350,6 +370,7 @@ extern host_item *rfc1413_except_hostlist; /* Munged data */
 extern char  *rfc1413_except_nets;    /* RFC 1413 exceptions */
 extern ip_net_item *rfc1413_except_netlist; /* Munged data */
 extern int    rfc1413_query_timeout;  /* Timeout on RFC 1413 calls */
+extern BOOL   rfc821_domains;         /* If set, syntax is 821, not 822 */
 extern uid_t  root_uid;               /* The uid for root */
 extern char  *route_option;           /* Option string for some routers */
 
@@ -366,6 +387,7 @@ extern char  *sender_accept_recipients; /* Ditto, reject RCPT TO */
 extern char  *sender_address;         /* Envelope sender */
 extern BOOL   sender_address_forced;  /* Set by -f */
 extern char  *sender_address_relay;   /* Only non-nearby relay from these addresses */
+extern BOOL   sender_address_rewritten; /* True if rewritten by verify */
 extern char  *sender_fullhost;        /* Sender host name + address */
 extern char  *sender_helo_name;       /* Host name from HELO/EHLO */
 extern char  *sender_host_accept;     /* Accept only from these hosts */
@@ -374,6 +396,7 @@ extern char  *sender_host_accept_relay; /* Accept non-nearby relay only from the
 extern host_item *sender_host_accept_relay_hosts; /* Munged data */
 extern char  *sender_host_address;    /* IP address of sender, as chars */
 extern char  *sender_host_name;       /* Host name from lookup, or "?" */
+extern BOOL   sender_host_notsocket;  /* Set for -bs and -bS */
 extern char  *sender_host_reject;     /* Reject from these hosts */
 extern char  *sender_host_reject_except; /* Except these hosts */
 extern host_item *sender_host_reject_hosts; /* Munged data */
@@ -382,6 +405,8 @@ extern char  *sender_host_reject_recipients; /* Reject RCPT from these hosts */
 extern host_item *sender_host_reject_recipients_hosts; /* Munged data */
 extern char  *sender_host_reject_relay; /* Reject non-nearby relay from these hosts */
 extern host_item *sender_host_reject_relay_hosts; /* Munged data */
+extern char  *sender_host_reject_relay_except; /* Exceptions */
+extern host_item *sender_host_reject_relay_except_hosts; /* Munged data */
 extern BOOL   sender_host_unknown;    /* TRUE for -bs and -bS except inetd */
 extern char  *sender_ident;           /* Sender identity via RFC 1413 */
 extern BOOL   sender_local;           /* TRUE for local senders */
@@ -397,6 +422,8 @@ extern char  *sender_net_reject_recipients; /* Reject RCPT from these nets */
 extern ip_net_item *sender_net_reject_recipients_nets; /* Munged data */
 extern char  *sender_net_reject_relay; /* Reject non-nearby relay from these nets */
 extern ip_net_item *sender_net_reject_relay_nets; /* Munged data */
+extern char  *sender_net_reject_relay_except; /* Exceptions */
+extern ip_net_item *sender_net_reject_relay_except_nets; /* Munged data */
 extern char  *sender_rcvhost;         /* Host data for Received: */
 extern char  *sender_reject;          /* Spam filter list */
 extern char  *sender_reject_except;   /* Spam filter exception list */
@@ -415,15 +442,18 @@ extern ip_net_item *sender_verify_except_netlist; /* Munged data */
 extern BOOL   sender_verify_fixup;    /* Fix broken senders from headers */
 extern BOOL   sender_verify_log_details; /* For debugging */
 extern BOOL   sender_verify_reject;   /* Reject if verify fails */
+extern int    sender_verify_max_retry_rate; /* Max retry rate */
 extern volatile BOOL sigalrm_seen;    /* Flag for sigalrm_handler */
 extern char **sighup_argv;            /* Args for re-execing after SIGHUP */
 extern int    smtp_accept_count;      /* Count of connections */
 extern int    smtp_accept_max;        /* Max SMTP connections */
 extern int    smtp_accept_queue;      /* Queue after so many connections */
+extern int    smtp_accept_queue_per_connection; /* Queue after so many msgs */
 extern int    smtp_accept_reserve;    /* Reserve these SMTP connections */
 extern char  *smtp_banner;            /* Banner string (to be expanded) */
 extern BOOL   smtp_batched_input;     /* TRUE if SMTP batch (no interaction) */
 extern int    smtp_connect_backlog;   /* Max backlog permitted */
+extern char  *smtp_etrn_command;      /* Command to run */
 extern host_item *smtp_etrn_hostlist; /* Munged data */
 extern char  *smtp_etrn_hosts;        /* Accept ETRN only from these hosts */
 extern ip_net_item *smtp_etrn_netlist; /* Munged data */
@@ -436,8 +466,10 @@ extern char  *smtp_expn_nets;         /* Networks for EXPN command */
 extern FILE  *smtp_in;                /* Incoming SMTP input file */
 extern BOOL   smtp_input;             /* TRUE if input is via SMTP */
 extern int    smtp_load_reserve;      /* Only from reserved if load > this */
+extern BOOL   smtp_log_connections;   /* Verbose connection logging */
 extern FILE  *smtp_out;               /* Incoming SMTP output file */
 extern int    smtp_port;              /* Alternate SMTP port number */
+extern BOOL   smtp_reading_data;      /* True if "." needed before response */
 extern int    smtp_receive_timeout;   /* Applies to each received line */
 extern host_item *smtp_reserve_hostlist; /* Munged data */
 extern char  *smtp_reserve_hosts;     /* Hosts for reserved slots */
@@ -448,6 +480,8 @@ extern BOOL   smtp_verify;            /* TRUE if VRFY permitted */
 extern BOOL   split_spool_directory;  /* TRUE to use multiple subdirs */
 extern char  *spool_directory;        /* Name of spool directory */
 extern char  *stderr_filename;        /* File for use with -df */
+extern int    store_max_malloc;       /* For keeping total */
+extern int    store_pool;             /* Current pool number */
 extern BOOL   strip_excess_angle_brackets; /* Surrounding route-addrs */
 extern BOOL   strip_trailing_dot;     /* Remove dots at ends of domains */
 
@@ -474,7 +508,7 @@ extern transport_info transports_available[]; /* Vector of available transports
 extern transport_instance *transports;  /* Chain of instantiated transports */
 extern transport_instance transport_defaults; /* Default values */
 
-extern int    transport_chunk_timeout;/* Set to time out individual writes */
+extern int    transport_write_timeout;/* Set to time out individual writes */
 
 extern tree_node *tree_duplicates;    /* Tree of duplicate addresses */
 extern tree_node *tree_nonrecipients; /* Tree of nonrecipient addresses */
@@ -496,6 +530,7 @@ extern char  *warnmsg_delay;          /* String form of delay time */
 extern char  *warnmsg_file;           /* Template for warning messages */
 extern char  *warnmsg_recipients;     /* Recipients of warning message */
 
+extern char  *verify_forced_errmsg;   /* Message from :defer: or :fail: */
 extern BOOL   verify_only;            /* TRUE for -bv */
 extern char  *version_copyright;      /* Copyright notice */
 extern char  *version_date;           /* Date of compilation */
