@@ -16,7 +16,8 @@
 /* header_last points to the last header during message reception and delivery;
 otherwise it is NULL. We add new headers only when header_last is not NULL.
 The function may get called sometimes when it is NULL (e.g. during address
-verification where rewriting options exist).
+verification where rewriting options exist). Headers are always added in the
+main storage pool.
 
 Arguments:
   type      Exim header type character
@@ -29,24 +30,27 @@ Returns:    nothing
 void
 header_add(int type, char *format, ...)
 {
-int slen;
+int pool_reset = store_pool;
 header_line *new;
 va_list ap;
+char buffer[8192];
 
 if (header_last == NULL) return;
 
 va_start(ap, format);
-if (!string_vformat(big_buffer, BIG_BUFFER_SIZE, format, ap))
-  log_write(0, LOG_PANIC_DIE, "string too long in header_add");
+if (!string_vformat(buffer, sizeof(buffer), format, ap))
+  log_write(0, LOG_PANIC_DIE, "string too long in header_add: %.100s ...",
+    buffer);
 va_end(ap);
-slen = (int)strlen(big_buffer);
 
-new = (header_line *)store_malloc(sizeof(header_line) + slen);
-strcpy(new->text, big_buffer);
+store_pool = POOL_MAIN;
+new = store_get(sizeof(header_line));
+new->text = string_copy(buffer);
+store_pool = pool_reset;
+
+new->slen = (int)strlen(buffer);
 new->next = NULL;
-new->prev = header_last;
 new->type = type;
-new->slen = slen;
 
 header_last->next = new;
 header_last = new;
