@@ -1,3 +1,4 @@
+
 /*************************************************
 *     Exim - an Internet mail transport agent    *
 *************************************************/
@@ -62,6 +63,49 @@ return (unsigned int)(seed >> 16) % limit;
 
 
 
+/*************************************************
+*       Build chain of host items from list      *
+*************************************************/
+
+/* This function builds a chain of host items from a textual list of host
+names. It does not do any lookups.
+
+Arguments:
+  anchor      anchor for the chain
+  list        text list
+
+Returns:      nothing
+*/
+
+void
+host_build_hostlist(host_item **anchor, char *list)
+{
+char *name;
+char buffer[1024];
+
+if (list == NULL) return;
+
+for (name = string_nextinlist(&list, ':', buffer, sizeof(buffer));
+     name != NULL;
+     name = string_nextinlist(&list, ':', buffer, sizeof(buffer)))
+  {
+  host_item *h = store_get(sizeof(host_item));
+  h->next = NULL;
+  h->name = string_copy(name);
+  h->address = NULL;
+  h->mx = -1;
+  h->status = hstatus_unknown;
+  h->why = hwhy_unknown;
+  h->last_try = 0;
+  *anchor = h;
+  anchor = &(h->next);
+  }
+}
+
+
+
+
+
 
 #ifndef STAND_ALONE    /* Omit when standalone testing */
 
@@ -85,6 +129,9 @@ syntax is a bit different. This value also includes the RFC 1413 identity.
 There wouldn't be two different variables if I had got all this right in the
 first place.
 
+Because this data may survive over more than one incoming SMTP message, it has
+to be copied into malloc store.
+
 Arguments:  none
 Returns:    nothing
 */
@@ -92,26 +139,27 @@ Returns:    nothing
 void
 host_build_sender_fullhost(void)
 {
+char *fullhost, *rcvhost;
+
 if (sender_host_address == NULL) return;
 
 if (sender_fullhost != NULL) store_free(sender_fullhost);
-if (sender_rcvhost != NULL) store_free(sender_rcvhost);
 
 if (sender_host_name == NULL)
   {
   if (sender_helo_name == NULL)
     {
-    sender_fullhost = string_sprintf("[%s]", sender_host_address);
-    sender_rcvhost = (sender_ident == NULL)?
+    fullhost = string_sprintf("[%s]", sender_host_address);
+    rcvhost = (sender_ident == NULL)?
       string_sprintf("[%s]", sender_host_address) :
       string_sprintf("[%s] (ident=%s)", sender_host_address, sender_ident);
     }
   else
     {
-    sender_fullhost = string_sprintf("(%s) [%s]", sender_helo_name,
+    fullhost = string_sprintf("(%s) [%s]", sender_helo_name,
       sender_host_address);
 
-    sender_rcvhost = (sender_ident == NULL)?
+    rcvhost = (sender_ident == NULL)?
       string_sprintf("[%s] (helo=%s)", sender_host_address, sender_helo_name) :
       string_sprintf("[%s] (helo=%s ident=%s)", sender_host_address,
         sender_helo_name, sender_ident);
@@ -126,18 +174,18 @@ else
          sender_helo_name[(len=(int)strlen(sender_helo_name))-1] == ']' &&
          strncmpic(sender_helo_name+1, sender_host_address, len - 2) == 0))
     {
-    sender_fullhost = string_sprintf("%s [%s]", sender_host_name,
+    fullhost = string_sprintf("%s [%s]", sender_host_name,
       sender_host_address);
-    sender_rcvhost = (sender_ident == NULL)?
+    rcvhost = (sender_ident == NULL)?
       string_sprintf("%s ([%s])", sender_host_name, sender_host_address) :
       string_sprintf("%s ([%s] ident=%s)", sender_host_name,
         sender_host_address, sender_ident);
     }
   else
     {
-    sender_fullhost = string_sprintf("%s (%s) [%s]", sender_host_name,
+    fullhost = string_sprintf("%s (%s) [%s]", sender_host_name,
       sender_helo_name, sender_host_address);
-    sender_rcvhost = (sender_ident == NULL)?
+    rcvhost = (sender_ident == NULL)?
       string_sprintf("%s ([%s] helo=%s)", sender_host_name,
         sender_host_address, sender_helo_name) :
       string_sprintf("%s\n\t([%s] helo=%s ident=%s)", sender_host_name,
@@ -145,6 +193,12 @@ else
     }
   }
 
+sender_fullhost =
+  store_malloc((int)strlen(fullhost) + (int)strlen(rcvhost) + 2);
+sender_rcvhost = sender_fullhost + (int)strlen(fullhost) + 1;
+strcpy(sender_fullhost, fullhost);
+strcpy(sender_rcvhost, rcvhost);
+
 DEBUG(5) debug_printf("sender_fullhost = %s\n", sender_fullhost);
 DEBUG(5) debug_printf("sender_rcvhost = %s\n", sender_rcvhost);
 }
@@ -324,17 +378,32 @@ return FAIL;   /* paranoia */
 
 /* This function finds the addresses of all the running interfaces on the
 machine. A chain of blocks containing the textual form of the addresses is
-returned. Solaris 2 has the SIOGIFNUM call to get the number of interfaces, but
-other OS (including Solaris 1) appear not to. So just screw in a largeish fixed
-number. Unfortunately, the www addressing scheme means that some hosts have a
-very large number of virtual interfaces. Such hosts are recommended to set
-local_interfaces to avoid problems with this (see host_find_interfaces below).
+returned.
+
+Problems:
+
+  (1) Solaris 2 has the SIOGIFNUM call to get the number of
+  interfaces, but other OS (including Solaris 1) appear not to. So just screw
+  in a largeish fixed number. Unfortunately, the www addressing scheme means
+  that some hosts have a very large number of virtual interfaces. Such hosts
+  are recommended to set local_interfaces to avoid problems with this (see
+  host_find_interfaces below).
+
+  (2) If the standard code is run on IRIX, it does not return any alias
+  interfaces. There is special purpose code for that operating system, which
+  uses the sysctl() function.
 
 Arguments:    none
 Returns:      a chain of ip_address_items, each pointing to a textual
               version of an IP address
 */
 
+
+#ifndef SYSCTL_IP_INTERFACES
+
+/* This is the form of the function that uses ioctl() in a way that works on
+most versions of Unix. */
+
 #define MAXINTERFACES 250
 
 static ip_address_item *
@@ -408,7 +477,7 @@ for (cp = buf; cp < buf + ifc.ifc_len; cp +=
       errno, strerror(errno));
 
   /* Create a data block for the address, fill in the data, and put it on the
-  chain. */
+  chain. This data has to survive for ever, so use malloc. */
 
   next = store_malloc(sizeof(ip_address_item));
   next->next = NULL;
@@ -420,7 +489,7 @@ for (cp = buf; cp < buf + ifc.ifc_len; cp +=
     last = next;
     }
 
-  DEBUG(8) debug_printf("Actual local interface address is %s\n",
+  DEBUG(9) debug_printf("Actual local interface address is %s\n",
     last->address);
   }
 
@@ -431,6 +500,115 @@ return yield;
 }
 
 
+/* This is the special form of the function using sysctl() which is the only
+form that returns all the aliases on IRIX systems. This code has its origins
+in a sample program that came from within SGI. */
+
+#else
+
+#include <sys/sysctl.h>
+#include <net/if_dl.h>
+#include <net/if_types.h>
+#include <net/soioctl.h>
+#include <net/route.h>
+
+
+#define ROUNDUP(a) ((a) > 0 ? (1 + (((a) - 1) | (sizeof(__uint64_t) -1))) \
+                    : sizeof(__uint64_t))
+#ifdef _HAVE_SA_LEN
+#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))
+#else
+#define ADVANCE(x, n) (x += ROUNDUP(_FAKE_SA_LEN_DST(n)))
+#endif
+
+
+static ip_address_item *
+host_find_running_interfaces(void)
+{
+ip_address_item *yield = NULL;
+ip_address_item *last = NULL;
+ip_address_item *next;
+
+size_t needed;
+int mib[6];
+char *buf, *nextaddr, *lim;
+register struct if_msghdr *ifm;
+
+mib[0] = CTL_NET;
+mib[1] = PF_ROUTE;
+mib[2] = 0;
+mib[3] = 0;
+mib[4] = NET_RT_IFLIST;
+mib[5] = 0;
+
+/* Get an estimate of the amount of store needed, then get the store and
+get the data into it. Any error causes a panic death. */
+
+if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
+  log_write(0, LOG_PANIC_DIE, "iflist-sysctl-estimate failed: %s",
+    strerror(errno));
+
+buf = store_get(needed);
+
+if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0)
+  log_write(0, LOG_PANIC_DIE, "sysctl of ifnet list failed: %s",
+    strerror(errno));
+
+/* Now fish out the data for each interface */
+
+lim  = buf + needed;
+for (nextaddr = buf; nextaddr < lim; nextaddr += ifm->ifm_msglen)
+  {
+  register struct sockaddr_dl *sdl;
+
+  ifm = (struct if_msghdr *)nextaddr;
+  sdl = (struct sockaddr_dl *)(ifm +1);
+
+  if (ifm->ifm_type != RTM_IFINFO)
+    {
+    struct ifa_msghdr *ifam = (struct ifa_msghdr *)ifm;
+    struct sockaddr_in *mask = NULL, *addr = NULL;
+
+    if ((ifam->ifam_addrs & RTA_NETMASK) != 0)
+      mask = (struct sockaddr_in *)(ifam + 1);
+
+    if ((ifam->ifam_addrs & RTA_IFA) != 0)
+      {
+      char *cp = (char *)mask;
+      struct sockaddr *sa = (struct sockaddr *)mask;
+      ADVANCE(cp, sa);
+      addr = (struct sockaddr_in *)cp;
+      }
+
+    /* Create a data block for the address, fill in the data, and put it on
+    the chain. This data has to survive for ever, so use malloc. */
+
+    if (addr != NULL)
+      {
+      next = store_malloc(sizeof(ip_address_item));
+      next->next = NULL;
+
+      (void)host_ntoa(-1, addr, next->address);
+
+      if (yield == NULL) yield = last = next; else
+        {
+        last->next = next;
+        last = next;
+        }
+
+      DEBUG(8) debug_printf("Actual local interface address is %s\n",
+        last->address);
+      }
+    }
+  }
+
+return yield;
+}
+
+#endif
+
+
+
 
 /*************************************************
 *         Find addresses on local interfaces     *
@@ -467,6 +645,8 @@ for (s = string_nextinlist(&listptr, ':', buffer, sizeof(buffer));
     log_write(0, LOG_MAIN|LOG_PANIC_DIE, "Malformed IP address \"%s\" in "
       "local_interfaces", s);
 
+  /* Data has to survive for ever, so use malloc. */
+
   next = store_malloc(sizeof(ip_address_item));
   next->next = NULL;
   strcpy(next->address, s);
@@ -739,6 +919,10 @@ have greater or equal MX values) from the chain. Note: marking them as unusable
 is NOT the right thing to do because it causes the hosts not to be used for
 other domains, for which they may well be correct.
 
+There is also a list of "pseudo-local" host names which are checked against the
+host names. Any match causes that host item to be treated the same as one which
+matches a local IP address.
+
 If the very first host is a local host, then all MX records had a precedence
 greater than or equal to that of the local host. Either there's a problem in
 the DNS, or an apparently remote name turned out to be an abbreviation for the
@@ -774,37 +958,19 @@ if (local_interface_data == NULL) local_interface_data = host_find_interfaces();
 
 for (h = host; h != last->next; h = h->next)
   {
+  #ifndef STAND_ALONE
+  if (hosts_treat_as_local != NULL &&
+      match_isinlist(h->name, hosts_treat_as_local, &re_hosts_treat_as_local,
+      TRUE))
+    goto FOUND_LOCAL;
+  #endif
+
   if (h->address != NULL)
     {
     ip_address_item *ip;
     for (ip = local_interface_data; ip != NULL; ip = ip->next)
-      {
-      if (strcmp(h->address, ip->address) == 0)
-        {
-        if (prev == NULL)
-          {
-          DEBUG(8) debug_printf((h->mx >= 0)?
-            "local host has lowest MX\n" :
-            "local host found for non-MX address\n");
-          return HOST_FOUND_LOCAL;
-          }
-
-        DEBUG(8)
-          {
-          debug_printf("local host in list - removed hosts:\n");
-          for (h = prev->next; h != last->next; h = h->next)
-            debug_printf("  %s %s %d\n", h->name, h->address, h->mx);
-          }
-
-        if (removed != NULL) *removed = TRUE;
-        prev->next = NULL;
-        return yield;
-        }
-      }
-
-    /* At least one remote address has been found */
-
-    yield = HOST_FOUND;
+      if (strcmp(h->address, ip->address) == 0) goto FOUND_LOCAL;
+    yield = HOST_FOUND;  /* At least one remote address has been found */
     }
 
   /* Update prev to point to the last host item before any that have
@@ -813,6 +979,31 @@ for (h = host; h != last->next; h = h->next)
   if (h->next == NULL || h->next->mx != h->mx) prev = h;
   }
 
+return yield;  /* No local hosts found: return HOST_FOUND or HOST_FIND_FAILED */
+
+
+/* A host whose IP address matches a local IP address, or whose name matches
+something in hosts_treat_as_local has been found. */
+
+FOUND_LOCAL:
+
+if (prev == NULL)
+  {
+  DEBUG(8) debug_printf((h->mx >= 0)?
+    "local host has lowest MX\n" :
+    "local host found for non-MX address\n");
+  return HOST_FOUND_LOCAL;
+  }
+
+DEBUG(8)
+  {
+  debug_printf("local host in list - removed hosts:\n");
+  for (h = prev->next; h != last->next; h = h->next)
+    debug_printf("  %s %s %d\n", h->name, h->address, h->mx);
+  }
+
+if (removed != NULL) *removed = TRUE;
+prev->next = NULL;
 return yield;
 }
 
@@ -823,21 +1014,26 @@ return yield;
 *        Find host name given an IP address      *
 *************************************************/
 
-/* For the moment, just use gethostbyaddr, and pass back a
-single string. Lower case it.
+/* For the moment, just use gethostbyaddr, and pass back a single string,
+either in permanent or pooled store.
 
-Argument: an IPv4 address in dotted-quad form OR
-          an IPv6 address in colon-separated form
-Returns:  the name of the host or NULL if not found
+Arguments:
+  address     an IPv4 address in dotted-quad form OR
+              an IPv6 address in colon-separated form
+  use_malloc  TRUE if permanent store is required
+
+Returns:      the name of the host or NULL if not found
 
 In addition, the variable host_findbyaddr_msg is set to an empty string on
 sucess, or to a reason for the failure otherwise, in a form suitable for
 tagging onto an error message. */
 
 char *
-host_find_byaddr(char *address)
+host_find_byaddr(char *address, BOOL use_malloc)
 {
 char *s, *t, *yield;
+host_item h, *hh;
+int len;
 struct hostent *hosts;
 struct in_addr addr;
 
@@ -873,19 +1069,49 @@ hosts = gethostbyaddr((char *)(&addr), sizeof(addr), AF_INET);
 
 if (hosts == NULL)
   {
-  HDEBUG(2) debug_printf("host lookup failed\n");
-  host_findbyaddr_msg = " (failed to find sending host name from IP address)";
+  HDEBUG(2) debug_printf("IP address lookup failed\n");
+  host_findbyaddr_msg = " (failed to find host name from IP address)";
   return NULL;
   }
 
-/* Copy and lowercase the name */
+/* Copy and lowercase the name. We do this before checking the forward lookup
+because gethostbyname() and gethostbyaddr() share static storage in some
+systems, so we need to take a copy of the name first. */
 
 s = (char *)hosts->h_name;
-t = yield = store_malloc((int)strlen(s) + 1);
+len = (int)strlen(s) + 1;
+t = yield = use_malloc? store_malloc(len) : store_get(len);
 while (*s != 0) *t++ = tolower(*s++);
 *t = 0;
-HDEBUG(2) debug_printf("host lookup yielded %s\n", yield);
-return yield;
+
+HDEBUG(2) debug_printf("IP address lookup yielded %s\n", yield);
+
+/* We need to verify that a forward lookup on the name we found does indeed
+correspond to the address. This is for security: in principle a malefactor who
+happened to own a reverse zone could set it to point to any names at all. */
+
+h.name = yield;
+h.address = NULL;
+h.next = NULL;
+
+if (host_find_byname(&h, NULL, FALSE) == HOST_FIND_FAILED)
+  {
+  HDEBUG(2) debug_printf("no IP address found for %s\n", yield);
+  host_findbyaddr_msg = string_sprintf(
+    " (no IP address found for %s, the host name obtained from %s)",
+    yield, address);
+  return NULL;
+  }
+
+else for (hh = &h; hh != NULL; hh = hh->next)
+  if (strcmp(hh->address, address) == 0) return yield;
+
+HDEBUG(2) debug_printf("%s does not match any IP address for %s\n",
+  address, yield);
+host_findbyaddr_msg =
+  string_sprintf(" (%s does not match any IP address for %s)", address, yield);
+
+return NULL;
 }
 
 
@@ -898,7 +1124,8 @@ return yield;
 /* The input is a host_item structure with the name filled in and the address
 field set to NULL. We use gethostbyname(). Of course, gethostbyname() may use
 the DNS, but it doesn't do MX processing. If more than one address is given,
-chain on additional host items, with other relevant fields copied.
+chain on additional host items, with other relevant fields copied. The type of
+store required for them is set by use_malloc.
 
 The lookup may result in a change of name. For compatibility with the dns
 lookup, return this via fully_qualified_name as well as updating the host item.
@@ -911,6 +1138,7 @@ Arguments:
                          host items to be chained on.
   fully_qualified_name   if not NULL, set to point to host name for
                          compatibility with host_find_bydns
+  use_malloc             TRUE if additional items must be in permanent store
 
 Returns:                 HOST_FIND_FAILED  Failed to find the host or domain
                          HOST_FOUND        Host found
@@ -918,7 +1146,7 @@ Returns:                 HOST_FIND_FAILED  Failed to find the host or domain
 */
 
 int
-host_find_byname(host_item *host, char **fully_qualified_name)
+host_find_byname(host_item *host, char **fully_qualified_name, BOOL use_malloc)
 {
 int af = AF_INET6;
 int i, yield;
@@ -953,7 +1181,8 @@ for (i = 1; i <= times; af = AF_INET, i++)
   the fully_qualified_name pointer. */
 
   if (strcmp(host->name, hostdata->h_name) != 0)
-    host->name = string_copy((char *)hostdata->h_name);
+    host->name = use_malloc? string_copy_malloc((char *)hostdata->h_name) :
+      string_copy((char *)hostdata->h_name);
   if (fully_qualified_name != NULL) *fully_qualified_name = host->name;
 
   /* Get the list of addresses. IPv4 and IPv6 addresses can be distinguished
@@ -979,7 +1208,8 @@ for (i = 1; i <= times; af = AF_INET, i++)
 
   while (*addrlist != NULL)
     {
-    host_item *next = store_malloc(sizeof(host_item));
+    host_item *next = use_malloc?
+      store_malloc(sizeof(host_item)) : store_get(sizeof(host_item));
     next->name = host->name;
     next->ident_string = host->ident_string;
     next->address = host_ntoa(ipv4_addr? AF_INET:AF_INET6, *addrlist++, NULL);
@@ -1104,7 +1334,7 @@ for (; i >= 0; type = T_A, i--)
       else
         {
         int new_sort_key = host->mx * 100 + random_number(100);
-        host_item *next = store_malloc(sizeof(host_item));
+        host_item *next = store_get(sizeof(host_item));
 
         /* New address goes first: insert the new block after the first one
         (so as not to disturb the original pointer) but put the new address
@@ -1318,7 +1548,7 @@ for (rr = dns_next_rr(RESET_ANSWERS); rr != NULL; rr = dns_next_rr(RESET_NEXT))
     {
     int sort_key = precedence * 100 + random_number(100);
 
-    host_item *next = store_malloc(sizeof(host_item));
+    host_item *next = store_get(sizeof(host_item));
     next->name = string_copy(data);
     next->address = NULL;
     next->mx = precedence;
@@ -1416,10 +1646,7 @@ for (rr = dns_next_rr(RESET_ADDITIONAL); rr != NULL;
     do
       {
       if (hh->address != NULL && strcmp(address, hh->address) == 0)
-        {
-        store_free(address);
         goto dns_next_rr;         /* Need goto to escape from inner loop */
-        }
       thishostlast = hh;
       hh = hh->next;
       }
@@ -1432,7 +1659,7 @@ for (rr = dns_next_rr(RESET_ADDITIONAL); rr != NULL;
     sorting key and do that. */
 
     new_sort_key = h->mx * 100 + random_number(100);
-    hh = store_malloc(sizeof(host_item));
+    hh = store_get(sizeof(host_item));
 
     /* New address goes first: insert the new block after the first one
     (so as not to disturb the original pointer) but put the new address
@@ -1557,9 +1784,7 @@ address in the blocking domain.
 Arguments:
   address      the string version of the address
   domain       the rbl domain to use
-  buffer       somewhere to put an error message if not NULL;
-               if no error, set to the null string
-  buflen       size of buffer
+  buffer       somewhere to point an error message
 
 Returns:    DNS_SUCCEED   successful lookup (i.e. the address is blocked)
             DNS_NOMATCH   name not found, or no data found for the given type
@@ -1568,7 +1793,7 @@ Returns:    DNS_SUCCEED   successful lookup (i.e. the address is blocked)
 */
 
 int
-host_check_rbl(char *address, char *domain, char *buffer, int buflen)
+host_check_rbl(char *address, char *domain, char **bufptr)
 {
 int bin[4];
 int rc;
@@ -1621,31 +1846,27 @@ strcpy(t, domain);
 HDEBUG(8) debug_printf("RBL lookup: %s\n", name);
 rc = dns_basic_lookup(name, type);
 
-/* If a buffer exists, set its content to the null string, or to a TXT record's
-data if present after success. */
+/* If a buffer pointer exists, set it to point to the null string, or to a TXT
+record's data if present after success. */
 
-if (buffer != NULL)
+if (rc == DNS_SUCCEED)
   {
-  buffer[0] = 0;
-  if (rc == DNS_SUCCEED)
+  *bufptr = "";
+  if (dns_basic_lookup(name, T_TXT) == DNS_SUCCEED)
     {
-    if (dns_basic_lookup(name, T_TXT) == DNS_SUCCEED)
+    dns_record *rr;
+
+    for (rr = dns_next_rr(RESET_ANSWERS); rr != NULL; rr = dns_next_rr(RESET_NEXT))
+      if (rr->type == T_TXT) break;
+
+    if (rr != NULL)
       {
-      dns_record *rr;
-
-      for (rr = dns_next_rr(RESET_ANSWERS); rr != NULL; rr = dns_next_rr(RESET_NEXT))
-        if (rr->type == T_TXT) break;
-
-      if (rr != NULL)
-        {
-        int len = (rr->data)[0];
-        if (len > buflen - 1) len = buflen - 1;
-        strncpy(buffer, (const char *)(rr->data + 1), len);
-        buffer[len] = 0;
-        }
-
-      DEBUG(8) debug_printf("RBL TXT: %s\n", buffer);
+      int len = (rr->data)[0];
+      if (len > 127) len = 127;
+      *bufptr = string_sprintf("%.*s", len, (const char *)(rr->data + 1));
       }
+
+    DEBUG(8) debug_printf("RBL TXT: %s\n", *bufptr);
     }
   }
 
@@ -1674,14 +1895,18 @@ BOOL search_parents = FALSE;
 char buffer[256];
 
 primary_hostname = "";
-debug_level = 8;
+debug_level = 9;
 debug_file = stdout;
 
+printf("Exim stand-alone host functions test\n");
+
 host_find_running_interfaces();
+debug_level = 8;
 
 if (argc > 1) primary_hostname = argv[1];
 
 /* So that debug level changes can be done first */
+
 dns_init(qualify_single, search_parents);
 
 printf("Testing host lookup\n");
@@ -1697,8 +1922,8 @@ while (fgets(buffer, 256, stdin) != NULL)
 
   if (strcmp(buffer, "q") == 0) break;
 
-  if (strcmp(buffer, "byname") ==0) byname = TRUE;
-  else if (strcmp(buffer, "no_byname") ==0) byname = FALSE;
+  if (strcmp(buffer, "byname") == 0) byname = TRUE;
+  else if (strcmp(buffer, "no_byname") == 0) byname = FALSE;
   else if (strcmp(buffer, "a_only") == 0) a_only = TRUE;
   else if (strcmp(buffer, "no_a_only") == 0) a_only = FALSE;
   else if (strcmp(buffer, "mx_only") == 0) mx_only = TRUE;
@@ -1730,9 +1955,9 @@ while (fgets(buffer, 256, stdin) != NULL)
     h.mx = -1;
     h.status = hstatus_unknown;
     h.why = hwhy_unknown;
-    h.address = "";
+    h.address = NULL;
 
-    rc = byname? host_find_byname(&h, &fully_qualified_name) :
+    rc = byname? host_find_byname(&h, &fully_qualified_name, FALSE) :
       host_find_bydns(&h, mx_only, a_only, qualify_single, search_parents,
         &fully_qualified_name, NULL);
 
@@ -1778,7 +2003,9 @@ while (fgets(buffer, 256, stdin) != NULL)
   while (len > 0 && isspace(buffer[len-1])) len--;
   buffer[len] = 0;
   if (strcmp(buffer, "q") == 0) break;
-  yield = host_find_byaddr(buffer);      /* Debug causes printing */
+  yield = host_find_byaddr(buffer, FALSE);      /* Debug causes printing */
+  if (yield == NULL)
+    printf("Lookup failed:%s\n", host_findbyaddr_msg);
   printf("\n> ");
   }
 
@@ -1788,12 +2015,12 @@ printf("Testing host_check_rbl\n");
 printf("> ");
 while (fgets(buffer, 256, stdin) != NULL)
   {
-  int yield;
+  char *bufptr;
   int len = (int)strlen(buffer);
   while (len > 0 && isspace(buffer[len-1])) len--;
   buffer[len] = 0;
   if (strcmp(buffer, "q") == 0) break;
-  yield = host_check_rbl(buffer, "rbl.maps.vix.com", buffer, sizeof(buffer));
+  (void) host_check_rbl(buffer, "rbl.maps.vix.com", &bufptr);
   printf("\n> ");
   }
 
