@@ -57,7 +57,7 @@ int id_query __P4(ident_t *, id,
 	FD_ZERO(&ws);
 	FD_SET(id->fd, &ws);
 
-#ifdef __hpux
+#if defined(__hpux) || defined(__H3050) || defined(__H3050R)
 	if ((res = select(FD_SETSIZE, (int *)0, (int *)&ws, (int *)0, timeout)) < 0)
 #else
 	if ((res = select(FD_SETSIZE, (fd_set *)0, &ws, (fd_set *)0, timeout)) < 0)
