@@ -1,3 +1,9 @@
+/*
+** support.c
+**
+** Author: Pr Emanuelsson <pell@lysator.liu.se>
+** Hacked by: Peter Eriksson <pen@lysator.liu.se>
+*/
 #include <stdio.h>
 #include <ctype.h>
 
@@ -18,7 +24,12 @@ char *id_strdup __P1(char *, str)
 
     cp = (char *) malloc(strlen(str)+1);
     if (cp == NULL)
+    {
+#ifdef DEBUG
+	perror("libident: malloc");
+#endif
         return NULL;
+    }
 
     strcpy(cp, str);
 
@@ -31,10 +42,10 @@ char *id_strtok __P3(char *, cp,
 		      char *, dc)
 {
     static char *bp = 0;
-    
+
     if (cp)
 	bp = cp;
-    
+
     /*
     ** No delimitor cs - return whole buffer and point at end
     */
@@ -44,23 +55,23 @@ char *id_strtok __P3(char *, cp,
 	    bp++;
 	return cs;
     }
-    
+
     /*
     ** Skip leading spaces
     */
     while (isspace(*bp))
 	bp++;
-    
+
     /*
     ** No token found?
     */
     if (!*bp)
 	return 0;
-    
+
     cp = bp;
     while (*bp && !strchr(cs, *bp))
 	bp++;
-    
+
     /*
     ** Remove trailing spaces
     */
@@ -68,8 +79,8 @@ char *id_strtok __P3(char *, cp,
     for (dc = bp-1; dc > cp && isspace(*dc); dc--)
 	;
     *++dc = '\0';
-    
+
     bp++;
-    
+
     return cp;
 }
