@@ -10,13 +10,14 @@
 
 #include "exim.h"
 
+#define LOG_NAME_SIZE 256
 
 
 /*************************************************
 *           Local static variables               *
 *************************************************/
 
-static char *mainlog_name;
+static char mainlog_name[LOG_NAME_SIZE];
 
 static int mainlogfd = -1;
 static int mainlog_inode = -1;
@@ -53,7 +54,7 @@ static void
 open_log(int *fd, char *name)
 {
 uid_t euid;
-char buffer[256];
+char buffer[LOG_NAME_SIZE];
 
 /* Get the log file name from the spool directory or the log file path. If
 the name is too long, it is a total and utter disaster. */
@@ -66,16 +67,20 @@ if (! ((log_file_path[0] == 0)?
   if (debug_file != NULL) fprintf(debug_file, "exim: %s\n", s);
   if (log_stderr != NULL && log_stderr != debug_file)
     fprintf(log_stderr, "exim: %s\n", s);
+  #ifndef NO_OPENLOG
   openlog("exim", LOG_PID|LOG_CONS, LOG_MAIL);
+  #endif
   syslog(LOG_CRIT, s);
   closelog();
   if (accept_call_bombout) accept_bomb_out("log writing failed");
   exit(EXIT_FAILURE);
   }
 
-/* Save the name of the main log, then open the file. */
+/* Save the name of the main log so that we can stat() it later to see if it
+has been cycled, then open the file. The static slot for saving it is the same
+size as buffer, and the text has been checked above to fit. */
 
-if (strcmp(name, "main") == 0) mainlog_name = string_copy(buffer);
+if (strcmp(name, "main") == 0) strcpy(mainlog_name, buffer);
 
 *fd = open(buffer, O_APPEND|O_WRONLY, LOG_MODE);
 if (*fd >= 0) return;
@@ -89,7 +94,7 @@ be the null string, attempt to build a log directory in the spool directory. */
 
 if (*fd < 0 && errno == ENOENT && log_file_path[0] == 0)
   {
-  directory_make(spool_directory, "log", LOG_DIRECTORY_MODE);
+  directory_make(spool_directory, "log", LOG_DIRECTORY_MODE, TRUE);
   *fd = open(buffer, O_CREAT|O_APPEND|O_WRONLY, LOG_MODE);
   }
 
@@ -109,9 +114,11 @@ if (*fd >= 0)
 /* Creation failed. There are some circumstances in which we get here when
 the effective uid is not root or exim, which is the problem. (For example, a
 non-setuid binary with log_arguments set, called in certain ways.) Rather than
-just bombing out, force the log to stderr and carry on. */
+just bombing out, force the log to stderr and carry on if stderr is available.
+*/
 
-if (euid != root_uid && (!exim_uid_set || euid != exim_uid))
+if (euid != root_uid && (!exim_uid_set || euid != exim_uid) &&
+    log_stderr != NULL)
   {
   *fd = fileno(log_stderr);
   return;
@@ -119,8 +126,8 @@ if (euid != root_uid && (!exim_uid_set || euid != exim_uid))
 
 /* Otherwise this is a disaster */
 
-log_write(0, LOG_PANIC_DIE, "Cannot open %s log file \"%s\": %s", name, buffer,
-  strerror(errno));
+log_write(0, LOG_PANIC_DIE, "Cannot open %s log file \"%s\": %s: "
+  "euid=%d egid=%d", name, buffer, strerror(errno), euid, getegid());
 /* Never returns */
 }
 
@@ -195,7 +202,9 @@ if (panic_recurseflag)
     fprintf(log_stderr, "exim: could not open panic log: aborting\n");
     fprintf(log_stderr, "exim: original error was: %s", log_buffer);
     }
+  #ifndef NO_OPENLOG
   openlog("exim", LOG_PID|LOG_CONS, LOG_MAIL);
+  #endif
   syslog(LOG_CRIT, "could not open panic log - original error was: %s",
     log_buffer);
   closelog();
@@ -326,10 +335,10 @@ if (!really_exim || testing)
   return;
   }
 
-/* Handle the main log. The log gets left open during a message's delivery
+/* Handle the main log. The log gets left open during reception or delivery
 once it has been opened, but we don't want to keep on writing to it for too
-long after it has been renamed. Therefore, do a stat() and see if the inode
-has changed, and if so, re-open. */
+long after it has been renamed. Therefore, do a stat() and see if the inode has
+changed, and if so, re-open. */
 
 if ((flags & LOG_MAIN) != 0 && log_level >= level)
   {
@@ -369,7 +378,7 @@ if ((flags & LOG_REJECT) != 0)
     /* The first recipient */
 
     string_format(ptr, LOG_BUFFER_SIZE - (ptr-log_buffer), "Recipients: %s\n",
-      recipients_list[0]);
+      recipients_list[0].address);
     while (*ptr) ptr++;
 
     /* And up to 2 additional recipients */
@@ -377,7 +386,7 @@ if ((flags & LOG_REJECT) != 0)
     for (i = 1; i < recipients_count && i < 3; i++)
       {
       string_format(ptr, LOG_BUFFER_SIZE - (ptr-log_buffer), "    %s\n",
-        recipients_list[i]);
+        recipients_list[i].address);
       while (*ptr) ptr++;
       }
 
