@@ -5,9 +5,13 @@
 /* Copyright (c) University of Cambridge 1995 - 1998 */
 /* See the file NOTICE for conditions of use and distribution. */
 
-/* Many thanks to Stuart Lynne for contributing the code for this driver. */
+/* Many thanks to Stuart Lynne for contributing the code for this driver.
 
-/* TODO
+"The driver should link against Netscape. It's tested with the Netscape LDAP
+SDK that is available in source. I don't know if it will work (but it should)
+with their binary LDAP SDK libraries."
+
+TODO
    1. We should support the bindname extension
    2. Should ldapm simply be an ldap url extension?
 */
@@ -174,8 +178,14 @@ if ((rc = ldap_bind_s(lcp->ld, NULL, NULL, LDAP_AUTH_SIMPLE)) != LDAP_SUCCESS)
 
 if (ldap_url_search(lcp->ld, ldap_url, 0) == -1)
   {
-  *errmsg = string_sprintf("ldap_url_search failed: %s",
-    ldap_err2string(lcp->ld->ld_errno));
+  char *matched, *error;
+  #if LDAP_OPT_SIZELIMIT        /* Netscape SDK */
+  (void)ldap_get_lderrno(lcp->ld, &matched, &error);
+  #else
+  matched = ld->ld_matched;
+  error = lcp->ld->ld_error;
+  #endif
+  *errmsg = string_sprintf("ldap_url_search failed: %s (%s)", error, matched);
   DEBUG(9) debug_printf("open: %s\n", *errmsg);
   return DEFER;
   }
@@ -187,8 +197,14 @@ rc = ldap_result(lcp->ld, LDAP_RES_ANY, 1, NULL, &result);
 
 if (rc == -1)
   {
-  *errmsg = string_sprintf("ldap_result: failed to yield any result: %s",
-    ldap_err2string(lcp->ld->ld_errno));
+  char *matched, *error;
+  #if LDAP_OPT_SIZELIMIT        /* Netscape SDK */
+  (void)ldap_get_lderrno(lcp->ld, &matched, &error);
+  #else
+  matched = lcp->ld->ld_matched;
+  error = lcp->ld->ld_error;
+  #endif
+  *errmsg = string_sprintf("ldap_url_search failed: %s (%s)", error, matched);
   DEBUG(9) debug_printf("%s\n", *errmsg);
   return DEFER;
   }
@@ -225,7 +241,7 @@ if (rc < 1)
 
 size = 100;
 ptr = 0;
-data = store_malloc(size);
+data = store_get(size);
 
 /* Loop through returned entries; we have already checked above for zero
 entries. */
