@@ -86,12 +86,16 @@ while (fgets(buffer, sizeof(buffer), f) != NULL)
   if (s-buffer != length || strncmpic(buffer, keystring, length) != 0)
     continue;
 
-  if (*s == ':') s++;
   while (isspace(*s)) s++;
+  if (*s == ':')
+    {
+    s++;
+    while (isspace(*s)) s++;
+    }
 
   size = 100;
   ptr = 0;
-  yield = store_malloc(size);
+  yield = store_get(size);
   if (*s != 0)
     yield = string_cat(yield, &size, &ptr, s, (int)strlen(s));
 
@@ -102,7 +106,10 @@ while (fgets(buffer, sizeof(buffer), f) != NULL)
     buffer[p] = 0;
     if (buffer[0] == 0 || buffer[0] == '#') continue;
     if (!isspace(buffer[0])) break;
-    yield = string_cat(yield, &size, &ptr, buffer, (int)strlen(buffer));
+    s = buffer;
+    while (isspace(*s)) s++;
+    *(--s) = ' ';
+    yield = string_cat(yield, &size, &ptr, s, (int)strlen(s));
     }
 
   yield[ptr] = 0;
