@@ -21,12 +21,43 @@ a string as a text string. This is sometimes useful for debugging output. */
 #define mac_string(s) # s
 #define mac_expanded_string(s) mac_string(s)
 
+
+/* Character-handling macros. It seems that the set of standard functions in
+ctype.h aren't actually all that useful. One reason for this is that email is
+international, so the concept of using a locale to vary what they do is not
+helpful. Another problem is that in different operating systems, the libraries
+yield different results, even in the default locale. For example, Linux yields
+TRUE for iscntrl() for all characters > 127, whereas many other systems yield
+FALSE. For these reasons we define our own set of macros for a number of
+character testing functions. Because it is so easy to define variables as char
+rather than unsigned char, ensure that all these tests treat their arguments as
+unsigned. */
+
+#define mac_iscntrl(c) \
+  ((uschar)(c) < 32 || (uschar)(c) == 127)
+
+#define mac_iscntrl_or_special(c) \
+  ((uschar)(c) < 32 || strchr(" ()<>@,;:\\\".[]\177", (uschar)(c)) != NULL)
+
+#define mac_isgraph(c) \
+  ((uschar)(c) > 32 && (uschar)(c) != 127)
+
+#define mac_isprint(c) \
+  (((uschar)(c) >= 32 && (uschar)(c) <= 126) || \
+  ((uschar)(c) > 127 && print_topbitchars))
+
+
 /* Define which ends of pipes are for reading and writing, as some systems
 don't make both fd's two-way. */
 
 #define pipe_read  0
 #define pipe_write 1
 
+/* Define symbols for identifying the two store pools. */
+
+#define POOL_MAIN 0
+#define POOL_TEMP 1
+
 /* The EDQUOT error code isn't universally available, though it is widespread.
 There is a particular shambles in SunOS5, where it did not exist originally,
 but got installed with a particular patch for Solaris 2.4. There is a
@@ -62,12 +93,6 @@ Most have sysconf(), but a few don't. */
 
 #define mac_islookup(a,b) ((lookup_list[a].type & (b)) != 0)
 
-/* A macro to test for a printing character. We don't use isprint() because
-although it can be affected by the locale, that often isn't set the way people
-actually want it. */
-
-#define mac_isprint(c) ((c >= 32 && c <= 126) || (c >= 128 && print_topbitchars))
-
 /* Some generalities */
 
 #define DEBUG(x)      if ((x) <= debug_level)
@@ -138,8 +163,14 @@ record. */
 /* Macros for calling the memory allocation routines with
 tracing information for debugging. */
 
+#define store_extend(addr,old,new) \
+  store_extend_3(addr, old, new, __FILE__, __LINE__)
+
 #define store_free(addr)   store_free_3(addr, __FILE__, __LINE__)
+#define store_get(size)    store_get_3(size, __FILE__, __LINE__)
 #define store_malloc(size) store_malloc_3(size, __FILE__, __LINE__)
+#define store_mark(flag)   store_mark_3(flag, __FILE__, __LINE__)
+#define store_reset(addr)  store_reset_3(addr, __FILE__, __LINE__)
 
 /* Options for dns_next_rr */
 
@@ -385,7 +416,7 @@ enum {
 
 enum { MSG_DELIVER, MSG_FREEZE, MSG_REMOVE, MSG_THAW, MSG_ADD_RECIPIENT,
        MSG_MARK_ALL_DELIVERED, MSG_MARK_DELIVERED, MSG_EDIT_SENDER,
-       MSG_EDIT_BODY };
+       MSG_EDIT_BODY, MSG_SHOW_BODY, MSG_SHOW_HEADER, MSG_SHOW_LOG };
 
 /* Options for local batched SMTP deliveries. Must be in the same order
 as for non-SMTP local batching below. */
@@ -423,7 +454,23 @@ enum {
 #define topt_add_envelope_to    0x04
 #define topt_from_hack          0x08
 #define topt_use_crlf           0x10
-#define topt_smtp_dots          0x20
-#define topt_writing_file       0x40
+#define topt_smtp_dots          0x20  /* Double start-line dots */
+#define topt_end_dot            0x40  /* Send terminting dot */
+
+/* Flags for recipient_block */
+
+#define rf_onetime              0x01  /* A one-time alias */
+#define rf_notify_never         0x02  /* NOTIFY= settings */
+#define rf_notify_success       0x04
+#define rf_notify_failure       0x08
+#define rf_notify_delay         0x10
+
+#define rf_dsnflags  (rf_notify_never | rf_notify_success | \
+                      rf_notify_failure | rf_notify_delay)
+
+/* DSN RET types */
+
+#define dsn_ret_full            1
+#define dsn_ret_hdrs            2
 
 /* End of macros.h */
