@@ -164,10 +164,10 @@ if (chain_ad == NULL || pattern[0] != '^')
     }
   else key = s;
 
-  /* Now do the actual lookup; throw away the actual data returned; partial
-  matching is all handled inside search_find(). However, partial matching is
-  not appropriate for certain lookups (e.g. when looking up user@domain for
-  sender rejection). */
+  /* Now do the actual lookup; throw away the actual data returned unless it
+  was asked for; partial matching is all handled inside search_find(). However,
+  partial matching is not appropriate for certain lookups (e.g. when looking
+  up user@domain for sender rejection). */
 
   if (!use_partial) partial = -1;
   if (pattern != NULL)                /* NULL => query type */
@@ -180,18 +180,16 @@ if (chain_ad == NULL || pattern[0] != '^')
   result = search_find(handle, pattern, key, search_type,
     partial, &expand_setup, &error);
 
-  /* Free the key if not the same as the original string; free the result,
-  unless it was asked for, and return appropriately. */
-
-  if (key != s) store_free(key);
   if (result == NULL) return FALSE;
-  if (value == NULL) store_free(result); else *value = result;
+  if (value != NULL) *value = result;
+
   expand_nmax = expand_setup;
   return TRUE;
   }
 
 
-/* Regular expression match: compile if necessary */
+/* Regular expression match: compile if necessary. Keep compiled blocks
+in malloc'd store as this is a cache for compiled strings. */
 
 p = *chain_ad;
 if (p == NULL)
@@ -423,7 +421,8 @@ return FALSE;
 /* This function is given a colon-separated list of files whose existence
 is to be tested. The string is first expanded, and the resulting file names
 must be absolute, but "!" may appear precede an (absolute) file name to
-indicate that non-existence is what is wanted. If an item in the list contains
+indicate that non-existence is what is wanted. An empty name, whether preceded
+by ! or not, has no effect. Otherwise, if an item in the list contains
 no / characters, it is interpreted as a user name or uid to be used when
 checking subsequent files. A gid can also be provided with a comma separator.
 
@@ -462,6 +461,8 @@ if (ss == NULL)
     expand_string_message);
   }
 
+DEBUG(9) debug_printf("require_files = %s\n", ss);
+
 listptr = ss;
 
 for (file = string_nextinlist(&listptr, ':', buffer, sizeof(buffer));
@@ -490,6 +491,10 @@ for (file = string_nextinlist(&listptr, ':', buffer, sizeof(buffer));
       }
     }
 
+  /* Empty items are just skipped */
+
+  if (*file == 0) continue;
+
   /* If the first character is not '/' then either we have a user name
   or uid (with optional group/gid), or else it's a non-absolute file name,
   which is a disaster. */
@@ -590,7 +595,6 @@ if (restore_uid)
     (int)getegid());
   }
 
-store_free(ss);
 errno = save_errno;
 return yield;
 }
@@ -735,7 +739,6 @@ for (test_address = string_nextinlist(&listptr, separator, buffer,
       while (key != NULL && watchdog-- > 0)
         {
         rc = match_check_string(key, sdomain, NULL, -1, TRUE, &result);
-        if (key != address + domain) store_free(key);
         if (!rc)
           {
           if (search_find_defer) return FALSE;
@@ -770,17 +773,14 @@ for (test_address = string_nextinlist(&listptr, separator, buffer,
           re_block *reblock = NULL;
           if (*buffer == 0) continue;
           rc =  match_check_string(localpart, buffer, &reblock, -1, TRUE, NULL);
-          if (reblock != NULL) store_free(reblock);
           if (rc) break;
           if (search_find_defer) return FALSE;
           }
 
-        /* Rejoin the local part to the domain, free the looked-up data,
-        return if we matched anything, else continue with another lookup
-        if there was an ">" item. */
+        /* Rejoin the local part to the domain, return if we matched anything,
+        else continue with another lookup if there was an ">" item. */
 
         localpart[llen] = '@';
-        store_free(result);
         if (rc) return TRUE;
         }
 
@@ -883,7 +883,6 @@ yield = match_address_list(
   ':');
 if (search_find_defer) return DEFER;
 
-store_free(ss);
 return yield? OK : FAIL;
 }
 
@@ -901,7 +900,7 @@ return yield? OK : FAIL;
 BOOL
 regex_match_and_setup(pcre *re, char *subject, int options, int setup)
 {
-int ovector[2*(EXPAND_MAXN+1)];
+int ovector[3*(EXPAND_MAXN+1)];
 int n = pcre_exec(re, NULL, subject, (int)strlen(subject), PCRE_EOPT|options,
   ovector, sizeof(ovector)/sizeof(int));
 BOOL yield = n >= 0;
@@ -934,6 +933,13 @@ if (yield == NULL)
 return yield;
 }
 
+char *
+parse_extract_address(char *mailbox, char **errorptr, int *start, int *end,
+  int *domain, BOOL allow_null)
+{
+return NULL;
+}
+
 
 
 
