@@ -21,9 +21,9 @@ redundant apparatus. */
 
 #ifdef STAND_ALONE
 
-address_item *deliver_make_addr(char *address)
+address_item *deliver_make_addr(char *address, BOOL copy)
 {
-address_item *addr = store_malloc(sizeof(address_item));
+address_item *addr = store_get(sizeof(address_item));
 addr->next = NULL;
 addr->parent = NULL;
 addr->orig = address;
@@ -57,7 +57,8 @@ Argument:
   s        pointer to the start of an address
   nl_ends  if TRUE, '\n' terminates an address
 
-Returns:   pointer past the end of the address (null or comma)
+Returns:   pointer past the end of the address
+           (i.e. points to null or comma)
 */
 
 char *
@@ -266,12 +267,23 @@ if (*s == '[')
 
 /* Handle a proper domain, which is a sequence of dot-separated atoms.
 An atom is a sequence of any characters except specials, space, and controls.
-The specials are ( ) < > @ , ; : \ " . [ and ]. */
+The specials are ( ) < > @ , ; : \ " . [ and ]. This is the rule for RFC 822.
+However, RFC 821 and its successor is tighter, allowing only letters, digits,
+and hyphens, not starting with a hyphen. There is a global flag that gets set
+when checking addresses that have come in over SMTP and should be checked
+according to the stricter rule. */
 
 for (;;)
   {
   uschar *tsave = t;
-  while (!iscntrl(*s) && strchr(" ()<>@,;:\\\".[]", *s) == NULL) *t++ = *s++;
+
+  if (rfc821_domains)
+    {
+    if (*s != '-') while (isalnum(*s) || *s == '-') *t++ = *s++;
+    }
+  else
+    while (!mac_iscntrl_or_special(*s)) *t++ = *s++;
+
   s = skip_comment(s);
   *t = 0;
 
@@ -341,15 +353,23 @@ for (;;)
       *t++ = c;
       if (c == '\\' && s[1] != 0) *t++ = *(++s);
       }
-    *t++ = '\"';
-    if (c == '\"') s++;
+    if (c == '\"')
+      {
+      s++;
+      *t++ = '\"';
+      }
+    else
+      {
+      *error = "unmatched doublequote in local part";
+      return s;
+      }
     }
 
   /* Handle an atom, but allow quoted pairs within it. */
 
-  else while (!iscntrl((c = *s)) && strchr(" ()<>@,;:\".[]", c) == NULL)
+  else while (!mac_iscntrl_or_special(*s) || *s == '\\')
     {
-    *t++ = *s++;
+    c = *t++ = *s++;
     if (c == '\\' && *s != 0) *t++ = *s++;
     }
 
@@ -446,15 +466,12 @@ read_addr_spec(uschar *s, uschar *t, int term, char **errorptr,
   uschar **domainptr)
 {
 s = read_local_part(s, t, errorptr, FALSE);
-if (*errorptr != NULL)
-  *errorptr = string_sprintf("%s in address", *errorptr);
-else
+if (*errorptr == NULL)
   {
   if (*s != term)
     {
     if (*s != '@')
-      *errorptr = string_sprintf(
-        "\"@\" or \".\" expected in address after \"%s\"", t);
+      *errorptr = string_sprintf("\"@\" or \".\" expected after \"%s\"", t);
     else
       {
       t += (int)strlen((const char *)t);
@@ -474,7 +491,7 @@ return s;
 *************************************************/
 
 /* This function extracts an operative address from a full RFC822 mailbox and
-returns it in a piece of malloc'ed store. We take the easy way and get a piece
+returns it in a piece of dynamic store. We take the easy way and get a piece
 of store the same size as the input, and then copy into it whatever is
 necessary. If we cannot find a valid address (syntax error), return NULL, and
 point the error pointer to the reason. The arguments "start" and "end" are used
@@ -512,7 +529,7 @@ char *
 parse_extract_address(char *mailbox, char **errorptr, int *start, int *end,
   int *domain, BOOL allow_null)
 {
-uschar *yield = store_malloc((int)strlen(mailbox) + 1);
+uschar *yield = store_get((int)strlen(mailbox) + 1);
 uschar *startptr, *endptr;
 uschar *s = (uschar *)mailbox;
 uschar *t = (uschar *)yield;
@@ -706,7 +723,6 @@ if (parse_found_group && *s == ';')
   parse_found_group = FALSE;
   parse_allow_group = TRUE;
   }
-store_free(yield);
 return NULL;
 }
 
@@ -987,6 +1003,7 @@ parse_extract_addresses(char *s, char *incoming_domain, address_item **anchor,
   char *directory, error_block **syntax_errors)
 {
 int count = 0;
+int pool_reset = store_pool;
 
 DEBUG(9) debug_printf("parse_extract_addresses: %s\n", s);
 
@@ -1203,7 +1220,7 @@ for (;;)
 
     /* Get a buffer and read the contents */
 
-    filebuf = store_malloc(statbuf.st_size + 1);
+    filebuf = store_get(statbuf.st_size + 1);
     if (fread(filebuf, 1, statbuf.st_size, f) != statbuf.st_size)
       {
       *error = string_sprintf("error while reading included file %s: %s",
@@ -1217,7 +1234,6 @@ for (;;)
     addr = NULL;
     extracted = parse_extract_addresses(filebuf, incoming_domain, &addr,
       error, optional, allow_specials, rewrite, directory, syntax_errors);
-    store_free(filebuf);
     if (extracted != 0) return extracted;    /* error already set */
 
     /* If addr == NULL the included file contained a permitted :blackhole: */
@@ -1269,12 +1285,8 @@ for (;;)
       recipient =
         parse_extract_address(s+1, error, &start, &end, &domain, FALSE);
       if (recipient != NULL)
-        {
-        char *new = (domain != 0)? NULL :
+        recipient = (domain != 0)? NULL :
           string_sprintf("%s@%s", recipient, incoming_domain);
-        store_free(recipient);
-        recipient = new;
-        }
       }
 
     /* Try parsing the item as an address. */
@@ -1288,7 +1300,7 @@ for (;;)
 
     if ((*s == '|' || *s == '/') && (recipient == NULL || domain == 0))
       {
-      char *t = store_malloc((int)strlen(s) + 1);
+      char *t = store_get((int)strlen(s) + 1);
       char *p = t;
       char *q = s;
       while (*q != 0)
@@ -1301,8 +1313,13 @@ for (;;)
         else *p++ = *q++;
         }
       *p = 0;
-      addr = deliver_make_addr(t);
+
+      /* Ensure that the created address is in the MAIN storage pool */
+
+      store_pool = POOL_MAIN;
+      addr = deliver_make_addr(t, TRUE);
       addr->pfr = TRUE;              /* indicates pipe/file/reply */
+      store_pool = pool_reset;
       }
 
     /* Item must be an address. Complain if not, else qualify, rewrite and set
@@ -1323,9 +1340,10 @@ for (;;)
           s = nexts;
           continue;
           }
+
         if (syntax_errors != NULL)
           {
-          error_block *e = store_malloc(sizeof(error_block));
+          error_block *e = store_get(sizeof(error_block));
           error_block *last = *syntax_errors;
           if (last == NULL) *syntax_errors = e; else
             {
@@ -1347,11 +1365,15 @@ for (;;)
           }
         }
 
-      /* Address was successfully parsed */
+      /* Address was successfully parsed. Rewrite, and then make an address
+      block in the MAIN storage pool. */
 
       recipient = rewrite? rewrite_address(recipient, TRUE, TRUE) :
         rewrite_address_qualify(recipient, TRUE);
-      addr = deliver_make_addr(recipient);
+
+      store_pool = POOL_MAIN;
+      addr = deliver_make_addr(recipient, TRUE);  /* TRUE => copy recipient */
+      store_pool = pool_reset;
       }
 
     /* Restore the final character in the original data, and add to the
@@ -1402,7 +1424,6 @@ while (fgets(buffer, sizeof(buffer), stdin) != NULL)
     strncpy(extract, buffer+start, end-start);
     extract[end-start] = 0;
     printf("%s %d %d %d \"%s\"\n", out, start, end, domain, extract);
-    store_free(out);
     }
   }
 
@@ -1430,7 +1451,6 @@ while (fgets(buffer, sizeof(buffer), stdin) != NULL)
       strncpy(extract, buffer+start, end-start);
       extract[end-start] = 0;
       printf("%s %d %d %d \"%s\"\n", out, start, end, domain, extract);
-      store_free(out);
       }
 
     s = ss + (terminator? 1:0);
@@ -1473,7 +1493,6 @@ while (fgets(buffer, sizeof(buffer), stdin) != NULL)
       address_item *addr = anchor;
       anchor = anchor->next;
       printf("%d %s\n", addr->pfr, addr->orig);
-      store_free(addr);
       }
     }
   else printf("Failed: %d %s\n", extracted, errmess);
