@@ -92,6 +92,7 @@ for (f = queue_get_spool_list(!queue_run_in_order &&
     if (deliver_selectstring != NULL)
       {
       BOOL found = FALSE;
+      void *reset_point = store_get(0);
       if (spool_read_header(f->text, FALSE, TRUE) != spool_read_OK) continue;
 
       if (recipients_list != NULL)
@@ -100,18 +101,18 @@ for (f = queue_get_spool_list(!queue_run_in_order &&
         for (i = 0; i < recipients_count; i++)
           {
           if (
-            strstric(recipients_list[i],deliver_selectstring,FALSE) != NULL &&
-              tree_search_addr(tree_nonrecipients,recipients_list[i],FALSE)
-                == NULL)
+            strstric(recipients_list[i].address, deliver_selectstring, FALSE)
+              != NULL &&
+                tree_search_addr(tree_nonrecipients, recipients_list[i].address,
+                  FALSE) == NULL)
             {
             found = TRUE;
             break;
             }
           }
-        accept_free_recipients();
         }
 
-      if (sender_address != NULL) store_free(sender_address);
+      store_reset(reset_point);
 
       /* If no addresses matched the string, skip this message. */
 
@@ -271,7 +272,7 @@ for (i = 0; i < subdir_max; i++)
         strcmp(name + SPOOL_NAME_LENGTH - 2, "-H") == 0)
       {
       queue_filename *next =
-        store_malloc(sizeof(queue_filename) + (int)strlen(name));
+        store_get(sizeof(queue_filename) + (int)strlen(name));
       strcpy(next->text, name);
       next->dir_char = subdirchar;
 
@@ -384,10 +385,8 @@ if (p->right != NULL) queue_list_extras(p->right);
 as quickly as possible, then scan each one for information to output. If any
 disappear while we are processing, just leave them out, but give an error if an
 explicit list was given. This function is a top-level function that is obeyed
-as a result of the -bp argument. There is no point wasting resources in freeing
-the store used for the list of names at the end. However, as there may be a lot
-of messages on the queue, we must tidy up the store after reading the headers
-for each one.
+as a result of the -bp argument. As there may be a lot of messages on the
+queue, we must tidy up the store after reading the headers for each one.
 
 Arguments:
    option     0 => list top-level recipients, with "D" for those delivered
@@ -404,6 +403,7 @@ queue_list(int option, char **list, int count)
 {
 int i;
 int now = (int)time(NULL);
+void *reset_point;
 queue_filename *f;
 
 /* If given a list of messages, build a chain containing their ids. */
@@ -414,7 +414,7 @@ if (count > 0)
   for (i = 0; i < count; i++)
     {
     queue_filename *next =
-      store_malloc(sizeof(queue_filename) + (int)strlen(list[i]) + 2);
+      store_get(sizeof(queue_filename) + (int)strlen(list[i]) + 2);
     sprintf(next->text, "%s-H", list[i]);
     next->dir_char = '*';
     next->next = NULL;
@@ -427,7 +427,11 @@ if (count > 0)
 
 else f = queue_get_spool_list(FALSE);
 
-/* Now scan the chain and print information. */
+/* Now scan the chain and print information, resetting store used
+each time. If queue_list_requires_admin is TRUE and the caller is not
+an admin user, list only those messages that were submitted by the caller. */
+
+reset_point = store_get(0);
 
 for (; f != NULL; f = f->next)
   {
@@ -435,11 +439,16 @@ for (; f != NULL; f = f->next)
   int size = 0;
   BOOL env_read;
 
+  store_reset(reset_point);
   message_size = 0;
   message_subdir[0] = f->dir_char;
   rc = spool_read_header(f->text, FALSE, count <= 0);
   if (rc == spool_read_notopen && errno == ENOENT && count <= 0) continue;
   save_errno = errno;
+
+  if (queue_list_requires_admin && !admin_user &&
+     (rc != spool_read_OK || real_uid != originator_uid)) continue;
+
   env_read = (rc == spool_read_OK || rc == spool_read_hdrerror);
 
   if (env_read)
@@ -466,7 +475,6 @@ for (; f != NULL; f = f->next)
     if (user_null_sender && sender_address[0] != 0)
       printf(" <> (%s)", sender_address);
     else printf(" <%s>", sender_address);
-    store_free(sender_address);
     }
 
   if (rc != spool_read_OK)
@@ -501,20 +509,16 @@ for (; f != NULL; f = f->next)
     for (i = 0; i < recipients_count; i++)
       {
       tree_node *delivered =
-        tree_search_addr(tree_nonrecipients, recipients_list[i], FALSE);
+        tree_search_addr(tree_nonrecipients, recipients_list[i].address, FALSE);
       if (!delivered || option != 1)
         printf("        %s %s\n", (delivered != NULL)? "D":" ",
-          recipients_list[i]);
+          recipients_list[i].address);
       if (delivered != NULL) delivered->data.val = TRUE;
-      store_free(recipients_list[i]);
       }
-    store_free(recipients_list);
     if (option == 2 && tree_nonrecipients != NULL)
       queue_list_extras(tree_nonrecipients);
     printf("\n");
     }
-
-  tree_free(tree_nonrecipients);
   }
 }
 
@@ -531,7 +535,6 @@ own authority checking.
 Arguments:
   id              id of the message to work on
   action          which action is required (MSG_xxx)
-  f               if not NULL, write messages to this file
   argv            the original argv for Exim
   argc            the original argc for Exim
   recipients_arg  offset to the list of recipients in argv
@@ -540,8 +543,7 @@ Returns:          FALSE if there was any problem
 */
 
 BOOL
-queue_action(char *id, int action, FILE *f, char **argv, int argc,
-  int recipients_arg)
+queue_action(char *id, int action, char **argv, int argc, int recipients_arg)
 {
 int i, j;
 BOOL yield = TRUE;
@@ -555,27 +557,80 @@ also causes message ids to be added to log messages. */
 
 strcpy(message_id, id);
 
-/* Open and lock the data file to ensure that no other process is working on
-this message. If the file does not exist, continue only if the action is remove
-and the user is an admin user, to allow for tidying up broken states. */
+/* The "actions" that just list the files do not require any locking to be
+done. Only admin users may read the spool files. */
+
+if (action >= MSG_SHOW_BODY)
+  {
+  int fd, i, rc;
+  char *subdirectory, *suffix;
+
+  if (!admin_user)
+    {
+    printf("Permission denied\n");
+    return FALSE;
+    }
+
+  if (recipients_arg < argc)
+    {
+    printf("*** Only one message can be listed at once\n");
+    return FALSE;
+    }
+
+  if (action == MSG_SHOW_BODY)
+    {
+    subdirectory = "input";
+    suffix = "-D";
+    }
+  else if (action == MSG_SHOW_HEADER)
+    {
+    subdirectory = "input";
+    suffix = "-H";
+    }
+  else
+    {
+    subdirectory = "msglog";
+    suffix = "";
+    }
+
+  for (i = 0; i < 2; i++)
+    {
+    message_subdir[0] = (split_spool_directory == (i == 0))? id[5] : 0;
+    sprintf(spoolname, "%s/%s/%s/%s%s", spool_directory, subdirectory,
+      message_subdir, id, suffix);
+    fd = open(spoolname, O_RDONLY);
+    if (fd >= 0) break;
+    if (i == 0) continue;
+    printf("Failed to open %s file for %s%s: %s\n", subdirectory, id, suffix,
+      strerror(errno));
+    return FALSE;
+    }
+
+  while((rc = read(fd, big_buffer, big_buffer_size)) > 0)
+    write(fileno(stdout), big_buffer, rc);
+
+  return TRUE;
+  }
+
+/* For actions that actually act, open and lock the data file to ensure that no
+other process is working on this message. If the file does not exist, continue
+only if the action is remove and the user is an admin user, to allow for
+tidying up broken states. */
 
 if (!spool_open_datafile(id))
   {
   if (errno == ENOENT)
     {
     yield = FALSE;
-    if (f != NULL) fprintf(f, "Spool data file for %s does not exist\n", id);
+    printf("Spool data file for %s does not exist\n", id);
     if (action != MSG_REMOVE || !admin_user) return FALSE;
-    if (f != NULL) fprintf(f, "Continuing, to ensure all files removed\n");
+    printf("Continuing, to ensure all files removed\n");
     }
   else
     {
-    if (f != NULL)
-      {
-      if (errno == 0) fprintf(f, "Message %s is locked\n", id);
-        else fprintf(f, "Couldn't open spool file for %s: %s\n", id,
-          strerror(errno));
-      }
+    if (errno == 0) printf("Message %s is locked\n", id);
+      else printf("Couldn't open spool file for %s: %s\n", id,
+        strerror(errno));
     return FALSE;
     }
   }
@@ -590,29 +645,27 @@ sprintf(spoolname, "%s-H", id);
 if (spool_read_header(spoolname, TRUE, FALSE) != spool_read_OK)
   {
   yield = FALSE;
-  if (f != NULL)
-    {
-    if (errno != ERRNO_SPOOLFORMAT)
-      fprintf(f, "Spool read error for %s: %s\n", spoolname, strerror(errno));
-    else
-      fprintf(f, "Spool format error for %s\n", spoolname);
-    }
+  if (errno != ERRNO_SPOOLFORMAT)
+    printf("Spool read error for %s: %s\n", spoolname, strerror(errno));
+  else
+    printf("Spool format error for %s\n", spoolname);
   if (action != MSG_REMOVE || !admin_user)
     {
     close(deliver_datafile);
     deliver_datafile = -1;
     return FALSE;
     }
-  if (f != NULL) fprintf(f, "Continuing to ensure all files removed\n");
+  printf("Continuing to ensure all files removed\n");
   }
 
 /* Check that the user running this process is entitled to operate on this
 message. Only admin users may freeze/thaw, add/cancel recipients, or otherwise
-mess about, but the original sender is permitted to remove a message. */
+mess about, but the original sender is permitted to remove a message. That's
+why we leave this check until after the headers are read. */
 
 if (!admin_user && (action != MSG_REMOVE || real_uid != originator_uid))
   {
-  if (f != NULL) fprintf(f, "Permission denied\n");
+  printf("Permission denied\n");
   close(deliver_datafile);
   deliver_datafile = -1;
   return FALSE;
@@ -625,7 +678,7 @@ username = (pw != NULL)? pw->pw_name : string_sprintf("uid %d", (int)real_uid);
 
 /* Take the necessary action. */
 
-if (f != NULL) fprintf(f, "Message %s ", id);
+printf("Message %s ", id);
 
 switch(action)
   {
@@ -633,7 +686,7 @@ switch(action)
   if (deliver_freeze)
     {
     yield = FALSE;
-    if (f != NULL) fprintf(f, "is already frozen\n");
+    printf("is already frozen\n");
     }
   else
     {
@@ -642,13 +695,13 @@ switch(action)
     deliver_frozen_at = time(NULL);
     if (spool_write_header(id) >= 0)
       {
-      if (f != NULL) fprintf(f, "is now frozen\n");
+      printf("is now frozen\n");
       log_write(0, LOG_MAIN, "frozen by %s", username);
       }
     else
       {
       yield = FALSE;
-      if (f != NULL) fprintf(f, "could not be frozen: failed to rewrite "
+      printf("could not be frozen: failed to rewrite "
         "spool header file\n");
       log_write(0, LOG_MAIN, "failed to rewrite header file");
       }
@@ -660,7 +713,7 @@ switch(action)
   if (!deliver_freeze)
     {
     yield = FALSE;
-    if (f != NULL) fprintf(f, "is not frozen\n");
+    printf("is not frozen\n");
     }
   else
     {
@@ -668,13 +721,13 @@ switch(action)
     deliver_manual_thaw = TRUE;
     if (spool_write_header(id) >= 0)
       {
-      if (f != NULL) fprintf(f, "is no longer frozen\n");
+      printf("is no longer frozen\n");
       log_write(0, LOG_MAIN, "unfrozen by %s", username);
       }
     else
       {
       yield = FALSE;
-      if (f != NULL) fprintf(f, "could not be unfrozen: failed to rewrite "
+      printf("could not be unfrozen: failed to rewrite "
         "spool header file\n");
       log_write(0, LOG_MAIN, "failed to rewrite header file");
       }
@@ -698,9 +751,8 @@ switch(action)
       if (errno != ENOENT)
         {
         yield = FALSE;
-        if (f != NULL)
-          fprintf(f, "Error while removing %s: %s\n", spoolname,
-            strerror(errno));
+        printf("Error while removing %s: %s\n", spoolname,
+          strerror(errno));
         }
       }
 
@@ -713,9 +765,8 @@ switch(action)
         if (errno != ENOENT)
           {
           yield = FALSE;
-          if (f != NULL)
-            fprintf(f, "Error while removing %s: %s\n", spoolname,
-              strerror(errno));
+          printf("Error while removing %s: %s\n", spoolname,
+            strerror(errno));
           }
         }
       }
@@ -724,12 +775,8 @@ switch(action)
   /* In the common case, the datafile is open (and locked), so give the
   obvious message. Otherwise be more specific. */
 
-  if (f != NULL)
-    {
-    if (deliver_datafile >= 0) fprintf(f, "has been removed\n");
-      else fprintf(f, "has been removed or did not exist\n");
-    }
-
+  if (deliver_datafile >= 0) printf("has been removed\n");
+    else printf("has been removed or did not exist\n");
   log_write(0, LOG_MAIN, "removed by %s", username);
   break;
 
@@ -737,19 +784,19 @@ switch(action)
   case MSG_MARK_ALL_DELIVERED:
   for (i = 0; i < recipients_count; i++)
     {
-    tree_add_nonrecipient(recipients_list[i], FALSE);
+    tree_add_nonrecipient(recipients_list[i].address, FALSE);
     }
   if (spool_write_header(id) >= 0)
     {
-    if (f != NULL) fprintf(f, "has been modified\n");
+    printf("has been modified\n");
     for (i = 0; i < recipients_count; i++)
       log_write(0, LOG_MAIN, "address <%s> marked delivered by %s",
-        recipients_list[i], username);
+        recipients_list[i].address, username);
     }
   else
     {
     yield = FALSE;
-    if (f != NULL) fprintf(f, "- failed to rewrite spool header file "
+    printf("- failed to rewrite spool header file "
       "while marking all recipients delivered\n");
     log_write(0, LOG_MAIN, "failed to rewrite header file while "
       "marking all recipients delivered");
@@ -761,8 +808,7 @@ switch(action)
   if (recipients_arg < argc - 1)
     {
     yield = FALSE;
-    if (f != NULL)
-      fprintf(f, "- only one sender address can be specified\n");
+    printf("- only one sender address can be specified\n");
     break;
     }
   doing = "editing sender";
@@ -780,8 +826,7 @@ switch(action)
   if (recipients_arg >= argc)
     {
     yield = FALSE;
-    if (f != NULL)
-      fprintf(f, "- error while %s: no address given\n", doing);
+    printf("- error while %s: no address given\n", doing);
     }
   else
     {
@@ -796,34 +841,31 @@ switch(action)
       if (receiver == NULL)
         {
         yield = FALSE;
-        if (f != NULL)
-          fprintf(f, "- error while %s:\n  bad address %s: %s\n",
-            doing, argv[recipients_arg], errmess);
+        printf("- error while %s:\n  bad address %s: %s\n",
+          doing, argv[recipients_arg], errmess);
         }
       else if (receiver[0] != 0 && domain == 0)
         {
         yield = FALSE;
-        if (f != NULL)
-          fprintf(f, "- error while %s:\n  bad address %s: "
-            "domain missing\n", doing, argv[recipients_arg]);
+        printf("- error while %s:\n  bad address %s: "
+          "domain missing\n", doing, argv[recipients_arg]);
         }
       else
         {
         if (action == MSG_ADD_RECIPIENT)
           {
-          accept_add_recipient(receiver);
+          accept_add_recipient(receiver, NULL, 0, 0);
           log_write(0, LOG_MAIN, "recipient <%s> added by %s",
             receiver, username);
           }
         else if (action == MSG_MARK_DELIVERED)
           {
           for (i = 0; i < recipients_count; i++)
-            if (strcmp(recipients_list[i], receiver) == 0) break;
+            if (strcmp(recipients_list[i].address, receiver) == 0) break;
           if (i >= recipients_count)
             {
-            if (f != NULL)
-              fprintf(f, "- error while %s:\n  %s is not a recipient:"
-                " message not updated\n", doing, receiver);
+            printf("- error while %s:\n  %s is not a recipient:"
+              " message not updated\n", doing, receiver);
             yield = FALSE;
             }
           else
@@ -847,13 +889,11 @@ switch(action)
     if (yield)
       {
       if (spool_write_header(id) >= 0)
-        {
-        if (f != NULL) fprintf(f, "has been modified\n");
-        }
+        printf("has been modified\n");
       else
         {
         yield = FALSE;
-        if (f != NULL) fprintf(f, "- failed to rewrite spool header file "
+        printf("- failed to rewrite spool header file "
           "while %s\n", doing);
         log_write(0, LOG_MAIN, "failed to rewrite header file while %s",
           doing);
@@ -867,8 +907,7 @@ switch(action)
   if (recipients_arg < argc)
     {
     yield = FALSE;
-    if (f != NULL)
-      fprintf(f, "- only one message can be edited at once\n");
+    printf("- only one message can be edited at once\n");
     }
 
   /* Make a copy of the body, and let the editor work on that. If the editor
@@ -892,7 +931,7 @@ switch(action)
     copy_fd = open(copyname, O_WRONLY|O_CREAT, SPOOL_MODE);
     if (copy_fd < 0)
       {
-      if (f != NULL) fprintf(f, "not modified: opening copy file failed: %s",
+      printf("not modified: opening copy file failed: %s",
         strerror(errno));
       yield = FALSE;
       break;
@@ -912,7 +951,7 @@ switch(action)
       {
       if (write(copy_fd, big_buffer, rc) != rc)
         {
-        if (f != NULL) fprintf(f, "not modified: copying failed (write): %s\n",
+        printf("not modified: copying failed (write): %s\n",
           strerror(errno));
         yield = FALSE;
         break;
@@ -921,7 +960,7 @@ switch(action)
 
     if (rc < 0)
       {
-      if (f != NULL) fprintf(f, "not modified: copying failed (read): %s\n",
+      printf("not modified: copying failed (read): %s\n",
         strerror(errno));
       yield = FALSE;
       }
@@ -941,19 +980,19 @@ switch(action)
           id);
         if ((rc = rename(copyname, spoolname)) == 0)
           {
-          if (f != NULL) fprintf(f, "has been modified\n");
+          printf("has been modified\n");
           log_write(0, LOG_MAIN, "body edited by %s", username);
           }
         else
           {
-          if (f != NULL) fprintf(f, "not modified: rename failed: %s\n",
+          printf("not modified: rename failed: %s\n",
             strerror(errno));
           yield = FALSE;
           }
         }
       else
         {
-        if (f != NULL) fprintf(f, "not modified: editing failed\n");
+        printf("not modified: editing failed\n");
         yield = FALSE;
         }
       }
