@@ -40,6 +40,7 @@ static optionlist optionlist_config[] = {
   { "always_bcc",               opt_bool,        &always_bcc },
   { "auto_thaw",                opt_time,        &auto_thaw },
   { "bi_command",               opt_stringptr,   &bi_command },
+  { "check_dns_names",          opt_bool|opt_hidden, &check_dns_names },
   { "check_log_inodes",         opt_int,         &check_log_inodes },
   { "check_log_space",          opt_int,         &check_log_space },
   { "check_spool_inodes",       opt_int,         &check_spool_inodes },
@@ -54,8 +55,12 @@ static optionlist optionlist_config[] = {
   { "deliver_queue_load_max",   opt_fixed,       &deliver_queue_load_max },
   { "delivery_date_remove",     opt_bool,        &delivery_date_remove },
   { "dns_again_means_nonexist", opt_stringptr,   &dns_again_means_nonexist },
+  { "dns_check_names",          opt_bool,        &check_dns_names },
   { "dns_retrans",              opt_time,        &dns_retrans },
   { "dns_retry",                opt_int,         &dns_retry },
+  #ifdef SUPPORT_DSN
+  { "dsn",                      opt_bool,        &dsn },
+  #endif
   { "envelope_to_remove",       opt_bool,        &envelope_to_remove },
   { "errmsg_file",              opt_stringptr,   &errmsg_file },
   { "errmsg_text",              opt_stringptr,   &errmsg_text },
@@ -65,6 +70,7 @@ static optionlist optionlist_config[] = {
   { "exim_group",               opt_gid,         &exim_gid },
   { "exim_path",                opt_stringptr,   &exim_path },
   { "exim_user",                opt_uid,         &exim_uid },
+  { "extract_addresses_remove_arguments", opt_bool, &extract_addresses_remove_arguments },
   { "finduser_retries",         opt_int,         &finduser_retries },
   { "forbid_domain_literals",   opt_bool,        &forbid_domain_literals },
   { "freeze_tell_mailmaster",   opt_bool,        &freeze_tell_mailmaster },
@@ -74,6 +80,8 @@ static optionlist optionlist_config[] = {
   { "headers_checks_fail",      opt_bool,        &headers_checks_fail },
   { "headers_sender_verify",    opt_bool,        &headers_sender_verify },
   { "headers_sender_verify_errmsg", opt_bool,    &headers_sender_verify_errmsg },
+  { "helo_accept_junk_hosts",   opt_stringptr,   &helo_accept_junk_hosts },
+  { "helo_accept_junk_nets",    opt_stringptr,   &helo_accept_junk_nets },
   { "helo_strict_syntax",       opt_bool,        &helo_strict_syntax },
   { "helo_verify",              opt_bool,        &helo_verify },
   { "helo_verify_except_hosts", opt_stringptr,   &helo_verify_except_hosts },
@@ -82,6 +90,7 @@ static optionlist optionlist_config[] = {
   { "hold_domains",             opt_domainlist,  &hold_domains },
   { "hold_domains_except",      opt_domainlist,  &hold_domains_except },
   { "host_lookup_nets",         opt_stringptr,   &host_lookup_nets },
+  { "hosts_treat_as_local",     opt_stringptr,   &hosts_treat_as_local },
   { "ignore_errmsg_errors",     opt_bool,        &ignore_errmsg_errors },
   { "ignore_errmsg_errors_after", opt_time,      &ignore_errmsg_errors_after },
   { "keep_malformed",           opt_time,        &keep_malformed },
@@ -90,6 +99,7 @@ static optionlist optionlist_config[] = {
   { "local_domains_include_host", opt_bool,      &local_domains_include_host },
   { "local_domains_include_host_literals", opt_bool, &local_domains_include_host_literals },
   { "local_interfaces",         opt_stringptr,   &local_interfaces },
+  { "localhost_number",         opt_stringptr,   &host_number_string },
   { "locally_caseless",         opt_bool,        &locally_caseless },
   { "log_all_parents",          opt_bool,        &log_all_parents },
   { "log_arguments",            opt_bool,        &log_arguments },
@@ -125,6 +135,7 @@ static optionlist optionlist_config[] = {
   { "prohibition_message",      opt_stringptr,   &prohibition_message },
   { "qualify_domain",           opt_stringptr,   &qualify_domain_sender },
   { "qualify_recipient",        opt_stringptr,   &qualify_domain_recipient },
+  { "queue_list_requires_admin",opt_bool,        &queue_list_requires_admin },
   { "queue_only",               opt_bool,        &queue_only },
   { "queue_only_load",          opt_fixed,       &queue_only_load },
   { "queue_remote",             opt_bool,        &queue_remote },
@@ -173,12 +184,14 @@ static optionlist optionlist_config[] = {
   { "sender_host_reject_except",opt_domainlist,  &sender_host_reject_except },
   { "sender_host_reject_recipients", opt_domainlist, &sender_host_reject_recipients },
   { "sender_host_reject_relay", opt_domainlist,  &sender_host_reject_relay },
+  { "sender_host_reject_relay_except", opt_domainlist, &sender_host_reject_relay_except },
   { "sender_net_accept",        opt_stringptr,   &sender_net_accept },
   { "sender_net_accept_relay",  opt_stringptr,   &sender_net_accept_relay },
   { "sender_net_reject",        opt_stringptr,   &sender_net_reject },
   { "sender_net_reject_except", opt_stringptr,   &sender_net_reject_except },
   { "sender_net_reject_recipients", opt_stringptr, &sender_net_reject_recipients },
   { "sender_net_reject_relay",  opt_stringptr,   &sender_net_reject_relay },
+  { "sender_net_reject_relay_except", opt_stringptr, &sender_net_reject_relay_except },
   { "sender_reject",            opt_stringptr,   &sender_reject },
   { "sender_reject_except",     opt_stringptr,   &sender_reject_except },
   { "sender_reject_recipients", opt_stringptr,   &sender_reject_recipients },
@@ -191,18 +204,22 @@ static optionlist optionlist_config[] = {
   { "sender_verify_except_nets",opt_stringptr,   &sender_verify_except_nets },
   { "sender_verify_fixup",      opt_bool,        &sender_verify_fixup },
   { "sender_verify_log_details", opt_bool,       &sender_verify_log_details },
+  { "sender_verify_max_retry_rate", opt_int,     &sender_verify_max_retry_rate },
   { "sender_verify_reject",     opt_bool,        &sender_verify_reject },
   { "smtp_accept_max",          opt_int,         &smtp_accept_max },
   { "smtp_accept_queue",        opt_int,         &smtp_accept_queue },
+  { "smtp_accept_queue_per_connection", opt_int, &smtp_accept_queue_per_connection },
   { "smtp_accept_reserve",      opt_int,         &smtp_accept_reserve },
   { "smtp_banner",              opt_stringptr,   &smtp_banner },
   { "smtp_connect_backlog",     opt_int,         &smtp_connect_backlog },
+  { "smtp_etrn_command",        opt_stringptr,   &smtp_etrn_command },
   { "smtp_etrn_hosts",          opt_domainlist,  &smtp_etrn_hosts },
   { "smtp_etrn_nets",           opt_stringptr,   &smtp_etrn_nets },
   { "smtp_etrn_serialize",      opt_bool,        &smtp_etrn_serialize },
   { "smtp_expn_hosts",          opt_domainlist,  &smtp_expn_hosts },
   { "smtp_expn_nets",           opt_stringptr,   &smtp_expn_nets },
   { "smtp_load_reserve",        opt_fixed,       &smtp_load_reserve },
+  { "smtp_log_connections",     opt_bool,        &smtp_log_connections },
   { "smtp_receive_timeout",     opt_time,        &smtp_receive_timeout },
   { "smtp_reserve_hosts",       opt_domainlist,  &smtp_reserve_hosts },
   { "smtp_reserve_nets",        opt_stringptr,   &smtp_reserve_nets },
@@ -225,6 +242,34 @@ static int optionlist_config_size =
 
 
 
+
+
+/*************************************************
+*         Find the name of an option             *
+*************************************************/
+
+/* This function is to aid debugging. Various functions take arguments that are
+variables in the options table above. For debugging output, it is useful to be
+able to find the name of the option which is currently being processed. This
+function finds it, if it exists, by searching the table.
+
+Arguments:   a value that is presumed to be in the table above
+Returns:     the option name, or an empty string
+*/
+
+char *
+readconf_find_option(void *p)
+{
+int i;
+for (i = 0; i < optionlist_config_size; i++)
+  if (p == optionlist_config[i].value) return optionlist_config[i].name;
+return "";
+}
+
+
+
+
+
 /*************************************************
 *            Read configuration line             *
 *************************************************/
@@ -260,6 +305,23 @@ for (;;)
   if (fgets(big_buffer, big_buffer_size, config_file) == NULL) return NULL;
   config_lineno++;
   n = (int)strlen(big_buffer);
+
+  /* Handle pathologically long lines - yes, it did happen - by extending
+  big_buffer. */
+
+  while (n == big_buffer_size - 1 && big_buffer[n-1] != '\n')
+    {
+    char *newbuffer = store_malloc(big_buffer_size += BIG_BUFFER_SIZE);
+    strcpy(newbuffer, big_buffer);
+    n = (int)strlen(newbuffer);
+    store_free(big_buffer);
+    big_buffer = newbuffer;
+    if (fgets(big_buffer+n, big_buffer_size-n, config_file) == NULL) break;
+    n = (int)strlen(big_buffer);
+    }
+
+  /* Remove trailing whitespace and find genuine start of line */
+
   while (n > 0 && isspace(big_buffer[n-1])) n--;
   big_buffer[n] = 0;
   s = big_buffer;
@@ -268,8 +330,9 @@ for (;;)
   }
 
 /* Handle continuation lines if requested to do so - this isn't used for
-strings, which handle continuation themselves. Otherwise the maximum is
-limited by the size of the big buffer. */
+strings, which handle continuation themselves. Some of this repeats code above,
+but I can't see how to pack them up tidily, and we do want this to be
+efficient, since Exim is forever reading its configuration file. */
 
 if (get_continued || (maybe_macro && isupper(*s)))
   {
@@ -278,12 +341,11 @@ if (get_continued || (maybe_macro && isupper(*s)))
     int c = fgetc(config_file);
     while (c == ' ' || c == '\t') c = fgetc(config_file);
     if (c == '\n' || c == EOF) { config_lineno++; break; }
+    big_buffer[n-1] = c;
 
     if (big_buffer_size - n < 256)
       {
-      char *newbuffer;
-      big_buffer_size += BIG_BUFFER_SIZE;
-      newbuffer = store_malloc(big_buffer_size);
+      char *newbuffer = store_malloc(big_buffer_size += BIG_BUFFER_SIZE);
       strcpy(newbuffer, s);
       s = newbuffer;
       n = (int)strlen(newbuffer);
@@ -293,9 +355,20 @@ if (get_continued || (maybe_macro && isupper(*s)))
 
     if (fgets(big_buffer+n, big_buffer_size-n, config_file) == NULL) break;
     config_lineno++;
-    if (c == '#') continue;
-    big_buffer[n-1] = c;
     n = (int)strlen(big_buffer);
+
+    while (n == big_buffer_size - 1 && big_buffer[n-1] != '\n')
+      {
+      char *newbuffer = store_malloc(big_buffer_size += BIG_BUFFER_SIZE);
+      strcpy(newbuffer, big_buffer);
+      n = (int)strlen(newbuffer);
+      store_free(big_buffer);
+      big_buffer = newbuffer;
+      if (fgets(big_buffer+n, big_buffer_size-n, config_file) == NULL) break;
+      n = (int)strlen(big_buffer);
+      }
+
+    if (c == '#') continue;
     while (n > 0 && isspace(big_buffer[n-1])) n--;
     big_buffer[n] = 0;
     }
@@ -602,6 +675,9 @@ BOOL freesptr = TRUE;
 optionlist *ol, *ol2;
 transport_instance *tp;
 struct passwd *pw;
+void *reset_point;
+char *intformat = "%i%n";
+char *inttype = "";
 char *sptr;
 char *s = buffer;
 char name[64];
@@ -712,8 +788,6 @@ switch (type)
 
   case opt_domainlist:
   case opt_stringptr:
-  freesptr = FALSE;
-
   case opt_uid:
   case opt_gid:
   case opt_expand_uid:
@@ -724,11 +798,17 @@ switch (type)
   case opt_transportptr:
   case opt_local_batch:
 
-  n = (int)strlen(s) + 1;
-  sptr = store_malloc(n);
-  *sptr = 0;                   /* insurance */
+  /* Strings are read into the normal store pool. As long we aren't too
+  near the end of the current block, the string will just use what is necessary
+  on the top of the stacking pool, because string_cat() uses the extension
+  mechanism. If keeping the actual string is not required, freesptr is set
+  TRUE, and at the end we reset the pool. */
 
-  if (*s == '\"')
+  n = (int)strlen(s) + 1;
+  reset_point = sptr = store_get(n);
+  *sptr = 0;                                /* insurance */
+
+  if (*s != '\"') strcpy(sptr, s); else
     {
     int count = 0;
 
@@ -784,10 +864,6 @@ switch (type)
     sptr[count] = 0;
     }
 
-  /* If not quoted, just take the rest of the line. */
-
-  else strcpy(sptr, s);
-
   /* Having read a string, we now have several different ways of using it,
   depending on the data type, so do another switch. */
 
@@ -816,13 +892,15 @@ switch (type)
       }
     /* Fall through */
 
-    /* If this was a string, set the variable to point to the new string*/
+    /* If this was a string, set the variable to point to the new string,
+    and set the flag so its store isn't reclaimed. */
 
     case opt_stringptr:
     if (data_block == NULL)
       *((char **)(ol->value)) = sptr;
     else
       *((char **)((char *)data_block + (long int)(ol->value))) = sptr;
+    freesptr = FALSE;
     break;
 
     /* If it was a transport - look it up in the transports list unless
@@ -975,7 +1053,8 @@ switch (type)
     break;
 
     /* If it was a uid list, look up each individual entry, and build
-    a vector of uids, with a count in the first element. */
+    a vector of uids, with a count in the first element. Put the vector
+    in permanent store so we can free the string. */
 
     case opt_uidlist:
       {
@@ -1007,7 +1086,8 @@ switch (type)
     break;
 
     /* If it was a gid list, look up each individual entry, and build
-    a vector of gids, with a count in the first element. */
+    a vector of gids, with a count in the first element. Put the vector
+    in permanent store so we can free the string. */
 
     case opt_gidlist:
       {
@@ -1145,9 +1225,9 @@ switch (type)
     break;
     }
 
-  /* Release store for temporary strings */
+  /* Release store if the value of the string doesn't need to be kept. */
 
-  if (freesptr) store_free(sptr);
+  if (freesptr) store_reset(reset_point);
   break;
 
   /* Boolean: if no characters follow, the value is boolvalue. Otherwise
@@ -1189,24 +1269,30 @@ switch (type)
     }
   break;
 
+  /* Octal integer */
+
+  case opt_octint:
+  intformat = "%o%n";
+  inttype = "octal ";
+
   /*  Integer: a simple(ish) case; allow octal and hex formats, and
   suffixes K and M. The different types affect output, not input. */
 
   case opt_mkint:
-  case opt_octint:
   case opt_int:
-  if (sscanf(s, "%i%n", &value, &count) != 1)
-    log_write(0, LOG_PANIC_DIE|LOG_CONFIG, "integer expected "
-      "for %s in line %d", name, config_lineno);
+  if (sscanf(s, intformat, &value, &count) != 1)
+    log_write(0, LOG_PANIC_DIE|LOG_CONFIG, "%sinteger expected "
+      "for %s in line %d", inttype, name, config_lineno);
 
   if (tolower(s[count]) == 'k') { value *= 1024; count++; }
   if (tolower(s[count]) == 'm') { value *= 1024*1024; count++; }
+
   while (isspace(s[count])) count++;
 
   if (s[count] != 0)
     log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
-      "extra characters follow integer value "
-      "for %s in line %d", name, config_lineno);
+      "extra characters follow %sinteger value "
+      "for %s in line %d", inttype, name, config_lineno);
 
   if (data_block == NULL)
     *((int *)(ol->value)) = value;
@@ -1399,7 +1485,7 @@ switch(ol->type & opt_mask)
   case opt_domainlist:
   case opt_stringptr:
   s = *((char **)value);
-  printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s, FALSE));
+  printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s));
   break;
 
   /* If fixed transport is unset, look for dynamic expanding string */
@@ -1416,7 +1502,7 @@ switch(ol->type & opt_mask)
       if (options_block != NULL)
         value2 = (void *)((char *)options_block + (long int)value2);
       s = *((char **)value2);
-      printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s, FALSE));
+      printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s));
       }
     else printf("%s =\n", name);
     }
@@ -1483,7 +1569,7 @@ switch(ol->type & opt_mask)
       if (options_block != NULL)
         value2 = (void *)((char *)options_block + (long int)value2);
       s = *((char **)value2);
-      printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s, FALSE));
+      printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s));
       break;
       }
     }
@@ -1495,7 +1581,7 @@ switch(ol->type & opt_mask)
     printf("%s =\n", name);
   else
     {
-    pw = getpwuid(*((int *)value));
+    pw = getpwuid(*((uid_t *)value));
     if (pw == NULL)
       printf("%s = %d\n", name, (int)(*((uid_t *)value)));
     else printf("%s = %s\n", name, pw->pw_name);
@@ -1516,7 +1602,7 @@ switch(ol->type & opt_mask)
       if (options_block != NULL)
         value2 = (void *)((char *)options_block + (long int)value2);
       s = *((char **)value2);
-      printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s, FALSE));
+      printf("%s = %s\n", name, (s == NULL)? "" : string_printing(s));
       break;
       }
     }
@@ -1935,14 +2021,14 @@ while ((s = get_config_line(TRUE, FALSE, TRUE)) != NULL &&
     macro_item *m;
     macro_item *mlast = NULL;
     char name[24];
-    int ptr = 0;
+    int namelen = 0;
 
     while (isalnum(*s) || *s == '_')
       {
-      if (ptr < sizeof(name)-1) name[ptr++] = *s;
+      if (namelen < sizeof(name)-1) name[namelen++] = *s;
       s++;
       }
-    name[ptr] = 0;
+    name[namelen] = 0;
     while (isspace(*s)) s++;
     if (*s++ != '=')
       log_write(0, LOG_CONFIG|LOG_PANIC_DIE, "malformed macro definition in "
@@ -1950,21 +2036,31 @@ while ((s = get_config_line(TRUE, FALSE, TRUE)) != NULL &&
     while (isspace(*s)) s++;
 
     /* If an existing macro of the same name was defined on the command line,
-    we just skip this definition. Otherwise it's an error. */
+    we just skip this definition. Otherwise it's an error to attempt to
+    redefine a macro. It is also an error to define a macro whose name begins
+    with the name of a previously-defined macro. */
 
     for (m = macros; m != NULL; m = m->next)
       {
+      int len = (int)strlen(m->name);
+
       if (strcmp(m->name, name) == 0)
         {
         if (m->command_line) break;
         log_write(0, LOG_CONFIG|LOG_PANIC_DIE, "macro \"%s\" is already "
           "defined", name);
         }
+
+      if (len < namelen && strstr(name, m->name) != NULL)
+        log_write(0, LOG_CONFIG|LOG_PANIC_DIE, "\"%s\" cannot be defined as "
+          "a macro because previously defined macro \"%s\" is a substring",
+          name, m->name);
+
       mlast = m;
       }
     if (m != NULL) continue;   /* Found an overriding command-line definition */
 
-    m = store_malloc(sizeof(macro_item) + (int)strlen(name));
+    m = store_get(sizeof(macro_item) + namelen);
     m->next = NULL;
     m->command_line = FALSE;
     if (mlast == NULL) macros = m; else mlast->next = m;
@@ -2059,15 +2155,14 @@ message. */
 regex_From = regex_must_compile(uucp_from_pattern);
 
 /* If there wasn't a -oX option and daemon_smtp_port was set, find the
-port number in network byte order. */
+port number from it. */
 
 if (daemon_listen && smtp_port < 0 && configured_smtp_service != NULL)
   {
   if (isdigit(*configured_smtp_service))
     {
     char *end;
-    smtp_port = (int)htons((unsigned short)strtol(configured_smtp_service,
-      &end, 0));
+    smtp_port = strtol(configured_smtp_service, &end, 0);
     if (end != configured_smtp_service + (int)strlen(configured_smtp_service))
       log_write(0, LOG_PANIC_DIE|LOG_CONFIG, "invalid SMTP service: %s",
         configured_smtp_service);
@@ -2113,6 +2208,7 @@ check_net_list(sender_net_accept_relay, "sender_net_accept_relay");
 check_net_list(sender_net_reject, "sender_net_reject");
 check_net_list(sender_net_reject_recipients, "sender_net_reject_recipients");
 check_net_list(sender_net_reject_relay, "sender_net_reject_relay");
+check_net_list(sender_net_reject_relay_except, "sender_net_reject_relay_except");
 check_net_list(sender_verify_except_nets, "sender_verify_except_nets");
 check_net_list(smtp_reserve_nets, "smtp_reserve_nets");
 
@@ -2289,8 +2385,6 @@ if (errors_reply_to != NULL)
   if (domain == 0)
     log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
       "errors_reply_to (%s) does not contain a domain", errors_reply_to);
-
-  store_free(recipient);
   }
 
 /* Kill_ip_options => log_ip_options */
@@ -2312,6 +2406,24 @@ if (debug_transport_file != NULL)
   log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
     "debug_transport is available only when TRANSPORT_DEBUG is defined");
 #endif
+
+/* Set up the host number if anything is specified. It is an expanded string
+so that it can be computed from the host name, for example. We do this last
+so as to ensure that everything else is set up before the expansion. */
+
+if (host_number_string != NULL)
+  {
+  char *end;
+  char *s = expand_string(host_number_string);
+  long int n = strtol(s, &end, 0);
+  if (*end != 0)
+    log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
+      "localhost_number value is not a number: %s", s);
+  if (n > 255)
+    log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
+      "localhost_number is greater than 255");
+  host_number = n;
+  }
 }
 
 
@@ -2349,7 +2461,7 @@ for (dd = drivers_available; dd->driver_name[0] != 0;
     int i;
     int len = dd->options_len;
     d->info = dd;
-    d->options_block = store_malloc(len);
+    d->options_block = store_get(len);
     memcpy(d->options_block, dd->options_block, len);
     for (i = 0; i < *(dd->options_count); i++)
       dd->options[i].type &= ~opt_set;
@@ -2435,7 +2547,7 @@ while ((buffer = get_config_line(TRUE, FALSE, FALSE)) != NULL &&
     /* Set up a new driver instance data block on the chain, with
     its default values installed. */
 
-    d = store_malloc(instance_size);
+    d = store_get(instance_size);
     memcpy(d, instance_default, instance_size);
     *p = d;
     p = &(d->next);
@@ -2687,7 +2799,7 @@ while ((p = get_config_line(TRUE, FALSE, FALSE)) != NULL &&
   retry_rule **rchain;
   char *pp, *error;
 
-  next = store_malloc(sizeof(retry_config));
+  next = store_get(sizeof(retry_config));
   next->next = NULL;
   *chain = next;
   chain = &(next->next);
@@ -2696,7 +2808,7 @@ while ((p = get_config_line(TRUE, FALSE, FALSE)) != NULL &&
   rchain = &(next->rules);
 
   pp = p;
-  while (isgraph(*p)) p++;
+  while (mac_isgraph(*p)) p++;
   if (p - pp <= 0)
     log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
       "missing domain name in line %d", config_lineno);
@@ -2717,7 +2829,7 @@ while ((p = get_config_line(TRUE, FALSE, FALSE)) != NULL &&
 
   while (isspace(*p)) p++;
   pp = p;
-  while (isgraph(*p)) p++;
+  while (mac_isgraph(*p)) p++;
   if (p - pp <= 0)
     log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
       "missing error type in line %d", config_lineno);
@@ -2734,7 +2846,7 @@ while ((p = get_config_line(TRUE, FALSE, FALSE)) != NULL &&
   while (isspace(*p)) p++;
   while (*p != 0)
     {
-    retry_rule *rule = store_malloc(sizeof(retry_rule));
+    retry_rule *rule = store_get(sizeof(retry_rule));
     *rchain = rule;
     rchain = &(rule->next);
     rule->next = NULL;
@@ -2807,7 +2919,7 @@ char *p, *pp;
 while ((p = get_config_line(TRUE, TRUE, FALSE)) != NULL &&
         strcmpic(p, "end") != 0)
   {
-  next = store_malloc(sizeof(rewrite_rule));
+  next = store_get(sizeof(rewrite_rule));
   next->next = NULL;
   *chain = next;
   chain = &(next->next);
@@ -2847,7 +2959,7 @@ while ((p = get_config_line(TRUE, TRUE, FALSE)) != NULL &&
     {
     int size = 128;
     int count = 0;
-    char *sptr = store_malloc(size);
+    char *sptr = store_get(size);
     while (*(++p) != 0)
       {
       if (*p == '\"') { p++; break; }
