@@ -34,13 +34,20 @@ already set (indicated by hstatus_unknown). The possible values are:
 If a retry record is retrieved from the hints database, the time of last
 trying is filled into the last_try field of the host block.
 
-Argument:   pointer to a host item
+Arguments:
+  host                pointer to a host item
+  include_ip_address  TRUE to include the address in the key - this is
+                        usual, but sometimes is not wanted
+  is_expired          set TRUE if the host has expired but is usable because
+                        its retry time has come; undefined except when the
+                        return is hstatus_usable
+
 Returns:    NULL if no retry record was read; otherwise
             pointer to the key name of the retry record
 */
 
 char *
-retry_check_address(host_item *host)
+retry_check_address(host_item *host, BOOL include_ip_address, BOOL *is_expired)
 {
 char *key;
 open_db dbblock;
@@ -57,7 +64,9 @@ host->status = hstatus_usable;
 
 /* Generate the key for the unusable tree and the retry database. */
 
-key = string_sprintf("T:%s:%s", host->name, host->address);
+key = include_ip_address?
+      string_sprintf("T:%s:%s", host->name, host->address) :
+      string_sprintf("T:%s", host->name);
 
 /* Search the tree of unusable addresses. This is filled in when deliveries
 fail, because the retry database itself is not updated until the end of all
@@ -69,8 +78,9 @@ node = tree_search(tree_unusable, key);
 if (node != NULL)
   {
   DEBUG(9) debug_printf("found in tree of unusables\n");
-  host->status = hstatus_unusable;
-  host->why = node->data.val;
+  host->status = (node->data.val > 255)?
+    hstatus_unusable_expired : hstatus_unusable;
+  host->why = node->data.val & 255;
   return NULL;
   }
 
@@ -81,16 +91,16 @@ if ((dbm_file = dbfn_open("retry", O_RDONLY, &dbblock)) == NULL) return NULL;
 retry_record = dbfn_read(dbm_file, key);
 dbfn_close(dbm_file);
 
-/* If there's no retry record, free the key and return. */
+/* If there's no retry record, return NULL. */
 
 if (retry_record == NULL)
   {
   DEBUG(9) debug_printf("no retry record\n");
-  store_free(key);
   return NULL;
   }
 
-/* Set the status according to the retry data, and return the key. */
+/* Set the status according to the retry data, and return the key. If
+it is usable, set is_expired. */
 
 if (time(NULL) < retry_record->next_try && !deliver_force)
   {
@@ -99,6 +109,7 @@ if (time(NULL) < retry_record->next_try && !deliver_force)
   host->why = hwhy_retry;
   host->last_try = retry_record->last_try;
   }
+else *is_expired = retry_record->expired;
 
 return key;
 }
@@ -134,7 +145,7 @@ void
 retry_add_item(address_item *addr, char *type, BOOL use_at, char *key,
   BOOL delete)
 {
-retry_item *rti = store_malloc(sizeof(retry_item));
+retry_item *rti = store_get(sizeof(retry_item));
 rti->next = addr->retries;
 addr->retries = rti;
 rti->destination = string_sprintf("%s:%s%s%s", type,
@@ -286,9 +297,10 @@ to the failed chain if all their destinations have timed out. */
 for (i = 0; i < 3; i++)
   {
   address_item *endaddr, *addr;
+  address_item *last_first = NULL;
   address_item **paddr = (i==0)? addr_succeed :
     (i==1)? addr_failed : addr_defer;
-  BOOL defer = (i == 2);
+  address_item **saved_paddr = NULL;
 
   DEBUG(5) debug_printf("%s addresses:\n", (i == 0)? "Succeeded" : (i == 1)?
     "Failed" : "Deferred");
@@ -387,7 +399,7 @@ for (i = 0; i < 3; i++)
         retry_record = dbfn_read(dbm_file, rti->destination);
         if (retry_record == NULL)
           {
-          retry_record = store_malloc(sizeof(dbdata_retry) + message_length);
+          retry_record = store_get(sizeof(dbdata_retry) + message_length);
           message_space = message_length;
           retry_record->first_failed = now;
           retry_record->last_try = now;
@@ -497,7 +509,7 @@ for (i = 0; i < 3; i++)
 
         if (message_length > message_space)
           {
-          dbdata_retry *newr = store_malloc(sizeof(dbdata_retry) + message_length);
+          dbdata_retry *newr = store_get(sizeof(dbdata_retry) + message_length);
           memcpy(newr, retry_record, sizeof(dbdata_retry));
           retry_record = newr;
           }
@@ -529,25 +541,56 @@ for (i = 0; i < 3; i++)
     means of one or more retry items, and no actions were skipped because a
     retry time had not been reached, and no retry item is still active,
     move the address to the failed queue, and restart this loop without
-    updating paddr. */
+    updating paddr.
 
-    if (endaddr->retries != NULL && !endaddr->retry_skipped &&
-        timed_out && defer)
+    If there were several addresses batched in the same delivery, only the top
+    one will have retry items attached to it, but we want to handle all the
+    same. Each will have a pointer back to its "top" address, and they will
+    precede the item with the retries. We have saved information about them
+    in passing (below) so they can all be cut out at once. */
+
+    if (i == 2)   /* Handling defers */
       {
-      *paddr = endaddr->next;
-      endaddr->retry_timedout = TRUE;
-      endaddr->next = *addr_failed;
-      *addr_failed = endaddr;
-      endaddr->message = (endaddr->message == NULL)? "retry timeout exceeded" :
-        string_sprintf("%s: retry timeout exceeded", endaddr->message);
-      log_write(0, LOG_MAIN, "** %s%s%s%s: retry timeout exceeded",
-        endaddr->orig,
-       (endaddr->parent == NULL)? "" : " <",
-       (endaddr->parent == NULL)? "" : endaddr->parent->orig,
-       (endaddr->parent == NULL)? "" : ">");
+      if (endaddr->retries != NULL && !endaddr->retry_skipped && timed_out)
+        {
+        if (last_first == endaddr) paddr = saved_paddr;
+        addr = *paddr;
+        *paddr = endaddr->next;
+
+        endaddr->next = *addr_failed;
+        *addr_failed = addr;
+
+        for (;; addr = addr->next)
+          {
+          addr->retry_timedout = TRUE;
+          addr->message = (addr->message == NULL)? "retry timeout exceeded" :
+            string_sprintf("%s: retry timeout exceeded", addr->message);
+          log_write(0, LOG_MAIN, "** %s%s%s%s: retry timeout exceeded",
+            addr->orig,
+           (addr->parent == NULL)? "" : " <",
+           (addr->parent == NULL)? "" : addr->parent->orig,
+           (addr->parent == NULL)? "" : ">");
+
+          if (addr == endaddr) break;
+          }
+
+        continue;                       /* Restart from changed *paddr */
+        }
+
+      /* This address is to remain on the defer chain. If it has a "first"
+      pointer, save the pointer to it in case we want to fail the set of
+      addresses when we get to the first one. */
+
+      if (endaddr->first != last_first)
+        {
+        last_first = endaddr->first;
+        saved_paddr = paddr;
+        }
       }
 
-    else paddr = &(endaddr->next);    /* Advance to next address */
+    /* All cases (succeed, fail, defer left on queue) */
+
+    paddr = &(endaddr->next);         /* Advance to next address */
     }                                 /* Loop for all addresses  */
   }                                   /* Loop for succeed, fail, defer */
 
