@@ -29,7 +29,7 @@ static char *rrname[] = {
 *            Ensure an address is qualified      *
 *************************************************/
 
-/* If a change is made, the store used by the incoming address is freed.
+/*
 
 Arguments:
   s              address to check
@@ -41,14 +41,9 @@ Returns:         fully-qualified address
 char *
 rewrite_address_qualify(char *s, BOOL is_recipient)
 {
-if (parse_find_at(s) == NULL)
-  {
-  char *new = string_sprintf("%s@%s", s,
+return (parse_find_at(s) != NULL)? s :
+  string_sprintf("%s@%s", s,
     is_recipient? qualify_domain_recipient : qualify_domain_sender);
-  store_free(s);
-  s = new;
-  }
-return s;
 }
 
 
@@ -57,13 +52,12 @@ return s;
 *               Rewrite a single address         *
 *************************************************/
 
-/* The incoming address is NOT freed. The yield is the input address if there
-is no rewriting to be done. For source-routed addresses, it is only the final
-route-addr that gets considered. Assume the input is a valid address, except in
-the case of SMTP-time rewriting, which is handled specially. When this function
-is called while processing filter and forward files, the uid may be that of the
-user. Ensure it is reset while expanding a replacement, in case that involves
-file lookups.
+/* The yield is the input address if there is no rewriting to be done. For
+source-routed addresses, it is only the final route-addr that gets considered.
+Assume the input is a valid address, except in the case of SMTP-time rewriting,
+which is handled specially. When this function is called while processing
+filter and forward files, the uid may be that of the user. Ensure it is reset
+while expanding a replacement, in case that involves file lookups.
 
 Arguments:
   s           address to rewrite
@@ -109,7 +103,6 @@ for (rule = rewrite_rules;
   int count = 0;
   char *save_localpart, *save_domain;
   char *error, *new, *newparsed;
-  char buffer[256];
 
   /* Ensure that the flag matches the flags in the rule. */
 
@@ -138,9 +131,9 @@ for (rule = rewrite_rules;
     {
     /* If this is the first rule we are trying on this address, set domain
     to point past the last '@' in the address and user to the start of the
-    address or following the last ':' if source-routed. If a source routed
-    address does not have an @ after the : then rewrite it using the
-    immediately preceding domain. */
+    address or following the last ':' if source-routed. Since Exim 0.56, the
+    parsing has insisted that there be a domain after the ':' in a source-
+    routed address. */
 
     if (domain == NULL)
       {
@@ -153,30 +146,6 @@ for (rule = rewrite_rules;
       else
         {
         user = strchr(subject, ':') + 1;
-
-        /* The rewriting to ensure there is an @ after the : always produces
-        a string that is one character shorter than it was before. Hence we
-        can malloc a piece of store of length strlen(subject). */
-
-        if (domain < user)
-          {
-          char *new = store_malloc((int)strlen(subject));
-          if (domain == subject+1)
-            {
-            sprintf(new, "%s@%.*s", user, user-domain-1, domain);
-            user = subject = new;
-            }
-          else
-            {
-            sprintf(new, "%.*s:%s@%.*s", domain - subject - 2, subject,
-              user, user - domain - 1, domain);
-            subject = new;
-            user = strchr(subject, ':') + 1;
-            }
-          domain = strrchr(subject, '@') + 1;
-          DEBUG(9)
-            debug_printf("re-wrote source routed address as %s\n", subject);
-          }
         pre_length = user - subject;
         }
       }
@@ -213,18 +182,17 @@ for (rule = rewrite_rules;
     save_localpart = deliver_localpart;
     save_domain = deliver_domain;
 
-    /* Paranoia */
-    if (domain - user > sizeof(buffer))
-      log_write(0, LOG_MAIN|LOG_PANIC_DIE, "Attempt to re-write a local "
-        "part that is longer than %d characters (process %s): <%s>",
-        sizeof(buffer) - 1, process_info, s);
+    /* We have user pointing to "localpart@domain" and domain pointing to
+    the domain. Temporarily terminate the local part so that it can be
+    set up as an expansion variable */
 
-    sprintf(buffer, "%.*s", domain - user - 1, user);
-    deliver_localpart = buffer;
+    domain[-1] = 0;
+    deliver_localpart = user;
     deliver_domain = domain;
 
     new = expand_string(rule->replacement);
 
+    domain[-1] = '@';
     deliver_localpart = save_localpart;
     deliver_domain = save_domain;
     }
@@ -248,9 +216,8 @@ for (rule = rewrite_rules;
 
   if (pre_length != 0)
     {
-    char *fullnew = store_malloc(pre_length + (int)strlen(new) + 1);
+    char *fullnew = store_get(pre_length + (int)strlen(new) + 1);
     sprintf(fullnew, "%.*s%s", pre_length, subject, new);
-    store_free(new);
     new = fullnew;
     }
 
@@ -262,10 +229,15 @@ for (rule = rewrite_rules;
   newparsed = parse_extract_address(new, &error, &start, &end, &pdomain,
     FALSE);
 
-  if (newparsed == NULL || pdomain == 0)
+  if (newparsed == NULL)
     {
-    log_write(0, LOG_MAIN|LOG_PANIC, "Rewrite of %s yielded unqualified or "
-      "unparseable address \"%s\"", subject, new);
+    log_write(0, LOG_MAIN|LOG_PANIC, "Rewrite of %s yielded unparseable "
+      "address: %s in address %s", subject, error, new);
+    }
+  else if (pdomain == 0)
+    {
+    log_write(0, LOG_MAIN|LOG_PANIC, "Rewrite of %s yielded unqualified "
+      "address \"%s\"", subject, new);
     }
   else
     {
@@ -289,13 +261,10 @@ for (rule = rewrite_rules;
 
     if (whole != NULL && (flag & rewrite_all_headers) != 0)
       {
-      /* Current rule has the w flag set; free previous yield and subject
-      if we are on 2nd or subsequent rewrite, and set up new values. */
+      /* Current rule has the w flag set */
 
       if ((rule->flags & rewrite_whole) != 0)
         {
-        if (yield != s) store_free(yield);
-        if (subject != yield) store_free(subject);
         yield = new;
         yield_start = start;
         yield_end = end;
@@ -314,13 +283,10 @@ for (rule = rewrite_rules;
 
       else
         {
-        store_free(subject);
         subject = newparsed;
-        store_free(new);
         new = string_sprintf("%.*s%s%n%s",
            yield_start, yield, subject, &end, yield + yield_end);
         yield_end = end;
-        store_free(yield);
         yield = new;
         }
       }
@@ -332,9 +298,7 @@ for (rule = rewrite_rules;
     else
       {
       NEVER_WHOLE:
-      if (yield != s) store_free(yield);
       subject = yield = newparsed;
-      store_free(new);
       }
 
     DEBUG(2) debug_printf("\"%s\"\n", yield);
@@ -368,7 +332,6 @@ if (uid_set)
   DEBUG(2) debug_printf("restored euid=%d egid=%d\n", (int)geteuid(),
     (int)getegid());
   }
-if (subject != yield) store_free(subject);
 expand_nmax = -1;
 return yield;
 }
@@ -438,7 +401,9 @@ Arguments:
                 this domain into routed_new
   routed_new  new routed domain if routed_old is not NULL
 
-Returns:      NULL if header unchanged; otherwise the rewritten header
+Returns:      NULL if header unchanged; otherwise the rewritten header, which
+              has been inserted into the header chain following the old one,
+              which in turn has been marked "old"
 */
 
 static header_line *
@@ -528,10 +493,8 @@ while (*s != 0)
         }
       }
 
-    /* Free the original; if there isn't a new address, do no more on this
-    iteration. */
+    /* If there isn't a new address, do no more on this iteration. */
 
-    store_free(recipient);
     if (new != NULL)
       {
       recipient = new;
@@ -542,11 +505,10 @@ while (*s != 0)
 
 
   /* This is not a router-inspired rewrite. Ensure the address is fully
-  qualified if that is permitted. (Note: rewrite_address-qualify frees the
-  old store.) If an unqualified address was received from a host that isn't
-  listed, do not continue rewriting this address. Sender, From or Reply-to
-  headers are treated as senders, the rest as recipients. This matters only
-  when there are different qualify strings. */
+  qualified if that is permitted. If an unqualified address was received
+  from a host that isn't listed, do not continue rewriting this address.
+  Sender, From or Reply-to headers are treated as senders, the rest as
+  recipients. This matters only when there are different qualify strings. */
 
   else
     {
@@ -560,10 +522,7 @@ while (*s != 0)
 
     if (changed && ((is_recipient && !allow_unqualified_recipient) ||
                     (!is_recipient && !allow_unqualified_sender)))
-      {
-      store_free(recipient);
       continue;
-      }
     }
 
   /* If there are rewrite rules for this type of header, apply
@@ -600,9 +559,10 @@ while (*s != 0)
 
     header_line *prev = (newh == NULL)? h : newh;
     header_line *newnewh =
-      store_malloc(sizeof(header_line) + prev->slen - oldlen + newlen + 4);
-    char *newt = newnewh->text;
+      store_get(sizeof(header_line) + prev->slen - oldlen + newlen + 4);
+    char *newt = store_get(prev->slen - oldlen + newlen + 4);
 
+    newnewh->text = newt;
     newnewh->type = prev->type;
     newnewh->slen = prev->slen - oldlen + newlen;
 
@@ -656,17 +616,25 @@ while (*s != 0)
 
     DEBUG(5) debug_printf("remainder: %s", (*s == 0)? "\n" : s);
 
-    if (newh != NULL) store_free(newh);
     newh = newnewh;
-    if (new != recipient) store_free(new);
     }
 
-  store_free(recipient);
   }
 
 parse_allow_group = FALSE;  /* Reset group flags */
 parse_found_group = FALSE;
 
+/* If a rewrite happened, put the new header into the chain following the old
+one, and mark the old one as replaced. */
+
+if (newh != NULL)
+  {
+  newh->next = h->next;
+  if (newh->next == NULL) header_last = newh;
+  h->type = htype_old;
+  h->next = newh;
+  }
+
 return newh;
 }
 
