@@ -16,10 +16,20 @@
 data blocks and which therefore have the opt_public flag set. */
 
 optionlist optionlist_routers[] = {
+  { "*expand_group",      opt_stringptr | opt_hidden | opt_public,
+                 (void *)(offsetof(router_instance, expand_gid)) },
   { "*expand_transport",  opt_stringptr|opt_public|opt_hidden,
                  (void *)offsetof(router_instance, expand_transport) },
+  { "*expand_user",       opt_stringptr | opt_hidden | opt_public,
+                 (void *)(offsetof(router_instance, expand_uid)) },
+  { "*set_group",         opt_bool | opt_hidden | opt_public,
+                 (void *)(offsetof(router_instance, gid_set)) },
+  { "*set_user",          opt_bool | opt_hidden | opt_public,
+                 (void *)(offsetof(router_instance, uid_set)) },
   { "condition",          opt_stringptr|opt_public,
                  (void *)offsetof(router_instance, condition) },
+  { "debug_print",        opt_stringptr | opt_public,
+                 (void *)offsetof(router_instance, debug_string) },
   { "domains",            opt_stringptr|opt_public,
                  (void *)offsetof(router_instance, domains) },
   { "driver",             opt_stringptr|opt_public,
@@ -38,10 +48,16 @@ optionlist optionlist_routers[] = {
                  (void *)offsetof(router_instance, fail_verify_recipient) },
   { "fail_verify_sender", opt_bool|opt_public,
                  (void *)offsetof(router_instance, fail_verify_sender) },
+  { "fallback_hosts",     opt_stringptr|opt_public,
+                 (void *)offsetof(router_instance, fallback_hosts) },
+  { "group",              opt_expand_gid | opt_public,
+                 (void *)(offsetof(router_instance, gid)) },
   { "headers_add",        opt_stringptr|opt_public,
                  (void *)offsetof(router_instance, extra_headers) },
   { "headers_remove",     opt_stringptr|opt_public,
                  (void *)offsetof(router_instance, remove_headers) },
+  { "initgroups",         opt_bool | opt_public,
+                 (void *)(offsetof(router_instance, initgroups)) },
   { "local_parts",        opt_stringptr|opt_public,
                  (void *)offsetof(router_instance, local_parts) },
   { "more",               opt_bool|opt_public,
@@ -58,6 +74,8 @@ optionlist optionlist_routers[] = {
                  (void *)offsetof(router_instance, transport) },
   { "unseen",             opt_bool|opt_public,
                  (void *)offsetof(router_instance, unseen) },
+  { "user",               opt_expand_uid | opt_public,
+                 (void *)(offsetof(router_instance, uid)) },
   { "verify",             opt_bool_verify|opt_hidden|opt_public,
                  (void *)offsetof(router_instance, verify_sender) },
   { "verify_only",        opt_bool|opt_public,
@@ -129,6 +147,10 @@ for (r = routers; r != NULL; r = r->next)
     }
   else log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s router:\n  "
       "%s is not valid for the self option", r->name, s);
+
+  /* Build a host list if fallback hosts is set */
+
+  host_build_hostlist(&(r->fallback_hostlist), r->fallback_hosts);
   }
 }
 
@@ -154,6 +176,59 @@ for (r = routers; r != NULL; r = r->next)
 
 
 
+/*************************************************
+*        Queue address for transport             *
+*************************************************/
+
+/* This function is called to put an address onto the local or remote transport
+queue, as appropriate. The same action is required by some (but not all) of the
+directors, and so they use this too. The fields in the router block that are
+used by this function have identical counterparts in the director block, to
+enable this to work. The generic uid/gid options are inspected and put into the
+address if they are set. For a remote transport, if there are fallback hosts,
+they are added to the address.
+
+When the driver is for verifying only, a transport need not be set, in which
+case it doesn't actually matter which queue the address gets put on.
+
+Arguments:
+  addr          the address, with the transport field set (if not verify only)
+  addr_local    the local transport chain
+  addr_remote   the remote transport chain,
+  rblock        the router block
+  type          "director" or "router"
+
+Returns:        FALSE on error
+*/
+
+BOOL
+route_queue(address_item *addr, address_item **addr_local,
+  address_item **addr_remote, router_instance *rblock, char *type)
+{
+if (addr->transport != NULL && addr->transport->info->local)
+  {
+  ugid_block ugid;
+  if (!direct_get_ugid((director_instance *)rblock, type, addr, &ugid))
+   return FALSE;
+  direct_set_ugid(addr, &ugid);
+  addr->next = *addr_local;
+  *addr_local = addr;
+  }
+else
+  {
+  addr->next = *addr_remote;
+  *addr_remote = addr;
+  addr->fallback_hosts = rblock->fallback_hostlist;
+  }
+
+DEBUG(2) debug_printf("queued for %s transport: local_part=%s domain=%s\n",
+  (addr->transport == NULL)? "<unset>" : addr->transport->name,
+  addr->local_part, addr->domain);
+
+return TRUE;
+}
+
+
 
 /*************************************************
 *             Set transport                      *
@@ -196,7 +271,6 @@ for (tp = transports; tp != NULL; tp = tp->next)
     {
     DEBUG(9) debug_printf("dynamic transport \"%s\" set\n", ss);
     dr->transport = tp;
-    store_free(ss);
     return TRUE;
     }
   }
@@ -235,20 +309,10 @@ route_skip_driver(router_instance *dr, char *local_part, char *domain,
   int verify, BOOL *pmore, char *driver_type)
 {
 
-/* Variables to hold data from lookup of a domain or local part. Free
-previous value and set empty by default. */
+/* Variables to hold data from lookup of a domain or local part. */
 
-if (domain_data != NULL)
-  {
-  store_free(domain_data);
-  domain_data = NULL;
-  }
-
-if (local_part_data != NULL)
-  {
-  store_free(local_part_data);
-  local_part_data = NULL;
-  }
+domain_data = NULL;
+local_part_data = NULL;
 
 /* Skip this driver if not verifying and it has verify_only set */
 
@@ -417,11 +481,12 @@ for (r = routers; more && r != NULL; r = r->next)
   if (loop_detected) continue;
 
   /* Before calling the router, set up values that may be used in string
-  expansions. Afterwards unset them, just in case. Also, unset the rewrite
-  flag; routers must explicitly request rewriting. */
+  expansions and output any debugging string. Afterwards unset them, just in
+  case. Also, unset the rewrite flag; routers must explicitly request
+  rewriting. */
 
-  deliver_domain = addr->route_domain;
-  deliver_localpart = addr->local_part;
+  deliver_set_expansions(addr);
+  debug_print_string(r->debug_string);
   addr->rewrite_headers = FALSE;
 
   /* Do file existence tests - must do after setting expansion values
@@ -480,8 +545,7 @@ for (r = routers; more && r != NULL; r = r->next)
     break;
     }
 
-  deliver_domain = NULL;
-  deliver_localpart = NULL;
+  deliver_set_expansions(NULL);
 
   /* If succeeded while verifying but fail_verify is set, convert into
   a failure. */
@@ -500,7 +564,7 @@ for (r = routers; more && r != NULL; r = r->next)
 
   if (yield == OK && r->unseen)
     {
-    address_item *new = deliver_make_addr(string_copy(addr->orig));
+    address_item *new = deliver_make_addr(addr->orig, TRUE);
     new->parent = addr;
     new->ignore_error |= addr->ignore_error;
     new->errors_address = addr->errors_address;
@@ -582,28 +646,19 @@ for (r = routers; more && r != NULL; r = r->next)
 
     /* Now the headers, except when verifying, when there aren't any! The
     first header is always "Received:" so we can skip it. There may not
-    always be headers, however - in "-bt" mode there won't be any. */
+    always be headers, however. In "-bt" mode for instance, there won't be
+    any. When a header gets rewritten, set the header_changed flag to force the
+    spool header file to be rewritten. */
 
     if (verify == v_none && header_list != NULL)
       {
       header_line *h;
       for (h = header_list->next; h != NULL; h = h->next)
         {
-        header_line *newh = rewrite_header(h, old_route_domain,
-          addr->route_domain);
-
-        /* The old header gets its type set to "old", and the new one is
-        inserted in the chain immediately following. The header_changed
-        flag is set to force the spool header file to be rewritten. */
-
+        header_line *newh =
+          rewrite_header(h, old_route_domain, addr->route_domain);
         if (newh != NULL)
           {
-          h->type = htype_old;
-          newh->prev = h;
-          newh->next = h->next;
-          h->next = newh;
-          if (newh->next == NULL) header_last = newh;
-            else newh->next->prev = newh;
           h = newh;
           header_changed = TRUE;
           }
@@ -673,21 +728,8 @@ for (r = routers; more && r != NULL; r = r->next)
   return OK;
   }
 
-/* Ensure store for domain and local part data is freed */
-
-if (domain_data != NULL)
-  {
-  store_free(domain_data);
-  domain_data = NULL;
-  }
-
-if (local_part_data != NULL)
-  {
-  store_free(local_part_data);
-  local_part_data = NULL;
-  }
-
-/* No routers accepted this address; fail it. */
+/* No routers accepted this address; fail it, with a default message if nothing
+is set. */
 
 if (addr->message == NULL)
   addr->message = string_sprintf("unrouteable mail domain \"%s\"",
