@@ -248,10 +248,9 @@ domainlist_router_entry(
   BOOL verify)                    /* TRUE if verifying */
 {
 BOOL matched = FALSE;
-BOOL free_routelist_item = FALSE;
 BOOL individual_transport_set = FALSE;
 int lookup_type, rc;
-host_item *h, *next_h, *prev, **hchain;
+host_item *h, *next_h, *prev;
 char *routelist_item = NULL;
 char *domain, *hostlist, *newhostlist, *options, *name, *listptr;
 domainlist_router_options_block *ob =
@@ -418,7 +417,6 @@ if (!matched && (ob->route_file != NULL || ob->route_query != NULL ||
   if (routelist_item != NULL)
     {
     (void) parse_routelist_item(routelist_item, NULL, &hostlist, &options);
-    free_routelist_item = TRUE;
     matched = TRUE;
     expand_nmax = expand_setup;
     }
@@ -449,7 +447,7 @@ it must not be present. */
 
 DEBUG(9) debug_printf("original hostlist=%s options=%s\n", hostlist, options);
 
-newhostlist = expand_string(hostlist);
+newhostlist = expand_string_copy(hostlist);
 expand_nmax = -1;
 
 /* If the expansion was forced to fail, just fail the router. Otherwise
@@ -503,8 +501,6 @@ while (*options != 0)
       log_write(0, LOG_MAIN, "Error in %s router: %s", rblock->name, s);
       addr->message = string_sprintf("error in router: %s", s);
       addr->special_action = SPECIAL_FREEZE;
-      if (free_routelist_item) store_free(routelist_item);
-      store_free(hostlist);
       return ERROR;
       }
     }
@@ -535,6 +531,8 @@ be used, for example, as an expansion in a file name). */
 
 if (transport != NULL && transport->info->local)
   {
+  ugid_block ugid;
+
   if (hostlist[0] != 0)
     {
     host_item *h;
@@ -546,12 +544,10 @@ if (transport != NULL && transport->info->local)
       addr->message = "error in router: attempt to pass more than one host "
         "name to a local transport";
       addr->special_action = SPECIAL_FREEZE;
-      if (free_routelist_item) store_free(routelist_item);
-      store_free(hostlist);
       return ERROR;
       }
 
-    addr->host_list = h = store_malloc(sizeof(host_item));
+    addr->host_list = h = store_get(sizeof(host_item));
     h->name = string_copy(hostlist);
     h->address = NULL;
     h->mx = -1;
@@ -562,15 +558,12 @@ if (transport != NULL && transport->info->local)
     }
 
   /* There is nothing more to do other than to queue the address for the
-  local transport, remembering to free the route item if it came from a file
-  lookup. */
+  local transport, filling in any uid/gid. This can be done by the common
+  route_queue() function. */
 
-  if (free_routelist_item) store_free(routelist_item);
-  store_free(hostlist);
   addr->transport = transport;
-  addr->next = *addr_local;
-  *addr_local = addr;
-  return OK;
+  return route_queue(addr, addr_local, addr_remote, rblock, "router")?
+    OK : ERROR;
   }
 
 
@@ -584,8 +577,6 @@ if (hostlist[0] == 0)
   addr->message =
     string_sprintf("error in router: no host(s) specified for %s", domain);
   addr->special_action = SPECIAL_FREEZE;
-  if (free_routelist_item) store_free(routelist_item);
-  store_free(hostlist);
   return ERROR;
   }
 
@@ -602,13 +593,10 @@ if (lookup_type == lk_pass)
     addr->message = string_sprintf("error in router: %s", s);
     log_write(0, LOG_MAIN, "Error in %s router: %s", rblock->name, s);
     addr->special_action = SPECIAL_FREEZE;
-    if (free_routelist_item) store_free(routelist_item);
-    store_free(hostlist);
     return ERROR;
     }
 
-  addr->route_domain = hostlist;  /* and don't free it! */
-  if (free_routelist_item) store_free(routelist_item);
+  addr->route_domain = hostlist;
   return PASS;
   }
 
@@ -617,34 +605,7 @@ if (lookup_type == lk_pass)
 for the list of configured hosts, and then finding their addresses. Typically
 there is only one host. */
 
-hchain = &(addr->host_list);
-listptr = hostlist;
-
-for (name = string_nextinlist(&listptr, ':', route_buffer,
-       sizeof(route_buffer));
-     name != NULL;
-     name = string_nextinlist(&listptr, ':', route_buffer,
-       sizeof(route_buffer)))
-  {
-  h = store_malloc(sizeof(host_item));
-  h->name = string_copy(name);
-  h->address = NULL;
-  h->mx = -1;
-  h->status = hstatus_unknown;
-  h->why = hwhy_unknown;
-  h->last_try = 0;
-
-  h->next = NULL;
-  *hchain = h;
-  hchain = &(h->next);
-  }
-
-
-/* Free the route item if it came from a file lookup, and free the textual
-hostlist. */
-
-if (free_routelist_item) store_free(routelist_item);
-store_free(hostlist);
+host_build_hostlist(&(addr->host_list), hostlist);
 
 
 /* Look up each host address. A lookup may add additional items into the chain
@@ -664,7 +625,7 @@ for (h = addr->host_list; h != NULL; prev = h, h = next_h)
   if (h->address != NULL) continue;
 
   if (lookup_type == lk_byname)
-    rc = host_find_byname(h, &canonical_name);
+    rc = host_find_byname(h, &canonical_name, FALSE);
   else
     rc = host_find_bydns(h, (lookup_type == lk_bydns_mx),
       (lookup_type == lk_bydns_a), FALSE, FALSE, &canonical_name,
@@ -738,6 +699,7 @@ if (transport == NULL)
 
 addr->transport = transport;
 addr->next = *addr_remote;
+addr->fallback_hosts = rblock->fallback_hostlist;
 *addr_remote = addr;
 
 return OK;
