@@ -101,7 +101,7 @@ if (!string_is_ip_address(domain+1))
 
 /* Set up a host item */
 
-addr->host_list = h = store_malloc(sizeof(host_item));
+addr->host_list = h = store_get(sizeof(host_item));
 
 h->next = NULL;
 h->address = string_copy(domain+1);
@@ -139,21 +139,12 @@ if (rc != OK) return rc;
 /* Fill in the transport, queue the address for local or remote delivery, and
 yield success. For local delivery, of course, the IP address won't be used. If
 just verifying, there need not be a transport, in which case it doesn't matter
-which queue we put the address on. */
+which queue we put the address on. This is all now handled by the route_queue()
+function. */
 
 addr->transport = transport;
-if (transport != NULL && addr->transport->info->local)
-  {
-  addr->next = *addr_local;
-  *addr_local = addr;
-  }
-else
-  {
-  addr->next = *addr_remote;
-  *addr_remote = addr;
-  }
-
-return OK;
+return route_queue(addr, addr_local, addr_remote, rblock, "router")?
+  OK : ERROR;
 }
 
 /* End of routers/ipliteral.c */
