@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 1997 */
+/* Copyright (c) University of Cambridge 1995 - 1998 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 
@@ -26,7 +26,7 @@ extern ipliteral_router_options_block ipliteral_router_option_defaults;
 /* The main and initialization entry points for the router */
 
 extern int ipliteral_router_entry(router_instance *, address_item *,
-  address_item **, address_item **);
+  address_item **, address_item **, BOOL);
 
 extern void ipliteral_router_init(router_instance *);
 
