@@ -128,7 +128,7 @@ char *query = NULL;
 char reply[256];
 char *hostname, *reroute, *domain, *listptr;
 char host_buffer[256];
-host_item *host = store_malloc(sizeof(host_item));
+host_item *host = store_get(sizeof(host_item));
 iplookup_router_options_block *ob =
   (iplookup_router_options_block *)(rblock->options_block);
 pcre *re = ob->re_response_pattern;
@@ -183,7 +183,7 @@ for (hostname = string_nextinlist(&listptr, ':', host_buffer,
   host->next = NULL;
 
   if (string_is_ip_address(host->name)) host->address = host->name;
-    else if (host_find_byname(host, NULL) == HOST_FIND_FAILED) continue;
+    else if (host_find_byname(host, NULL, FALSE) == HOST_FIND_FAILED) continue;
 
   /* Loop for possible multiple IP addresses for the given name. */
 
@@ -251,7 +251,7 @@ for (hostname = string_nextinlist(&listptr, ':', host_buffer,
     works (the router will timeout later on the read call). */
 
     sigalrm_seen = FALSE;
-    signal(SIGALRM, sigalrm_handler);
+    os_non_restarting_signal(SIGALRM, sigalrm_handler);
     alarm(ob->timeout);
 
     rc = connect(query_socket, (struct sockaddr *)&sin, sizeof(sin));
@@ -349,9 +349,6 @@ for (hostname = string_nextinlist(&listptr, ':', host_buffer,
   if (h != NULL) break;
   }
 
-/* Free the query. */
-
-if (query != NULL) store_free(query);
 
 /* If hostname is NULL, we have failed to find any host, or failed to
 connect to any of the IP addresses, or timed out while reading or writing to
@@ -473,9 +470,7 @@ pass back this address to the directors. */
 if (match_isinlist(domain, local_domains, &re_local_domains, TRUE))
   yield = ISLOCAL;
 
-/* Free dynamic store, cancel numerical expansion variables, and return */
-
-if (reroute != reply) store_free(reroute);
+/* Cancel numerical expansion variables, and return */
 
 expand_nmax = -1;
 return yield;
