@@ -15,19 +15,23 @@
 /* Options specific to the lookuphost router. */
 
 optionlist lookuphost_router_options[] = {
-  { "gethostbyname",   opt_bool,
+  { "check_secondary_mx", opt_bool,
+      (void *)(offsetof(lookuphost_router_options_block, check_secondary_mx)) },
+  { "gethostbyname",      opt_bool,
       (void *)(offsetof(lookuphost_router_options_block, gethostbyname)) },
   { "mx_domains",         opt_stringptr,
       (void *)(offsetof(lookuphost_router_options_block, mx_domains)) },
-  { "non_mx_domains",     opt_stringptr,
-      (void *)(offsetof(lookuphost_router_options_block, non_mx_domains)) },
-  { "qualify_single",  opt_bool,
+  { "mx_domains_except",  opt_stringptr,
+      (void *)(offsetof(lookuphost_router_options_block, mx_domains_except)) },
+  { "non_mx_domains",     opt_stringptr|opt_hidden,
+      (void *)(offsetof(lookuphost_router_options_block, mx_domains_except)) },
+  { "qualify_single",     opt_bool,
       (void *)(offsetof(lookuphost_router_options_block, qualify_single)) },
-  { "rewrite_headers", opt_bool,
+  { "rewrite_headers",    opt_bool,
       (void *)(offsetof(lookuphost_router_options_block, rewrite_headers)) },
-  { "search_parents",  opt_bool,
+  { "search_parents",     opt_bool,
       (void *)(offsetof(lookuphost_router_options_block, search_parents)) },
-  { "widen_domains",   opt_stringptr,
+  { "widen_domains",      opt_stringptr,
       (void *)(offsetof(lookuphost_router_options_block, widen_domains)) }
 };
 
@@ -40,15 +44,16 @@ int lookuphost_router_options_count =
 /* Default private options block for the lookuphost router. */
 
 lookuphost_router_options_block lookuphost_router_option_defaults = {
+  FALSE,           /* check_secondary_mx */
   FALSE,           /* gethostbyname */
   TRUE,            /* qualify_single */
   FALSE,           /* search_parents */
   TRUE,            /* rewrite_headers */
   NULL,            /* widen_domains */
   NULL,            /* mx_domains */
-  NULL,            /* non_mx_domains */
+  NULL,            /* mx_domains_except */
   NULL,            /* re_mx_domains */
-  NULL             /* re_non_mx_domains */
+  NULL             /* re_mx_domains_except */
 };
 
 
@@ -131,7 +136,7 @@ if (ob->widen_domains != NULL)
 for (;;)
   {
   if (ob->gethostbyname)
-    rc = host_find_byname(&h, &fully_qualified_name);
+    rc = host_find_byname(&h, &fully_qualified_name, FALSE);
 
   /* Unfortunately, we cannot set the mx_only option in advance, because the
   DNS lookup may extend an unqualified name. Therefore, we must do the test
@@ -147,9 +152,9 @@ for (;;)
         h.mx < 0 &&                            /* Found A records only */
         addr->local_part[0] != ',' &&          /* Not source routed */
         addr->local_part[0] != ':' &&
-        (ob->non_mx_domains == NULL ||         /* Not in non_mx_domains list */
-        !match_isinlist(fully_qualified_name, ob->non_mx_domains,
-        &(ob->re_non_mx_domains), TRUE)) &&
+        (ob->mx_domains_except == NULL ||      /* Not in mx_domains_except list */
+        !match_isinlist(fully_qualified_name, ob->mx_domains_except,
+        &(ob->re_mx_domains_except), TRUE)) &&
         (ob->mx_domains != NULL &&             /* In mx-domains list */
         match_isinlist(fully_qualified_name, ob->mx_domains,
         &(ob->re_mx_domains), TRUE)) &&
@@ -184,11 +189,18 @@ for (;;)
   otherwise "unknown mail domain" is used, which is confusing. Also, in this
   case don't do the widening. We need check only the first host to see if
   its MX has been filled in, but there is no address, because if there were
-  any usable addresses returned, we would not have had HOST_FIND_FAILED. */
+  any usable addresses returned, we would not have had HOST_FIND_FAILED.
+
+  As a common cause of this problem is MX records with IP addresses on the
+  RHS, give a special message in this case. */
 
   if (h.mx >= 0 && h.address == NULL)
     {
     addr->message = "all relevant MX records point to non-existent hosts";
+    if (string_is_ip_address(h.name))
+      addr->message = string_sprintf("%s: it appears that the DNS operator for "
+         "this domain has installed an invalid MX record with an IP address "
+         "instead of a domain name on the right hand side", addr->message);
     return FAIL;
     }
 
@@ -247,6 +259,14 @@ if (rc == HOST_FOUND_LOCAL)
   if (rc != OK) return rc;
   }
 
+/* Otherwise, insist on being a secondary MX if so configured */
+
+else if (ob->check_secondary_mx && !addr->local_host_removed)
+  {
+  DEBUG(2) debug_printf("check_secondary_mx set and local host not secondary\n");
+  return FAIL;
+  }
+
 /* Set up the errors address, if any. The function uses only fields that
 are in the same place in director and router blocks. */
 
@@ -264,27 +284,17 @@ if (rc != OK) return rc;
 /* Get store in which to preserve the original host item, chained on
 to the address. */
 
-addr->host_list = store_malloc(sizeof(host_item));
+addr->host_list = store_get(sizeof(host_item));
 addr->host_list[0] = h;
 
 /* Fill in the transport, queue the address for local or remote delivery, and
 yield success. If just verifying, there may not be a transport, in which case
-it doesn't matter which queue is used. */
+it doesn't matter which queue is used. This is all now handled by the
+route_queue() function. */
 
 addr->transport = rblock->transport;
-if (addr->transport != NULL && addr->transport->info->local)
-  {
-  addr->next = *addr_local;
-  *addr_local = addr;
-  }
-else
-  {
-  addr->next = *addr_remote;
-  *addr_remote = addr;
-  }
-
-DEBUG(9) debug_printf("%s router succeeded\n", rblock->name);
-return OK;
+return route_queue(addr, addr_local, addr_remote, rblock, "router")?
+  OK : ERROR;
 }
 
 /* End of routers/lookuphost.c */
