@@ -8,15 +8,16 @@
 /* Private structure for the private options. */
 
 typedef struct {
+  BOOL check_secondary_mx;
   BOOL gethostbyname;
   BOOL qualify_single;
   BOOL search_parents;
   BOOL rewrite_headers;
   char *widen_domains;
   char *mx_domains;
-  char *non_mx_domains;
+  char *mx_domains_except;
   re_block *re_mx_domains;
-  re_block *re_non_mx_domains;
+  re_block *re_mx_domains_except;
 } lookuphost_router_options_block;
 
 /* Data for reading the private options. */
