@@ -8,7 +8,12 @@
 /* A set of functions to search databases in various formats. An open
 database is represented by a void * value which is returned from a lookup-
 specific "open" function. These are now all held in individual modules in the
-lookups subdirectory and the functions here form a generic interface. */
+lookups subdirectory and the functions here form a generic interface.
+
+Caching is used to improve performance. Open files are cached until a tidyup
+function is called, and for each file the result of the last lookup is cached.
+All this data is held in malloc'd store so as to be independent of the stacking
+store pools. */
 
 #include "exim.h"
 
@@ -17,13 +22,6 @@ lookups subdirectory and the functions here form a generic interface. */
 
 tree_node *search_tree = NULL;
 
-/* Place to cache lookups for efficiency. */
-
-static char cached_filename[256] = "";
-static char cached_keystring[256] = "";
-static int  cached_type = -1;
-static char *cached_data = NULL;
-
 
 
 /*************************************************
@@ -147,12 +145,12 @@ return 0;
 *               Release cached resources         *
 *************************************************/
 
-/* When search_open is called for single-key lookup types, it caches the file
-that it opens in search_tree. The name of the tree node is a concatenation of
-the search type with the file name. These files are closed only when this
-tidyup routine is called, typically at the end of sections of code where a
-number of lookups might occur. There is also a general tidyup function which is
-called for the lookup driver, if it exists.
+/* When search_open is called it caches the "file" that it opens in
+search_tree. The name of the tree node is a concatenation of the search type
+with the file name. For query-style lookups, the file name is empty. Real
+files are closed only when this tidyup routine is called, typically at the end
+of sections of code where a number of lookups might occur. There is also a
+general tidyup function which is called for the lookup driver, if it exists.
 
 First, there is an internal, recursive subroutine.
 
@@ -163,15 +161,22 @@ Returns:    nothing
 static void
 tidyup_subtree(tree_node *t)
 {
-int search_type;
+int search_type = t->name[0] - '0';
+search_cache *c = (search_cache *)(t->data.ptr);
+
 if (t->left != NULL) tidyup_subtree(t->left);
 if (t->right != NULL) tidyup_subtree(t->right);
-search_type = t->name[0] - '0';
+
 if (lookup_list[search_type].close != NULL)
-  lookup_list[search_type].close(t->data.ptr);
+  lookup_list[search_type].close(c->handle);
+
+if (c->key != NULL) store_free(c->key);
+if (c->data != NULL) store_free(c->data);
+store_free(c);
 store_free(t);
 }
 
+
 /* The external entry point
 
 Argument: none
@@ -185,7 +190,7 @@ lookup_info *li;
 
 DEBUG(9) debug_printf("search_tidyup called\n");
 
-/* Close individually each cached handle */
+/* Close individually each cached open file. */
 
 if (search_tree != NULL)
   {
@@ -199,15 +204,6 @@ for (li = lookup_list; li->name[0] != 0; li++)
   {
   if (li->tidy != NULL) (li->tidy)();
   }
-
-/* Forget any cached lookup data */
-
-if (cached_data != NULL)
-  {
-  store_free(cached_data);
-  cached_data = NULL;
-  }
-cached_type = -1;
 }
 
 
@@ -235,10 +231,10 @@ files, or for NIS. (2) In the "query" style, no "filename" is given; instead
 just a single query string is passed. This applies to multiple-key lookup
 types such as NIS+.
 
-Before opening, scan the tree of cached files to see if this "file" is
-already open for the correct search type. If so, return the saved handle. If
-not, put the handle in the tree for possible subsequent use. See
-search_tidyup above for closing the cached files.
+Before opening, scan the tree of cached files to see if this file is already
+open for the correct search type. If so, return the saved handle. If not, put
+the handle in the tree for possible subsequent use. See search_tidyup above for
+closing the cached files.
 
 Arguments:
   filename       the name of the file for single-key+file style lookups,
@@ -253,7 +249,9 @@ Arguments:
   message        points to a char * which can be set to point to an error
                  message when something goes wrong
 
-Returns:         an identifying handle for the open database
+Returns:         an identifying handle for the open database; for single-key
+                 search types this is the pointer to the tree block in the
+                 cache of open files
 */
 
 void *
@@ -262,6 +260,7 @@ search_open(char *filename, int search_type, int modemask, int *owners,
 {
 void *handle;
 tree_node *t;
+search_cache *c;
 char keybuffer[256];
 
 DEBUG(9) debug_printf("search_open: %s \"%s\"\n",
@@ -269,44 +268,43 @@ DEBUG(9) debug_printf("search_open: %s \"%s\"\n",
   (filename == NULL)? "NULL" : filename);
 
 /* See if we already have this open for this type of search, and if so,
-pass back the previous handle. The key for the tree node is the search type
-plus '0' concatenated with the file name. */
+pass back the tree block as the handle. The key for the tree node is the search
+type plus '0' concatenated with the file name. */
 
-if (filename != NULL)
+sprintf(keybuffer, "%c%.254s", search_type + '0',
+  (filename == NULL)? "" : filename);
+
+if ((t = tree_search(search_tree, keybuffer)) != NULL)
   {
-  sprintf(keybuffer, "%c%.254s", search_type + '0', filename);
-  if ((t = tree_search(search_tree, keybuffer)) != NULL)
-    {
-    DEBUG(9) debug_printf("  cached\n");
-    return t->data.ptr;
-    }
+  DEBUG(9) debug_printf("  cached\n");
+  return t;
   }
 
-/* Otherwise, open the file or database - each search type has its own
-code, which is now split off into separately compiled modules. If opening is
-successful for a type that has a filename, call the file-checking function if
-there is one, and if all is still well, enter the file name, with type
-prefixed, into the tree. */
+/* Otherwise, open the file or database - each search type has its own code,
+which is now split off into separately compiled modules. If opening is
+successful, call the file-checking function if there is one, and if all is
+still well, enter the open database into the tree. */
 
 handle = lookup_list[search_type].open(filename, message);
+if (handle == NULL) return handle;
 
-if (handle != NULL && filename != NULL)
+if (lookup_list[search_type].check != NULL &&
+   !lookup_list[search_type].check(handle, filename, modemask, owners,
+     owngroups, message))
   {
-  if (lookup_list[search_type].check != NULL &&
-     !lookup_list[search_type].check(handle, filename, modemask, owners,
-       owngroups, message))
-    {
-    lookup_list[search_type].close(handle);
-    return NULL;
-    }
-
-  t = store_malloc(sizeof(tree_node) + (int)strlen(keybuffer));
-  strcpy(t->name, keybuffer);
-  t->data.ptr = handle;
-  tree_insertnode(&search_tree, t);
+  lookup_list[search_type].close(handle);
+  return NULL;
   }
 
-return handle;
+t = store_malloc(sizeof(tree_node) + (int)strlen(keybuffer));
+strcpy(t->name, keybuffer);
+
+t->data.ptr = c = store_malloc(sizeof(search_cache));
+c->handle = handle;
+c->key = NULL;
+c->data = NULL;
+tree_insertnode(&search_tree, t);
+return t;
 }
 
 
@@ -317,10 +315,10 @@ return handle;
 *  Internal function: Find one item in database  *
 *************************************************/
 
-/* The answer is always put into dynamic store. If the key contains a colon,
-then it is treated as a double key: the first part is the key for the record
-in the file, and the remainder is a subkey that is used to extract a subfield
-from the main data. Subfields are specified as subkey=value in the records.
+/* The answer is always put into ???. If the key contains a colon, then it is
+treated as a double key: the first part is the key for the record in the file,
+and the remainder is a subkey that is used to extract a subfield from the main
+data. Subfields are specified as subkey=value in the records.
 
 The last lookup is cached by file name and key - using the handle is no good
 as it isn't unique enough.
@@ -344,6 +342,7 @@ internal_search_find(void *handle, char *filename, char *keystring, int type,
   char **errmsg)
 {
 int length;
+search_cache *c = (search_cache *)(((tree_node *)handle)->data.ptr);
 char *subkey = NULL;
 char *colon = NULL;
 
@@ -363,29 +362,24 @@ if (filename != NULL)
     }
   }
 
-/* Turn a NULL filename into an empty string so that it gets copied and
-compared in the cache without special code. */
-
-else filename = "";
-
 DEBUG(9) debug_printf("internal_search_find: file=\"%s\"\n  "
-  "type=%s key=\"%s\"\n", filename, lookup_list[type].name, keystring);
+  "type=%s key=\"%s\"\n", (filename == NULL)? "NULL" : filename,
+  lookup_list[type].name, keystring);
 
 /* Insurance. If the keystring is empty, just fail. */
 
 if (keystring[0] == 0) return NULL;
 
-/* If there is no cached record of the right type, or if this lookup is for a
-new key, or in a different file, we must search the file and set up new cached
-data. */
+/* Look up the data for the key, unless it is already in the cache for this
+file. */
 
-if (type != cached_type ||
-    strcmp(keystring, cached_keystring) != 0 ||
-    strcmp(filename, cached_filename) != 0)
+if (c->key == NULL || strcmp(c->key, keystring) != 0)
   {
+  char *data = NULL;
+
   DEBUG(9)
     {
-    if (filename[0] != 0)
+    if (filename != NULL)
       debug_printf("file lookup required for %s%s%s\n  in %s\n",
         keystring,
         (subkey == NULL)? "" : ":",
@@ -395,35 +389,37 @@ if (type != cached_type ||
       debug_printf("database lookup required for %s\n", keystring);
     }
 
-  if (cached_data != NULL)
-    {
-    store_free(cached_data);    /* free previous */
-    cached_data = NULL;
-    }
-  cached_type = -1;             /* for FAIL or DEFER */
+  /* Free any previous cache */
+
+  if (c->key != NULL) store_free(c->key);
+  if (c->data != NULL) store_free(c->data);
+  c->key = c->data = NULL;
 
   /* Call the code for the different kinds of search. DEFER is handled
   like FAIL, except that search_find_defer is set so the caller can
   distinguish if necessary. */
 
   length = (int)strlen(keystring);
-  if (lookup_list[type].find(handle, filename, keystring, length,
-      &cached_data, errmsg) == DEFER)
+  if (lookup_list[type].find(c->handle, filename, keystring, length,
+      &data, errmsg) == DEFER)
     {
     search_find_defer = TRUE;
     }
 
-  /* A record that has been found is now in cached_data, which is either NULL
-  or points to a bit of dynamic store. Remember the file name, main key, and
-  lookup type, but only if the file name and main key are < 256 characters
-  long (the size of the cache slots). Longer keys are presumably exceedingly
-  rare... */
+  /* A record that has been found is now in data, which is either NULL
+  or points to a bit of dynamic store. Cache the result of the lookup. */
 
-  else if ((int)strlen(filename) < 256 && length < 256)
+  else
     {
-    strcpy(cached_filename, filename);
-    strcpy(cached_keystring, keystring);
-    cached_type = type;
+    int len = (int)strlen(keystring) + 1;
+    c->key = store_malloc(len);
+    memcpy(c->key, keystring, len);
+    if (data != NULL)
+      {
+      int len = (int)strlen(data) + 1;
+      c->data = store_malloc(len);
+      memcpy(c->data, data, len);
+      }
     }
   }
 
@@ -431,19 +427,19 @@ else DEBUG(9) debug_printf("cached data used for lookup of %s%s%s%s%s\n",
   keystring,
   (subkey == NULL)? "" : ":",
   (subkey == NULL)? "" : subkey,
-  (filename[0] == 0)? "" : "\n  in ",
-  (filename[0] == 0)? "" : filename);
+  (filename == NULL)? "" : "\n  in ",
+  (filename == NULL)? "" : filename);
 
 /* Debug: output the main record */
 
 DEBUG(9)
   {
-  if (cached_data == NULL)
+  if (c->data == NULL)
     {
     if (search_find_defer) debug_printf("lookup deferred\n");
       else debug_printf("lookup failed\n");
     }
-  else debug_printf("lookup yielded: %s\n", cached_data);
+  else debug_printf("lookup yielded: %s\n", c->data);
   }
 
 /* Put back the colon if it was overwritten */
@@ -451,12 +447,11 @@ DEBUG(9)
 if (colon != NULL) *colon = ':';
 
 /* If we have found data, pick out the subfield if required. Otherwise
-make a fresh copy of the whole cached string. */
+return the whole cached string. */
 
-if (subkey != NULL && cached_data != NULL)
-  return expand_getkeyed(subkey, cached_data);
-    else if (cached_data != NULL) return string_copy(cached_data);
-      else return NULL;
+return (c->data == NULL)? NULL :
+  (subkey != NULL)?
+    expand_getkeyed(subkey, c->data) : string_copy(c->data);
 }
 
 
@@ -527,8 +522,7 @@ else partial = -1;
 /* First of all, try to match the key string verbatim. If matched a complete
 entry but could have been partial, flag to set up variables. */
 
-yield = internal_search_find(handle, filename, keystring, type,
-  errmsg);
+yield = internal_search_find(handle, filename, keystring, type, errmsg);
 if (search_find_defer) return NULL;
 if (yield != NULL) { if (partial >= 0) set_null_wild = TRUE; }
 
@@ -578,13 +572,11 @@ else if (partial >= 0)
           }
         break;
         }
-      keystring3 += 2;
+      keystring3++;    /* Must work for null components */
       }
     }
 
   else set_null_wild = TRUE; /* Matched a wild entry without any wild part */
-
-  store_free(keystring2);
   }
 
 /* If nothing has been matched, but the option to look for "*@" is set, try
