@@ -130,15 +130,17 @@ return 1 + ((nleft > nright)? nleft : nright);
 /* This is the real function...
 
 Arguments:
-  connect   pointer to the root of the tree
-  f         FILE to read data from
-  buffer    contains next input line; further lines read into it
+  connect      pointer to the root of the tree
+  f            FILE to read data from
+  buffer       contains next input line; further lines read into it
+  buffer_size  size of the buffer
 
-Returns:    FALSE on format error
+Returns:       FALSE on format error
 */
 
 static BOOL
-read_nonrecipients_tree(tree_node **connect, FILE *f, char *buffer)
+read_nonrecipients_tree(tree_node **connect, FILE *f, char *buffer,
+  int buffer_size)
 {
 tree_node *node;
 int n = (int)strlen(buffer);
@@ -146,22 +148,24 @@ BOOL right = buffer[1] == 'Y';
 
 if (n < 5) return FALSE;    /* malformed line */
 buffer[n-1] = 0;            /* Remove \n */
-node = store_malloc(sizeof(tree_node) + n - 3);
+node = store_get(sizeof(tree_node) + n - 3);
 *connect = node;
 strcpy(node->name, buffer + 3);
 node->data.ptr = NULL;
 
 if (buffer[0] == 'Y')
   {
-  if (fgets(buffer, 256, f) == NULL ||
-    !read_nonrecipients_tree(&node->left, f, buffer)) return FALSE;
+  if (fgets(buffer, buffer_size, f) == NULL ||
+    !read_nonrecipients_tree(&node->left, f, buffer, buffer_size))
+      return FALSE;
   }
 else node->left = NULL;
 
 if (right)
   {
-  if (fgets(buffer, 256, f) == NULL ||
-    !read_nonrecipients_tree(&node->right, f, buffer)) return FALSE;
+  if (fgets(buffer, buffer_size, f) == NULL ||
+    !read_nonrecipients_tree(&node->right, f, buffer, buffer_size))
+      return FALSE;
   }
 else node->right = NULL;
 
@@ -277,7 +281,7 @@ n = (int)strlen(big_buffer);
 if (n < 3 || big_buffer[0] != '<' || big_buffer[n-2] != '>')
   goto SPOOL_FORMAT_ERROR;
 
-sender_address = store_malloc(n-2);
+sender_address = store_get(n-2);
 strncpy(sender_address, big_buffer+1, n-3);
 sender_address[n-3] = 0;
 
@@ -300,6 +304,11 @@ deliver_frozen_at = 0;
 local_error_message = FALSE;
 user_null_sender = FALSE;
 
+#ifdef SUPPORT_DSN
+dsn_ret = 0;
+dsn_envid = NULL;
+#endif
+
 for (;;)
   {
   if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
@@ -312,19 +321,27 @@ for (;;)
     else if (strcmp(big_buffer, "-resent") == 0)
       header_names = header_names_resent;
     else if (strncmp(big_buffer, "-body_linecount", 15) == 0)
-      {
       body_linecount = atoi(big_buffer + 15);
-      }
     else if (strncmp(big_buffer, "-frozen", 7) == 0)
       {
       deliver_freeze = TRUE;
       deliver_frozen_at = atoi(big_buffer + 7);
       }
+    else if (strcmp(big_buffer, "-deliver_firsttime") == 0)
+      deliver_firsttime = TRUE;
     else if (strcmp(big_buffer, "-manual_thaw") == 0)
       deliver_manual_thaw = TRUE;
     else if (strcmp(big_buffer, "-user_null_sender") == 0)
       user_null_sender = TRUE;
 
+    #ifdef SUPPORT_DSN
+    else if (strncmp(big_buffer, "-dsn_ret", 8) == 0)
+      dsn_ret= atoi(big_buffer + 8);
+
+    else if (strncmp(big_buffer, "-dsn_envid", 10) == 0)
+      dsn_envid = string_copy(big_buffer + 11);
+    #endif
+
     else if (strncmp(big_buffer, "-host_address", 13) == 0)
       sender_host_address = string_copy(big_buffer + 14);
     else if (strncmp(big_buffer, "-host_name", 10) == 0)
@@ -333,6 +350,8 @@ for (;;)
       sender_helo_name = string_copy(big_buffer + 11);
     else if (strncmp(big_buffer, "-ident", 6) == 0)
       sender_ident = string_copy(big_buffer + 7);
+    else if (strncmp(big_buffer, "-received_protocol", 18) == 0)
+      received_protocol = string_copy(big_buffer + 19);
 
   /* To allow new versions of Exim that add additional flags to interwork
   with older versions that do not understand them, just ignore any flagged
@@ -356,7 +375,7 @@ DEBUG(5)
 containing "XX", indicating no tree. */
 
 if (strncmp(big_buffer, "XX\n", 3) != 0 &&
-  !read_nonrecipients_tree(&tree_nonrecipients, f, big_buffer))
+  !read_nonrecipients_tree(&tree_nonrecipients, f, big_buffer, big_buffer_size))
     goto SPOOL_FORMAT_ERROR;
 
 DEBUG(5)
@@ -374,19 +393,46 @@ if (sscanf(big_buffer, "%d", &recipients_count) != 1) goto SPOOL_FORMAT_ERROR;
 DEBUG(5) debug_printf("recipients_count=%d\n", recipients_count);
 
 recipients_list_max = recipients_count;
-recipients_list = store_malloc(recipients_count * sizeof(char *));
+recipients_list = store_get(recipients_count * sizeof(recipient_item));
 
 for (n = 0; n < recipients_count; n++)
   {
   int nn;
-  char *receiver;
+  int flags = 0;
+  int pno = 0;
+  char *orcpt = NULL;
+  char *p;
+
   if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
   nn = (int)strlen(big_buffer);
   if (nn < 2) goto SPOOL_FORMAT_ERROR;
-  receiver = store_malloc(nn);
-  big_buffer[nn-1] = 0;
-  strcpy(receiver, big_buffer);
-  recipients_list[n] = receiver;
+
+  /* Look back from the end for digits/commas and a space, indicating the
+  presence of flags, pno data, and possibly the DSN orcpt field. */
+
+  p = big_buffer + nn - 2;
+  while (isdigit(*p) || *p == ',') p--;
+  if (*p == ' ')
+    {
+    int orc;
+    *p++ = 0;
+    sscanf(p, "%d,%d,%d", &flags, &pno, &orc);
+    if (orc > 0)
+      {
+      p -= orc + 1;
+      p[-1] = 0;
+      orcpt = string_copy(p);
+      }
+    }
+  else big_buffer[nn-1] = 0;
+
+  recipients_list[n].address = string_copy(big_buffer);
+  recipients_list[n].flags = flags;
+  recipients_list[n].pno = pno;
+
+  #ifdef SUPPORT_DSN
+  recipients_list[n].orcpt = orcpt;
+  #endif
   }
 
 /* The remainder of the spool header file contains the headers for the message,
@@ -416,23 +462,16 @@ while ((n = fgetc(f)) != EOF)
 
   if (read_headers)
     {
-    h = store_malloc(sizeof(header_line) + n);
+    h = store_get(sizeof(header_line));
     h->next = NULL;
     h->type = flag[0];
     h->slen = n;
+    h->text = store_get(n+1);
 
     if (h->type == htype_received) received_count++;
 
-    if (header_list == NULL)
-      {
-      header_list = h;
-      h->prev = NULL;
-      }
-    else
-      {
-      header_last->next = h;
-      h->prev = header_last;
-      }
+    if (header_list == NULL) header_list = h;
+      else header_last->next = h;
     header_last = h;
 
     for (i = 0; i < n; i++)
