@@ -327,18 +327,14 @@ return ch;
 
 /* This function is called for critical strings. It checks for any
 non-printing characters, and if any are found, it makes a new copy
-of the string with suitable escape sequences. The old string is
-freed if the flag is true.
-
-Arguments:
-  s          input string
-  free_old   if TRUE and new string generated, free the old one
+of the string with suitable escape sequences.
 
+Argument:    input string
 Returns:     string with non-printers encoded as printing sequences
 */
 
 char *
-string_printing(char *s, BOOL free_old)
+string_printing(char *s)
 {
 int nonprintcount = 0;
 int length = 0;
@@ -357,7 +353,7 @@ if (nonprintcount == 0) return s;
 /* Get a new block of store guaranteed big enough to hold the
 expanded string. */
 
-ss = store_malloc(length + nonprintcount * 4 + 1);
+ss = store_get(length + nonprintcount * 4 + 1);
 
 /* Copy everying, escaping non printers. The unsigned char thing is
 for systems that have signed chars by default. */
@@ -385,10 +381,6 @@ while (*t != 0)
     }
   }
 *tt = 0;
-
-/* Free the old if requested, and return the new. */
-
-if (free_old) store_free(s);
 return ss;
 }
 
@@ -407,7 +399,26 @@ Returns:  copy of string in new store
 char *
 string_copy(char *s)
 {
-char *ss = (char *)store_malloc((int)strlen(s) + 1);
+char *ss = store_get((int)strlen(s) + 1);
+strcpy(ss, s);
+return ss;
+}
+
+
+
+/*************************************************
+*     Copy and save string in malloc'd store     *
+*************************************************/
+
+/*
+Argument: string to copy
+Returns:  copy of string in new store
+*/
+
+char *
+string_copy_malloc(char *s)
+{
+char *ss = store_malloc((int)strlen(s) + 1);
 strcpy(ss, s);
 return ss;
 }
@@ -426,7 +437,7 @@ Returns:  copy of string in new store, with letters lowercased
 char *
 string_copylc(char *s)
 {
-char *ss = (char *)store_malloc((int)strlen(s) + 1);
+char *ss = store_get((int)strlen(s) + 1);
 char *p = ss;
 while (*s != 0) *p++ = tolower(*s++);
 *p = 0;
@@ -452,7 +463,7 @@ Returns:    copy of string in new store
 char *
 string_copyn(char *s, int n)
 {
-char *ss = (char *)store_malloc(n + 1);
+char *ss = store_get(n + 1);
 strncpy(ss, s, n);
 ss[n] = 0;
 return ss;
@@ -476,7 +487,7 @@ Returns:    copy of string in new store, with letters lowercased
 char *
 string_copynlc(char *s, int n)
 {
-char *ss = (char *)store_malloc(n + 1);
+char *ss = store_get(n + 1);
 char *p = ss;
 while (n-- > 0) *p++ = tolower(*s++);
 *p = 0;
@@ -502,8 +513,8 @@ Returns:    pointer to fresh piece of store containing sprintf'ed string
 char *
 string_sprintf(char *format, ...)
 {
-char buffer[8192];
 va_list ap;
+char buffer[8192];
 va_start(ap, format);
 if (!string_vformat(buffer, sizeof(buffer), format, ap))
   log_write(0, LOG_PANIC_DIE, "string_sprintf expansion was longer than %d",
@@ -679,12 +690,14 @@ string_cat(char *string, int *size, int *ptr, char *s, int len)
 int p = *ptr;
 if (p + len >= *size)
   {
-  char *newstring;
+  int oldsize = *size;
   while (*size <= p + len) *size += 50;
-  newstring = (char *)store_malloc(*size);
-  memcpy(newstring, string, p);
-  store_free(string);
-  string = newstring;
+  if (!store_extend(string, oldsize, *size))
+    {
+    char *newstring = store_get(*size);
+    memcpy(newstring, string, p);
+    string = newstring;
+    }
   }
 strncpy(string + p, s, len);
 *ptr = p + len;
@@ -943,9 +956,10 @@ return (eno == EACCES)?
 *          Generate log address list             *
 *************************************************/
 
-/* This function generates a list of and address and its parents, for use in
-logging lines. It is passed the address and its ultimate parent, and a flag
-indicating a directed local address.
+/* This function generates a list consisting of an address and its parents, for
+use in logging lines. It is passed the address and its ultimate parent, and a
+flag indicating a directed local address. For saved onetime aliased addresses,
+the onetime parent field is used.
 
 Arguments:
   addr          bottom (ultimate) address
@@ -963,7 +977,7 @@ string_log_address(address_item *addr, address_item *topaddr, BOOL dlocal,
 int size = 64;
 int ptr = 0;
 BOOL add_topaddr = TRUE;
-char *yield = store_malloc(size);
+char *yield = store_get(size);
 
 /* A local delivery that was directed starts with just the local part. */
 
@@ -997,7 +1011,7 @@ else
     }
   yield[ptr] = 0;
 
-  if (strcmp(yield, topaddr->orig) == 0 &&
+  if (strcmp(yield, topaddr->orig) == 0 && addr->onetime_parent == NULL &&
       (!all_parents || addr->parent == NULL || addr->parent == topaddr))
     add_topaddr = FALSE;
   }
@@ -1023,13 +1037,19 @@ if ((all_parents || addr->pfr) &&
   yield = string_cat(yield, &size, &ptr, ")", 1);
   }
 
-/* Finally, add the top address if it is required */
+/* Add the top address if it is required */
 
 if (add_topaddr)
   {
   yield = string_cat(yield, &size, &ptr, " <", 2);
-  yield = string_cat(yield, &size, &ptr, topaddr->orig,
-    (int)strlen(topaddr->orig));
+
+  if (addr->onetime_parent == NULL)
+    yield = string_cat(yield, &size, &ptr, topaddr->orig,
+      (int)strlen(topaddr->orig));
+  else
+    yield = string_cat(yield, &size, &ptr, addr->onetime_parent,
+      (int)strlen(addr->onetime_parent));
+
   yield = string_cat(yield, &size, &ptr, ">", 1);
   }
 
