@@ -42,6 +42,36 @@ typedef struct re_block {
 } re_block;
 
 
+/* Block for holding a uid and gid, possibly unset, and an initgroups flag. */
+
+typedef struct ugid_block {
+  uid_t uid;
+  gid_t gid;
+  BOOL  uid_set;
+  BOOL  gid_set;
+  BOOL  initgroups;
+} ugid_block;
+
+
+/* Structure for holding information about a host for use mainly by
+routers, but also used when checking lists of hosts. Looking up host
+addresses is done in both cases, using this structure. */
+
+typedef struct host_item {
+  struct host_item *next;
+  char *name;
+  char *address;
+  /* Used when routing */
+  int   mx;                      /* MX value if found via MX records */
+  int   sort_key;                /* MX*100 plus random "fraction" */
+  int   status;                  /* Usable, unusable, or unknown */
+  int   why;                     /* Why host is unusable */
+  int   last_try;                /* Time of last try if known */
+  /* Used when checking */
+  char *ident_string;            /* RFC 1413 ident string */
+  re_block *compiled_name;       /* When name is wildcarded */
+} host_item;
+
 /* Directors, routers, and transports have similar data blocks. Each driver
 that is compiled into the code is represented by a xxx_info block; the active
 drivers are represented by a chain of xxx_instance blocks. To make it possible
@@ -87,6 +117,7 @@ typedef struct transport_instance {
   char *driver_name;             /* Must be first */
   int  (*setup)(                 /* Setup entry point for local transports */
     struct transport_instance *,
+    struct address_item *,
     char **);
                                  /**************************************/
   int   local_batch;             /* )                                  */
@@ -111,6 +142,10 @@ typedef struct transport_instance {
   char *filter_command;          /* For on-the-fly-filtering */
   char *add_headers;             /* Add these headers */
   char *remove_headers;          /* Remove these headers */
+  char *debug_string;            /* Debugging output */
+  BOOL  delivery_date_add;       /* Add Delivery-Date header */
+  BOOL  envelope_to_add;         /* Add Envelope-To header */
+  BOOL  return_path_add;         /* Add Return-Path header */
   BOOL  return_output;           /* TRUE if output should always be returned */
   BOOL  return_fail_output;      /* ditto, but only on failure */
   BOOL  log_output;              /* Similarly for logging */
@@ -164,6 +199,9 @@ typedef struct director_instance {
   char *errors_to;               /* Errors address */
   char *extra_headers;           /* Additional headers */
   char *remove_headers;          /* Those to remove */
+  char *debug_string;            /* Debugging output */
+  char *fallback_hosts;          /* For remote transports */
+  host_item *fallback_hostlist;
   re_block *re_domains;          /* Used when domains are wildcarded */
   re_block *re_except_domains;
   re_block *re_local_parts;
@@ -174,6 +212,13 @@ typedef struct director_instance {
   BOOL  verify_recipient;        /* Use this director when verifying a recipient*/
   transport_instance *transport; /* Assigned transport */
   char *expand_transport;        /* Dynamic transport string */
+  char *expand_uid;              /* Expanded uid string */
+  char *expand_gid;              /* Expanded gid string */
+  uid_t uid;                     /* Fixed uid value */
+  gid_t gid;                     /* Fixed gid value */
+  BOOL  uid_set;                 /* Flag to indicate uid is set */
+  BOOL  gid_set;                 /* Flag to indicate gid is set */
+  BOOL  initgroups;              /* True if initgroups is required */
   /********************************************************************/
 
   char *condition;               /* General condition */
@@ -240,17 +285,26 @@ typedef struct router_instance {
   char *errors_to;               /* Errors address */
   char *extra_headers;           /* Additional headers */
   char *remove_headers;          /* Removed headers */
+  char *debug_string;            /* Debugging output */
+  char *fallback_hosts;          /* For remote transports */
+  host_item *fallback_hostlist;
   re_block *re_domains;          /* Used when domains are wildcarded */
   re_block *re_except_domains;
   re_block *re_local_parts;
   re_block *re_except_local_parts;
-
   BOOL  more;                    /* If FALSE, do no more if this one fails */
   BOOL  verify_only;             /* Skip this router if not verifying */
   BOOL  verify_sender;           /* Use this router when verifying a sender */
   BOOL  verify_recipient;        /* Use this router when verifying a recipient*/
   transport_instance *transport; /* Assigned transport */
   char *expand_transport;        /* Dynamic transport string */
+  char *expand_uid;              /* Expanded uid string */
+  char *expand_gid;              /* Expanded gid string */
+  uid_t uid;                     /* Fixed uid value */
+  gid_t gid;                     /* Fixed gid value */
+  BOOL  uid_set;                 /* Flag to indicate uid is set */
+  BOOL  gid_set;                 /* Flag to indicate gid is set */
+  BOOL  initgroups;              /* True if initgroups is required */
   /********************************************************************/
 
   char *condition;               /* General condition */
@@ -318,25 +372,6 @@ typedef struct lookup_info {
 } lookup_info;
 
 
-/* Structure for holding information about a host for use mainly by
-routers, but also used when checking lists of hosts. Looking up host
-addresses is done in both cases, using this structure. */
-
-typedef struct host_item {
-  struct host_item *next;
-  char *name;
-  char *address;
-  /* Used when routing */
-  int   mx;                      /* MX value if found via MX records */
-  int   sort_key;                /* MX*100 plus random "fraction" */
-  int   status;                  /* Usable, unusable, or unknown */
-  int   why;                     /* Why host is unusable */
-  int   last_try;                /* Time of last try if known */
-  /* Used when checking */
-  char *ident_string;            /* RFC 1413 ident string */
-  re_block *compiled_name;       /* When name is wildcarded */
-} host_item;
-
 /* Structure for holding a single IP address; used for the chain of
 addresses and ports for the local host. Make the char string large enough
 to hold an IPv6 address. */
@@ -397,17 +432,28 @@ typedef struct reply_item {
 } reply_item;
 
 
-/* Chain of header lines; the basic structure can contain an empty
-string; a bigger block is obtained for longer strings. */
+/* Structure for maintaining a chain of header lines. */
 
 typedef struct header_line {
   struct header_line *next;
-  struct header_line *prev;
   int type;
   int slen;
-  char text[1];
+  char *text;
 } header_line;
 
+
+/*Structure for holding information about an envelope address */
+
+typedef struct recipient_item {
+  char *address;                 /* the recipient address */
+  int   flags;                   /* whose names start rf_ */
+  int   pno;                     /* parent number when it's a "fixed" alias */
+  #ifdef SUPPORT_DSN
+  char *orcpt;                   /* DSN orcpt field */
+  #endif
+} recipient_item;
+
+
 /* The address_item structure contains many fields which are used at various
 times while delivering a message. Some are used only for remote deliveries;
 some only for local. */
@@ -415,6 +461,8 @@ some only for local. */
 typedef struct address_item {
   struct address_item *next;     /* for chaining addresses */
   struct address_item *parent;   /* parent address */
+  struct address_item *first;    /* point to first after group delivery */
+  char *onetime_parent;          /* saved original parent for onetime */
   char *orig;                    /* as read from the spool */
   char *unique;                  /* used for disambiguating .forwarded */
   char *local_part;              /* local part of envelope address */
@@ -441,6 +489,7 @@ typedef struct address_item {
   transport_instance *transport; /* transport to use */
   host_item *host_list;          /* host data for the transport */
   host_item *transported;        /* host that took delivery or failed hard */
+  host_item *fallback_hosts;     /* to try if delivery defers */
   int  child_count;              /* number of child addresses */
   uid_t uid;                     /* uid for local transporting */
   gid_t gid;                     /* gid for local transporting */
@@ -470,6 +519,10 @@ typedef struct address_item {
   char *return_filename;         /* ) some transports write text into a file */
   int   return_file;             /* ) for return to the sender */
   reply_item *reply;             /* and some generate reply messages */
+  #ifdef SUPPORT_DSN
+  char *dsn_orcpt;               /* DSN orcpt value */
+  int  dsn_flags;                /* DSN flags */
+  #endif
 } address_item;
 
 
@@ -539,6 +592,15 @@ typedef struct tree_node {
   char name[1];                   /* node name - variable length */
 } tree_node;
 
+/* Structure for holding the handle and the cached last lookup for searches.
+This block is pointed to by the tree entry for the open file. */
+
+typedef struct search_cache {
+  void *handle;                   /* lookup handle */
+  char *key;                      /* the key of the last lookup */
+  char *data;                     /* the data of the last lookup */
+} search_cache;
+
 /* Structure for holding a partially decoded DNS record; the name has been
 uncompressed, but the data pointer is into the raw data. */
 
