@@ -12,10 +12,11 @@ transports. */
 #include "exim.h"
 
 
-/* Static variable to control whether to retry or not when writing
-a block writes only part of it (for correct EDQUOT handling). */
+/* Newline status flag and current pointer for write_chunk() */
+
+static BOOL was_nl;
+static char *chunk_ptr;
 
-static int write_block_count;
 
 /* Generic options for transports, all of which live inside transport_instance
 data blocks and which therefore have the opt_public flag set. Note that there
@@ -25,14 +26,22 @@ certain transports. */
 optionlist optionlist_transports[] = {
   { "add_headers",      opt_stringptr|opt_public|opt_hidden,
                  (void *)offsetof(transport_instance, add_headers) },
+  { "debug_print",      opt_stringptr | opt_public,
+                 (void *)offsetof(transport_instance, debug_string) },
+  { "delivery_date_add", opt_bool|opt_public,
+                 (void *)(offsetof(transport_instance, delivery_date_add)) },
   { "driver",           opt_stringptr|opt_public,
                  (void *)offsetof(transport_instance, driver_name) },
+  { "envelope_to_add",   opt_bool|opt_public,
+                 (void *)(offsetof(transport_instance, envelope_to_add)) },
   { "headers_add",      opt_stringptr|opt_public,
                  (void *)offsetof(transport_instance, add_headers) },
   { "headers_remove",   opt_stringptr|opt_public,
                  (void *)offsetof(transport_instance, remove_headers) },
   { "remove_headers",   opt_stringptr|opt_public|opt_hidden,
                  (void *)offsetof(transport_instance, remove_headers) },
+  { "return_path_add",   opt_bool|opt_public,
+                 (void *)(offsetof(transport_instance, return_path_add)) },
   { "shadow_condition", opt_stringptr|opt_public,
                  (void *)offsetof(transport_instance, shadow_condition) },
   { "shadow_transport", opt_transportptr|opt_public,
@@ -96,36 +105,64 @@ for (t = transports; t != NULL; t = t->next)
 *             Write block of data                *
 *************************************************/
 
-/* Subroutine called by write_chunk actually to write a data block. It
-applies a timeout if transport_chunk_timeout is greater than zero. Also,
-on some systems at least, if a quota is exceeded *during* the write, the yield
-is the number of bytes written and errno is not set. Therefore, in this case,
-we want to try to write a second time to output the remainder of the data after
-a non-negative return from write(), except after a timeout. This time, no bytes
-should get written, and a proper error should get put into errno. This is
-managed by having transport_write_message set the value of write_block_count to
-2 if writing to a file and to 1 otherwise.
+/* Subroutine called by write_chunk() and at the end of the message actually
+to write a data block. Also called directly by some transports to write
+additional data to the file descriptor (e.g. prefix, suffix).
+
+If a transport wants data transfers to be timed, it sets a non-zero value in
+transport_write_timeout and sets up sigalrm_handler() as the SIGALRM handler.
+A non-zero transport_write_timeout causes a timer to be set for each block of
+data written from here. If time runs out, then write() fails and provokes an
+error return. The caller can then inspect sigalrm_seen to check for a timeout.
+
+On some systems, if a quota is exceeded during the write, the yield is the
+number of bytes written rather than an immediate error code. This also happens
+on some systems in other cases, for example a pipe that goes away because the
+other end's process terminates (Linux). On other systems, (e.g. Solaris 2) you
+get the error codes the first time.
+
+The write() function is also interruptible; the Solaris 2.6 man page says:
+
+     If write() is interrupted by a signal before it writes any
+     data, it will return -1 with errno set to EINTR.
+
+     If write() is interrupted by a signal after it successfully
+     writes some data, it will return the number of bytes written.
+
+To handle these cases, we want to restart the write() to output the remainder
+of the data after a non-negative return from write(), except after a timeout.
+In the error cases (EDQUOT, EPIPE) no bytes get written the second time, and a
+proper error then occurs. In principle, after an interruption, the second
+write() could suffer the same fate, but we do not want to continue for
+evermore, so stick a maximum repetition count on the loop to act as a
+longstop.
 
 Arguments:
   fd        file descriptor to write to
   block     block of bytes to write
   len       number of bytes to write
 
-Returns:    TRUE on success, FALSE on failure (with errno preserved)
+Returns:    TRUE on success, FALSE on failure (with errno preserved);
+              transport_count is incremented by the count of bytes written
 */
 
-static BOOL
-write_block(int fd, char *block, int len)
+BOOL
+transport_write_block(int fd, char *block, int len)
 {
-int i, rc;
-for (i = 0; i < write_block_count; i++)
+int i, rc, save_errno;
+
+for (i = 0; i < 100; i++)
   {
   DEBUG(9)
     debug_printf("writing data block size=%d timeout=%d\n",
-      len, transport_chunk_timeout);
-  if (transport_chunk_timeout > 0) alarm(transport_chunk_timeout);
+      len, transport_write_timeout);
+  if (transport_write_timeout > 0) alarm(transport_write_timeout);
   rc = write(fd, block, len);
-  if (transport_chunk_timeout > 0)
+  save_errno = errno;
+
+  /* Cancel the alarm and deal with a timeout */
+
+  if (transport_write_timeout > 0)
     {
     alarm(0);
     if (sigalrm_seen)
@@ -134,13 +171,39 @@ for (i = 0; i < write_block_count; i++)
       return FALSE;
       }
     }
+
+  /* Hopefully, the most common case is success, so test that first. */
+
   if (rc == len) { transport_count += len; return TRUE; }
-  if (rc < 0) return FALSE;
-  len -= rc;
-  block += rc;
-  transport_count += rc;
-  DEBUG(9) debug_printf("write incomplete (%d)\n", rc);
+
+  /* A non-negative return code is an incomplete write. Try again. */
+
+  if (rc >= 0)
+    {
+    len -= rc;
+    block += rc;
+    transport_count += rc;
+    DEBUG(9) debug_printf("write incomplete (%d)\n", rc);
+    continue;
+    }
+
+  /* A negative return code with an EINTR error is another form of
+  incomplete write, zero bytes having been written */
+
+  if (save_errno == EINTR)
+    {
+    DEBUG(9) debug_printf("write interrupted before anything written\n");
+    continue;
+    }
+
+  /* Otherwise there's been an error */
+
+  errno = save_errno;
+  return FALSE;
   }
+
+/* We've tried and tried and tried but still failed */
+
 errno = ERRNO_WRITEINCOMPLETE;
 return FALSE;
 }
@@ -148,25 +211,48 @@ return FALSE;
 
 
 
+/*************************************************
+*             Write formatted string             *
+*************************************************/
+
+/* This is called by various transports. It is a convenience function.
+
+Arguments:
+  fd          file descriptor
+  format      string format
+  ...         arguments for format
+
+Returns:      the yield of transport_write_block()
+*/
+
+BOOL
+transport_write_string(int fd, char *format, ...)
+{
+va_list ap;
+va_start(ap, format);
+if (!string_vformat(big_buffer, big_buffer_size, format, ap))
+  log_write(0, LOG_PANIC_DIE, "overlong formatted string in appendfile");
+va_end(ap);
+return transport_write_block(fd, big_buffer, strlen(big_buffer));
+}
+
+
+
+
 /*************************************************
 *              Write character chunk             *
 *************************************************/
 
-/* Subroutine used by transport_write_message to scan character chunks for
+/* Subroutine used by transport_write_message() to scan character chunks for
 newlines and act appropriately. The object is to minimise the number of writes.
 In practice, either from_hack or (use_crlf & smtp_dots) is likely to be set,
-but not both. However, the code works with any combination.
+but not both. However, the code works with any combination. The output byte
+stream is buffered up in deliver_out_buffer, which is written only when it gets
+full, thus minimizing write operations and TCP packets.
 
 If the from hack or smtp_dots options are requested, we must do the necessary
-at the start of the chunk if it is the start of the data section, or if the
-previous last char was '\n'. This is handled by a static flag.
-
-If a transport wants data transfers to be timed, it sets a non-zero value
-in transport_chunk_timeout and sets up sigalrm_handler() as the SIGALRM
-handler. A non-zero transport_chunk_timeout causes a timer to be set for each
-chunk of data written from here via write_block() above. If time runs out, then
-a write() fails and provokes an error return. The caller can then inspect
-sigalrm_seen to check for a timeout.
+at the start of the chunk if the previous last char was '\n'. This is handled
+by a static flag.
 
 Arguments:
   fd         file descript to write to
@@ -179,32 +265,23 @@ Arguments:
 Returns:     TRUE on success, FALSE on failure (with errno preserved)
 */
 
-static BOOL was_nl;     /* NL status flag */
-
 static BOOL
 write_chunk(int fd, char *chunk, int len, BOOL from_hack, BOOL use_crlf,
   BOOL smtp_dots)
 {
 char *start = chunk;
 char *end = chunk + len;
-char *outptr = deliver_out_buffer;
-char *ptr;
+register char *ptr;
 
-/* If none of the processing options is set, we can just write out the
-block in one go without further ado. */
-
-if (!(from_hack || use_crlf || smtp_dots)) return write_block(fd, chunk, len);
-
-/* Otherwise, scan the buffer for newlines while copying into the output
-buffer. This is to ensure we write the main data output in big chunks. Assume
-len > 0. First handle the case when the preceding char was NL. */
+/* Scan the buffer for newlines while copying into the output
+buffer. First handle the case when the preceding char was NL. */
 
 if (was_nl)
   {
   if (from_hack && len >= 5 && strncmp(start, "From ", 5) == 0)
-    *outptr++ = '>';
+    *chunk_ptr++ = '>';
   else if (smtp_dots && *start == '.')
-    *outptr++ = '.';
+    *chunk_ptr++ = '.';
   was_nl = FALSE;
   }
 
@@ -216,10 +293,10 @@ for (ptr = start; ptr < end; ptr++)
 
   /* If there isn't room for another 4 characters, flush the buffer. */
 
-  if ((len = outptr - deliver_out_buffer) > DELIVER_OUT_BUFFER_SIZE - 5)
+  if ((len = chunk_ptr - deliver_out_buffer) > DELIVER_OUT_BUFFER_SIZE - 5)
     {
-    if (!write_block(fd, deliver_out_buffer, len)) return FALSE;
-    outptr = deliver_out_buffer;
+    if (!transport_write_block(fd, deliver_out_buffer, len)) return FALSE;
+    chunk_ptr = deliver_out_buffer;
     }
 
   if ((ch = *ptr) == '\n')
@@ -228,11 +305,11 @@ for (ptr = start; ptr < end; ptr++)
 
     /* Insert CR before NL if required */
 
-    if (use_crlf) *outptr++ = '\r';
+    if (use_crlf) *chunk_ptr++ = '\r';
 
     /* Now the NL */
 
-    *outptr++ = '\n';
+    *chunk_ptr++ = '\n';
 
     /* If it happens that \n is the last character in the chunk, set the
     flag for next time (and for checking the final char). */
@@ -243,21 +320,18 @@ for (ptr = start; ptr < end; ptr++)
     is a case-sensitive test. */
 
     else if (from_hack && left >= 5 && strncmp(ptr+1, "From ", 5) == 0)
-      *outptr++ = '>';
+      *chunk_ptr++ = '>';
 
     /* The smtp dot convention inserts a dot before any line starting with a
     dot. (We don't need to test left >= 1 because we know it isn't <= 0.) */
 
-    else if (smtp_dots && ptr[1] == '.') *outptr++ = '.';
+    else if (smtp_dots && ptr[1] == '.') *chunk_ptr++ = '.';
     }
 
-  else *outptr++ = ch;
+  else *chunk_ptr++ = ch;
   }
 
-/* Write out any remaining data in the buffer before returning. */
-
-return (len = outptr - deliver_out_buffer) <= 0 ||
-  write_block(fd, deliver_out_buffer, len);
+return TRUE;
 }
 
 
@@ -292,7 +366,7 @@ file descriptor deliver_datafile. Make sure we start it at the beginning.
 The yield is TRUE if all went well, and FALSE if not. Exit *immediately* after
 any writing or reading error, leaving the code in errno intact. Error exits
 can include timeouts for certain transports, which are requested by setting
-transport_chunk_timeout non-zero.
+transport_write_timeout non-zero.
 
 Arguments:
   addr                  (chain of) addresses (for extra headers), or NULL;
@@ -305,8 +379,7 @@ Arguments:
     from_hack             if TRUE, apply the "from hack"
     use_crlf              if TRUE, turn NL into CR LF
     smtp_dots             if TRUE, implement the SMTP dot escaping algorithm
-    writing_file          if TRUE, set write_block_count to 2, else 1
-                                   (see comment in write_block() above)
+    dot_end               if TRUE, send a terminating "." line at the end
   errors_to             if not NULL, used for the return-path header
   size_limit            if > 0, this is a limit to the size of message written;
                           it is used when returning messages to their senders,
@@ -314,12 +387,13 @@ Arguments:
                           buffering
   add_headers           a string containing one or more headers to add; it is
                           expanded, and must be in correct RFC 822 format as
-                          it is transmitted verbatim; NULL => no additions
+                          it is transmitted verbatim; NULL => no additions,
+                          and so does empty string or forced expansion fail
   remove_headers        a colon-separated list of headers to remove, or NULL.
 
 Returns:                TRUE on success; FALSE (with errno) on failure.
                         In addition, the global variable transport_count
-                        contains the number of bytes written.
+                        is incremented by the number of bytes written.
 */
 
 static BOOL
@@ -333,13 +407,10 @@ BOOL from_hack = (options & topt_from_hack) != 0;
 BOOL use_crlf  = (options & topt_use_crlf)  != 0;
 BOOL smtp_dots = (options & topt_smtp_dots) != 0;
 
-transport_count = 0;
+/* Initialize newline flag, and pointer in output buffer. */
 
-/* Arrange for 2 attempts at writing a block if the first partially succeeds
-when writing to a file, in order to be sure of getting EDQUOT when a quota is
-exceeded. */
-
-write_block_count = ((options & topt_writing_file) == 0)? 1 : 2;
+was_nl = TRUE;
+chunk_ptr = deliver_out_buffer;
 
 /* Add return-path: if requested, and there is one. */
 
@@ -357,7 +428,6 @@ any relevant duplicates. */
 
 if ((options & topt_add_envelope_to) != 0)
   {
-  int len;
   address_item *p;
   char buffer[256];
   strcpy(buffer, "Envelope-to: ");
@@ -468,32 +538,36 @@ if (addr != NULL)
 
 /* If a string containing additional headers exists, expand it and write
 out the result. This is done last so that if it (deliberately or accidentally)
-isn't in header format, it won't mess up any other headers. */
+isn't in header format, it won't mess up any other headers. An empty string or
+a forced expansion failure are noops. */
 
 if (add_headers != NULL)
   {
-  int len;
   char *s = expand_string(add_headers);
   if (s == NULL)
     {
-    errno = ERRNO_ADDHEADER_FAIL;
-    return FALSE;
+    if (!expand_string_forcedfail)
+      {
+      errno = ERRNO_ADDHEADER_FAIL;
+      return FALSE;
+      }
+    }
+  else
+    {
+    int len = (int)strlen(s);
+    if (len > 0)
+      {
+      if (!write_chunk(fd, s, len, FALSE, use_crlf, smtp_dots))
+        return FALSE;
+      if (s[len-1] != '\n' && !write_chunk(fd, "\n", 1, FALSE, use_crlf,
+        smtp_dots)) return FALSE;
+      }
     }
-  len = (int)strlen(s);
-  if (!write_chunk(fd, s, len, FALSE, use_crlf, smtp_dots))
-    return FALSE;
-  if (s[len-1] != '\n' && !write_chunk(fd, "\n", 1, FALSE, use_crlf,
-    smtp_dots)) return FALSE;
   }
 
-/* Separate headers from body with a blank line; set the static flag for
-the chunk writer. Note: use_crlf won't be in use when writing to a file, so we
-are only every trying to write one byte to a file. If this fails because of
-quote excession, a proper errno should result. */
+/* Separate headers from body with a blank line */
 
-if (use_crlf? (write(fd, "\r\n", 2) != 2) : (write(fd, "\n", 1) != 1))
-  return FALSE;
-was_nl = TRUE;
+if (!write_chunk(fd, "\n", 1, FALSE, use_crlf, smtp_dots)) return FALSE;
 
 /* Ensure the body is positioned at the start of its file, following the
 message id if present, then write it, applying the size limit if required. */
@@ -515,18 +589,21 @@ while ((len = read(deliver_datafile, deliver_in_buffer,
 
 if (len != 0) return FALSE;
 
-/* If the last character of the body was not '\n', manufacture one. Note:
-use_crlf won't be in use when writing to a file, so we are only every trying to
-write one byte to a file. If this fails because of quote excession, a proper
-errno should result. */
+/* If the last character of the body was not '\n', manufacture one. */
 
-if (!was_nl)
-  {
-  if (use_crlf? (write(fd, "\r\n", 2) != 2) : (write(fd, "\n", 1) != 1))
-    return FALSE;
-  }
+if (!was_nl && !write_chunk(fd, "\n", 1, FALSE, use_crlf, smtp_dots))
+  return FALSE;
 
-return TRUE;
+/* If requested, add a terminating "." line (SMTP output). */
+
+if ((options & topt_end_dot) != 0 &&
+    !write_chunk(fd, ".\n", 2, FALSE, use_crlf, FALSE))
+  return FALSE;
+
+/* Write out any remaining data in the buffer before returning. */
+
+return (len = chunk_ptr - deliver_out_buffer) <= 0 ||
+  transport_write_block(fd, deliver_out_buffer, len);
 }
 
 
@@ -536,12 +613,11 @@ return TRUE;
 *    External interface to write the message     *
 *************************************************/
 
-/* If there is no filtering required, call the internal function
-above to do the real work, passing over all the arguments from this
-function. Otherwise, set up a filtering process, fork another process to call
-the internal function to write to the filter, and in this process just suck
-from the filter and write down the given fd. At the end, tidy up the pipes and
-the processes.
+/* If there is no filtering required, call the internal function above to do
+the real work, passing over all the arguments from this function. Otherwise,
+set up a filtering process, fork another process to call the internal function
+to write to the filter, and in this process just suck from the filter and write
+down the given fd. At the end, tidy up the pipes and the processes.
 
 Arguments:     as for internal_transport_write_message() above
 
@@ -609,7 +685,8 @@ if ((write_pid = fork()) == 0)
   close(pfd[pipe_read]);
   rc = internal_transport_write_message(addr, fd_write,
     (options &
-      ~(topt_writing_file|topt_from_hack|topt_use_crlf|topt_smtp_dots)),
+      ~(topt_from_hack | topt_use_crlf |
+        topt_smtp_dots | topt_end_dot)),
     errors_to, size_limit, add_headers, remove_headers);
   save_errno = errno;
   write(pfd[pipe_write], (void *)&rc, sizeof(BOOL));
@@ -636,27 +713,30 @@ if (write_pid < 0)
 /* Copy the message from the filter to the output fd. A read error leaves len
 == -1 and errno set. We need to apply a timeout to the read, to cope with
 the case when the filter gets stuck, but it can be quite a long one. If
-transport_chunk_timeout is positive, the SIGALRM handler is already set
+transport_write_timeout is positive, the SIGALRM handler is already set
 up for writing timeouts. If not, we must set it up here. */
 
 DEBUG(9) debug_printf("copying from the filter\n");
-write_block_count = ((options & topt_writing_file) == 0)? 1 : 2;
 
 from_hack = (options & topt_from_hack) != 0;
 use_crlf  = (options & topt_use_crlf)  != 0;
 smtp_dots = (options & topt_smtp_dots) != 0;
 
+transport_count = 0;
+was_nl = TRUE;
+chunk_ptr = deliver_out_buffer;
+
 for (;;)
   {
-  if (transport_chunk_timeout <= 0)  /* Yes, that is "<=" see above */
+  if (transport_write_timeout <= 0)  /* Yes, that is "<=" see above */
     {
     sigalrm_seen = FALSE;
-    signal(SIGALRM, sigalrm_handler);
+    os_non_restarting_signal(SIGALRM, sigalrm_handler);
     }
   alarm(300);
   len = read(fd_read, deliver_in_buffer, DELIVER_IN_BUFFER_SIZE);
   alarm(0);
-  if (transport_chunk_timeout <= 0) signal(SIGALRM, SIG_IGN);
+  if (transport_write_timeout <= 0) signal(SIGALRM, SIG_IGN);
   if (sigalrm_seen)
     {
     save_errno = ETIMEDOUT;
@@ -680,9 +760,20 @@ for (;;)
 
 if (len != 0) { save_errno = errno; goto TIDY_UP; }
 
-/* Message successfully written */
+/* If requested, add a terminating "." line (SMTP output). */
 
-yield = TRUE;
+if ((options & topt_end_dot) != 0 &&
+    !write_chunk(fd, ".\n", 2, FALSE, use_crlf, FALSE))
+  {
+  save_errno = errno;
+  goto TIDY_UP;
+  }
+
+/* Write out any remaining data in the buffer. */
+
+yield = (len = chunk_ptr - deliver_out_buffer) <= 0 ||
+  transport_write_block(fd, deliver_out_buffer, len);
+save_errno = errno;
 
 /* Tidying up code: ensure the pipes are all closed and the processes are
 removed. If there has been an error, kill the processes before waiting for
@@ -826,7 +917,7 @@ for (host = hostlist; host!= NULL; host = host->next)
   host_record = dbfn_read(dbm_file, host->name);
   if (host_record == NULL)
     {
-    host_record = store_malloc(sizeof(dbdata_wait) + MESSAGE_ID_LENGTH);
+    host_record = store_get(sizeof(dbdata_wait) + MESSAGE_ID_LENGTH);
     host_record->count = host_record->sequence = 0;
     }
 
@@ -859,7 +950,6 @@ for (host = hostlist; host!= NULL; host = host->next)
         if (strncmp(s, message_id, MESSAGE_ID_LENGTH) == 0)
           { already = TRUE; break; }
         }
-      store_free(cont);
       }
     }
 
@@ -887,7 +977,7 @@ for (host = hostlist; host!= NULL; host = host->next)
   else
     {
     dbdata_wait *newr =
-      store_malloc(sizeof(dbdata_wait) + host_length + MESSAGE_ID_LENGTH);
+      store_get(sizeof(dbdata_wait) + host_length + MESSAGE_ID_LENGTH);
     memcpy(newr, host_record, sizeof(dbdata_wait) + host_length);
     host_record = newr;
     }
@@ -1015,7 +1105,11 @@ for (;;)
     strncpy(new_message_id, host_record->text + host_length,
       MESSAGE_ID_LENGTH);
     new_message_id[MESSAGE_ID_LENGTH] = 0;
-    sprintf(buffer + path_len, "%s-D", new_message_id);
+
+    if (split_spool_directory)
+      sprintf(buffer + path_len, "%c/%s-D", new_message_id[5], new_message_id);
+    else
+      sprintf(buffer + path_len, "%s-D", new_message_id);
 
     /* The listed message may be the one we are currently processing. If
     so, we want to remove it from the list without doing anything else.
@@ -1061,7 +1155,6 @@ for (;;)
     /* Else replace the current with the continuation */
 
     dbfn_delete(dbm_file, buffer);
-    store_free(host_record);
     host_record = newr;
     host_length = host_record->count * MESSAGE_ID_LENGTH;
     }
@@ -1127,7 +1220,7 @@ int status;
 if ((pid = vfork()) == 0)
   {
   int fd, i;
-  char *argv[12 + MAX_CLMACROS];
+  char *argv[13 + MAX_CLMACROS];
 
   /* Disconnect entirely from the parent process. */
 
@@ -1151,6 +1244,11 @@ if ((pid = vfork()) == 0)
     argv[i++] = "-C";
     argv[i++] = config_filename;
     }
+
+  #ifdef SUPPORT_DSN
+  if (smtp_use_dsn) argv[i++] = "-MCD";
+  #endif
+
   if (smtp_use_size) argv[i++] = "-MCS";
   argv[i++] = "-MC";
   argv[i++] = transport_name;
@@ -1310,15 +1408,19 @@ dbfn_close(dbm_file);
 /* This function is called when a command line is to be parsed and executed
 directly, without the use of /bin/sh. It is called by the pipe transport,
 and also from the main delivery code when setting up a transport filter
-process.
+process. The code for ETRN also makes use of this; in that case, no addresses
+are passed.
 
 Arguments:
   argvptr            pointer to anchor for argv vector
   cmd                points to the command string
   expand_arguments   true if expansion is to occur
-  expand_failed      error value to set if expansion fails
-  addr               chain of addresses
+  expand_failed      error value to set if expansion fails; not relevant if
+                     addr == NULL
+  addr               chain of addresses, or NULL
   etext              text for use in error messages
+  errptr             where to put error message if addr is NULL;
+                     otherwise it is put in the first address
 
 Returns:             TRUE if all went well; otherwise an error will be
                      set in the first address and FALSE returned
@@ -1326,7 +1428,7 @@ Returns:             TRUE if all went well; otherwise an error will be
 
 BOOL
 transport_set_up_command(char ***argvptr, char *cmd, BOOL expand_arguments,
-  int expand_failed, address_item *addr, char *etext)
+  int expand_failed, address_item *addr, char *etext, char **errptr)
 {
 address_item *ad;
 char **argv;
@@ -1342,7 +1444,7 @@ the batch options is set. */
 
 for (ad = addr; ad != NULL; ad = ad->next) address_count++;
 max_args = address_count + 60;
-*argvptr = argv = (char **)store_malloc((max_args+1)*sizeof(char *));
+*argvptr = argv = store_get((max_args+1)*sizeof(char *));
 
 /* Split the command up into arguments terminated by white space. Lose
 trailing space at the start and end. Double-quoted arguments can contain \\ and
@@ -1350,7 +1452,7 @@ trailing space at the start and end. Double-quoted arguments can contain \\ and
 total length can be no longer than the length of the original. */
 
 s = cmd;
-ss = store_malloc((int)strlen(s) + 1);
+ss = store_get((int)strlen(s) + 1);
 while (isspace(*s)) s++;
 
 while (*s != 0 && argcount < max_args)
@@ -1414,9 +1516,14 @@ argv[argcount] = (char *)0;
 
 if (*s != 0)
   {
-  addr->transport_return = FAIL;
-  addr->message = string_sprintf("Too many arguments in command \"%s\" in "
+  char *msg = string_sprintf("Too many arguments in command \"%s\" in "
     "%s", cmd, etext);
+  if (addr != NULL)
+    {
+    addr->transport_return = FAIL;
+    addr->message = msg;
+    }
+  else *errptr = msg;
   return FALSE;
   }
 
@@ -1440,12 +1547,12 @@ DEBUG(7)
   {
   debug_printf("direct command:\n");
   for (i = 0; argv[i] != (char *)0; i++)
-    debug_printf("  argv[%d] = %s\n", i, string_printing(argv[i], FALSE));
+    debug_printf("  argv[%d] = %s\n", i, string_printing(argv[i]));
   }
 
 if (expand_arguments)
   {
-  BOOL allow_dollar_recipients =
+  BOOL allow_dollar_recipients = addr != NULL &&
     addr->parent != NULL && strcmp(addr->parent->orig, "message filter") == 0;
 
   for (i = 0; argv[i] != (char *)0; i++)
@@ -1453,7 +1560,8 @@ if (expand_arguments)
 
     /* Handle special fudge for passing an address list */
 
-    if (strcmp(argv[i], "$pipe_addresses") == 0 ||
+    if (addr != NULL &&
+        strcmp(argv[i], "$pipe_addresses") == 0 ||
         strcmp(argv[i], "${pipe_addresses}") == 0)
       {
       int additional;
@@ -1486,10 +1594,15 @@ if (expand_arguments)
 
       if (expanded_arg == NULL)
         {
-        addr->transport_return = expand_failed;
-        addr->message = string_sprintf("Expansion of \"%s\" "
+        char *msg = string_sprintf("Expansion of \"%s\" "
           "from command \"%s\" in %s failed: %s",
           argv[i], cmd, etext, expand_string_message);
+        if (addr != NULL)
+          {
+          addr->transport_return = expand_failed;
+          addr->message = msg;
+          }
+        else *errptr = msg;
         return FALSE;
         }
       argv[i] = expanded_arg;
@@ -1500,37 +1613,11 @@ if (expand_arguments)
     {
     debug_printf("direct command after expansion:\n");
     for (i = 0; argv[i] != (char *)0; i++)
-      debug_printf("  argv[%d] = %s\n", i, string_printing(argv[i], FALSE));
+      debug_printf("  argv[%d] = %s\n", i, string_printing(argv[i]));
     }
   }
 
 return TRUE;
 }
 
-
-
-
-/*************************************************
-*           Free command list                    *
-*************************************************/
-
-/* Free the store set up for a command list
-
-Argument:   argvptr   pointer to anchor for argv vector
-Returns:    nothing
-*/
-
-void
-transport_free_command(char ***argvptr)
-{
-int i = 0;
-char **argv = *argvptr;
-if (argv == NULL) return;
-
-while (argv[i] != (char *)0) store_free(argv[i++]);
-store_free(argv);
-*argvptr = NULL;
-}
-
-
 /* End of transport.c */
