@@ -43,14 +43,10 @@ optionlist appendfile_transport_options[] = {
       (void *)(offsetof(appendfile_transport_options_block, create_file_string)) },
   { "current_directory", opt_stringptr | opt_public,
       (void *)(offsetof(transport_instance, current_dir)) },
-  { "delivery_date_add", opt_bool,
-      (void *)(offsetof(appendfile_transport_options_block, delivery_date_add)) },
   { "directory",         opt_stringptr,
       (void *)(offsetof(appendfile_transport_options_block, dirname)) },
   { "directory_mode",    opt_octint,
       (void *)(offsetof(appendfile_transport_options_block, dirmode)) },
-  { "envelope_to_add",   opt_bool,
-      (void *)(offsetof(appendfile_transport_options_block, envelope_to_add)) },
   { "file",              opt_stringptr,
       (void *)(offsetof(appendfile_transport_options_block, filename)) },
   { "file_must_exist",   opt_bool,
@@ -67,10 +63,22 @@ optionlist appendfile_transport_options[] = {
       (void *)(offsetof(appendfile_transport_options_block, lockfile_mode)) },
   { "lockfile_timeout",  opt_time,
       (void *)(offsetof(appendfile_transport_options_block, lockfile_timeout)) },
+#ifdef SUPPORT_MAILDIR
   { "maildir_format",    opt_bool,
       (void *)(offsetof(appendfile_transport_options_block, maildir_format )) } ,
   { "maildir_retries",   opt_int,
       (void *)(offsetof(appendfile_transport_options_block, maildir_retries)) },
+  { "maildir_tag",       opt_stringptr,
+      (void *)(offsetof(appendfile_transport_options_block, maildir_tag)) },
+#endif
+#ifdef SUPPORT_MAILSTORE
+  { "mailstore_format",  opt_bool,
+      (void *)(offsetof(appendfile_transport_options_block, mailstore_format )) },
+  { "mailstore_prefix",  opt_stringptr,
+      (void *)(offsetof(appendfile_transport_options_block, mailstore_prefix )) },
+  { "mailstore_suffix",  opt_stringptr,
+      (void *)(offsetof(appendfile_transport_options_block, mailstore_suffix )) },
+#endif
   { "mode",              opt_octint,
       (void *)(offsetof(appendfile_transport_options_block, mode)) },
   { "mode_fail_narrower",opt_bool,
@@ -85,8 +93,6 @@ optionlist appendfile_transport_options[] = {
       (void *)(offsetof(appendfile_transport_options_block, require_lockfile)) },
   { "retry_use_local_part", opt_bool | opt_public,
       (void *)offsetof(transport_instance, retry_use_local_part) },
-  { "return_path_add",   opt_bool,
-      (void *)(offsetof(appendfile_transport_options_block, return_path_add)) },
   { "suffix",            opt_stringptr,
       (void *)(offsetof(appendfile_transport_options_block, suffix)) },
   { "use_crlf",          opt_bool,
@@ -115,7 +121,11 @@ appendfile_transport_options_block appendfile_transport_option_defaults = {
   "\n",           /* suffix */
   "anywhere",     /* create_file_string (string value for create_file) */
   NULL,           /* quota */
+  NULL,           /* maildir_tag */
+  NULL,           /* mailstore_prefix */
+  NULL,           /* mailstore_suffix */
   0,              /* quota_value */
+
   0600,           /* mode */
   0700,           /* dirmode */
   0600,           /* lockfile_mode */
@@ -133,12 +143,10 @@ appendfile_transport_options_block appendfile_transport_option_defaults = {
   TRUE,           /* use_fcntl */
   FALSE,          /* use_crlf */
   TRUE,           /* from_hack */
-  TRUE,           /* return_path_add */
-  TRUE,           /* delivery_date_add */
-  TRUE,           /* envelope_to_add */
   FALSE,          /* file_must_exist */
   TRUE,           /* mode_fail_narrower */
-  FALSE           /* maildir format */
+  FALSE,          /* maildir_format */
+  FALSE           /* mailstore_format */
 };
 
 
@@ -154,13 +162,15 @@ by the user.
 
 Arguments:
   tblock     points to the transport instance
+  addrlist   addresses about to be delivered
   errmsg     where to put an error message
 
 Returns:     OK, FAIL, or DEFER
 */
 
 static int
-appendfile_transport_setup(transport_instance *tblock, char **errmsg)
+appendfile_transport_setup(transport_instance *tblock, address_item *addrlist,
+  char **errmsg)
 {
 appendfile_transport_options_block *ob =
   (appendfile_transport_options_block *)(tblock->options_block);
@@ -248,11 +258,18 @@ if (ob->filename != NULL)
   }
 
 /* If a directory name was specified, it must be an absolute path. Can check
-here only if there are no expansions. */
+here only if there are no expansions. Only one of maildir or mailstore may be
+specified. */
 
-if (ob->dirname != NULL && ob->dirname[0] != '/' && ob->dirname[0] != '$')
-  log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s transport:\n  "
-  "the directory option must specify an absolute path", tblock->name);
+if (ob->dirname != NULL)
+  {
+  if (ob->dirname[0] != '/' && ob->dirname[0] != '$')
+    log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s transport:\n  "
+      "the directory option must specify an absolute path", tblock->name);
+  if (ob->maildir_format && ob->mailstore_format)
+    log_write(0, LOG_PANIC_DIE|LOG_CONFIG2, "%s transport:\n "
+      "only one of maildir and mailstore may be specified", tblock->name);
+  }
 
 /* If a fixed uid field is set, then a gid field must also be set. */
 
@@ -358,55 +375,65 @@ close(fd);
 
 
 /*************************************************
-*       Write block and check for errors         *
+*    Check size of directory's files for quota   *
 *************************************************/
 
-/* On failing to write the given number of bytes, but having written
-some, try to write the rest in order to force an over-quota error to
-appear. Otherwise no error is set.
+/* This function is called if quota is set for one of the delivery modes that
+delivers into a specific directory. It scans the directory and stats all the
+files in order to get a total size. This is an expensive thing to do, but some
+people are prepared to bear the cost.
 
 Arguments:
-  fd        file descriptor to write to
-  buffer    pointer to block
-  count     number of bytes
+  dirname       the name of the directory
 
-Returns:    TRUE if write successful
+Returns:        the sum of the sizes of the stattable files
+                zero if the directory cannot be opened
 */
 
-static BOOL
-write_block(int fd, char *buffer, int count)
+static int
+check_dir_size(char *dirname)
 {
-int rc;
-if ((rc = write(fd, buffer, count)) != count &&
-    (rc < 0 || write(fd, buffer+rc, count-rc) != count-rc)) return FALSE;
-return TRUE;
+DIR *dir;
+int sum = 0;
+struct dirent *ent;
+struct stat statbuf;
+
+dir = opendir(dirname);
+if (dir == NULL) return 0;
+
+while ((ent = readdir(dir)) != NULL)
+  {
+  char *name = ent->d_name;
+  char buffer[1024];
+
+  if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0) continue;
+  if (!string_format(buffer, sizeof(buffer), "%s/%s", dirname, name))
+    {
+    log_write(0, LOG_PANIC|LOG_MAIN, "name too long in check_dir_size: "
+      "dir=%s name=%s", dirname, name);
+    continue;
+    }
+
+  if (stat(buffer, &statbuf) < 0)
+    {
+    DEBUG(9) debug_printf("check_dir_size: stat error %d for %s: %s\n",
+      errno, buffer, strerror(errno));
+    continue;
+    }
+
+  if ((statbuf.st_mode & S_IFREG) != 0)
+    sum += statbuf.st_size;
+  else if ((statbuf.st_mode & S_IFDIR) != 0)
+    sum += check_dir_size(buffer);
+  }
+
+closedir(dir);
+DEBUG(9) debug_printf("check_dir_size: dir=%s sum=%d\n", dirname, sum);
+return sum;
 }
 
 
 
-/*************************************************
-*             Write formatted string             *
-*************************************************/
-
-/*
-Arguments:
-  fd          file descriptor
-  format      string format
-  ...         arguments for format
-
-Returns:      the yield of write_block()
-*/
-
-static BOOL
-write_string(int fd, char *format, ...)
-{
-va_list ap;
-va_start(ap, format);
-if (!string_vformat(big_buffer, BIG_BUFFER_SIZE, format, ap))
-  log_write(0, LOG_PANIC_DIE, "overlong formatted string in appendfile");
-va_end(ap);
-return write_block(fd, big_buffer, strlen(big_buffer));
-}
 
 
 
@@ -433,7 +460,10 @@ directory name) is given:
 
 (2f) If the file name to be expanded contains a reference to $local_part,
      check that the user name doesn't contain a / because that would cause
-     a reference to another directory.
+     a reference to another directory. **** This test is no longer applied. It
+     is a pretty weak test, and can cause trouble if $local_part appears in the
+     expanded string, but isn't in fact used as part of the file name (e.g. it
+     is a lookup key). ****
 
 (3f) Expand the file name.
 
@@ -507,7 +537,9 @@ The following items apply in the case when a directory name is given:
 
 (2d) Create a new file in the directory using a temporary name, by opening for
      writing and with O_CREAT. If maildir format is being used, the file
-     is created in a temporary subdirectory with a prescribed name.
+     is created in a temporary subdirectory with a prescribed name. If
+     mailstore format is being used, the envelope file is first created with a
+     temporary name, then the data file.
 
 The following items apply in all cases:
 
@@ -534,7 +566,8 @@ The following applies after writing a unique file in a directory:
 
 (12d)For non-maildir-format: Generate a unique name for the file from the time
      and the inode number, and rename the file after writing it. For maildir
-     format, rename the file into the new directory.
+     format, rename the file into the new directory. For mailstore format,
+     rename the envelope file to its correct name.
 
 This transport yields FAIL only when a file name is generated by an alias or
 forwarding operation and attempting to open it gives EPERM, ENOENT, or EACCES.
@@ -552,15 +585,16 @@ struct stat statbuf;
 char *path;
 char *filename;
 char *hitchname;
+char *dataname;
 char *lockname = NULL;     /* keep picky compiler happy */
 char *newname = NULL;      /* ditto */
 char *cr = "";
 time_t now = time(NULL);
 struct utimbuf times;
 BOOL smtp_dots = FALSE;
-BOOL return_path_add = ob->return_path_add;
-BOOL delivery_date_add = ob->delivery_date_add;
-BOOL envelope_to_add = ob->envelope_to_add;
+BOOL return_path_add = tblock->return_path_add;
+BOOL delivery_date_add = tblock->delivery_date_add;
+BOOL envelope_to_add = tblock->envelope_to_add;
 BOOL isdirectory = FALSE;
 uid_t uid = getuid();
 gid_t gid = getgid();
@@ -619,6 +653,11 @@ else
   doesn't attempt to change directory. Note that deliver_localpart may be
   unset during some deliveries (e.g. batched SMTP). */
 
+  /**** This check is probably not very helpful; people don't usually have
+  slashes in user names, and if they are processing random local parts they
+  can either check explicitly, or just let it fail because the directory
+  doesn't exist.
+
   if (deliver_localpart != NULL &&
       strchr(deliver_localpart, '/') != NULL &&
      (strstr(fdname, "${local_part}") != NULL ||
@@ -629,6 +668,7 @@ else
     addr->special_action = SPECIAL_FREEZE;
     return;
     }
+  ****/
 
   /* Check that the expanded name is absolute. This check is needed in case
   it started with an expansion item. If not, the check is done at
@@ -651,8 +691,8 @@ DEBUG(9) debug_printf("appendfile: mode=%o notify_comsat=%d quota=%d\n"
   mode, ob->notify_comsat, ob->quota_value,
   isdirectory? "directory" : "file",
   path,
-  (ob->prefix == NULL)? "null" : string_printing(ob->prefix, FALSE),
-  (ob->suffix == NULL)? "null" : string_printing(ob->suffix, FALSE));
+  (ob->prefix == NULL)? "null" : string_printing(ob->prefix),
+  (ob->suffix == NULL)? "null" : string_printing(ob->suffix));
 
 
 /* If the -N option is set, can't do any more. */
@@ -679,9 +719,10 @@ if (!isdirectory)
     return;
     }
 
-  /* The name of the file to be opened */
+  /* The name of the file to be opened, and the file to which the data
+  is written. */
 
-  filename = path;
+  dataname = filename = path;
 
   /* The locking of mailbox files is worse than the naming of cats, which is
   known to be "a difficult matter" (T.S. Eliot) and just as cats must have
@@ -764,11 +805,8 @@ if (!isdirectory)
 
   if (ob->use_lockfile)
     {
-    len = (int)strlen(filename);
-    lockname = store_malloc(len + 8);
-    sprintf(lockname, "%s.lock", filename);
-    hitchname = store_malloc(len + 32 + (int)strlen(primary_hostname));
-    sprintf(hitchname, "%s.%s.%08x.%08x", lockname, primary_hostname,
+    lockname = string_sprintf("%s.lock", filename);
+    hitchname = string_sprintf( "%s.%s.%08x.%08x", lockname, primary_hostname,
       (unsigned int)now, (unsigned int)getpid());
 
     DEBUG(9) debug_printf("lock name: %s\nhitch name: %s\n", lockname,
@@ -1171,21 +1209,22 @@ if (!isdirectory)
   saved_size = statbuf.st_size;
   }
 
-/* Handle the case of creating a unique file in a given directory. A temporary
-name is used to create the file. Later, when it is written, the name is changed
-to a unique one. There is no need to lock the file. An attempt is made to
-create the directory if it does not exist. */
+/* Handle the case of creating a unique file in a given directory (not in
+maildir or mailstore format). A temporary name is used to create the file.
+Later, when it is written, the name is changed to a unique one. There is no
+need to lock the file. An attempt is made to create the directory if it does
+not exist. */
 
-else if (!ob->maildir_format)
+else if (!ob->maildir_format && !ob->mailstore_format)
   {
   DEBUG(9) debug_printf("delivering to unique file (non-maildir)\n");
-  filename = string_sprintf("%s/temp.%d.%s", path, (int)getpid(),
-    primary_hostname);
+  filename = dataname =
+    string_sprintf("%s/temp.%d.%s", path, (int)getpid(), primary_hostname);
   fd = open(filename, O_WRONLY|O_CREAT, mode);
   if (fd < 0 &&                                    /* failed to open, and */
-      (errno != ENOENT ||                          /* either not exists */
+      (errno != ENOENT ||                          /* either not non-exist */
        !ob->create_directory ||                    /* or not allowed to make */
-       mkdir(path, ob->dirmode) < 0 ||             /* or failed to create dir */
+       !directory_make(NULL, path, ob->dirmode, FALSE) ||  /* or failed to create dir */
        (fd = open(filename, O_WRONLY|O_CREAT, mode)) < 0)) /* or then failed to open */
     {
     addr->basic_errno = errno;
@@ -1215,12 +1254,15 @@ else if (!ob->maildir_format)
     string_base62(statbuf.st_ino));
   }
 
+#ifdef SUPPORT_MAILDIR
+
 /* Handle the case of a unique file in maildir format. The file is written to
 the tmp subdirectory, with a prescribed form of name. */
 
-else
+else if (ob->maildir_format)
   {
   char *subdirs[] = { "/tmp", "/new", "/cur" };
+  char *tag = "";
 
   DEBUG(9) debug_printf("delivering in maildir format in %s\n", path);
 
@@ -1267,10 +1309,10 @@ else
 
       if (errno == ENOENT && ob->create_directory)
         {
-        if (mkdir(dir, ob->dirmode) != 0)
+        if (!directory_make(NULL, dir, ob->dirmode, FALSE))
           {
           if (errno == EEXIST) continue;     /* repeat the race loop */
-          addr->message = string_sprintf ("cannot mkdir %s%s", path, mdir);
+          addr->message = string_sprintf("cannot create %s%s", path, mdir);
 	  addr->basic_errno = errno;
 	  addr->special_action = SPECIAL_FREEZE;
 	  return;
@@ -1281,7 +1323,8 @@ else
 
       /* stat() error other than ENOENT, or ENOENT and not creatable */
 
-      addr->message = string_sprintf("stat() error for %s%s", path, mdir);
+      addr->message = string_sprintf("stat() error for %s%s: %s", path, mdir,
+        strerror(errno));
       addr->basic_errno = errno;
       addr->special_action = SPECIAL_FREEZE;
       return;
@@ -1292,7 +1335,7 @@ else
 
     if (j >= 10)
       {
-      addr->message = string_sprintf("existence of %s%s unclear\n");
+      addr->message = string_sprintf("existence of %s%s unclear\n", path, mdir);
       addr->basic_errno = errno;
       addr->special_action = SPECIAL_FREEZE;
       return;
@@ -1309,17 +1352,64 @@ else
       }
     }
 
-  /* All relevant directories now exist. Attempt to open the temporary
-  file a limited number of times. I think this rather scary-looking for
-  statement is actually OK. If open succeeds, the loop is broken; if not,
-  there is a test on the value of i. */
+  /* All relevant directories now exist. Set up the maildir tag, if it is
+  configured, by copying the expanded, validated string into the tag variable,
+  preceded by a colon. Otherwise leave tag as the empty string. */
+
+  if (ob->maildir_tag != NULL)
+    {
+    char *optr;
+    char *iptr = expand_string(ob->maildir_tag);
+
+    /* Expansion failure is soft if forced */
+
+    if (iptr == NULL)
+     {
+     if (!expand_string_forcedfail)
+       {
+       addr->transport_return = PANIC;
+       addr->message = string_sprintf("Expansion of \"%s\" (maildir_tag "
+	 "for %s transport) failed: %s", ob->maildir_tag, tblock->name,
+	 expand_string_message);
+       return;
+       }
+     }
+
+    /* The validated string will never be longer than the input string + 1.
+    Ignore characters other than [A-Za-z0-9.:_-,], and put a colon at the
+    start. */
+
+    else
+      {
+      optr = tag = store_get(strlen(iptr) + 2);
+      *optr++ = ':';
+
+      while (*iptr != 0)
+	{
+	if (isalnum(*iptr) ||
+	    *iptr == '.' || *iptr == ':' ||
+	    *iptr == '_' || *iptr == '-' || *iptr == ',')
+	  *optr++ = *iptr;
+	iptr++;
+	}
+      *optr = 0;
+
+      /* If the string ends up empty, scrap it. */
+
+      if (tag[1] == 0) *tag = 0;
+      }
+    }
+
+  /* Attempt to open the temporary file a limited number of times. I think this
+  rather scary-looking for statement is actually OK. If open succeeds, the loop
+  is broken; if not, there is a test on the value of i. */
 
   for (i = 1;; i++)
     {
-    char *basename = string_sprintf("%lu.%lu.%s", time(NULL), getpid(),
-      primary_hostname);
+    char *basename = string_sprintf("%lu.%lu.%s%s", time(NULL), getpid(),
+      primary_hostname, tag);
 
-    filename = string_sprintf("tmp/%s", basename);
+    filename = dataname = string_sprintf("tmp/%s", basename);
     newname = string_sprintf("new/%s", basename);
 
     if (stat(filename, &statbuf) == 0)
@@ -1332,20 +1422,27 @@ else
         filename, strerror(errno));
       }
 
+    /* If the name is too long, try again without the tag */
+
+    else if (errno == ENAMETOOLONG)
+      {
+      DEBUG(2) debug_printf("maildir_tag (%s) makes filename too long\n", tag);
+      *tag = 0;
+      continue;    /* Repeat without delay */
+      }
+
+    /* Too many retries - give up */
+
     if (i >= ob->maildir_retries)
       {
       addr->message = string_sprintf ("failed to open %s (%d tr%s)",
         filename, i, (i == 1)? "y" : "ies");
       addr->basic_errno = errno;
-      addr->transport_return = DEFER;
       return;
       }
 
     /* Open or stat failed but we haven't tried too many times yet. */
 
-    store_free(basename);
-    store_free(filename);
-    store_free(newname);
     sleep(2);
     }
 
@@ -1356,19 +1453,154 @@ else
   chmod(filename, mode);
   }
 
+#endif
+
+#ifdef SUPPORT_MAILSTORE
+
+/* Handle the case of a unique file in mailstore format. First write the
+envelope to a temporary file, then open the main file. The unique base name for
+the files consists of the message id plus the pid of this delivery process. */
+
+else
+  {
+  FILE *env_file;
+  address_item *taddr;
+  char *basename = string_sprintf("%s/%s-%s", path, message_id,
+    string_base62((long int)getpid()));
+
+  DEBUG(9) debug_printf("delivering in mailstore format in %s\n", path);
+
+  filename = string_sprintf("%s.tmp", basename);
+  newname  = string_sprintf("%s.env", basename);
+  dataname = string_sprintf("%s.msg", basename);
+
+  fd = open(filename, O_WRONLY|O_CREAT|O_EXCL, mode);
+  if (fd < 0 &&                                    /* failed to open, and */
+      (errno != ENOENT ||                          /* either not non-exist */
+       !ob->create_directory ||                    /* or not allowed to make */
+       !directory_make(NULL, path, ob->dirmode, FALSE) ||  /* or failed to create dir */
+       (fd = open(filename, O_WRONLY|O_CREAT|O_EXCL, mode)) < 0)) /* or then failed to open */
+    {
+    addr->basic_errno = errno;
+    addr->message = string_sprintf("while creating file %s", filename);
+    return;
+    }
+
+  /* Why are these here? Put in because they are present in the non-maildir
+  directory case above. */
+
+  chown(filename, uid, gid);
+  chmod(filename, mode);
+
+  /* Write the envelope file, then close it. */
+
+  env_file = fdopen(fd, "w+");
+
+  if (ob->mailstore_prefix != NULL)
+    {
+    char *s = expand_string(ob->mailstore_prefix);
+    if (s == NULL)
+      {
+      if (!expand_string_forcedfail)
+        {
+        addr->transport_return = PANIC;
+        addr->message = string_sprintf("Expansion of \"%s\" (mailstore prefix "
+          "for %s transport) failed: %s", ob->mailstore_prefix, tblock->name,
+          expand_string_message);
+        fclose(env_file);
+        unlink(filename);
+        return;
+        }
+      }
+    else
+      {
+      int n = (int)strlen(s);
+      fprintf(env_file, "%s", s);
+      if (n == 0 || s[n-1] != '\n') fprintf(env_file, "\n");
+      }
+    }
+
+  fprintf(env_file, "%s\n", sender_address);
+
+  for (taddr = addr; taddr!= NULL; taddr = taddr->next)
+    {
+    if (taddr->local_part[0] == ',' || taddr->local_part[0] == ':')
+      fprintf(env_file, "@%s%s\n", taddr->domain, taddr->local_part);
+    else
+      fprintf(env_file, "%s@%s\n", taddr->local_part, taddr->domain);
+    }
+
+  if (ob->mailstore_suffix != NULL)
+    {
+    char *s = expand_string(ob->mailstore_suffix);
+    if (s == NULL)
+      {
+      if (!expand_string_forcedfail)
+        {
+        addr->transport_return = PANIC;
+        addr->message = string_sprintf("Expansion of \"%s\" (mailstore suffix "
+          "for %s transport) failed: %s", ob->mailstore_suffix, tblock->name,
+          expand_string_message);
+        fclose(env_file);
+        unlink(filename);
+        return;
+        }
+      }
+    else
+      {
+      int n = (int)strlen(s);
+      fprintf(env_file, "%s", s);
+      if (n == 0 || s[n-1] != '\n') fprintf(env_file, "\n");
+      }
+    }
+
+  fflush(env_file);
+  if (close(fd) < 0)
+    {
+    addr->basic_errno = errno;
+    addr->message = string_sprintf("while closing %s", filename);
+    unlink(filename);
+    return;
+    }
+
+  DEBUG(9) debug_printf("Envelope file %s written\n", filename);
+
+  /* Now open the data file */
+
+  fd = open(dataname, O_WRONLY|O_CREAT|O_EXCL, mode);
+  if (fd < 0)
+    {
+    addr->basic_errno = errno;
+    addr->message = string_sprintf("while creating file %s", dataname);
+    unlink(filename);
+    return;
+    }
+
+  /* Why are these here? Put in because they are present in the non-maildir
+  directory case above. */
+
+  chown(dataname, uid, gid);
+  chmod(dataname, mode);
+  }
+
+#endif
+
+/* If quota is set for any of the directory formats, find the current usage
+within that directory. This is expensive, but happens only if configures. */
+
+if (isdirectory && ob->quota_value > 0) saved_size = check_dir_size(path);
 
 /* At last we can write the message to the file, preceded by any configured
 prefix line, and followed by any configured suffix line. If there are any
 writing errors, we must defer. */
 
-DEBUG(9) debug_printf("writing to file %s\n", filename);
+DEBUG(9) debug_printf("writing to file %s\n", dataname);
 
 yield = OK;
 errno = 0;
 
 /* If there is a local quota setting, check that we are not going to exceed it
-with this message. This of course applies only file appending, not when writing
-in a directory. */
+with this message. */
 
 if (ob->quota_value > 0)
   {
@@ -1388,11 +1620,10 @@ if (tblock->local_smtp != local_smtp_off && yield == OK)
   {
   if (ob->use_crlf) cr = "\r";
   smtp_dots = TRUE;
-  return_path_add = delivery_date_add = envelope_to_add = FALSE;
   if (tblock->bsmtp_helo &&
-      !write_string(fd, "HELO %s%s\n", primary_hostname, cr))
+      !transport_write_string(fd, "HELO %s%s\n", primary_hostname, cr))
     yield = DEFER;
-  else if (!write_string(fd, "MAIL FROM: <%s>%s\n",
+  else if (!transport_write_string(fd, "MAIL FROM:<%s>%s\n",
     ((addr->errors_address != NULL)? addr->errors_address : sender_address),
       cr)) yield = DEFER;
   else
@@ -1402,13 +1633,13 @@ if (tblock->local_smtp != local_smtp_off && yield == OK)
       {
       address_item *b = (a->pfr)? a->parent: a;
       if ((b->local_part[0] == ',' || b->local_part[0] == ':')?
-        !write_string(fd,
-          "RCPT TO: <@%s%s>%s\n", b->domain, b->local_part, cr) :
-        !write_string(fd,
-          "RCPT TO: <%s@%s>%s\n", b->local_part, b->domain, cr))
+        !transport_write_string(fd,
+          "RCPT TO:<@%s%s>%s\n", b->domain, b->local_part, cr) :
+        !transport_write_string(fd,
+          "RCPT TO:<%s@%s>%s\n", b->local_part, b->domain, cr))
         { yield = DEFER; break; }
       }
-    if (yield == OK && !write_string(fd, "DATA%s\n", cr)) yield = DEFER;
+    if (yield == OK && !transport_write_string(fd, "DATA%s\n", cr)) yield = DEFER;
     }
   }
 
@@ -1424,7 +1655,7 @@ if (yield == OK && ob->prefix != NULL)
       "transport) failed", ob->prefix, tblock->name);
     goto RETURN;
     }
-  if (!write_string(fd, "%s", prefix)) yield = DEFER;
+  if (!transport_write_string(fd, "%s", prefix)) yield = DEFER;
   }
 
 /* The options to write_message request a return-path header if configured, the
@@ -1435,13 +1666,12 @@ off for local_smtp). */
 
 if (yield == OK &&
   !transport_write_message(addr, fd,
-    topt_writing_file |
-      (return_path_add? topt_add_return_path : 0) |
-      (delivery_date_add? topt_add_delivery_date : 0) |
-      (envelope_to_add? topt_add_envelope_to : 0) |
-      (ob->from_hack? topt_from_hack : 0) |
-      (ob->use_crlf? topt_use_crlf : 0) |
-      (smtp_dots? topt_smtp_dots : 0),
+    (return_path_add? topt_add_return_path : 0) |
+    (delivery_date_add? topt_add_delivery_date : 0) |
+    (envelope_to_add? topt_add_envelope_to : 0) |
+    (ob->from_hack? topt_from_hack : 0) |
+    (ob->use_crlf? topt_use_crlf : 0) |
+    (smtp_dots? topt_smtp_dots : 0),
     addr->errors_address, 0, tblock->add_headers, tblock->remove_headers))
   yield = DEFER;
 
@@ -1457,13 +1687,13 @@ if (yield == OK && ob->suffix != NULL)
       "transport) failed", ob->suffix, tblock->name);
     goto RETURN;
     }
-  if (!write_string(fd, "%s", suffix)) yield = DEFER;
+  if (!transport_write_string(fd, "%s", suffix)) yield = DEFER;
   }
 
 /* If local_smtp, write the terminating dot. */
 
 if (yield == OK && tblock->local_smtp != local_smtp_off &&
-  !write_string(fd, ".%s\n", cr)) yield = DEFER;
+  !transport_write_string(fd, ".%s\n", cr)) yield = DEFER;
 
 /* Force out the remaining data to check for any errors */
 
@@ -1491,7 +1721,7 @@ if (yield != OK)
       "writing to file %s", filename);
     #endif
     DEBUG(9) debug_printf("System quota exceeded for %s: time since file "
-      "read = %s\n", filename, readconf_printtime(addr->more_errno));
+      "read = %s\n", dataname, readconf_printtime(addr->more_errno));
     }
 
   /* Handle Exim's own quota-imposition */
@@ -1500,9 +1730,9 @@ if (yield != OK)
     {
     addr->more_errno = time(NULL) - times.actime;
     addr->message = string_sprintf("MTA-imposed quota exceeded while "
-      "writing to file %s", filename);
+      "writing to file %s", dataname);
     DEBUG(9) debug_printf("Exim quota exceeded for %s: time since file "
-      "read = %s\n", filename, readconf_printtime(addr->more_errno));
+      "read = %s\n", dataname, readconf_printtime(addr->more_errno));
     }
 
   /* Handle a process failure while writing via a filter; the return
@@ -1510,7 +1740,7 @@ if (yield != OK)
 
   else if (errno == ERRNO_FILTER_FAIL)
     addr->message = string_sprintf("filter process failure %d while writing "
-      "to %s", addr->more_errno, filename);
+      "to %s", addr->more_errno, dataname);
 
   /* Handle failure to add a header - usually expansion failure. */
 
@@ -1518,13 +1748,21 @@ if (yield != OK)
     {
     yield = PANIC;
     addr->message = string_sprintf("failed to expand added headers while "
-      "writing to %s: %s", filename, expand_string_message);
+      "writing to %s: %s", dataname, expand_string_message);
+    }
+
+  /* Handle failure to complete writing of a data block */
+
+  else if (errno == ERRNO_WRITEINCOMPLETE)
+    {
+    addr->message = string_sprintf("failed to write data block while "
+      "writing to %s", dataname);
     }
 
   /* For other errors, a general-purpose explanation */
 
   else addr->message = string_sprintf("error while writing to %s",
-    filename);
+    dataname);
 
   /* For a file, reset the file size to what it was before we started, leaving
   the last modification time unchanged, so it will get reset also. All systems
@@ -1546,7 +1784,7 @@ else
   addr->basic_errno = 0;
 
   /* Handle the case of writing to a new file in a directory. This applies
-  both to maildir and non-maildir formats. */
+  to all single-file formats. */
 
   if (isdirectory)
     {
@@ -1561,7 +1799,8 @@ else
     else if (close(fd) < 0)
       {
       addr->basic_errno = errno;
-      addr->message = string_sprintf("close() error for %s", filename);
+      addr->message = string_sprintf("close() error for %s",
+        (ob->mailstore_format)? dataname : filename);
       yield = DEFER;
       }
 
@@ -1578,7 +1817,7 @@ else
       else
         {
         DEBUG(9) debug_printf("renamed %s as %s\n", filename, newname);
-        filename = NULL;   /* Prevents attempt to unlink at end */
+        filename = dataname = NULL;   /* Prevents attempt to unlink at end */
         }
       }
     }
@@ -1605,7 +1844,7 @@ addr->transport_return = yield;
 
 /* Close the file, which will release the fcntl lock. For a directory write it
 is closed above, except in cases of error which goto RETURN, when we also need
-to remove the original file. If everything has gone right but close fails,
+to remove the original file(s). If everything has gone right but close fails,
 defer the message. Then unlink the lock file, if present. This point in the
 code is jumped to from a number of places when errors are detected, in order to
 get the file closed and the lock file tidied away. */
@@ -1620,7 +1859,11 @@ if (fd >= 0 && close(fd) < 0 && yield == OK)
   }
 
 if (hd >= 0) unlink(lockname);
-if (isdirectory && filename != NULL) unlink(filename);
+if (isdirectory && filename != NULL)
+  {
+  unlink(filename);
+  if (dataname != filename) unlink(dataname);
+  }
 }
 
 /* End of transport/appendfile.c */
