@@ -41,12 +41,8 @@ optionlist pipe_transport_options[] = {
       (void *)(offsetof(pipe_transport_options_block, cmd)) },
   { "current_directory", opt_stringptr | opt_public,
       (void *)(offsetof(transport_instance, current_dir)) },
-  { "delivery_date_add", opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, delivery_date_add)) },
   { "directory",         opt_stringptr | opt_public | opt_hidden,
       (void *)(offsetof(transport_instance, home_dir)) },
-  { "envelope_to_add", opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, envelope_to_add)) },
   { "freeze_exec_fail",  opt_bool,
       (void *)(offsetof(pipe_transport_options_block, freeze_exec_fail)) },
   { "from_hack",         opt_bool,
@@ -81,10 +77,10 @@ optionlist pipe_transport_options[] = {
       (void *)(offsetof(transport_instance, return_fail_output)) },
   { "return_output",     opt_bool | opt_public,
       (void *)(offsetof(transport_instance, return_output)) },
-  { "return_path_add",   opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, return_path_add)) },
   { "suffix",            opt_stringptr,
       (void *)(offsetof(pipe_transport_options_block, suffix)) },
+  { "temp_errors",       opt_stringptr,
+      (void *)(offsetof(pipe_transport_options_block, temp_errors)) },
   { "timeout",           opt_time,
       (void *)(offsetof(pipe_transport_options_block, timeout)) },
   { "umask",             opt_octint,
@@ -112,15 +108,14 @@ pipe_transport_options_block pipe_transport_option_defaults = {
   "From ${if def:return_path{$return_path}{MAILER-DAEMON}} ${tod_bsdinbox}\n",
                   /* prefix */
   "\n",           /* suffix */
+  mac_expanded_string(EX_TEMPFAIL) ":"    /* temp_errors */
+  mac_expanded_string(EX_CANTCREAT),
   022,            /* umask */
   20480,          /* max_output */
   60*60,          /* timeout */
   FALSE,          /* freeze_exec_fail */
   FALSE,          /* from_hack */
   FALSE,          /* ignore_status */
-  FALSE,          /* return_path_add */
-  FALSE,          /* delivery_date_add */
-  FALSE,          /* envelope_to_add */
   FALSE,          /* restrict_to_path */
   FALSE,          /* use_shell */
 };
@@ -159,6 +154,17 @@ if (tblock->uid_set && !tblock->gid_set && tblock->expand_gid == NULL)
   log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
     "user set without group for the %s transport", tblock->name);
 
+/* Temp_errors must consist only of digits and colons */
+
+if (ob->temp_errors != NULL)
+  {
+  size_t p = strspn(ob->temp_errors, "0123456789:");
+  if (ob->temp_errors[p] != 0)
+    log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
+      "temp_errors option must be a list of numbers in the %s transport",
+        tblock->name);
+  }
+
 /* Only one of return_output/return_fail_output or log_output/log_fail_output
 should be set. */
 
@@ -229,7 +235,7 @@ is in the addresses). */
 sprintf(buffer, "%.50s transport", tname);
 if (!transport_set_up_command(argvptr, cmd, expand_arguments,
       (addr->local_part[0]== '|')? FAIL : PANIC,
-      addr, buffer))
+      addr, buffer, NULL))
   return FALSE;
 
 /* Point to the set-up arguments. */
@@ -348,7 +354,7 @@ set_up_shell_command(char ***argvptr, char *cmd, BOOL expand_arguments,
 {
 char **argv;
 
-*argvptr = argv = (char **)store_malloc((4)*sizeof(char *));
+*argvptr = argv = store_get((4)*sizeof(char *));
 
 argv[0] = "/bin/sh";
 argv[1] = "-c";
@@ -374,7 +380,7 @@ if (expand_arguments)
 
     if (p[-1] == '{') { q++; p--; }
 
-    s = store_malloc(size);
+    s = store_get(size);
     offset = p - cmd - 1;
     strncpy(s, cmd, offset);
 
@@ -427,14 +433,16 @@ int envcount = 0;
 pipe_transport_options_block *ob =
   (pipe_transport_options_block *)(tblock->options_block);
 BOOL smtp_dots = FALSE;
-BOOL return_path_add = ob->return_path_add;
-BOOL delivery_date_add = ob->delivery_date_add;
-BOOL envelope_to_add = ob->envelope_to_add;
+BOOL written_ok = FALSE;
+BOOL return_path_add = tblock->return_path_add;
+BOOL delivery_date_add = tblock->delivery_date_add;
+BOOL envelope_to_add = tblock->envelope_to_add;
 BOOL expand_arguments;
 char **argv;
 char *envp[20];
 char *env_local_part;
 char *cmd;
+char *eol = (ob->use_crlf)? "\r\n" : "\n";
 
 DEBUG(2) debug_printf("%s transport entered\n", tblock->name);
 
@@ -637,42 +645,31 @@ if (tblock->local_smtp != local_smtp_off)
     addr->errors_address : sender_address;
 
   smtp_dots = TRUE;
-  return_path_add = delivery_date_add = envelope_to_add = FALSE;
 
-  if (tblock->bsmtp_helo)
-    {
-    write(fd, "HELO ", 5);
-    write(fd, primary_hostname, (int)strlen(primary_hostname));
-    if (ob->use_crlf) write(fd, "\r\n", 2);
-      else write(fd, "\n", 1);
-    }
+  if (tblock->bsmtp_helo &&
+      !transport_write_string(fd, "HELO %s%s", primary_hostname, eol))
+    goto END_WRITE;
 
-  write(fd, "MAIL FROM: <", 12);
-  write(fd, sender, (int)strlen(sender));
-  if (ob->use_crlf) write(fd, ">\r\n", 3);
-    else write(fd, ">\n", 2);
+  if (!transport_write_string(fd, "MAIL FROM:<%s>%s", sender, eol))
+    goto END_WRITE;
 
   for (a = addr; a != NULL; a = a->next)
     {
-    write(fd, "RCPT TO: <", 10);
     if (a->local_part[0] == ',' || a->local_part[0] == ':')
       {
-      write(fd, "@", 1);
-      write(fd, a->domain, (int)strlen(a->domain));
-      write(fd, a->local_part, (int)strlen(a->local_part));
+      if (!transport_write_string(fd,
+          "RCPT TO:<@%s%s>%s", a->domain, a->local_part, eol))
+        goto END_WRITE;
       }
     else
       {
-      write(fd, a->local_part, (int)strlen(a->local_part));
-      write(fd, "@", 1);
-      write(fd, a->domain, (int)strlen(a->domain));
+      if (!transport_write_string(fd,
+          "RCPT TO:<%s@%s>%s", a->local_part, a->domain, eol))
+        goto END_WRITE;
       }
-    if (ob->use_crlf) write(fd, ">\r\n", 3);
-      else write(fd, ">\n", 2);
     }
 
-  if (ob->use_crlf) write(fd, "DATA\r\n", 6);
-    else write(fd, "DATA\n", 5);
+  if (!transport_write_string(fd, "DATA%s", eol)) goto END_WRITE;
   }
 
 /* Now any other configured prefix. */
@@ -687,9 +684,11 @@ if (ob->prefix != NULL)
       "transport) failed: %s", ob->prefix, tblock->name, expand_string_message);
     return;
     }
-  write(fd, prefix, (int)strlen(prefix));
+  if (!transport_write_block(fd, prefix, (int)strlen(prefix))) goto END_WRITE;
   }
 
+/* Now the actual message */
+
 if (!transport_write_message(addr, fd,
     (return_path_add? topt_add_return_path : 0) |
     (delivery_date_add? topt_add_delivery_date : 0) |
@@ -698,25 +697,7 @@ if (!transport_write_message(addr, fd,
     (ob->use_crlf? topt_use_crlf : 0) |
     (smtp_dots? topt_smtp_dots : 0),
     addr->errors_address, 0, tblock->add_headers, tblock->remove_headers))
-  {
-  if (errno != EPIPE && errno != ERRNO_WRITEINCOMPLETE)
-    {
-    addr->transport_return = PANIC;
-    addr->basic_errno = errno;
-    if (errno == ERRNO_ADDHEADER_FAIL)
-      addr->message = string_sprintf("Failed to expand added headers");
-    else if (errno == ERRNO_FILTER_FAIL)
-      addr->message = string_sprintf("Filter process failure");
-    else
-      addr->message = string_sprintf("Error %d", errno);
-    return;
-    }
-  else DEBUG(9)
-    {
-    debug_printf("transport error %s ignored\n",
-      (errno == EPIPE)? "EPIPE" : "WRITEINCOMPLETE");
-    }
-  }
+  goto END_WRITE;
 
 /* Now any configured suffix */
 
@@ -730,21 +711,57 @@ if (ob->suffix != NULL)
       "transport) failed: %s", ob->suffix, tblock->name, expand_string_message);
     return;
     }
-  write(fd, suffix, (int)strlen(suffix));
+  if (!transport_write_block(fd, suffix, (int)strlen(suffix))) goto END_WRITE;
   }
 
 /* If local_smtp, write the terminating dot. */
 
-if (tblock->local_smtp != local_smtp_off)
-  {
-  if (ob->use_crlf) write(fd, ".\r\n", 3);
-    else write(fd, ".\n", 2);
-  }
+if (tblock->local_smtp != local_smtp_off &&
+    !transport_write_string(fd, ".%s", eol))
+  goto END_WRITE;
+
+/* Flag all writing completed successfully. */
+
+written_ok = TRUE;
+
+/* Come here if there are errors during writing. */
+
+END_WRITE:
 
 /* OK, the writing is now all done. Close the pipe. */
 
 (void) close(fd);
 
+/* Handle errors during writing. If the process at the far end of the pipe
+died without reading all of it, we expect an EPIPE error, which should be
+ignored. We used also to ignore WRITEINCOMPLETE but the writing function is
+now cleverer at handling OS where the death of a pipe doesn't give EPIPE
+immediately. See comments therein. This change made 04-Sep-98. Clean up this
+code in a year or so. */
+
+if (!written_ok)
+  {
+  if (errno == EPIPE /* || errno == ERRNO_WRITEINCOMPLETE */ )
+    {
+    debug_printf("transport error %s ignored\n",
+      (errno == EPIPE)? "EPIPE" : "WRITEINCOMPLETE");
+    }
+  else
+    {
+    addr->transport_return = PANIC;
+    addr->basic_errno = errno;
+    if (errno == ERRNO_ADDHEADER_FAIL)
+      addr->message = string_sprintf("Failed to expand added headers");
+    else if (errno == ERRNO_FILTER_FAIL)
+      addr->message = string_sprintf("Filter process failure");
+    else if (errno == ERRNO_WRITEINCOMPLETE)
+      addr->message = string_sprintf("Failed repeatedly to write data");
+    else
+      addr->message = string_sprintf("Error %d", errno);
+    return;
+    }
+  }
+
 /* Wait for the child process to complete and take action if the returned
 status is nonzero or if something went wrong. */
 
@@ -797,8 +814,9 @@ if ((rc = child_close(pid, ob->timeout)) != 0)
 	 that a mailer (e.g.) could not create a connection,
 	 and the request should be reattempted later.
 
-  Based on this, use exit code EX_TEMPFAIL to mean "defer" when not ignoring
-  the returned status.
+  Based on this, we use exit code EX_TEMPFAIL as a default to mean "defer" when
+  not ignoring the returned status. However, there is now an option that
+  contains a list of temporary codes, with TEMPFAIL and CANTCREAT as defaults.
 
   Another case that needs special treatment is if execve() failed (typically
   the command that was given is a non-existent path). By default this is
@@ -833,8 +851,22 @@ if ((rc = child_close(pid, ob->timeout)) != 0)
 
     else if (!ob->ignore_status)
       {
+      char *s = ob->temp_errors;
+      char *p;
       int size, ptr, i;
-      addr->transport_return = (rc == EX_TEMPFAIL)? DEFER : FAIL;
+      char buffer[64];
+
+      addr->transport_return = FAIL;
+
+      /* Scan the list of temporary failure codes; if any match, the
+      result becomes DEFER. */
+
+      for (p = string_nextinlist(&s, ':', buffer, sizeof(buffer));
+           p != NULL;
+           p = string_nextinlist(&s, ':', buffer, sizeof(buffer)))
+        {
+        if (rc == atoi(p)) { addr->transport_return = DEFER; break; }
+        }
 
       /* Ensure the message contains the expanded command and arguments. This
       doesn't have to be brilliantly efficient - it is an error situation. */
