@@ -13,15 +13,13 @@ typedef struct {
   char *path;
   char *prefix;
   char *suffix;
+  char *temp_errors;
   int   umask;
   int   max_output;
   int   timeout;
   BOOL  freeze_exec_fail;
   BOOL  from_hack;
   BOOL  ignore_status;
-  BOOL  return_path_add;
-  BOOL  delivery_date_add;
-  BOOL  envelope_to_add;
   BOOL  restrict_to_path;
   BOOL  use_shell;
   BOOL  use_crlf;
