@@ -15,9 +15,9 @@ typedef struct {
   char *interface;
   char *service;
   char *mx_domains;
-  char *non_mx_domains;
+  char *mx_domains_except;
   re_block *re_mx_domains;
-  re_block *re_non_mx_domains;
+  re_block *re_mx_domains_except;
   char *serialize_hosts;
   re_block *re_serialize_hosts;
   char *serialize_nets;
@@ -28,11 +28,14 @@ typedef struct {
   int   data_timeout;
   int   final_timeout;
   int   size_addition;
+  int   max_rcpt;
   BOOL  allow_localhost;
   BOOL  gethostbyname;
   BOOL  dns_qualify_single;
   BOOL  dns_search_parents;
   BOOL  delay_after_cutoff;
+  BOOL  max_rcpt_serialize;
+  BOOL  retry_include_ip_address;
 } smtp_transport_options_block;
 
 /* Data for reading the private options. */
