@@ -59,7 +59,7 @@ if (addr->local_part_quoted)
   int i;
   int size = 64;
   int ptr = 0;
-  char *s = store_malloc(size);
+  char *s = store_get(size);
   char *p = prefix;
   s = string_cat(s, &size, &ptr, "\"", 1);
   for (i = 0; i < 3; i++)
@@ -151,7 +151,6 @@ address_item *addr_local = NULL;
 address_item *addr_succeed = NULL;
 char *ko_prefix, *cr;
 char *errmess;
-char *orig_receiver;
 char *receiver =
   parse_extract_address(s, &errmess, &start, &end, &domain, FALSE);
 
@@ -187,13 +186,10 @@ if (parse_find_at(receiver) == NULL)
   receiver = rewrite_address_qualify(receiver, is_recipient);
   }
 
-orig_receiver = receiver;
 DEBUG(9)
   debug_printf("%s %s\n", address_test? "Testing" : "Verifying", receiver);
 
-/* Rewrite and report on it. It is guaranteed that the address of a new
-string will be different to the old, even though the old one is freed in the
-process. */
+/* Rewrite and report on it. */
 
 if (rewrite_rules != NULL)
   {
@@ -207,7 +203,7 @@ if (rewrite_rules != NULL)
 
 /* Set up an initial address structure. */
 
-addr_new = addr_orig = deliver_make_addr(receiver);
+addr_new = addr_orig = deliver_make_addr(receiver, FALSE);
 
 /* We need a loop, since a directed address might generate a number of new
 addresses. We must also cope with generated pipes and files at the top
@@ -304,6 +300,8 @@ while (addr_new != NULL)
   f will not be NULL. Don't output anything for success in that case at this
   stage. Note that a director may set up local or remote delivery. */
 
+  verify_forced_errmsg = NULL;
+
   rc = (addr->local)?
     direct_address(addr, &addr_local, &addr_remote, &addr_new, &addr_succeed,
       verify_type) :
@@ -406,16 +404,16 @@ while (addr_new != NULL)
     if (addr_new == NULL)
       {
       if (addr_local == NULL)
-        fprintf(f, "250 mail to %s is discarded\r\n", s);
+        fprintf(f, "250 mail to <%s> is discarded\r\n", s);
       else
-        fprintf(f, "250 %s\r\n", addr_local->orig);
+        fprintf(f, "250 <%s>\r\n", addr_local->orig);
       }
     else while (addr_new != NULL)
       {
       address_item *addr = addr_new;
       addr_new = addr->next;
       if (addr_new == NULL) ok_prefix = "250 ";
-      fprintf(f, "%s%s\r\n", ok_prefix, addr->orig);
+      fprintf(f, "%s<%s>\r\n", ok_prefix, addr->orig);
       }
     return OK;
     }
@@ -554,7 +552,8 @@ of strings of the form [ident@]host, to scan the strings it contains and build
 a chain of control blocks of the broken-down or compiled data for subsequent
 use. If the host is precisely "@" it refers to the primary hostname. The daemon
 calls this at its start if it is listening for SMTP calls. It is also called
-from verify_sender_preliminary for other kinds of SMTP input.
+from verify_check_host for other kinds of call. The list is then re-used if
+necessary, and for that reason, any store that it gets must be got by malloc.
 
 Arguments:
   list      colon-separated list of items
@@ -570,7 +569,6 @@ char *s, *t;
 char buffer[1024];
 
 DEBUG(2) debug_printf("verify_setup_hostlist called\n");
-if (*anchor != NULL) return;                  /* Just in case */
 
 for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
      s != NULL;
@@ -581,17 +579,13 @@ for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
   *anchor = h;
   anchor = &(h->next);
 
-  /* Make a permanent copy of the string. */
-
-  s = string_copy(s);
-
   /* If there's an '@' in the string that is not the last character, it starts
   with an ident string; carve that off. */
 
   if ((t = strchr(s, '@')) != NULL && t[1] != 0)
     {
     *t = 0;
-    h->ident_string = s;
+    h->ident_string = string_copy_malloc(s);
     s = t+1;
     }
   else h->ident_string = NULL;
@@ -599,7 +593,7 @@ for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
   /* If the host name string is precisely "@", then substitute the primary
   host name. */
 
-  h->name = (strcmp(s, "@") == 0)? string_copy(primary_hostname) : s;
+  h->name = string_copy_malloc((strcmp(s, "@") == 0)? primary_hostname : s);
   h->compiled_name = NULL;
   h->address = NULL;
 
@@ -608,6 +602,15 @@ for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
 
   if (string_is_ip_address(h->name)) h->address = h->name;
 
+  /* For non-wildcarded names, the IP address used to be looked up here.
+  However, this meant that the daemon hung onto the address for an indefinite
+  amount of time, which is Considered Harmful. For non-daemon input, it also
+  meant that names that weren't actually going to be tested got looked up.
+  So we cut this code out (27-Aug-98). Tidy up the source once this has
+  established itself in service. */
+
+
+  #ifdef NEVER
   /* Else if the name is not wildcarded (starting with * or ^) or a file lookup
   (contains ';') then we can look up the IP address(es) once and for all.
   Ignore failure - this will just cause slower reverse lookups to happen later.
@@ -616,9 +619,12 @@ for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
   else if (s[0] != '*' && s[0] != '^' && strchr(s, ';') == NULL)
     {
     char *fully_qualified_name;
-    (void) host_find_byname(h, &fully_qualified_name);
+    (void) host_find_byname(h, &fully_qualified_name, TRUE);
     while (*anchor != NULL) anchor = &((*anchor)->next);  /* Multihomed */
     }
+  #endif
+
+
   }
 }
 
@@ -633,7 +639,7 @@ for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
 calling host (plus ident) is in a list of hosts + idents.
 
 Arguments:
-  list                 host list for checking
+  listptr              pointer to host list for checking
   anchor               anchor for broken down data (may already be done if
                          via daemon)
   find_failed_return   the value to return if the host name has to be found
@@ -646,12 +652,19 @@ Returns:     TRUE if current host+ident is in the list;
 */
 
 BOOL
-verify_check_host(char *list, host_item **anchor, BOOL find_failed_return)
+verify_check_host(char **listptr, host_item **anchor, BOOL find_failed_return)
 {
 host_item *h;
+char *list = *listptr;
 BOOL ffr = find_failed_return;
 char *sender_host_ipv4;
 
+DEBUG(2)
+  {
+  debug_printf("verify_check_host(%s) called\n", readconf_find_option(listptr));
+  DEBUG(9) debug_printf("  anchor=%d list=%s\n", *anchor, list);
+  }
+
 /* We can't do anything if the calling host's address is not available or if
 the list is empty. */
 
@@ -670,11 +683,10 @@ sender_host_ipv4 = (strncmp(sender_host_address, "::ffff:", 7) == 0)?
   sender_host_address + 7 : sender_host_address;
 
 /* Scan each listed item, checking the ident string if present, and then
-checking the address. For single names, the address will be in the host item.
-If the address isn't set in the host item, we have to match on the textual name
-of the host, which should normally be wildcarded, though this case can also
-arise if the forward lookup failed. Textual name matching requires a reverse
-lookup of the sender address, if that hasn't already been done.
+checking the address. If the address isn't already set in the host item, we
+can look it up directly only if the name isn't wildcarded. Otherwise we have to
+do a reverse lookup of the IP address in order to get the host name, if this
+hasn't already been done. Confirm this with a forward lookup afterwards.
 
 If a reverse lookup fails, we normally return find_failed_return. This is set
 true for calls to test reject lists and false for calls to test accept lists.
@@ -707,6 +719,26 @@ for (h = *anchor; h != NULL; h = h->next)
     if ((strcmp(ident, sender_ident) == 0) != match) continue;
     }
 
+  /* Optimize for the special case when the name is "*". */
+
+  if (strcmp(h->name, "*") == 0) return TRUE;
+
+  /* If we don't already have the IP address (from a previous check on this
+  list or as an additional address for a multihomed host), and the name
+  contains no wildcards or a lookup, try to find the IP address from a forward
+  lookup. Multihomed hosts will cause additional host items to get inserted
+  into the chain. That's fine - they will all get scanned by the loop. Ensure
+  they go into permanent store, though, by setting the last arg to
+  host_find_byname() TRUE. This means they will persist over a number of
+  incoming SMTP messages. */
+
+  if (h->address == NULL && h->name[0] != '*' && h->name[0] != '^' &&
+      strchr(h->name, ';') == NULL)
+    {
+    char *fully_qualified_name;
+    (void) host_find_byname(h, &fully_qualified_name, TRUE);
+    }
+
   /* If we have the IP address already, test that. If it is an IPv4 address
   ensure that IPv6 compatibility addresses are correctly compared. */
 
@@ -717,15 +749,14 @@ for (h = *anchor; h != NULL; h = h->next)
     if (strcmp(h->address, s) == 0) return TRUE; else continue;
     }
 
-  /* Otherwise match on the name. Optimise for the special case of "*".
-  If we can't find the host name, return the value given in find_failed_
-  return. */
-
-  if (strcmp(h->name, "*") == 0) return TRUE;
+  /* Otherwise we are going to need to match on the name. If it hasn't already
+  been found, look it up here, ensuring that it goes into permanent store, and
+  rebuilt the fullhost values. If we can't find the host name, return the
+  value given in find_failed_return. */
 
   if (sender_host_name == NULL)
     {
-    sender_host_name = host_find_byaddr(sender_host_address);
+    sender_host_name = host_find_byaddr(sender_host_address, TRUE);
     if (sender_host_name == NULL)
       {
       HDEBUG(2) debug_printf("failed to get host name for %s in "
@@ -736,6 +767,8 @@ for (h = *anchor; h != NULL; h = h->next)
     host_build_sender_fullhost();
     }
 
+  /* Match on the host name */
+
   if (match_check_string(sender_host_name, h->name, &(h->compiled_name), -1,
     TRUE, NULL)) return TRUE;
   if (search_find_defer)
@@ -762,7 +795,8 @@ build a chain of control blocks of the broken-down or compiled data for
 subsequent use. For IPv4 addresses, the mask is an integer, or in the form of
 another v4 address. For IPv6 addresses the mask is always an integer. The
 daemon calls this at its start if it is listening for SMTP calls. The string
-has already been syntax checked at read-in time.
+has already been syntax checked at read-in time. Because the resulting control
+blocks are saved for subsequent use, they must be put in malloc'd store.
 
 Arguments:
   list      colon-separated list of items
@@ -778,7 +812,6 @@ char *s;
 char buffer[256];
 
 DEBUG(2) debug_printf("verify_setup_netlist called\n");
-if (*anchor != NULL) return;                  /* Just in case */
 
 for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
      s != NULL;
@@ -797,7 +830,7 @@ for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
 
   if (*s == '/')
     {
-    n->filename = string_copy(s);
+    n->filename = string_copy_malloc(s);
     continue;
     }
 
@@ -821,22 +854,29 @@ for (s = string_nextinlist(&list, ':', buffer, sizeof(buffer));
 address of the current calling host is in a list of networks.
 
 Arguments:
-  list       net list for checking
+  listptr    pointer to net list for checking
   anchor     anchor for broken down data (may already be done if via daemon)
 
 Returns:     TRUE if IP address of current host is in the list
 */
 
 BOOL
-verify_check_net(char *list, ip_net_item **anchor)
+verify_check_net(char **listptr, ip_net_item **anchor)
 {
+char *list = *listptr;
+
+DEBUG(2)
+  {
+  debug_printf("verify_check_net(%s) called\n", readconf_find_option(listptr));
+  DEBUG(9) debug_printf("  anchor=%d list=%s\n", *anchor, list);
+  }
+
 /* We can't do anything if the calling host's address is not available, or if
 the list is empty. */
 
 if (sender_host_address == NULL || list == NULL) return FALSE;
 
-/* Initialize the munged data if not already done so (for calls via the
-daemon it should already be set up). */
+/* Initialize the munged data if not already done so. */
 
 if (*anchor == NULL) verify_setup_netlist(list, anchor);
 
@@ -871,20 +911,23 @@ fields, especially if they are resent- fields, check all of them if there is
 more than one.
 
 Arguments:
-  newaddr    address of pointer to point at the verified address, or
-               NULL, if value of the address not wanted
-  header_id  address of pointer to point at name of header used,
-               or NULL if not wanted
+  newaddr        address of pointer to point at the verified address, or
+                   NULL, if value of the address not wanted
+  header_id      address of pointer to point at name of header used,
+                   or NULL if not wanted
+  sender_yield   the yield to give if the sender_address is found (saves
+                   verifying it again)
 
-Returns:     result of the verification attempt: OK, FAIL, or DEFER;
-             FAIL is given if no appropriate headers are found
+Returns:         result of the verification attempt: OK, FAIL, or DEFER;
+                 FAIL is given if no appropriate headers are found
 */
 
 static int
-verify_get_header_address(char **newaddr, char **header_id)
+verify_get_header_address(char **newaddr, char **header_id, int sender_yield)
 {
 static int header_types[] = { htype_sender, htype_replyto, htype_from };
 static int header_name_offsets[] = { hn_sender, hn_replyto, hn_from };
+int senderlen = (int)strlen(sender_address);
 int yield = FAIL;
 int i;
 
@@ -904,9 +947,17 @@ for (i = 0; i < 3; i++)
     ss = parse_find_address_end(s, FALSE);
     terminator = *ss;
     *ss = 0;
-    new_ok =
+
+    /* Sender address in header may have whitespace on the end; if this
+    test doesn't work, it isn't really harmful other than that a redundant
+    verification gets done. */
+
+    new_ok = (strncmp(s, sender_address, senderlen) == 0 &&
+              (s[senderlen] == 0 || isspace(s[senderlen])))?
+      sender_yield :
       verify_address(s, NULL, &is_local, newaddr,
         sender_verify_log_details? vopt_log_details : 0);
+
     *ss = terminator;
 
     if (new_ok == OK)
@@ -991,11 +1042,13 @@ then MAIL FROM is accepted, but a flag is set to cause all subsequent RCPT TO
 commands to be rejected.
 
 Arguments:
-  errcode    set this to an SMTP error code on failure
+  errcode    set this to an SMTP error code on failure, or if there is a
+               warning message on success
   errmess    set this to point to an error message on failure
 
 Returns:     TRUE if address verified, or did not fail recently, or host
-               is in the exception list, i.e. if is OK to proceed
+               is in the exception list, i.e. if is OK to proceed. The
+               variable sender_ok indicates why it was allowed to proceed.
 */
 
 BOOL
@@ -1005,7 +1058,8 @@ char *newaddr;
 open_db dbblock;
 open_db *dbm_file;
 dbdata_reject *reject;
-int rejectlen;
+int rc, rejectlen, yield;
+time_t now;
 char buffer[SENDER_ADDRESS_MAXLENGTH + 256];
 
 /* If the sender address is empty, it's an error message with, in effect,
@@ -1020,9 +1074,9 @@ if (sender_address[0] == 0)
 /* See if this is one of the trusted hosts/identd combinations or nets
 for which we accept all addresses. If so, do no further checking. */
 
-if (verify_check_host(sender_verify_except_hosts,
+if (verify_check_host(&sender_verify_except_hosts,
       &sender_verify_except_hostlist, FALSE) ||
-    verify_check_net(sender_verify_except_nets,
+    verify_check_net(&sender_verify_except_nets,
       &sender_verify_except_netlist))
   {
   HDEBUG(2) debug_printf("host matched entry in sender_verify_except list\n");
@@ -1030,126 +1084,189 @@ if (verify_check_host(sender_verify_except_hosts,
   return TRUE;
   }
 
+/* Try to verify the address */
+
 HDEBUG(2) debug_printf("verifying %s\n", sender_address);
 
-switch(verify_address(sender_address, NULL, &sender_is_local, &newaddr,
-    sender_verify_log_details? vopt_log_details : 0))
-  {
-  /* After a successful return, the address may have been changed (typically
-  a domain will be canonicized or expanded by a router). A new copy is always
-  given, but not if the orginal was specifically rewritten. */
+rc = verify_address(sender_address, NULL, &sender_is_local, &newaddr,
+    sender_verify_log_details? vopt_log_details : 0);
 
-  case OK:
+/* After a successful return, the address may have been changed (typically
+a domain will be canonicized or expanded by a router). */
+
+if (rc == OK)
+  {
   DEBUG(2) debug_printf("%s verified ok as %s\n", sender_address, newaddr);
-  store_free(sender_address);
   sender_address = newaddr;
+  sender_address_rewritten = TRUE;  /* Prevent accept from doing it again */
   sender_ok = SENDER_OK_OK;
   return TRUE;
+  }
 
-  /* Defer is usually a DNS time out. If try_verify switch is set, which means
-  accept if soft error, accept it with an appropriate message. Otherwise, pass
-  it on, leaving sender_ok set to DEFER, leaving possible (temporary) rejection
-  to the second verification function. */
+/* If we are not going to reject permanent failures or if we are going to
+accept temporary failures and it is one of them, accept the address, and note
+it is a "warning" state. */
 
-  case DEFER:
-  sender_ok = SENDER_OK_DEFER;
-  if (sender_try_verify)
+if (!sender_verify_reject || (rc == DEFER && sender_try_verify))
+  {
+  if (rc == DEFER)
     {
-    DEBUG (2) debug_printf("%s verification deferred: accepted unverified\n",
-      sender_address);
-    *errcode = 250;
     *errmess = "warning: temporarily unable to resolve sender address: "
       "accepted unverified";
     if (sender_verify_log_details)
       log_write(0, LOG_REJECT, "%s verification deferred but sender_try_verify "
         "is true", sender_address);
-    sender_ok = SENDER_OK_WARNING;
     }
-  return TRUE;
-  }
-
-/* Otherwise it's a hard failure. If rejection is not required, accept with
-a warning. */
-
-if (!sender_verify_reject)
-  {
+  else
+    {
+    *errmess = sender_is_local?
+      "warning: unknown local-part in sender address" :
+      "warning: cannot route to sender address";
+    if (sender_verify_log_details)
+      log_write(0, LOG_REJECT, "%s failed verification but "
+        "sender_verify_reject is false", sender_address);
+    }
   *errcode = 250;
-  *errmess = sender_is_local?
-    "warning: unknown local-part in sender address" :
-    "warning: cannot route to sender address";
-
-  if (sender_verify_log_details)
-    log_write(0, LOG_REJECT, "%s failed verification but sender_verify_reject "
-      "is false", sender_address);
-
   sender_ok = SENDER_OK_WARNING;
   return TRUE;
   }
 
-/* It's really unacceptable */
-
-sender_ok = SENDER_OK_NOT;
-
-/* See if this address from this host has recently been rejected. If it has
-not, pass now, leaving rejection to the second function, which will create
-a DBM entry if necessary (it might not be necessary if the sender is fixed
-up from the headers). */
-
-dbm_file = dbfn_open("reject", O_RDWR, &dbblock);
-if (dbm_file == NULL)
-  {
-  HDEBUG(2)
-    debug_printf("%s verification failed - will give error after DATA\n",
-      sender_address);
-  return TRUE;
-  }
+/* Set up the key for the reject hints database, and attempt to open it.
+If successful, read the record. */
 
 sprintf(buffer, "%s:%.200s", sender_address,
   (sender_host_name != NULL)? sender_host_name :
   (sender_host_address != NULL)? sender_host_address : "");
-reject = dbfn_read_with_length(dbm_file, buffer, &rejectlen);
 
-if (reject == NULL || time(NULL) - reject->time_stamp > 24*60*60)
+dbm_file = dbfn_open("reject", O_RDWR, &dbblock);
+reject = (dbm_file == NULL)?
+  NULL : dbfn_read_with_length(dbm_file, buffer, &rejectlen);
+
+/* Ignore data that is more than 24 hours old */
+
+now = time(NULL);
+if (reject != NULL && now - reject->time_stamp > 24*60*60) reject = NULL;
+
+/* Defer is usually a DNS time out. If we have given this temporary rejection
+too often recently, convert it into a permanent failure. Once the failure is
+permanent, subsequent temporary failures are all made permanent, until the hint
+expires (currently 24 hours). Otherwise accept the address for now, leaving
+sender_ok set to DEFER, thus postponing possible (temporary) rejection to the
+second verification function. */
+
+if (rc == DEFER)
+  {
+  float temp_rate;
+
+  sender_ok = SENDER_OK_DEFER;
+
+  /* If there's no hints file, or no (recent) record in it, accept the address
+  for now. This enables the headers to be read (for logging) and a rejection
+  to be given later - or maybe the address can be fixed up from the headers. */
+
+  if (dbm_file == NULL || reject == NULL)
+    {
+    HDEBUG(2) debug_printf("temporary error while verifying %s - will give "
+      "error after DATA\n", sender_address);
+    if (dbm_file != NULL) dbfn_close(dbm_file);
+    return TRUE;
+    }
+
+  /* Control gets here if there has been a temporary or permanent rejection of
+  this address recently. If we read a short record, this is old data from a
+  previous version of Exim, which did not store temp_rate fields because it
+  kept hints for permanent rejection only. Treat as a permanent rejection. This
+  temporary code was installed in September 1998. Remove in a couple of years'
+  time. */
+
+  temp_rate = (rejectlen < sizeof(dbdata_reject))? 0.0 : reject->temp_rate;
+
+  /* If temp_rate is zero, the previous rejection was permanent, so this one
+  must be as well. Otherwise, if the max retry rate is set, we convert to a
+  permanent error if it is exceeded. The test here is for continuing: failure
+  is falling through. The first failure of an address will have succeeded here
+  and given the error later, so as to log the headers. Subsequent failures can
+  therefore give the failure here, updating the rate as they do so. */
+
+  if (temp_rate > 0.0 &&
+       (sender_verify_max_retry_rate <= 0 ||
+        (int)temp_rate < sender_verify_max_retry_rate))
+    {
+    reject->temp_rate = ((reject->temp_rate + 1.0) * 3600) /
+                        (now - reject->time_stamp + 3600);
+    dbfn_write(dbm_file, buffer, reject, sizeof(dbdata_reject));
+    dbfn_close(dbm_file);
+    DEBUG(2) debug_printf("%s verification deferred\n", sender_address);
+    *errcode = 451;
+    *errmess = "temporarily unable to verify sender address (try again later)";
+    return FALSE;
+    }
+
+  HDEBUG(2)
+    {
+    if (temp_rate == 0.0)
+      debug_printf("%s verification deferred after previous failure - "
+        "treating as permanent error\n", sender_address);
+    else
+      debug_printf("%s verification deferred too frequently - "
+        "forcing a permanent error\n", sender_address);
+    }
+  }
+
+/* If control gets here it's a hard failure, possibly converted from a
+temporary failure that happened too often. */
+
+sender_ok = SENDER_OK_NOT;
+
+/* If there has been no recent rejection, pass now, leaving rejection to the
+second function, which will create a DBM entry if necessary (it might not be
+necessary if the sender is fixed up from the headers). */
+
+if (dbm_file == NULL || reject == NULL)
   {
-  dbfn_close(dbm_file);
   HDEBUG(2)
     debug_printf("%s verification failed - will give error after DATA\n",
       sender_address);
+  if (dbm_file != NULL) dbfn_close(dbm_file);
   return TRUE;
   }
 
-/* There's been a recent rejection. If there has only been one, reject now
-(i.e. reject the MAIL FROM) and up the count. For compatibility with older
-versions of Exim that didn't have the count, inspect the length of the
-record. This change made in September 1996; after a few months this could
-be removed. */
+/* There's been a recent rejection. If the previous rejection was after DATA,
+reject now (i.e. reject the MAIL FROM), and set the flag to say so. */
 
-if (rejectlen < sizeof(dbdata_reject) || !reject->rejected_mail_from)
+if (!reject->rejected_mail_from)
   {
-  dbdata_reject newreject;
   *errcode = 550;
   *errmess = sender_is_local?
     "unknown local-part in sender" : "cannot route to sender";
-  newreject.rejected_mail_from = TRUE;
-  dbfn_write(dbm_file, buffer, &newreject, sizeof(dbdata_reject));
-  dbfn_close(dbm_file);
+  reject->rejected_mail_from = TRUE;
   DEBUG(2) debug_printf("%s verification failed after MAIL FROM\n",
     sender_address);
-  return FALSE;
+  yield = FALSE;
   }
 
 /* There has been a previous recent rejection after MAIL FROM; the mailer
 at the far end is horribly broken. Allow through this MAIL FROM with warning
 text, but set refuse_all_rcpts to cause all RCPT TO commands to be failed
-with 550 - which seems to be the only thing some mailers understand. */
+with 550 - which seems to be the only thing some mailers understand. We still
+need to rewrite the hint record to update the timestamp. */
 
+else
+  {
+  refuse_all_rcpts = TRUE;
+  *errcode = 250;
+  *errmess = "reject all recipients: 3 times bad sender";
+  HDEBUG(2) debug_printf("%s verification failed - will reject all recipients\n",
+    sender_address);
+  yield = TRUE;
+  }
+
+/* Update the hint and return; reject_rate = 0.0 => permanent error */
+
+reject->temp_rate = 0.0;
+dbfn_write(dbm_file, buffer, reject, sizeof(dbdata_reject));
 dbfn_close(dbm_file);
-refuse_all_rcpts = TRUE;
-*errcode = 250;
-*errmess = "reject all recipients: 3 times bad sender";
-HDEBUG(2) debug_printf("%s verification failed - will reject all recipients\n",
-  sender_address);
-return TRUE;
+return yield;
 }
 
 
@@ -1179,7 +1296,8 @@ verify_sender(int *errcode, char **errmess)
 {
 open_db dbblock;
 open_db *dbm_file;
-dbdata_reject reject;
+dbdata_reject newreject;
+int rejectlen;
 char buffer[256];
 
 /* The sender verified correctly on the preliminary check. However, sometimes
@@ -1195,7 +1313,7 @@ switch(sender_ok)
   if (sender_verify_fixup)
     {
     char *newaddr, *header_id;
-    if (verify_get_header_address(&newaddr, &header_id) == OK)
+    if (verify_get_header_address(&newaddr, &header_id, FAIL) == OK)
       log_write(0, LOG_MAIN, "return-path %s is rewritable as %s using %s",
         sender_address, newaddr, header_id);
     }
@@ -1209,7 +1327,7 @@ switch(sender_ok)
   if (headers_sender_verify ||
       (sender_address[0] == 0 && headers_sender_verify_errmsg))
     {
-    switch (verify_get_header_address(NULL, NULL))
+    switch (verify_get_header_address(NULL, NULL, OK))
       {
       case DEFER:
       if (headers_checks_fail)
@@ -1264,7 +1382,7 @@ address. */
 if (sender_verify_fixup)
   {
   char *newaddr, *header_id;
-  if (verify_get_header_address(&newaddr, &header_id) == OK)
+  if (verify_get_header_address(&newaddr, &header_id, FAIL) == OK)
     {
     log_write(0, LOG_MAIN, "return-path %s rewritten as %s using %s",
       sender_address, newaddr, header_id);
@@ -1276,39 +1394,43 @@ if (sender_verify_fixup)
     }
   }
 
-/* If the original verification attempt deferred, give a temporary error
-return. */
+/* Set up for temporary or permanent rejection */
 
 if (sender_ok == SENDER_OK_DEFER)
   {
   DEBUG(2) debug_printf("%s verification deferred\n", sender_address);
   *errcode = 451;
   *errmess = "temporarily unable to verify sender address (try again later)";
-  return FALSE;
+  }
+else
+  {
+  DEBUG(2) debug_printf("%s verification failed after data\n", sender_address);
+  *errcode = 550;
+  *errmess = sender_is_local?
+    "unknown local-part in sender" : "cannot route to sender";
   }
 
-/* Otherwise it's a hard failure. Update the database to record this rejection.
-O_RDWR (rather than O_WRONLY) is needed by Berkeley native DB. If opening
-fails, don't worry. */
+/* Set up the key name and open the hints database. O_RDWR (rather than
+O_WRONLY) is needed by Berkeley native DB even when reading only. If the
+database won't open, we can do no more. */
+
+sprintf(buffer, "%s:%.200s", sender_address,
+  (sender_host_name != NULL)? sender_host_name :
+  (sender_host_address != NULL)? sender_host_address : "");
 
 dbm_file = dbfn_open("reject", O_RDWR|O_CREAT, &dbblock);
+
+/* Write a new record for this failure. A temporary failure rate of 0.0
+indicates a permanent error. Otherwise initialize to "1 per hour". */
+
 if (dbm_file != NULL)
   {
-  sprintf(buffer, "%s:%.200s", sender_address,
-    (sender_host_name != NULL)? sender_host_name :
-    (sender_host_address != NULL)? sender_host_address : "");
-  reject.rejected_mail_from = FALSE;
-  dbfn_write(dbm_file, buffer, &reject, sizeof(dbdata_reject));
+  newreject.rejected_mail_from = FALSE;
+  newreject.temp_rate = (sender_ok == SENDER_OK_DEFER)? 1.0 : 0.0;
+  dbfn_write(dbm_file, buffer, &newreject, sizeof(dbdata_reject));
   dbfn_close(dbm_file);
   }
 
-/* Now give a hard error */
-
-*errcode = 550;
-*errmess = sender_is_local?
-  "unknown local-part in sender" : "cannot route to sender";
-
-DEBUG(2) debug_printf("%s verification failed after data\n", sender_address);
 return FALSE;
 }
 
@@ -1340,17 +1462,17 @@ verify_get_ident(int socket)
 sender_ident = NULL;
 if (rfc1413_query_timeout > 0 &&
     (rfc1413_except_hosts == NULL ||
-      !verify_check_host(rfc1413_except_hosts, &rfc1413_except_hostlist, FALSE))
+      !verify_check_host(&rfc1413_except_hosts, &rfc1413_except_hostlist, FALSE))
     &&
     (rfc1413_except_nets == NULL ||
-      !verify_check_net(rfc1413_except_nets, &rfc1413_except_netlist)))
+      !verify_check_net(&rfc1413_except_nets, &rfc1413_except_netlist)))
   {
   char *ident_ptr = ident_id(socket, rfc1413_query_timeout);
   if (ident_ptr != NULL)
     {
     strncpy(ident_buffer, ident_ptr, sizeof(ident_buffer));
     ident_buffer[sizeof(ident_buffer) - 1] = 0;
-    sender_ident = string_printing(ident_buffer, FALSE);
+    sender_ident = string_printing(ident_buffer);
     /* NB: free() not store_free() as it wasn't got by store_malloc() */
     free(ident_ptr);
     }
